//----------------------------------------------------------------------------
//  EDGE Automap Functions
//----------------------------------------------------------------------------
// 
//  Copyright (c) 1999-2000  The EDGE Team.
// 
//  This program is free software; you can redistribute it and/or
//  modify it under the terms of the GNU General Public License
//  as published by the Free Software Foundation; either version 2
//  of the License, or (at your option) any later version.
//
//  This program is distributed in the hope that it will be useful,
//  but WITHOUT ANY WARRANTY; without even the implied warranty of
//  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//  GNU General Public License for more details.
//
//----------------------------------------------------------------------------
//
//  Based on the DOOM source code, released by Id Software under the
//  following copyright:
//
//    Copyright (C) 1993-1996 by id Software, Inc.
//
//----------------------------------------------------------------------------

#include "i_defs.h"
#include "am_map.h"

#include "con_cvar.h"
#include "con_main.h"
#include "dm_defs.h"
#include "dm_state.h"
#include "dstrings.h"
#include "m_argv.h"
#include "m_cheat.h"
#include "m_misc.h"
#include "p_local.h"
#include "r_state.h"
#include "st_stuff.h"
#include "w_wad.h"
#include "z_zone.h"

#define DEBUG_TRUEBSP  0

// For use if I do walls with outsides/insides
#define REDS        (256-5*16)
#define REDRANGE    16
#define BLUES       (256-4*16+8)
#define BLUERANGE   8
#define GREENS      (7*16)
#define GREENRANGE  16
#define GRAYS       (6*16)
#define GRAYSRANGE  16
#define BROWNS      (4*16)
#define BROWNRANGE  16
#define YELLOWS     (256-32+7)
#define YELLOWRANGE 1
#define BLACK       0
#define WHITE       (256-47)

// Automap colours
#define BACKGROUND  BLACK
#define YOURCOLOURS WHITE
#define YOURRANGE   0
#define WALLCOLOURS REDS
#define WALLRANGE   REDRANGE
#define TSWALLCOLOURS GRAYS
#define TSWALLRANGE   GRAYSRANGE
#define FDWALLCOLOURS BROWNS
#define FDWALLRANGE   BROWNRANGE
#define CDWALLCOLOURS YELLOWS
#define CDWALLRANGE   YELLOWRANGE
#define THINGCOLOURS  GREENS
#define THINGRANGE    GREENRANGE
#define SECRETWALLCOLOURS WALLCOLOURS
#define SECRETWALLRANGE   WALLRANGE
#define GRIDCOLOURS   (GRAYS + GRAYSRANGE/2)
#define GRIDRANGE     0
#define XHAIRCOLOURS  GRAYS

#define AM_PANDOWNKEY KEYD_DOWNARROW
#define AM_PANUPKEY   KEYD_UPARROW
#define AM_PANRIGHTKEY    KEYD_RIGHTARROW
#define AM_PANLEFTKEY     KEYD_LEFTARROW
#define AM_ZOOMINKEY  '='
#define AM_ZOOMOUTKEY '-'
#define AM_STARTKEY   key_map
#define AM_ENDKEY     key_map
#define AM_GOBIGKEY   '0'
#define AM_FOLLOWKEY  'f'
#define AM_GRIDKEY    'g'
#define AM_MARKKEY    'm'
#define AM_CLEARMARKKEY    'c'

#define AM_NUMMARKPOINTS 10

#define PLAYERRADIUS 16.0

// scale on entry
#define INITSCALEMTOF (0.2)
// how much the automap moves window per tic in frame-buffer coordinates
// moves 140 pixels in 1 second
#define F_PANINC 4
// how much zoom-in per tic
// goes to 2x in 1 second
#define M_ZOOMIN (1.02)
// how much zoom-out per tic
// pulls out to 0.5x in 1 second
#define M_ZOOMOUT (1/M_ZOOMIN)

// translates between frame-buffer and map distances - moved down with rest of functions
#define FTOM(x) ((float_t)((x) * scale_ftom))
#define MTOF(x) ((int)((x) * scale_mtof))

// translates between frame-buffer and map coordinates
#define CXMTOF(x)  (f_x + MTOF((x) - m_x))
#define CYMTOF(y)  (f_y + f_h - MTOF((y) - m_y))

#define CXFTOM(x)  (m_x + FTOM((x) - f_x))
#define CYFTOM(y)  (m_y + FTOM(f_y + f_h - (y)))

//
// The vector graphics for the automap.
//  A line drawing of the player pointing right,
//   starting from the middle.
//
#define R ((8*PLAYERRADIUS)/7)
mline_t player_arrow[] =
{
    {
        {-R + R / 8, 0},
        {R, 0}},  // -----
     {
        {R, 0},
        {R - R / 2, R / 4}},  // ----->
     {
        {R, 0},
        {R - R / 2, -R / 4}},
    {
        {-R + R / 8, 0},
        {-R - R / 8, R / 4}},  // >---->
     {
        {-R + R / 8, 0},
        {-R - R / 8, -R / 4}},
    {
        {-R + 3 * R / 8, 0},
        {-R + R / 8, R / 4}},  // >>--->
     {
        {-R + 3 * R / 8, 0},
        {-R + R / 8, -R / 4}}
};

#undef R
#define NUMPLYRLINES (sizeof(player_arrow)/sizeof(mline_t))

#define R ((8*PLAYERRADIUS)/7)
mline_t cheat_player_arrow[] =
{
    {
        {-R + R / 8, 0},
        {R, 0}},  // -----
     {
        {R, 0},
        {R - R / 2, R / 6}},  // ----->
     {
        {R, 0},
        {R - R / 2, -R / 6}},
    {
        {-R + R / 8, 0},
        {-R - R / 8, R / 6}},  // >----->
     {
        {-R + R / 8, 0},
        {-R - R / 8, -R / 6}},
    {
        {-R + 3 * R / 8, 0},
        {-R + R / 8, R / 6}},  // >>----->
     {
        {-R + 3 * R / 8, 0},
        {-R + R / 8, -R / 6}},
    {
        {-R / 2, 0},
        {-R / 2, -R / 6}},  // >>-d--->
     {
        {-R / 2, -R / 6},
        {-R / 2 + R / 6, -R / 6}},
    {
        {-R / 2 + R / 6, -R / 6},
        {-R / 2 + R / 6, R / 4}},
    {
        {-R / 6, 0},
        {-R / 6, -R / 6}},  // >>-dd-->
     {
        {-R / 6, -R / 6},
        {0, -R / 6}},
    {
        {0, -R / 6},
        {0, R / 4}},
    {
        {R / 6, R / 4},
        {R / 6, -R / 7}},  // >>-ddt->
     {
        {R / 6, -R / 7},
        {R / 6 + R / 32, -R / 7 - R / 32}},
    {
        {R / 6 + R / 32, -R / 7 - R / 32},
        {R / 6 + R / 10, -R / 7}}
};

#undef R
#define NUMCHEATPLYRLINES (sizeof(cheat_player_arrow)/sizeof(mline_t))

#define R (1.0)
// -ES- 1999/10/16 Cos(60) and Sin(60)
#define C (0.5 * R)
#define S (0.8660254038 * R)
mline_t triangle_guy[] =
{
  { {-S, -C}, {S, -C}  },
  { {S, -C}, {0, R} },
  { {0, R}, {-S, -C} }
};

#undef C
#undef S
#define NUMTRIANGLEGUYLINES (sizeof(triangle_guy)/sizeof(mline_t))

#define R (1.0)
mline_t thintriangle_guy[] =
{
  { {-.5 * R, -.7 * R}, {R, 0}             },
  { {R, 0},             {-.5 * R, .7 * R}  },
  { {-.5 * R, .7 * R},  {-.5 * R, -.7 * R} }
};

#undef R
#define NUMTHINTRIANGLEGUYLINES (sizeof(thintriangle_guy)/sizeof(mline_t))

static int cheating = 0;
static int grid = 0;

static int leveljuststarted = 1;  // kluge until LevelInit() is called

int automapactive = 0;
static int finit_width;
static int finit_height;

// location of window on screen
static int f_x;
static int f_y;

// size of window on screen
static int f_w;
static int f_h;

static int lightlev;  // used for funky strobing effect

static int amclock;

static mpoint_t m_paninc;  // how far the window pans each tic (map coords)

static float_t mtof_zoommul;  // how far the window zooms in each tic (map coords)

static float_t ftom_zoommul;  // how far the window zooms in each tic (fb coords)

static float_t m_x, m_y;  // LL x,y where the window is on the map (map coords)

static float_t m_x2, m_y2;  // UR x,y where the window is on the map (map coords)

//
// width/height of window on map (map coords)
//
static float_t m_w;
static float_t m_h;

// based on level size
static float_t min_x;
static float_t min_y;
static float_t max_x;
static float_t max_y;

static float_t max_w;  // max_x-min_x,

static float_t max_h;  // max_y-min_y

// based on player size
static float_t min_w;
static float_t min_h;

static float_t min_scale_mtof;  // used to tell when to stop zooming out

static float_t max_scale_mtof;  // used to tell when to stop zooming in

// old stuff for recovery later
static float_t old_m_w, old_m_h;
static float_t old_m_x, old_m_y;

// old location used by the Follower routine
static mpoint_t f_oldloc;

// used by MTOF to scale from map-to-frame-buffer coords
static float_t scale_mtof = INITSCALEMTOF;

// used by FTOM to scale from frame-buffer-to-map coords (=1/scale_mtof)
static float_t scale_ftom;

static player_t *plr;  // the player represented by an arrow

static const patch_t *marknums[10];  // numbers used for marking by the automap

static mpoint_t markpoints[AM_NUMMARKPOINTS];  // where the points are

static int markpointnum = 0;  // next point to be assigned

static int followplayer = 1;  // specifies whether to follow the player around

//static unsigned char cheat_amap_seq[] = { 0xb2, 0x26, 0x26, 0x2e, 0xff };
//static unsigned char cheat_amap_seq[] = { 'i','d','d','t', 0xff };
cheatseq_t cheat_amap =
{0, 0};

static boolean_t stopped = true;

boolean_t newhud = false;

// Calculates the slope and slope according to the x-axis of a line
// segment in map coordinates (with the upright y-axis n' all) so
// that it can be used with the brain-dead drawing stuff.

static void GetIslope(mline_t * ml, islope_t * is)
{
  float_t dx, dy;

  dy = ml->a.y - ml->b.y;
  dx = ml->b.x - ml->a.x;
  if (dy == 0)
    is->islp = (dx < 0 ? INT_MIN : INT_MAX);
  else
    is->islp = dx / dy;
  if (dx == 0)
    is->slp = (dy < 0 ? INT_MIN : INT_MAX);
  else
    is->slp = dy / dx;

}

//
//
//
static void ActivateNewScale(void)
{
  m_x += m_w / 2;
  m_y += m_h / 2;
  m_w = FTOM(f_w);
  m_h = FTOM(f_h);
  m_x -= m_w / 2;
  m_y -= m_h / 2;
  m_x2 = m_x + m_w;
  m_y2 = m_y + m_h;
}

//
//
//
static void SaveScaleAndLoc(void)
{
  old_m_x = m_x;
  old_m_y = m_y;
  old_m_w = m_w;
  old_m_h = m_h;
}

//
//
//
static void RestoreScaleAndLoc(void)
{

  m_w = old_m_w;
  m_h = old_m_h;
  if (!followplayer)
  {
    m_x = old_m_x;
    m_y = old_m_y;
  }
  else
  {
    m_x = plr->mo->x - m_w / 2;
    m_y = plr->mo->y - m_h / 2;
  }
  m_x2 = m_x + m_w;
  m_y2 = m_y + m_h;

  // Change the scaling multipliers
  scale_mtof = (float_t)f_w / m_w;
  scale_ftom = 1 / scale_mtof;
}

//
// adds a marker at the current location
//
static void AddMark(void)
{
  markpoints[markpointnum].x = m_x + m_w / 2;
  markpoints[markpointnum].y = m_y + m_h / 2;
  markpointnum = (markpointnum + 1) % AM_NUMMARKPOINTS;
}

//
// Determines bounding box of all vertices,
// sets global variables controlling zoom range.
//
static void FindMinMaxBoundaries(void)
{
  int i;
  float_t a;
  float_t b;

  min_x = min_y = INT_MAX;
  max_x = max_y = INT_MIN;

  for (i = 0; i < numvertexes; i++)
  {
    if (vertexes[i].x < min_x)
      min_x = vertexes[i].x;
    else if (vertexes[i].x > max_x)
      max_x = vertexes[i].x;

    if (vertexes[i].y < min_y)
      min_y = vertexes[i].y;
    else if (vertexes[i].y > max_y)
      max_y = vertexes[i].y;
  }

  max_w = max_x - min_x;
  max_h = max_y - min_y;

  min_w = 2 * PLAYERRADIUS;  // const? never changed?

  min_h = 2 * PLAYERRADIUS;

  a = (float_t)f_w / max_w;
  b = (float_t)f_h / max_h;

  min_scale_mtof = a < b ? a : b;
  max_scale_mtof = (float_t)f_h / (2 * PLAYERRADIUS);
}

//
//
//
static void ChangeWindowLoc(void)
{
  if (m_paninc.x != 0 || m_paninc.y != 0)
  {
    followplayer = 0;
    f_oldloc.x = INT_MAX;
  }

  m_x += m_paninc.x;
  m_y += m_paninc.y;

  if (m_x + m_w / 2 > max_x)
    m_x = max_x - m_w / 2;
  else if (m_x + m_w / 2 < min_x)
    m_x = min_x - m_w / 2;

  if (m_y + m_h / 2 > max_y)
    m_y = max_y - m_h / 2;
  else if (m_y + m_h / 2 < min_y)
    m_y = min_y - m_h / 2;

  m_x2 = m_x + m_w;
  m_y2 = m_y + m_h;
}

//
//
//
static void InitVariables(void)
{
  if (newhud == true)
    automapactive = 1;
  else
    automapactive = 2;

  f_oldloc.x = INT_MAX;
  amclock = 0;
  lightlev = 0;

  m_paninc.x = m_paninc.y = 0;
  ftom_zoommul = 1.0;
  mtof_zoommul = 1.0;

  m_w = FTOM(f_w);
  m_h = FTOM(f_h);

  // find player to center on initially
  plr = consoleplayer;
  if (!plr->in_game)
    for (plr = players; plr; plr = plr->next)
      if (plr->in_game)
        break;

        
  DEV_ASSERT2(plr);

  m_x = plr->mo->x - m_w / 2;
  m_y = plr->mo->y - m_h / 2;

  ChangeWindowLoc();

  // for saving & restoring
  old_m_x = m_x;
  old_m_y = m_y;
  old_m_w = m_w;
  old_m_h = m_h;

  // inform the status bar of the change
  stbar_update = true;
}

//
// 
//
static void LoadPics(void)
{
  int i;
  char namebuf[9];

  for (i = 0; i < 10; i++)
  {
    sprintf(namebuf, "AMMNUM%d", i);
    marknums[i] = W_CacheLumpName(namebuf);
  }
}

static void UnloadPics(void)
{
  int i;

  for (i = 0; i < 10; i++)
    W_DoneWithLump(marknums[i]);

}

static void ClearMarks(void)
{
  int i;

  for (i = 0; i < AM_NUMMARKPOINTS; i++)
    markpoints[i].x = -1;  // means empty

  markpointnum = 0;
}

//
// should be called at the start of every level
// right now, i figure it out myself
//
static void LevelInit(void)
{
  if (!cheat_amap.sequence)
    cheat_amap.sequence = DDF_LanguageLookup("iddt");

  leveljuststarted = 0;

  f_x = f_y = 0;
  f_w = finit_width;
  f_h = finit_height;

  ClearMarks();

  FindMinMaxBoundaries();

  scale_mtof = min_scale_mtof * (10.0/7);

  if (scale_mtof > max_scale_mtof)
    scale_mtof = min_scale_mtof;

  scale_ftom = 1 / scale_mtof;
}

//
//
//
void AM_Stop(void)
{
  UnloadPics();
  automapactive = 0;
  stopped = true;
}

//
//
//
static void StartAM(void)
{
  // static int lastlevel = -1, lastepisode = -1;

  if (!stopped)
    AM_Stop();

  LevelInit();

  InitVariables();
  LoadPics();

  stopped = false;
}

void AM_InitResolution(void)
{
  finit_width = SCREENWIDTH;
  finit_height = SCREENHEIGHT - 32;

  LevelInit();  // -ES- 1998/08/20

  CON_CreateCVarBool("newhud", cf_normal, &newhud);

  if (M_CheckParm("-newmap"))
    newhud = true;
}

//
// Hides the map.
//
static void AM_Hide(void)
{
  automapactive = 0;
  viewactive = true;
}

static void AM_Show(void)
{
  if (stopped)
    StartAM();

  if (newhud == true)
    automapactive = 1;
  else
    automapactive = 2;

  viewactive = false;
}

//
// set the window scale to the maximum size
//
static void MinOutWindowScale(void)
{
  scale_mtof = min_scale_mtof;
  scale_ftom = 1 / scale_mtof;
  ActivateNewScale();
}

//
// set the window scale to the minimum size
//
static void MaxOutWindowScale(void)
{
  scale_mtof = max_scale_mtof;
  scale_ftom = 1 / scale_mtof;
  ActivateNewScale();
}

//
// Handle events (user inputs) in automap mode
//
boolean_t AM_Responder(event_t * ev)
{

  int rc;
  static int bigstate = 0;

  rc = false;

  if (!automapactive)
  {
    if (ev->type == ev_keydown && ((ev->value.key == (AM_STARTKEY >> 16)) || (ev->value.key == (AM_STARTKEY & 0xffff))))
    {
      AM_Show();
      rc = true;
    }
  }
  else if (ev->type == ev_keydown)
  {

    rc = true;
    switch (ev->value.key)
    {
      case AM_PANRIGHTKEY:
        // pan right
        if (!followplayer)
          m_paninc.x = FTOM(F_PANINC);
        else
          rc = false;
        break;
      case AM_PANLEFTKEY:
        // pan left
        if (!followplayer)
          m_paninc.x = -FTOM(F_PANINC);
        else
          rc = false;
        break;
      case AM_PANUPKEY:
        // pan up
        if (!followplayer)
          m_paninc.y = FTOM(F_PANINC);
        else
          rc = false;
        break;
      case AM_PANDOWNKEY:
        // pan down
        if (!followplayer)
          m_paninc.y = -FTOM(F_PANINC);
        else
          rc = false;
        break;
      case AM_ZOOMOUTKEY:
        // zoom out
        mtof_zoommul = M_ZOOMOUT;
        ftom_zoommul = M_ZOOMIN;
        break;
      case AM_ZOOMINKEY:
        // zoom in
        mtof_zoommul = M_ZOOMIN;
        ftom_zoommul = M_ZOOMOUT;
        break;
      case AM_GOBIGKEY:
        bigstate = !bigstate;
        if (bigstate)
        {
          SaveScaleAndLoc();
          MinOutWindowScale();
        }
        else
          RestoreScaleAndLoc();
        break;
      case AM_FOLLOWKEY:
        followplayer = !followplayer;
        f_oldloc.x = INT_MAX;
        // -ACB- 1998/08/10 Use DDF Lang Reference
        if (followplayer)
          CON_PlayerMessageLDF(plr, "AutoMapFollowOn");
        else
          CON_PlayerMessageLDF(plr, "AutoMapFollowOff");
        break;
      case AM_GRIDKEY:
        grid = !grid;
        // -ACB- 1998/08/10 Use DDF Lang Reference
        if (grid)
          CON_PlayerMessageLDF(plr, "AutoMapGridOn");
        else
          CON_PlayerMessageLDF(plr, "AutoMapGridOff");
        break;
      case AM_MARKKEY:
        // -ACB- 1998/08/10 Use DDF Lang Reference
        CON_PlayerMessage(plr, "%s %d",
            DDF_LanguageLookup("AutoMapMarkedSpot"),
            markpointnum);
        AddMark();
        break;
      case AM_CLEARMARKKEY:
        ClearMarks();
        // -ACB- 1998/08/10 Use DDF Lang Reference
        CON_PlayerMessageLDF(plr, "AutoMapMarksClear");
        break;
      default:
        if (ev->value.key == (AM_ENDKEY >> 16) || ev->value.key == (AM_ENDKEY & 0xffff))
        {
          if (automapactive == 1)
          {
            stbar_update = true;
            automapactive = 2;
          }
          else
          {
            bigstate = 0;
            viewactive = true;
            AM_Hide();
          }
        }
        else
        {
          rc = false;
        }
    }
    // -ACB- 1999/09/28 Proper casting
    if (!deathmatch && M_CheckCheat(&cheat_amap, (char)ev->value.key))
    {
      rc = false;
      cheating = (cheating + 1) % 3;
    }
  }

  else if (ev->type == ev_keyup)
  {
    rc = false;
    switch (ev->value.key)
    {
      case AM_PANRIGHTKEY:
        if (!followplayer)
          m_paninc.x = 0;
        break;
      case AM_PANLEFTKEY:
        if (!followplayer)
          m_paninc.x = 0;
        break;
      case AM_PANUPKEY:
        if (!followplayer)
          m_paninc.y = 0;
        break;
      case AM_PANDOWNKEY:
        if (!followplayer)
          m_paninc.y = 0;
        break;
      case AM_ZOOMOUTKEY:
      case AM_ZOOMINKEY:
        mtof_zoommul = 1.0;
        ftom_zoommul = 1.0;
        break;
    }
  }

  return rc;

}

//
// Zooming
//
static void ChangeWindowScale(void)
{

  // Change the scaling multipliers
  scale_mtof *= mtof_zoommul;
  scale_ftom = 1.0 / scale_mtof;

  if (scale_mtof < min_scale_mtof)
    MinOutWindowScale();
  else if (scale_mtof > max_scale_mtof)
    MaxOutWindowScale();
  else
    ActivateNewScale();
}

//
//
//
static void DoFollowPlayer(void)
{

  if (f_oldloc.x != plr->mo->x || f_oldloc.y != plr->mo->y)
  {
    m_x = FTOM(MTOF(plr->mo->x)) - m_w / 2;
    m_y = FTOM(MTOF(plr->mo->y)) - m_h / 2;
    m_x2 = m_x + m_w;
    m_y2 = m_y + m_h;
    f_oldloc.x = plr->mo->x;
    f_oldloc.y = plr->mo->y;
  }
}

//
//
//
static void UpdateLightLev(void)
{
  static int nexttic = 0;

  //static int litelevels[] = { 0, 3, 5, 6, 6, 7, 7, 7 };
  static int litelevels[] =
  {0, 4, 7, 10, 12, 14, 15, 15};
  static int litelevelscnt = 0;

  // Change light level
  if (amclock > nexttic)
  {
    lightlev = litelevels[litelevelscnt++];
    if (litelevelscnt == sizeof(litelevels) / sizeof(int))
       litelevelscnt = 0;

    nexttic = amclock + 6 - (amclock % 6);
  }

}

//
// Updates on Game Tick
//
void AM_Ticker(void)
{

  if (!automapactive)
    return;

  amclock++;

  if (followplayer)
    DoFollowPlayer();

  // Change the zoom if necessary
  if (ftom_zoommul != 1.0)
    ChangeWindowScale();

  // Change x,y location
  if (m_paninc.x != 0 || m_paninc.y != 0)
    ChangeWindowLoc();

  // Update light level
  // UpdateLightLev();

}

//
// Rotation in 2D.
// Used to rotate player arrow line character.
//
static void Rotate(float_t * x, float_t * y, angle_t a)
{
  float_t tmpx;

  tmpx = *x * M_Cos(a) - *y * M_Sin(a);

  *y = *x * M_Sin(a) + *y * M_Cos(a);

  *x = tmpx;
}

//
// Clear automap frame buffer.
//
static void ClearFB(int colour)
{
  // -AJA- 1999/07/04: now uses V_DrawBox().

  V_DrawBox(main_scr, 0, 0, f_w, f_h, colour);
}

//
// Automap clipping of lines.
//
// Based on Cohen-Sutherland clipping algorithm but with a slightly
// faster reject and precalculated slopes.  If the speed is needed,
// use a hash algorithm to handle  the common cases.
//
static boolean_t ClipMline(mline_t * ml, fline_t * fl)
{
  enum
  {
    LEFT = 1,
    RIGHT = 2,
    BOTTOM = 4,
    TOP = 8
  };

  register int outcode1 = 0;
  register int outcode2 = 0;
  register int outside;

  fpoint_t tmp;
  int dx;
  int dy;

#define DOOUTCODE(oc, mx, my)\
        (oc) = 0;\
        if ((my) < 0) (oc) |= TOP;\
        else if ((my) >= f_h) (oc) |= BOTTOM;\
        if ((mx) < 0) (oc) |= LEFT;\
        else if ((mx) >= f_w) (oc) |= RIGHT;

  // do trivial rejects and outcodes
  if (ml->a.y > m_y2)
    outcode1 = TOP;
  else if (ml->a.y < m_y)
    outcode1 = BOTTOM;

  if (ml->b.y > m_y2)
    outcode2 = TOP;
  else if (ml->b.y < m_y)
    outcode2 = BOTTOM;

  if (outcode1 & outcode2)
    return false;  // trivially outside

  if (ml->a.x < m_x)
    outcode1 |= LEFT;
  else if (ml->a.x > m_x2)
    outcode1 |= RIGHT;

  if (ml->b.x < m_x)
    outcode2 |= LEFT;
  else if (ml->b.x > m_x2)
    outcode2 |= RIGHT;

  if (outcode1 & outcode2)
    return false;  // trivially outside

  // transform to frame-buffer coordinates.
  fl->a.x = CXMTOF(ml->a.x);
  fl->a.y = CYMTOF(ml->a.y);
  fl->b.x = CXMTOF(ml->b.x);
  fl->b.y = CYMTOF(ml->b.y);

  DOOUTCODE(outcode1, fl->a.x, fl->a.y);
  DOOUTCODE(outcode2, fl->b.x, fl->b.y);

  if (outcode1 & outcode2)
    return false;

  while (outcode1 | outcode2)
  {
    // may be partially inside box
    // find an outside point
    if (outcode1)
      outside = outcode1;
    else
      outside = outcode2;

    // clip to each side
    if (outside & TOP)
    {
      dy = fl->a.y - fl->b.y;
      dx = fl->b.x - fl->a.x;
      tmp.x = fl->a.x + (dx * (fl->a.y)) / dy;
      tmp.y = 0;
    }
    else if (outside & BOTTOM)
    {
      dy = fl->a.y - fl->b.y;
      dx = fl->b.x - fl->a.x;
      tmp.x = fl->a.x + (dx * (fl->a.y - f_h)) / dy;
      tmp.y = f_h - 1;
    }
    else if (outside & RIGHT)
    {
      dy = fl->b.y - fl->a.y;
      dx = fl->b.x - fl->a.x;
      tmp.y = fl->a.y + (dy * (f_w - 1 - fl->a.x)) / dx;
      tmp.x = f_w - 1;
    }
    else if (outside & LEFT)
    {
      dy = fl->b.y - fl->a.y;
      dx = fl->b.x - fl->a.x;
      tmp.y = fl->a.y + (dy * (-fl->a.x)) / dx;
      tmp.x = 0;
    }

    if (outside == outcode1)
    {
      fl->a = tmp;
      DOOUTCODE(outcode1, fl->a.x, fl->a.y);
    }
    else
    {
      fl->b = tmp;
      DOOUTCODE(outcode2, fl->b.x, fl->b.y);
    }

    if (outcode1 & outcode2)
      return false;  // trivially outside

  }

  return true;
}
#undef DOOUTCODE

// -AJA- 1999/07/04: moved the line drawing code to v_video1/2.c, and
// cleaned up the following two routines.

static void DrawFline(fline_t * fl, int colour)
{
#ifdef DEVELOPERS
  static int fuck = 0;

  // For debugging only
  if (fl->a.x < 0 || fl->a.x >= f_w
      || fl->a.y < 0 || fl->a.y >= f_h
      || fl->b.x < 0 || fl->b.x >= f_w
      || fl->b.y < 0 || fl->b.y >= f_h)
  {
    fprintf(stderr, "fuck %d \r", fuck++);
    return;
  }
#endif

  V_DrawLine(main_scr, fl->a.x, fl->a.y, fl->b.x, fl->b.y, colour);
}

//
// Clip lines, draw visible parts of lines.
//
static void DrawMline(mline_t * ml, int colour)
{
  static fline_t fl;

  if (automapactive == 1)
  {
    colour = GREENS + (GREENRANGE / 2) + (colour % (GREENRANGE / 2));
  }

  if (ClipMline(ml, &fl))
  {
    DrawFline(&fl, colour);
  }
}

//
// Draws flat (floor/ceiling tile) aligned grid lines.
//
static void DrawGrid(int colour)
{
  float_t x, y;
  float_t start, end;
  mline_t ml;

  // Figure out start of vertical gridlines
  start = m_x + fmod(MAPBLOCKUNITS - (m_x - bmaporgx), MAPBLOCKUNITS);
  end = m_x + m_w;

  // draw vertical gridlines
  ml.a.y = m_y;
  ml.b.y = m_y + m_h;
  for (x = start; x < end; x += MAPBLOCKUNITS)
  {
    ml.a.x = x;
    ml.b.x = x;
    DrawMline(&ml, colour);
  }

  // Figure out start of horizontal gridlines
  start = m_y + fmod(MAPBLOCKUNITS - (m_y - bmaporgy), MAPBLOCKUNITS);
  end = m_y + m_h;

  // draw horizontal gridlines
  ml.a.x = m_x;
  ml.b.x = m_x + m_w;
  for (y = start; y < end; y += MAPBLOCKUNITS)
  {
    ml.a.y = y;
    ml.b.y = y;
    DrawMline(&ml, colour);
  }

}

//
// CheckSimiliarRegions
//
// Checks whether the two sectors' regions are similiar.  If they are
// different enough, a line will be drawn on the automap.
//
// -AJA- 1999/12/07: written.

static boolean_t CheckSimiliarRegions(sector_t *front, sector_t *back)
{
  vert_region_t *F, *B;

  F = front->regions;
  B = back->regions;

  while (F && B)
  {
    if (F->floor->h != B->floor->h)
      return false;
    
    if (F->ceil->h != B->ceil->h)
      return false;
    
    F = F->sec_next;
    B = B->sec_next;
  }

  return (F || B) ? false : true;
}

//
// Determines visible lines, draws them.
// This is LineDef based, not LineSeg based.
//
static void DrawWalls(void)
{
  int i;
  static mline_t l;

  for (i = 0; i < numlines; i++)
  {
    sector_t *front = lines[i].frontsector;
    sector_t *back  = lines[i].backsector;
    
    l.a.x = lines[i].v1->x;
    l.a.y = lines[i].v1->y;
    l.b.x = lines[i].v2->x;
    l.b.y = lines[i].v2->y;

    if (rotatemap)
    {
      // rotate map point around player location
      // negative angles rotate 2D points counter clockwise
      l.a.x -= plr->mo->x;
      l.b.x -= plr->mo->x;
      l.a.y -= plr->mo->y;
      l.b.y -= plr->mo->y;
      
      Rotate(&l.a.x, &l.a.y, (plr->mo->angle*-1) + ANG90);
      Rotate(&l.b.x, &l.b.y, (plr->mo->angle*-1) + ANG90);
      
      l.a.x += plr->mo->x;
      l.b.x += plr->mo->x;
      l.a.y += plr->mo->y;
      l.b.y += plr->mo->y;
    }

    if (cheating || (lines[i].flags & ML_MAPPED))
    {
      if ((lines[i].flags & ML_DONTDRAW) && !cheating)
        continue;

      if (!front || !back)
      {
        DrawMline(&l, WALLCOLOURS + lightlev);
      }
      else
      {
        if (lines[i].special && lines[i].special->singlesided)
        {  
          // teleporters
          DrawMline(&l, WALLCOLOURS + WALLRANGE / 2);
        }
        else if (lines[i].flags & ML_SECRET)
        {  
          // secret door
          if (cheating)
            DrawMline(&l, SECRETWALLCOLOURS + lightlev);
          else
            DrawMline(&l, WALLCOLOURS + lightlev);
        }
        else if (back->floor.h != front->floor.h)
        {
          // floor level change
          DrawMline(&l, FDWALLCOLOURS + lightlev);
        }
        else if (back->ceil.h != front->ceil.h)
        {
          // ceiling level change
          DrawMline(&l, CDWALLCOLOURS + lightlev);
        }
        else if ((front->regions->sec_next || back->regions->sec_next) &&
            ! CheckSimiliarRegions(front, back))
        {
          // -AJA- 1999/10/09: extra floor change.
          DrawMline(&l, CDWALLCOLOURS);
        }
        else if (cheating)
        {
          DrawMline(&l, TSWALLCOLOURS + lightlev);
        }
      }
    }
    else if (plr->powers[PW_AllMap])
    {
      if (!(lines[i].flags & ML_DONTDRAW))
        DrawMline(&l, GRAYS + 3);
    }
  }
}


// -AJA- for debugging
#if DEBUG_TRUEBSP
static void DEBUG_ShowSubSecs(void)
{
  int x, y;

  for (y=0;     y < f_h; y += 3)
  for (x=(y&1); x < f_w; x += 3)
  {
    float_t mx = CXFTOM(x);
    float_t my = CYFTOM(y);

    int subsec = R_PointInSubsector(mx, my) - subsectors;

    V_DrawPixel(main_scr, x, y, subsec * 17 + (subsec/256) * 11);
  }
}
static void DEBUG_DrawSegs(void)
{
  int i;
  mline_t l;
  seg_t *seg;

  for (i=0; i < numsegs; i++)
  {
    seg = &segs[i];

    if (! seg->miniseg)
      continue;

    // draw partnered segs only once
    if (seg->partner && i > (seg->partner - segs))
      continue;

    l.a.x = seg->x1;
    l.a.y = seg->y1;
    l.b.x = seg->x2;
    l.b.y = seg->y2;

    DrawMline(&l, TSWALLCOLOURS);
  }
}
#endif

static void DrawLineCharacter(mline_t *lineguy, int lineguylines, float_t scale, angle_t angle, int colour, float_t x, float_t y)
{
  int i;
  mline_t l;

  for (i = 0; i < lineguylines; i++)
  {
    l.a.x = lineguy[i].a.x;
    l.a.y = lineguy[i].a.y;

    if (scale != 0)
    {
      l.a.x *= scale;
      l.a.y *= scale;
    }

    if (angle)
      Rotate(&l.a.x, &l.a.y, angle);

    l.a.x += x;
    l.a.y += y;

    l.b.x = lineguy[i].b.x;
    l.b.y = lineguy[i].b.y;

    if (scale != 0)
    {
      l.b.x *= scale;
      l.b.y *= scale;
    }

    if (angle)
      Rotate(&l.b.x, &l.b.y, angle);

    l.b.x += x;
    l.b.y += y;

    DrawMline(&l, colour);
  }
}

static void DrawPlayers(void)
{
  int i;
  player_t *p;
  static int their_colours[] =
  {GREENS, GRAYS, BROWNS, REDS, GREENS, GRAYS, BROWNS, REDS};
  int their_colour = -1;
  int colour;

  if (!netgame)
  {
    if (rotatemap)
      i = ANG90;
    else
      i = plr->mo->angle;

    if (cheating)
      DrawLineCharacter(cheat_player_arrow, NUMCHEATPLYRLINES, 0, i,
          WHITE, plr->mo->x, plr->mo->y);
    else
      DrawLineCharacter(player_arrow, NUMPLYRLINES, 0, i,
          WHITE, plr->mo->x, plr->mo->y);
    return;
  }

  for (p = players; p; p = p->next)
  {
    their_colour++;

    if ((deathmatch && !singledemo) && p != plr)
      continue;

    if (!p->in_game)
      continue;

    if (p->powers[PW_PartInvis])
      colour = 246;  // *close* to black

    else
      colour = their_colours[their_colour];

    DrawLineCharacter(player_arrow, NUMPLYRLINES, 0, p->mo->angle,
        colour, p->mo->x, p->mo->y);
  }

}

static void DrawThings(int colours, int colourrange)
{
  int i;
  mobj_t *t;
  float_t x, y;

  for (i = 0; i < numsubsectors; i++)
  {
    t = subsectors[i].thinglist;

    while (t)
    {
      x = t->x;
      y = t->y;

      if (rotatemap)
      {
        // rotate coordinates so they are on the map correctly
        x -= plr->mo->x;
        y -= plr->mo->y;
        Rotate(&x, &y, ANG90 - plr->mo->angle);
        x += plr->mo->x;
        y += plr->mo->y;
      }
      DrawLineCharacter(thintriangle_guy, NUMTHINTRIANGLEGUYLINES,
          16, t->angle, colours + lightlev, x, y);
      t = t->snext;
    }
  }
}

static void DrawMarks(void)
{
  int i, fx, fy, w, h;

  for (i = 0; i < AM_NUMMARKPOINTS; i++)
  {
    if (markpoints[i].x != -1)
    {
      //      w = SHORT(marknums[i]->width);
      //      h = SHORT(marknums[i]->height);
      w = 5;  // because something's wrong with the wad, i guess

      h = 6;  // because something's wrong with the wad, i guess

      fx = CXMTOF(markpoints[i].x);
      fy = CYMTOF(markpoints[i].y);
      if (fx >= f_x && fx <= f_w - w && fy >= f_y && fy <= f_h - h)
        V_DrawPatch(main_scr, fx, fy, marknums[i]);
    }
  }

}

static void DrawCrosshair(int colour)
{
  // -AJA- 1999/07/04: now uses V_DrawPixel().

  V_DrawPixel(main_scr, f_w / 2, f_h / 2, colour);  // single point for now
}

void AM_Drawer(void)
{
  if (!automapactive)
    return;

  if (automapactive == 2)
    ClearFB(BACKGROUND);

  if (grid)
    DrawGrid(GRIDCOLOURS);

#if DEBUG_TRUEBSP
  DEBUG_DrawSegs();
#endif

  DrawWalls();
  DrawPlayers();

  if (cheating == 2)
    DrawThings(THINGCOLOURS, THINGRANGE);

  DrawCrosshair(XHAIRCOLOURS);
  DrawMarks();

  V_MarkRect(f_x, f_y, f_w, f_h);
}
