//----------------------------------------------------------------------------
//  DOOM Definition File Codes (Animated textures)
//----------------------------------------------------------------------------
// 
//  Copyright (c) 1999-2000  The EDGE Team.
// 
//  This program is free software; you can redistribute it and/or
//  modify it under the terms of the GNU General Public License
//  as published by the Free Software Foundation; either version 2
//  of the License, or (at your option) any later version.
//
//  This program is distributed in the hope that it will be useful,
//  but WITHOUT ANY WARRANTY; without even the implied warranty of
//  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//  GNU General Public License for more details.
//
//----------------------------------------------------------------------------
//
// Animated Texture/Flat Setup and Parser Code
//

#include "i_defs.h"

#include "ddf_locl.h"
#include "ddf_main.h"
#include "dm_state.h"
#include "p_spec.h"
#include "r_local.h"
#include "z_zone.h"

static animdef_t buffer_anim;
static animdef_t *dynamic_anim;

static const animdef_t template_anim =
{
  DDF_BASE_NIL,  // ddf

  true,  // istexture
  "",    // endname
  "",    // startname
  8,     // speed
};

static void DDF_AnimGetType(const char *info, void *storage, void *priv);

// -ACB- 1998/08/10 Use DDF_MainGetLumpName for getting the..lump name.
// -KM- 1998/09/27 Use DDF_MainGetTime for getting tics
static const commandlist_t anim_commands[] =
{
  {"TYPE",   DDF_AnimGetType, &buffer_anim.istexture, NULL},
  {"SPEED",  DDF_MainGetTime, &buffer_anim.speed,     NULL},
  {"FIRST",  DDF_MainGetInlineStr10, &buffer_anim.startname, NULL},
  {"LAST",   DDF_MainGetInlineStr10, &buffer_anim.endname,   NULL},
  {NULL, NULL, NULL, NULL}
};

// Floor/ceiling animation sequences, defined by first and last frame,
// i.e. the flat (64x64 tile) name or texture name to be used.
//
// The full animation sequence is given using all the flats between
// the start and end entry, in the order found in the WAD file.

animdef_t ** animdefs = NULL;
int numanimdefs = 0;

static stack_array_t animdefs_a;


//
//  DDF PARSE ROUTINES
//

static boolean_t AnimStartEntry(const char *name)
{
  int i;
  boolean_t replaces = false;

  if (name && name[0])
  {
    for (i=0; i < numanimdefs; i++)
    {
      if (DDF_CompareName(animdefs[i]->ddf.name, name) == 0)
      {
        dynamic_anim = animdefs[i];
        replaces = true;
        break;
      }
    }
  }

  // not found, create a new one
  if (! replaces)
  {
    Z_SetArraySize(&animdefs_a, ++numanimdefs);

    dynamic_anim = animdefs[numanimdefs - 1];
    dynamic_anim->ddf.name = (name && name[0]) ? Z_StrDup(name) :
        DDF_MainCreateUniqueName("UNNAMED_ANIM", numanimdefs);
  }

  dynamic_anim->ddf.number = 0;

  return replaces;
}

static void AnimCopyOld(void)
{
  buffer_anim = dynamic_anim[0];
}

static void AnimInstantiate(const char *template_name)
{
  int i;

  if (! template_name)
  {
    // use built-in template
    buffer_anim = template_anim;
    return;
  }

  for (i=0; i < numanimdefs; i++)
  {
    // ignore current entry
    if (animdefs[i] == dynamic_anim)
      continue;
    
    if (DDF_CompareName(animdefs[i]->ddf.name, template_name) == 0)
    {
      buffer_anim = animdefs[i][0];
      return;
    }
  }

  I_Error("Template `%s' not found for anims.ddf\n", template_name);
}

static void AnimParseField(const char *field, const char *contents,
    int index GCCATTR(unused), boolean_t is_last GCCATTR(unused))
{
  I_WriteDebug("ANIM_PARSE: [%s] [%s]\n", field, contents);

  if (! DDF_MainParseField(anim_commands, field, contents))
    DDF_Error("Unknown anims.ddf command: %s\n", field);
}

static void AnimFinishEntry(void)
{
  ddf_base_t base;
  
  if (buffer_anim.speed <= 0)
    DDF_Error("Bad TICS value for anim: %d\n", buffer_anim.speed);
  
  if (!buffer_anim.startname || !buffer_anim.startname[0])
    DDF_Error("Missing first name for anim.\n");

  if (!buffer_anim.startname || !buffer_anim.startname[0])
    DDF_Error("Missing last name for anim.\n");
  
  // transfer static entry to dynamic entry
  
  base = dynamic_anim->ddf;
  dynamic_anim[0] = buffer_anim;
  dynamic_anim->ddf = base;

  // Compute CRC.  In this case, there is no need, since animations
  // have zero impact on the game simulation.
  dynamic_anim->ddf.crc = 0;
}


void DDF_ReadAnims(void *data, int size)
{
  readinfo_t anims;

  anims.memfile = data;
  anims.memsize = size;
  anims.tag = "ANIMATIONS";
  anims.entries_per_dot = 2;

  if (anims.memfile)
  {
    anims.message = NULL;
    anims.filename = NULL;
    anims.lumpname = "DDFANIM";
  }
  else
  {
    anims.message = "DDF_InitAnimations";
    anims.filename = "anims.ddf";
    anims.lumpname = NULL;
  }

  anims.start_entry  = AnimStartEntry;
  anims.copy_old     = AnimCopyOld;
  anims.instantiate  = AnimInstantiate;
  anims.parse_field  = AnimParseField;
  anims.finish_entry = AnimFinishEntry;

  DDF_MainReadFile(&anims);
}

void DDF_AnimInit(void)
{
  Z_InitStackArray(&animdefs_a, (void ***)&animdefs, sizeof(animdef_t), 0);
}

void DDF_AnimCleanUp(void)
{
}

//
// DDF_AnimGetType
//
static void DDF_AnimGetType(const char *info, void *storage, void *priv GCCATTR(unused))
{
  boolean_t *is_tex = (boolean_t *) storage;
  
  DEV_ASSERT2(storage);

  if (DDF_CompareName(info, "FLAT") == 0)
    (*is_tex) = false;
  else if (DDF_CompareName(info, "TEXTURE") == 0)
    (*is_tex) = true;
  else
    DDF_Error("Unknown animation type: %s\n", info);
}

