//----------------------------------------------------------------------------
//  Data Definition File Code (Level Defines)
//----------------------------------------------------------------------------
// 
//  Copyright (c) 1999-2000  The EDGE Team.
// 
//  This program is free software; you can redistribute it and/or
//  modify it under the terms of the GNU General Public License
//  as published by the Free Software Foundation; either version 2
//  of the License, or (at your option) any later version.
//
//  This program is distributed in the hope that it will be useful,
//  but WITHOUT ANY WARRANTY; without even the implied warranty of
//  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//  GNU General Public License for more details.
//
//----------------------------------------------------------------------------
//
// Level Setup and Parser Code
//

#include "i_defs.h"

#include "ddf_locl.h"
#include "ddf_main.h"
#include "dm_state.h"
#include "m_fixed.h"
#include "m_math.h"
#include "p_mobj.h"
#include "w_wad.h"
#include "z_zone.h"

static mapstuff_t buffer_map;
static mapstuff_t *dynamic_map;

static const mapstuff_t template_map =
{
  DDF_BASE_NIL,  // ddf

  NULL,  // next
  NULL,  // description
  "",    // namegraphic
  "",    // lump
  "",    // sky
  "",    // surround
  0,     // music

  0,     // partime
  NULL,  // interscr
  0,     // force_on
  0,     // force_off
  "",    // nextmapname
  "",    // secretmapname
  0,     // autotag

  {
    // finale[0]
    {
      NULL,  // text
      "",    // text_back
      "",    // text_flat
      3,     // text_speed
      150,   // text_wait
      0,     // numpics
      0,     // picwait
      NULL,  // pics
      false, // docast
      false, // dobunny
      0      // music
    },

    // finale[1]
    {
      NULL,  // text
      "",    // text_back
      "",    // text_flat
      3,     // text_speed
      150,   // text_wait
      0,     // numpics
      0,     // picwait
      NULL,  // pics
      false, // docast
      false, // dobunny
      0      // music
    }
  }
};

mapstuff_t ** level_maps = NULL;
int num_level_maps = 0;

static stack_array_t level_maps_a;


static void DDF_LevelGetSpecials(const char *info, void *storage, void *priv);
static void DDF_LevelGetPic(const char *info, void *storage, void *priv);
static void DDF_LevelGetEpisode(const char *info, void *storage, void *priv);

// -KM- 1998/11/25 Finales are all go.
static const commandlist_t level_commands[] =
{
    {"LUMPNAME", DDF_MainGetInlineStr10, &buffer_map.lump, NULL},
    {"DESCRIPTION", DDF_MainGetString, &buffer_map.description, NULL},
    {"NAME GRAPHIC", DDF_MainGetInlineStr10, &buffer_map.namegraphic, NULL},
    {"SKY TEXTURE", DDF_MainGetInlineStr10, &buffer_map.sky, NULL},
    {"MUSIC ENTRY", DDF_MainGetNumeric, &buffer_map.music, NULL},
    {"SURROUND FLAT", DDF_MainGetInlineStr10, &buffer_map.surround, NULL},
    {"SPECIAL", DDF_LevelGetSpecials, NULL, NULL},
    {"NEXT MAP", DDF_MainGetInlineStr10, &buffer_map.nextmapname, NULL},
    {"SECRET MAP", DDF_MainGetInlineStr10, &buffer_map.secretmapname, NULL},
    {"AUTOTAG", DDF_MainGetNumeric, &buffer_map.autotag, NULL},
    {"PARTIME", DDF_MainGetTime, &buffer_map.partime, NULL},
    {"EPISODE", DDF_LevelGetEpisode, &buffer_map.interscr, NULL},

    // -AJA- FIXME: needs a generic system for subfields...
    {"END.TEXT", DDF_MainGetString, &buffer_map.f[0].text, NULL},
    {"END.TEXT GRAPHIC", DDF_MainGetInlineStr10, &buffer_map.f[0].text_back, NULL},
    {"END.TEXT FLAT", DDF_MainGetInlineStr10, &buffer_map.f[0].text_flat, NULL},
    {"END.TEXT SPEED", DDF_MainGetFloat, &buffer_map.f[0].text_speed, NULL},
    {"END.TEXT WAIT", DDF_MainGetNumeric, &buffer_map.f[0].text_wait, NULL},
    {"END.GRAPHIC", DDF_LevelGetPic, &buffer_map.f[0], NULL},
    {"END.GRAPHIC WAIT", DDF_MainGetTime, &buffer_map.f[0].picwait, NULL},
    {"END.CAST", DDF_MainGetBoolean, &buffer_map.f[0].docast, NULL},
    {"END.BUNNY", DDF_MainGetBoolean, &buffer_map.f[0].dobunny, NULL},
    {"END.MUSIC", DDF_MainGetNumeric, &buffer_map.f[0].music, NULL},

    {"PRE.TEXT", DDF_MainGetString, &buffer_map.f[1].text, NULL},
    {"PRE.TEXT GRAPHIC", DDF_MainGetInlineStr10, &buffer_map.f[1].text_back, NULL},
    {"PRE.TEXT FLAT", DDF_MainGetInlineStr10, &buffer_map.f[1].text_flat, NULL},
    {"PRE.TEXT SPEED", DDF_MainGetFloat, &buffer_map.f[1].text_speed, NULL},
    {"PRE.TEXT WAIT", DDF_MainGetNumeric, &buffer_map.f[1].text_wait, NULL},
    {"PRE.GRAPHIC", DDF_LevelGetPic, &buffer_map.f[1], NULL},
    {"PRE.GRAPHIC WAIT", DDF_MainGetTime, &buffer_map.f[1].picwait, NULL},
    {"PRE.CAST", DDF_MainGetBoolean, &buffer_map.f[1].docast, NULL},
    {"PRE.BUNNY", DDF_MainGetBoolean, &buffer_map.f[1].dobunny, NULL},
    {"PRE.MUSIC", DDF_MainGetNumeric, &buffer_map.f[1].music, NULL},

    {NULL, NULL, NULL, NULL}
};

static specflags_t map_specials[] =
{
    {"JUMPING", MPF_Jumping, 0},
    {"MLOOK", MPF_Mlook, 0},
    {"FREELOOK", MPF_Mlook, 0},  // -AJA- backwards compat.
    {"TRANSLUCENCY", MPF_Translucency, 0},
    {"CHEATS", MPF_Cheats, 0},
    {"ITEM RESPAWN", MPF_ItemRespawn, 0},
    {"FAST MONSTERS", MPF_FastParm, 0},
    {"RESURRECT RESPAWN", MPF_ResRespawn, 0},
    {"TELEPORT RESPAWN", MPF_ResRespawn, 1},
    {"STRETCH SKY", MPF_StretchSky, 0},
    {"NORMAL SKY", MPF_StretchSky, 1},
    {"TRUE3D", MPF_True3D, 0},
    {"ENEMY STOMP", MPF_Stomp, 0},
    {"MORE BLOOD", MPF_MoreBlood, 0},
    {"NORMAL BLOOD", MPF_MoreBlood, 1},
    {"RESPAWN", MPF_Respawn, 0},
    {"AUTOAIM", MPF_AutoAim, 0},
    {"AA MLOOK", MPF_AutoAimMlook, 0},
    {"EXTRAS", MPF_Extras, 0},
    {"RESET PLAYER", MPF_ResetPlayer, 0},
    {"LIMIT ZOOM", MPF_LimitZoom, 0},
    {NULL, 0, 0}
};


//
//  DDF PARSE ROUTINES
//

static boolean_t LevelStartEntry(const char *name)
{
  int i;
  boolean_t replaces = false;

  if (name && name[0])
  {
    for (i=0; i < num_level_maps; i++)
    {
      if (DDF_CompareName(level_maps[i]->ddf.name, name) == 0)
      {
        dynamic_map = level_maps[i];
        replaces = true;
        break;
      }
    }
  }

  // not found, create a new one
  if (! replaces)
  {
    Z_SetArraySize(&level_maps_a, ++num_level_maps);
    
    dynamic_map = level_maps[num_level_maps - 1];
    dynamic_map->ddf.name = (name && name[0]) ? Z_StrDup(name) :
        DDF_MainCreateUniqueName("UNNAMED_LEVEL_MAP", num_level_maps);
  }

  dynamic_map->ddf.number = 0;

  return replaces;
}

static void LevelCopyOld(void)
{
  buffer_map = dynamic_map[0];
}

static void LevelInstantiate(const char *template_name)
{
  int i;

  if (! template_name)
  {
    // use built-in template
    buffer_map = template_map;
    return;
  }

  for (i=0; i < num_level_maps; i++)
  {
    // ignore current entry
    if (level_maps[i] == dynamic_map)
      continue;
    
    if (DDF_CompareName(level_maps[i]->ddf.name, template_name) == 0)
    {
      buffer_map = level_maps[i][0];
      return;
    }
  }

  I_Error("Template `%s' not found for levels.ddf\n", template_name);
}

static void LevelParseField(const char *field, const char *contents,
    int index, boolean_t is_last)
{
  I_WriteDebug("LEVEL_PARSE: [%s] [%s]\n", field, contents);

  if (! DDF_MainParseField(level_commands, field, contents))
    DDF_Error("Unknown levels.ddf command: %s\n", field);
}

static void LevelFinishEntry(void)
{
  ddf_base_t base;
  
  // FIXME: check stuff here...
  
  // transfer static entry to dynamic entry
  
  base = dynamic_map->ddf;
  dynamic_map[0] = buffer_map;
  dynamic_map->ddf = base;

  // compute CRC...
  CRC32_Init(&dynamic_map->ddf.crc);

  // FIXME: add stuff...
  
  CRC32_Done(&dynamic_map->ddf.crc);

  // Initial Hack
  currentmap = dynamic_map;
}


void DDF_ReadLevels(void *data, int size)
{
  readinfo_t levels;

  levels.memfile = data;
  levels.memsize = size;
  levels.tag = "LEVELS";
  levels.entries_per_dot = 2;

  if (levels.memfile)
  {
    levels.message = NULL;
    levels.filename = NULL;
    levels.lumpname = "DDFLEVL";
  }
  else
  {
    levels.message = "DDF_InitLevels";
    levels.filename = "levels.ddf";
    levels.lumpname = NULL;
  }

  levels.start_entry  = LevelStartEntry;
  levels.copy_old     = LevelCopyOld;
  levels.instantiate  = LevelInstantiate;
  levels.parse_field  = LevelParseField;
  levels.finish_entry = LevelFinishEntry;

  DDF_MainReadFile(&levels);
}

void DDF_LevelInit(void)
{
  Z_InitStackArray(&level_maps_a, (void ***)&level_maps, sizeof(mapstuff_t), 0);
}

void DDF_LevelCleanUp(void)
{
}


//
// DDF_LevelMapLookup
//
// Changes the current map: this is globally visible to all those who ref
// ddf_main.h. It will check there is an map with that entry name exists
// and the map lump name also exists. Returns map if it exists.

const mapstuff_t *DDF_LevelMapLookup(const char *refname)
{
  int i;

  if (!refname)
    return NULL;

  for (i=num_level_maps-1; i >= 0; i--)
  {
    if (DDF_CompareName(level_maps[i]->ddf.name, refname) != 0)
      continue;

    if (W_CheckNumForName(level_maps[i]->lump) == -1)
      continue;
    
    return level_maps[i];
  }

  return NULL;
}

//
// DDF_LevelGetPic
//
// Adds finale pictures to the level's list.

void DDF_LevelGetPic(const char *info, void *storage, void *priv)
{
  finale_t *f = (finale_t *)storage;

  Z_Resize(f->pics, char, 9 * (f->numpics + 1));
  Z_StrNCpy(f->pics + f->numpics * 9, info, 8);
  f->numpics++;
}

static void DDF_LevelGetEpisode(const char *info, void *storage, void *priv)
{
  const wi_map_t **dest = (const wi_map_t **)storage;

  *dest = DDF_GameLookup(info);
}

void DDF_LevelGetSpecials(const char *info, void *storage, void *priv)
{
  // -AJA- 2000/02/02: reworked this for new system.

  int flag_value;

  switch (DDF_MainCheckSpecialFlag(info, map_specials, &flag_value,
      true, true))
  {
    case CHKF_Positive:
      buffer_map.force_on  |=  flag_value;
      buffer_map.force_off &= ~flag_value;
      break;
    
    case CHKF_Negative:
      buffer_map.force_on  &= ~flag_value;
      buffer_map.force_off |=  flag_value;
      break;

    case CHKF_User:
      buffer_map.force_on  &= ~flag_value;
      buffer_map.force_off &= ~flag_value;
      break;
    
    case CHKF_Unknown:
      DDF_Error("DDF_LevelGetSpecials: Unknown level special: %s", info);
      break;
  }
}

