//----------------------------------------------------------------------------
//  Data Definition File Codes (Linedefs)
//----------------------------------------------------------------------------
// 
//  Copyright (c) 1999-2000  The EDGE Team.
// 
//  This program is free software; you can redistribute it and/or
//  modify it under the terms of the GNU General Public License
//  as published by the Free Software Foundation; either version 2
//  of the License, or (at your option) any later version.
//
//  This program is distributed in the hope that it will be useful,
//  but WITHOUT ANY WARRANTY; without even the implied warranty of
//  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//  GNU General Public License for more details.
//
//----------------------------------------------------------------------------
//
// Line Definitions Setup and Parser Code
//
// -KM- 1998/09/01 Written.
// -ACB- 1998/09/06 Beautification: cleaned up so I can read it :).
// -KM- 1998/10/29 New types of linedefs added: colourmap, sound, friction, gravity
//                  auto, singlesided, music, lumpcheck
//                  Removed sector movement to ddf_main.c, so can be accesed by
//                  ddf_sect.c
//

#include "i_defs.h"

#include "ddf_locl.h"
#include "ddf_main.h"
#include "dm_state.h"
#include "m_fixed.h"
#include "p_mobj.h"
#include "p_local.h"
#include "z_zone.h"

#include <ctype.h>
#include <stdio.h>
#include <stdlib.h>

#define DDF_LineHashFunc(x)  (((x) + 211) % 211)

// -KM- 1999/01/29 Improved scrolling.
// Scrolling
typedef enum
{
  dir_none = 0,
  dir_vert = 1,
  dir_up = 2,
  dir_horiz = 4,
  dir_left = 8
}
scrolldirs_e;


static linedeftype_t buffer_line;
static linedeftype_t *dynamic_line;

// these bits logically belong with buffer_line:
static float_t s_speed;
static scrolldirs_e s_dir;

const linedeftype_t template_line =
{
  DDF_BASE_NIL,  // ddf

  0,        // newtrignum
  0,        // how triggered (walk/switch/shoot)
  0,        // things that can trigger it
  KF_NONE,  // keys needed
  -1,       // can be activated repeatedly
  -1,       // special type: don't change.
  false,    // crushing

  // Floor
  {
    mov_undefined,  // type
    false,          // floorOrCeiling
    false,          // crush
    -1, -1,         // speed up/down
    ref_absolute,   // dest ref
    0,              // dest
    "",             // texture
    0, 0,           // wait, prewait
    sfx_None, sfx_None, sfx_None, sfx_None,  // SFX start/up/down/stop
    0, 0.0          // scroll_angle, scroll_speed
  },

  // Ceiling
  {
    mov_undefined,  // type
    true,           // floorOrCeiling
    false,          // crush
    -1, -1,         // speed up/down
    ref_absolute,   // dest ref
    0,              // dest
    "",             // texture
    0, 0,           // wait, prewait
    sfx_None, sfx_None, sfx_None, sfx_None,  // SFX start/up/down/stop
    0, 0.0          // scroll_angle, scroll_speed
  },

  // Donut
  {
    false,  //  dodonut
    sfx_None, sfx_None,  // SFX down/stop inner circle
    sfx_None, sfx_None   // SFX up/stop outer loop
  },

  // Teleport
  {
    false,    // is a teleporter
    NULL,     // effect in object
    NULL,     // in object's name
    NULL,     // effect out object
    NULL,     // out object's name
    0,        // delay
    0,        // special
  },

  // Lights
  {
    lite_none, //  lights action
    64,        //  Set light to this level
    0,0,0      //  dark/brighttime, sync
  },

  EXIT_None,   // exit type
  0, 0,        // scrolling X/Y
  SCPT_RIGHT,  // scroll parts
  NULL,        // security message
  NULL,        // colourmap
  M_PI,        // gravity     \ 
  M_PI,        // friction     M_PI means "not changed"
  M_PI,        // viscosity   /
  sfx_None,    // sfx
  0,           // music
  false,       // automatic line
  false,       // single sided

  // Extra floor:
  {
    0,           // type
    VISIBLE,     // translucency
    EFCTL_None   // control
  },

  VISIBLE,  // translucency
  DEFAULT_APPEAR,  // appear
  0,    // special_flags
  0     // trigger_effect
};

const moving_plane_t donut_floor =
{
  mov_Once,        //  Type
  false,           //  floorOrCeiling (floor)
  false,           //  crush
  FLOORSPEED / 2,  //  speed up
  FLOORSPEED / 2,  //  speed down
  ref_absolute,    //  dest ref
  (float_t)INT_MAX,//  dest
  "",              //  texture
  0, 0,            //  wait, prewait
  0, 0, 0, 0,      //  SFX start/up/down/stop
  0, 0             //  Scroll angle/speed
};

linedeftype_t ** ddf_linetypes;
int num_ddf_linetypes;

static stack_array_t ddf_linetypes_a;

static linedeftype_t *line_lookup_cache[211];


static void DDF_LineGetTrigType(const char *info, void *storage, void *priv);
static void DDF_LineGetActivators(const char *info, void *storage, void *priv);
static void DDF_LineGetSecurity(const char *info, void *storage, void *priv);
static void DDF_LineGetScroller(const char *info, void *storage, void *priv);
static void DDF_LineGetScrollPart(const char *info, void *storage, void *priv);
static void DDF_LineGetExtraFloor(const char *info, void *storage, void *priv);
static void DDF_LineGetEFControl(const char *info, void *storage, void *priv);
static void DDF_LineGetTeleportSpecial(const char *info, void *storage, void *priv);
static void DDF_LineGetRadTrig(const char *info, void *storage, void *priv);
static void DDF_LineGetSpecialFlags(const char *info, void *storage, void *priv);


static const commandlist_t linedef_commands[] =
{
    {"NEWTRIGGER", DDF_MainGetNumeric, &buffer_line.newtrignum, NULL},
    {"TYPE", DDF_LineGetTrigType, &buffer_line.type, NULL},
    {"ACTIVATORS", DDF_LineGetActivators, &buffer_line.obj, NULL},
    {"KEYS", DDF_LineGetSecurity, &buffer_line.keys, NULL},
    {"FAILED MESSAGE", DDF_MainGetString, &buffer_line.failedmessage, NULL},
    {"COUNT", DDF_MainGetNumeric, &buffer_line.count, NULL},
    {"SECSPECIAL", DDF_MainGetNumeric, &buffer_line.specialtype, NULL},
    {"CRUSH", DDF_MainGetBoolean, &buffer_line.crush, NULL},

    // -AJA- FIXME: use a generic system for subfields...
    {"FLOOR.TYPE", DDF_MainGetMType, &buffer_line.f.type, NULL},
    {"FLOOR.SPEED UP", DDF_MainGetFloat, &buffer_line.f.speed_up, NULL},
    {"FLOOR.SPEED DOWN", DDF_MainGetFloat, &buffer_line.f.speed_down, NULL},
    {"FLOOR.DEST REF", DDF_MainGetDestRef, &buffer_line.f.destref, NULL},
    {"FLOOR.DEST OFFSET", DDF_MainGetFloat, &buffer_line.f.dest, NULL},
    {"FLOOR.TEXTURE", DDF_MainGetInlineStr10, &buffer_line.f.tex, NULL},
    {"FLOOR.PAUSE TIME", DDF_MainGetTime, &buffer_line.f.wait, NULL},
    {"FLOOR.WAIT TIME", DDF_MainGetTime, &buffer_line.f.prewait, NULL},
    {"FLOOR.SFX START", DDF_MainLookupSound, &buffer_line.f.sfxstart, NULL},
    {"FLOOR.SFX UP", DDF_MainLookupSound, &buffer_line.f.sfxup, NULL},
    {"FLOOR.SFX DOWN", DDF_MainLookupSound, &buffer_line.f.sfxdown, NULL},
    {"FLOOR.SFX STOP", DDF_MainLookupSound, &buffer_line.f.sfxstop, NULL},

    {"CEILING.TYPE", DDF_MainGetMType, &buffer_line.c.type, NULL},
    {"CEILING.SPEED UP", DDF_MainGetFloat, &buffer_line.c.speed_up, NULL},
    {"CEILING.SPEED DOWN", DDF_MainGetFloat, &buffer_line.c.speed_down, NULL},
    {"CEILING.DEST REF", DDF_MainGetDestRef, &buffer_line.c.destref, NULL},
    {"CEILING.DEST OFFSET", DDF_MainGetFloat, &buffer_line.c.dest, NULL},
    {"CEILING.TEXTURE", DDF_MainGetInlineStr10, &buffer_line.c.tex, NULL},
    {"CEILING.PAUSE TIME", DDF_MainGetTime, &buffer_line.c.wait, NULL},
    {"CEILING.WAIT TIME", DDF_MainGetTime, &buffer_line.c.prewait, NULL},
    {"CEILING.SFX START", DDF_MainLookupSound, &buffer_line.c.sfxstart, NULL},
    {"CEILING.SFX UP", DDF_MainLookupSound, &buffer_line.c.sfxup, NULL},
    {"CEILING.SFX DOWN", DDF_MainLookupSound, &buffer_line.c.sfxdown, NULL},
    {"CEILING.SFX STOP", DDF_MainLookupSound, &buffer_line.c.sfxstop, NULL},

    {"DONUT", DDF_MainGetBoolean, &buffer_line.d.dodonut, NULL},
    {"DONUT IN SFX", DDF_MainLookupSound, &buffer_line.d.d_sfxin, NULL},
    {"DONUT IN SFXSTOP", DDF_MainLookupSound, &buffer_line.d.d_sfxinstop, NULL},
    {"DONUT OUT SFX", DDF_MainLookupSound, &buffer_line.d.d_sfxout, NULL},
    {"DONUT OUT SFXSTOP", DDF_MainLookupSound, &buffer_line.d.d_sfxoutstop, NULL},
    {"TELEPORT", DDF_MainGetBoolean, &buffer_line.t.teleport, NULL},
    {"TELEPORT DELAY", DDF_MainGetTime, &buffer_line.t.delay, NULL},
    {"TELEIN EFFECTOBJ", DDF_MainGetString, &buffer_line.t.inspawnobj_ref, NULL},
    {"TELEOUT EFFECTOBJ", DDF_MainGetString, &buffer_line.t.outspawnobj_ref, NULL},
    {"TELEPORT SPECIAL", DDF_LineGetTeleportSpecial, &buffer_line.t.special, NULL},
    {"LIGHT TYPE", DDF_MainGetLighttype, &buffer_line.l.type, NULL},
    {"LIGHT LEVEL", DDF_MainGetNumeric, &buffer_line.l.light, NULL},
    {"LIGHT DARK TIME", DDF_MainGetTime, &buffer_line.l.darktime, NULL},
    {"LIGHT BRIGHT TIME", DDF_MainGetTime, &buffer_line.l.brighttime, NULL},
    {"LIGHT PROBABILITY", DDF_MainGetNumeric, &buffer_line.l.light, NULL},
    {"LIGHT SYNC", DDF_MainGetTime, &buffer_line.l.sync, NULL},
    {"EXIT", DDF_MainGetExit, &buffer_line.e_exit, NULL},
    {"SCROLL", DDF_LineGetScroller, &s_dir, NULL},
    {"SCROLLING SPEED", DDF_MainGetFloat, &s_speed, NULL},
    {"SCROLL XSPEED", DDF_MainGetFloat, &buffer_line.s_xspeed, NULL},
    {"SCROLL YSPEED", DDF_MainGetFloat, &buffer_line.s_yspeed, NULL},
    {"SCROLL PARTS", DDF_LineGetScrollPart, &buffer_line.scroll_parts, NULL},
    {"USE COLOURMAP", DDF_MainGetColourmap, &buffer_line.use_colourmap, NULL},
    {"GRAVITY", DDF_MainGetFloat, &buffer_line.gravity, NULL},
    {"FRICTION", DDF_MainGetFloat, &buffer_line.friction, NULL},
    {"VISCOSITY", DDF_MainGetFloat, &buffer_line.viscosity, NULL},
    {"SOUND", DDF_MainLookupSound, &buffer_line.sfx, NULL},
    {"MUSIC", DDF_MainGetNumeric, &buffer_line.music, NULL},
    {"AUTO", DDF_MainGetBoolean, &buffer_line.autoline, NULL},
    {"SINGLESIDED", DDF_MainGetBoolean, &buffer_line.singlesided, NULL},
    {"EXTRAFLOOR TYPE", DDF_LineGetExtraFloor, NULL, NULL},
    {"EXTRAFLOOR TRANSLUCENCY", DDF_MainGetFloat, &buffer_line.ef.translucency, NULL},
    {"EXTRAFLOOR CONTROL", DDF_LineGetEFControl, NULL, NULL},
    {"TRANSLUCENCY", DDF_MainGetFloat, &buffer_line.translucency, NULL},
    {"WHEN APPEAR", DDF_MainGetWhenAppear, &buffer_line.appear, NULL},
    {"SPECIAL", DDF_LineGetSpecialFlags, &buffer_line.special_flags, NULL},
    {"RADIUS_TRIGGER", DDF_LineGetRadTrig, NULL, NULL},
    {NULL, NULL, NULL, NULL}
};


// -AJA- FIXME

static struct
{
  char *s;
  int n;
}

s_scroll[] =
{
  { "NONE", dir_none | dir_none } ,
  { "UP", (dir_vert | dir_up) | ((~dir_none) << 16) } ,
  { "DOWN", (dir_vert) | ((~dir_up) << 16) } ,
  { "LEFT", (dir_horiz | dir_left) | ((~dir_none) << 16) } ,
  { "RIGHT", (dir_horiz) | ((~dir_left) << 16) }
}
,

s_keys[] =
{
  { "NONE", KF_NONE } ,
  { "BLUE CARD", KF_BLUECARD } ,
  { "YELLOW CARD", KF_YELLOWCARD } ,
  { "RED CARD", KF_REDCARD } ,
  { "BLUE SKULL", KF_BLUESKULL } ,
  { "YELLOW SKULL", KF_YELLOWSKULL } ,
  { "RED SKULL", KF_REDSKULL } ,
  { "REQUIRES ALL", KF_REQUIRESALL }
}
,

s_trigger[] =
{
  { "WALK", line_walkable } ,
  { "PUSH", line_pushable } ,
  { "SHOOT", line_shootable }
}
,

s_activators[] =
{
  { "PLAYER", trig_player } ,
  { "MONSTER", trig_monster } ,
  { "MISSILE", trig_projectile }
};



//
//  DDF PARSE ROUTINES
//

static boolean_t LinedefStartEntry(const char *name)
{
  int i;
  boolean_t replaces = false;
  int number = MAX(0, atoi(name));

  if (number == 0)
    DDF_Error("Bad linedef number in lines.ddf: %s\n", name);

  for (i=0; i < num_ddf_linetypes; i++)
  {
    if (ddf_linetypes[i]->ddf.number == number)
    {
      dynamic_line = ddf_linetypes[i];
      replaces = true;
      break;
    }
  }

  // not found, create a new one
  if (! replaces)
  {
    Z_SetArraySize(&ddf_linetypes_a, ++num_ddf_linetypes);
    
    dynamic_line = ddf_linetypes[num_ddf_linetypes - 1];
  }

  dynamic_line->ddf.name   = NULL;
  dynamic_line->ddf.number = number;

  s_speed = 1.0;
  s_dir = dir_none;

  return replaces;
}

static void LinedefCopyOld(void)
{
  buffer_line = dynamic_line[0];
}

static void LinedefInstantiate(const char *template_name)
{
  int i;
  int number;

  if (! template_name)
  {
    // use built-in template
    buffer_line = template_line;
    return;
  }

  number = MAX(0, atoi(template_name));

  if (number == 0)
    DDF_Error("Bad template reference in lines.ddf: %s\n", template_name);

  for (i=0; i < num_ddf_linetypes; i++)
  {
    // ignore current entry
    if (ddf_linetypes[i] == dynamic_line)
      continue;
    
    if (ddf_linetypes[i]->ddf.number == number)
    {
      buffer_line = ddf_linetypes[i][0];
      return;
    }
  }

  I_Error("Template not found for lines.ddf: %s\n", template_name);
}

static void LinedefParseField(const char *field, const char *contents,
    int index, boolean_t is_last)
{
  I_WriteDebug("LINEDEF_PARSE: [%s] [%s]\n", field, contents);

  // handle `SCROLL_PARTS' specially, reset them to 0 at start
  if (index == 0 && DDF_CompareName("SCROLL_PARTS", field) == 0)
    buffer_line.scroll_parts = 0;
  
  if (! DDF_MainParseField(linedef_commands, field, contents))
    DDF_Error("Unknown lines.ddf command: %s\n", field);
}

static void LinedefFinishEntry(void)
{
  ddf_base_t base;
  
  buffer_line.c.crush = buffer_line.f.crush = buffer_line.crush;

  // -KM- 1999/01/29 Convert old style scroller to new.
  if (s_dir & dir_vert)
  {
    if (s_dir & dir_up)
      buffer_line.s_yspeed = s_speed;
    else
      buffer_line.s_yspeed = -s_speed;
  }
  if (s_dir & dir_horiz)
  {
    if (s_dir & dir_left)
      buffer_line.s_xspeed = s_speed;
    else
      buffer_line.s_xspeed = -s_speed;
  }

  // check stuff...

  if (buffer_line.ef.type)
  {
    if (! (buffer_line.ef.type & EXFL_Present))
      DDF_Error("Extrafloor type missing THIN, THICK or LIQUID.\n");

    if (buffer_line.ef.translucency != 1.0 && 
        ! (buffer_line.ef.type & EXFL_SeeThrough))
      DDF_Error("Extrafloor must have SEE_THROUGH tag if translucent.\n");
  }

  if (buffer_line.friction != (float_t)M_PI && 
      buffer_line.friction < 0.0001)
  {
    DDF_Error("Friction value too low (%1.2f), it would prevent "
        "all movement.\n", buffer_line.friction);
  }
  
  if (buffer_line.viscosity != (float_t)M_PI && 
      buffer_line.viscosity >= 0.9999)
  {
    DDF_Error("Viscosity value too high (%1.2f), it would prevent "
        "all movement.\n", buffer_line.viscosity);
  }
  
  // FIXME: check more stuff...
  
  // transfer static entry to dynamic entry
  
  base = dynamic_line->ddf;
  dynamic_line[0] = buffer_line;
  dynamic_line->ddf = base;

  // compute CRC...
  CRC32_Init(&dynamic_line->ddf.crc);

  // FIXME: add stuff...
  
  CRC32_Done(&dynamic_line->ddf.crc);
}


void DDF_ReadLines(void *data, int size)
{
  readinfo_t lines;

  lines.memfile = data;
  lines.memsize = size;
  lines.tag = "LINES";
  lines.entries_per_dot = 6;

  if (lines.memfile)
  {
    lines.message = NULL;
    lines.filename = NULL;
    lines.lumpname = "DDFLINE";
  }
  else
  {
    lines.message = "DDF_InitLinedefs";
    lines.filename = "lines.ddf";
    lines.lumpname = NULL;
  }

  lines.start_entry  = LinedefStartEntry;
  lines.copy_old     = LinedefCopyOld;
  lines.instantiate  = LinedefInstantiate;
  lines.parse_field  = LinedefParseField;
  lines.finish_entry = LinedefFinishEntry;

  DDF_MainReadFile(&lines);
}

//
// DDF_LinedefInit
//
void DDF_LinedefInit(void)
{
  Z_InitStackArray(&ddf_linetypes_a, (void ***)&ddf_linetypes, 
      sizeof(linedeftype_t), 0);

  // clear lookup cache
  memset(line_lookup_cache, 0, sizeof(line_lookup_cache));
}

//
// DDF_LinedefCleanUp
//
void DDF_LinedefCleanUp(void)
{
  int i;
  
  for (i=0; i < num_ddf_linetypes; i++)
  {
    linedeftype_t *line = ddf_linetypes[i];
    
    line->t.inspawnobj = line->t.inspawnobj_ref ?
        DDF_MobjLookup(line->t.inspawnobj_ref) : NULL;
    
    line->t.outspawnobj = line->t.outspawnobj_ref ?
        DDF_MobjLookup(line->t.outspawnobj_ref) : NULL;
  }
}

//
// DDF_LineGetScroller
//
// Check for scroll types

void DDF_LineGetScroller(const char *info, void *storage, void *priv)
{
  int i;

  for (i = sizeof(s_scroll) / sizeof(s_scroll[0]); i--;)
  {
    if (!stricmp(info, s_scroll[i].s))
    {
      s_dir &= s_scroll[i].n >> 16;
      s_dir |= s_scroll[i].n & 0xffff;
      return;
    }
  }
  DDF_Error("Unknown scroll direction %s\n", info);
}

//
// DDF_LineGetSecurity
//
// Get Red/Blue/Yellow
//
void DDF_LineGetSecurity(const char *info, void *storage, void *priv)
{
  int i;
  boolean_t required = false;

  if (info[0] == '+')
  {
    required = true;
    info++;
  }

  for (i = sizeof(s_keys) / sizeof(s_keys[0]); i--;)
  {
    if (!stricmp(info, s_keys[i].s))
    {
      buffer_line.keys |= s_keys[i].n;
      if (required)
        buffer_line.keys |= s_keys[i].n << 8;
      return;
    }
  }

  DDF_Error("Unknown key type %s\n", info);
}

//
// DDF_LineGetTrigType
//
// Check for walk/push/shoot
//
void DDF_LineGetTrigType(const char *info, void *storage, void *priv)
{
  int i;

  for (i = sizeof(s_trigger) / sizeof(s_trigger[0]); i--;)
  {
    if (!stricmp(info, s_trigger[i].s))
    {
      buffer_line.type = s_trigger[i].n;
      return;
    }
  }

  DDF_Error("Unknown Trigger type %s\n", info);
}

//
// DDF_LineGetActivators
//
// Get player/monsters/missiles
//
void DDF_LineGetActivators(const char *info, void *storage, void *priv)
{
  int i;

  for (i = sizeof(s_activators) / sizeof(s_activators[0]); i--;)
  {

    if (!stricmp(info, s_activators[i].s))
    {
      buffer_line.obj |= s_activators[i].n;
      return;
    }
  }

  DDF_Error("Unknown Activator type %s\n", info);
}

static specflags_t extrafloor_types[] =
{
    // definers:
    {"THIN",   EF_DEF_THIN,   0},
    {"THICK",  EF_DEF_THICK,  0},
    {"LIQUID", EF_DEF_LIQUID, 0},

    // modifiers:
    {"FALL THROUGH", EXFL_Liquid, 0},
    {"SEE THROUGH", EXFL_SeeThrough, 0},
    {"SHOOT THROUGH", EXFL_ShootThrough, 0},
    {"BOOMTEX", EXFL_BoomTex, 0},
    {"WATER", EXFL_Water, 0},
    {"SHADE", EXFL_NoShade, 1},
    {"FLOODER", EXFL_Flooder, 0},
    {"SIDE UPPER", EXFL_SideUpper, 0},
    {"SIDE LOWER", EXFL_SideLower, 0},
    {NULL, 0, 0}
};

//
// DDF_LineGetExtraFloor
//
// Gets the extra floor type(s).
//
// -AJA- 1999/06/21: written.
// -AJA- 2000/03/27: updated for simpler system.

void DDF_LineGetExtraFloor(const char *info, void *storage, void *priv)
{
  int flag_value;

  if (DDF_CompareName(info, "NONE") == 0)
  {
    buffer_line.ef.type = 0;
    return;
  }

  switch (DDF_MainCheckSpecialFlag(info, extrafloor_types,
      &flag_value, true, false))
  {
    case CHKF_Positive:
      buffer_line.ef.type |= flag_value;
      break;
    
    case CHKF_Negative:
      buffer_line.ef.type &= ~flag_value;
      break;

    case CHKF_User:
    case CHKF_Unknown:
      DDF_Error("Unknown Extrafloor Type: %s", info);
      break;
  }
}

static specflags_t ef_control_types[] =
{
    {"NONE",   EFCTL_None,   0},
    {"REMOVE", EFCTL_Remove, 0},
    {NULL, 0, 0}
};

//
// DDF_LineGetEFControl
//
void DDF_LineGetEFControl(const char *info, void *storage, void *priv)
{
  int flag_value;

  switch (DDF_MainCheckSpecialFlag(info, ef_control_types,
      &flag_value, false, false))
  {
    case CHKF_Positive:
    case CHKF_Negative:
      buffer_line.ef.control = flag_value;
      break;
    
    case CHKF_User:
    case CHKF_Unknown:
      DDF_Error("Unknown CONTROL_EXTRAFLOOR tag: %s", info);
      break;
  }
}

//
// DDF_LineGetTeleportSpecial
//
// Gets the teleporter special flags.
//
// -AJA- 1999/07/12: written.

static specflags_t teleport_specials[] =
{
    {"SAME DIR", TELSP_SameDir, 0},
    {"SAME HEIGHT", TELSP_SameHeight, 0},
    {"SAME SPEED", TELSP_SameSpeed, 0},
    {"SAME OFFSET", TELSP_SameOffset, 0},
    {"PRESERVE", TELSP_Preserve, 0},
    {"ROTATE", TELSP_Rotate, 0},
    {NULL, 0, 0}
};

void DDF_LineGetTeleportSpecial(const char *info, void *storage, void *priv)
{
  int flag_value;

  switch (DDF_MainCheckSpecialFlag(info, teleport_specials,
      &flag_value, true, false))
  {
    case CHKF_Positive:
      buffer_line.t.special |= flag_value;
      break;
    
    case CHKF_Negative:
      buffer_line.t.special &= ~flag_value;
      break;

    case CHKF_User:
    case CHKF_Unknown:
      DDF_Error("DDF_LineGetTeleportSpecial: Unknown Special: %s", info);
      break;
  }
}

//
// DDF_LineGetScrollPart
//
// Gets the scroll part flags.
//
// -AJA- 1999/07/12: written.

static specflags_t scrollpart_specials[] =
{
    {"RIGHT UPPER", SCPT_RightUpper, 0},
    {"RIGHT MIDDLE", SCPT_RightMiddle, 0},
    {"RIGHT LOWER", SCPT_RightLower, 0},
    {"RIGHT", SCPT_RIGHT, 0},
    {"LEFT UPPER", SCPT_LeftUpper, 0},
    {"LEFT MIDDLE", SCPT_LeftMiddle, 0},
    {"LEFT LOWER", SCPT_LeftLower, 0},
    {"LEFT", SCPT_LEFT, 0},
    {"LEFT REVERSE X", SCPT_LeftRevX, 0},
    {"LEFT REVERSE Y", SCPT_LeftRevY, 0},
    {NULL, 0, 0}
};

void DDF_LineGetScrollPart(const char *info, void *storage, void *priv)
{
  int flag_value;

  switch (DDF_MainCheckSpecialFlag(info, scrollpart_specials,
      &flag_value, true, false))
  {
    case CHKF_Positive:
      buffer_line.scroll_parts |= flag_value;
      break;
    
    case CHKF_Negative:
      buffer_line.scroll_parts &= ~flag_value;
      break;

    case CHKF_User:
    case CHKF_Unknown:
      DDF_Error("DDF_LineGetScrollPart: Unknown Part: %s", info);
      break;
  }
}

//
// DDF_GetSecHeightReference
//
// Finds a sector height, using the reference provided; will select
// the approriate method of obtaining this value, if it cannot
// get it directly.
//
// -KM-  1998/09/01 Wrote Procedure.
// -ACB- 1998/09/06 Remarked and Reformatted.
//
float_t DDF_GetSecHeightReference(heightref_e ref, sector_t * sec)
{
  switch (ref)
  {
    case ref_absolute:
      return 0;

    case ref_floorHeight:
      return sec->floor.h;

    case ref_ceilingHeight:
      return sec->ceil.h;

    case ref_LowestSurroundingCeiling:
      return P_FindLowestCeilingSurrounding(sec);

    case ref_HighestSurroundingCeiling:
      return P_FindHighestCeilingSurrounding(sec);

    case ref_LowestSurroundingFloor:
      return P_FindLowestFloorSurrounding(sec);

    case ref_HighestSurroundingFloor:
      return P_FindHighestFloorSurrounding(sec);

    case ref_NextHighestFloor:
      return P_FindNextHighestFloor(sec, sec->floor.h);

    case ref_NextLowestFloor:
      return P_FindNextLowestFloor(sec, sec->floor.h);

    case ref_NextHighestCeiling:
      return P_FindNextHighestCeiling(sec, sec->ceil.h);

    case ref_NextLowestCeiling:
      return P_FindNextLowestCeiling(sec, sec->ceil.h);

    case ref_LowestLoTexture:
      return P_FindRaiseToTexture(sec);

    default:
      I_Error("DDF_GetSecHeightReference: undefined reference %d\n", ref);
  }

  return -1;
}

//
// DDF_LineLookupNum
//
// Returns the special linedef properties from given specialtype
//
// -KM-  1998/09/01 Wrote Procedure
// -ACB- 1998/09/06 Remarked and Reformatted....

const linedeftype_t *DDF_LineLookupNum(int number)
{
  int slot = DDF_LineHashFunc(number);
  int i;

  // check the cache
  if (line_lookup_cache[slot] &&
      line_lookup_cache[slot]->ddf.number == number)
  {
    return line_lookup_cache[slot];
  }

  // find line by number
  // NOTE: go backwards, so newer ones are found first
  for (i=num_ddf_linetypes-1; i >= 0; i--)
  {
    if (ddf_linetypes[i]->ddf.number == number)
      break;
  }
  
  if (i < 0)
  {
    // -AJA- 1999/06/19: Don't crash out if the line type is unknown, just
    // print a message and ignore it (like for unknown thing types).

    I_Warning("Unknown line type %i\n", number);

    return &template_line;
  }

  // update the cache
  line_lookup_cache[slot] = ddf_linetypes[i];

  return ddf_linetypes[i];
}

static specflags_t line_specials[] =
{
  {"MUST REACH", LINSP_MustReach, 0},
  {"SWITCH SEPARATE", LINSP_SwitchSeparate, 0},
  {NULL, 0, 0}
};

//
// DDF_LineGetSpecialFlags
//
// Gets the line special flags.

void DDF_LineGetSpecialFlags(const char *info, void *storage, void *priv)
{
  int flag_value;

  switch (DDF_MainCheckSpecialFlag(info, line_specials, &flag_value,
      true, false))
  {
    case CHKF_Positive:
      buffer_line.special_flags |= flag_value;
      break;
    
    case CHKF_Negative:
      buffer_line.special_flags &= ~flag_value;
      break;

    case CHKF_User:
    case CHKF_Unknown:
      DDF_Error("Unknown line special: %s", info);
      break;
  }
}

//
// DDF_LineGetRadTrig
//
// Gets the line's radius trigger effect.

static void DDF_LineGetRadTrig(const char *info, void *storage, void *priv)
{
  if (DDF_CompareName(info, "ENABLE_TAGGED") == 0)
  {
    buffer_line.trigger_effect = +1;
    return;
  }
  if (DDF_CompareName(info, "DISABLE_TAGGED") == 0)
  {
    buffer_line.trigger_effect = -1;
    return;
  }

  DDF_Error("DDF_LineGetRadTrig: Unknown effect: %s\n", info);
}
