//----------------------------------------------------------------------------
//  DOOM Definition File Codes (Sectors)
//----------------------------------------------------------------------------
// 
//  Copyright (c) 1999-2000  The EDGE Team.
// 
//  This program is free software; you can redistribute it and/or
//  modify it under the terms of the GNU General Public License
//  as published by the Free Software Foundation; either version 2
//  of the License, or (at your option) any later version.
//
//  This program is distributed in the hope that it will be useful,
//  but WITHOUT ANY WARRANTY; without even the implied warranty of
//  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//  GNU General Public License for more details.
//
//----------------------------------------------------------------------------
//
// Sector Setup and Parser Code
//
// -KM- 1998/09/27 Written.
//

#include "i_defs.h"

#include "ddf_locl.h"
#include "ddf_main.h"
#include "dm_state.h"
#include "m_fixed.h"
#include "p_mobj.h"
#include "p_local.h"
#include "z_zone.h"

#define DDF_SectHashFunc(x)  (((x) + 211) % 211)

static specialsector_t buffer_sect;
static specialsector_t *dynamic_sect;

static const specialsector_t template_sect =
{
  DDF_BASE_NIL,  // ddf

  false,      // secret
  GRAVITY,    // gravity
  FRICTION,   // friction
  VISCOSITY,  // viscosity
  false,      // crush

  // Floor
  {
    mov_undefined,  // type
    false,  // floorOrCeiling
    false,  // crush
    0,      // speed_up
    0,      // speed_down
    0,      // destref
    0,      // dest
    "",     // tex
    0,      // wait
    0,      // prewait
    NULL,   // sfxstart
    NULL,   // sfxup
    NULL,   // sfxdown
    NULL,   // sfxstop
    0,      // scroll_angle
    0.0     // scroll_speed
  },

  // Ceiling
  {
    mov_undefined,  // type
    true,   // floorOrCeiling
    false,  // crush
    0,      // speed_up
    0,      // speed_down
    0,      // destref
    0,      // dest
    "",     // tex
    0,      // wait
    0,      // prewait
    NULL,   // sfxstart
    NULL,   // sfxup
    NULL,   // sfxdown
    NULL,   // sfxstop
    0,      // scroll_angle
    0.0     // scroll_speed
  },

  {lite_none, 0,0,0,0},  // lights
  0,            // damage
  0,            // damagetime
  0,            // special_flags
  EXIT_None,    // exit type
  NULL,         // colourmap
  0,            // sfx
  DEFAULT_APPEAR,  // appear
  4,            // crush_time
  10.0,         // crush_damage
  0,            // push_angle
  0.0,          // push_speed
  0.0           // push_zspeed
};

specialsector_t ** ddf_sectors = NULL;
int num_ddf_sectors = 0;

static stack_array_t ddf_sectors_a;

static specialsector_t *sector_lookup_cache[211];


void DDF_SectGetSpecialFlags(const char *info, void *storage, void *priv);

static const commandlist_t sect_commands[] =
{
    {"SECRET", DDF_MainGetBoolean, &buffer_sect.secret, NULL},
    {"DAMAGE", DDF_MainGetFloat, &buffer_sect.damage, NULL},
    {"DAMAGETIME", DDF_MainGetTime, &buffer_sect.damagetime, NULL},
    {"SPECIAL", DDF_SectGetSpecialFlags, &buffer_sect.special_flags, NULL},
    {"CRUSH", DDF_MainGetBoolean, &buffer_sect.crush, NULL},

    // -AJA- FIXME: use a generic system for subfields...
    {"FLOOR.TYPE", DDF_MainGetMType, &buffer_sect.f.type, NULL},
    {"FLOOR.SPEED UP", DDF_MainGetFloat, &buffer_sect.f.speed_up, NULL},
    {"FLOOR.SPEED DOWN", DDF_MainGetFloat, &buffer_sect.f.speed_down, NULL},
    {"FLOOR.DEST REF", DDF_MainGetDestRef, &buffer_sect.f.destref, NULL},
    {"FLOOR.DEST OFFSET", DDF_MainGetFloat, &buffer_sect.f.dest, NULL},
    {"FLOOR.TEXTURE", DDF_MainGetInlineStr10, &buffer_sect.f.tex, NULL},
    {"FLOOR.PAUSE TIME", DDF_MainGetTime, &buffer_sect.f.wait, NULL},
    {"FLOOR.WAIT TIME", DDF_MainGetTime, &buffer_sect.f.prewait, NULL},
    {"FLOOR.SFX START", DDF_MainLookupSound, &buffer_sect.f.sfxstart, NULL},
    {"FLOOR.SFX UP", DDF_MainLookupSound, &buffer_sect.f.sfxup, NULL},
    {"FLOOR.SFX DOWN", DDF_MainLookupSound, &buffer_sect.f.sfxdown, NULL},
    {"FLOOR.SFX STOP", DDF_MainLookupSound, &buffer_sect.f.sfxstop, NULL},
    {"FLOOR.SCROLL ANGLE", DDF_MainGetAngle, &buffer_sect.f.scroll_angle, NULL},
    {"FLOOR.SCROLL SPEED", DDF_MainGetFloat, &buffer_sect.f.scroll_speed, NULL},

    {"CEILING.TYPE", DDF_MainGetMType, &buffer_sect.c.type, NULL},
    {"CEILING.SPEED UP", DDF_MainGetFloat, &buffer_sect.c.speed_up, NULL},
    {"CEILING.SPEED DOWN", DDF_MainGetFloat, &buffer_sect.c.speed_down, NULL},
    {"CEILING.DEST REF", DDF_MainGetDestRef, &buffer_sect.c.destref, NULL},
    {"CEILING.DEST OFFSET", DDF_MainGetFloat, &buffer_sect.c.dest, NULL},
    {"CEILING.TEXTURE", DDF_MainGetInlineStr10, &buffer_sect.c.tex, NULL},
    {"CEILING.PAUSE TIME", DDF_MainGetTime, &buffer_sect.c.wait, NULL},
    {"CEILING.WAIT TIME", DDF_MainGetTime, &buffer_sect.c.prewait, NULL},
    {"CEILING.SFX START", DDF_MainLookupSound, &buffer_sect.c.sfxstart, NULL},
    {"CEILING.SFX UP", DDF_MainLookupSound, &buffer_sect.c.sfxup, NULL},
    {"CEILING.SFX DOWN", DDF_MainLookupSound, &buffer_sect.c.sfxdown, NULL},
    {"CEILING.SFX STOP", DDF_MainLookupSound, &buffer_sect.c.sfxstop, NULL},
    {"CEILING.SCROLL ANGLE", DDF_MainGetAngle, &buffer_sect.c.scroll_angle, NULL},
    {"CEILING.SCROLL SPEED", DDF_MainGetFloat, &buffer_sect.c.scroll_speed, NULL},

    {"LIGHT TYPE", DDF_MainGetLighttype, &buffer_sect.l.type, NULL},
    {"LIGHT LEVEL", DDF_MainGetNumeric, &buffer_sect.l.light, NULL},
    {"LIGHT DARKTIME", DDF_MainGetTime, &buffer_sect.l.darktime, NULL},
    {"LIGHT BRIGHTTIME", DDF_MainGetTime, &buffer_sect.l.brighttime, NULL},
    {"LIGHT PROBABILITY", DDF_MainGetNumeric, &buffer_sect.l.light, NULL},
    {"LIGHT SYNC", DDF_MainGetTime, &buffer_sect.l.sync, NULL},
    {"EXIT", DDF_MainGetExit, &buffer_sect.e_exit, NULL},
    {"USE COLOURMAP", DDF_MainGetColourmap, &buffer_sect.use_colourmap, NULL},
    {"GRAVITY", DDF_MainGetFloat, &buffer_sect.gravity, NULL},
    {"FRICTION", DDF_MainGetFloat, &buffer_sect.friction, NULL},
    {"VISCOSITY", DDF_MainGetFloat, &buffer_sect.viscosity, NULL},
    {"SOUND", DDF_MainLookupSound, &buffer_sect.sfx, NULL},
    {"WHEN APPEAR", DDF_MainGetWhenAppear, &buffer_sect.appear, NULL},
    {"CRUSH TIME", DDF_MainGetNumeric, &buffer_sect.crush_time, NULL},
    {"CRUSH DAMAGE", DDF_MainGetFloat, &buffer_sect.crush_damage, NULL},
    {"PUSH ANGLE", DDF_MainGetAngle, &buffer_sect.push_angle, NULL},
    {"PUSH SPEED", DDF_MainGetFloat, &buffer_sect.push_speed, NULL},
    {"PUSH ZSPEED", DDF_MainGetFloat, &buffer_sect.push_zspeed, NULL},
    {NULL, NULL, NULL, NULL}
};


//
//  DDF PARSE ROUTINES
//

static boolean_t SectorStartEntry(const char *name)
{
  int i;
  boolean_t replaces = false;
  int number = MAX(0, atoi(name));

  if (number == 0)
    DDF_Error("Bad sector number in sectors.ddf: %s\n", name);

  for (i=0; i < num_ddf_sectors; i++)
  {
    if (ddf_sectors[i]->ddf.number == number)
    {
      dynamic_sect = ddf_sectors[i];
      replaces = true;
      break;
    }
  }

  // not found, create a new one
  if (! replaces)
  {
    Z_SetArraySize(&ddf_sectors_a, ++num_ddf_sectors);
    
    dynamic_sect = ddf_sectors[num_ddf_sectors - 1];
  }

  dynamic_sect->ddf.name   = NULL;
  dynamic_sect->ddf.number = number;

  return replaces;
}

static void SectorCopyOld(void)
{
  buffer_sect = dynamic_sect[0];
}

static void SectorInstantiate(const char *template_name)
{
  int i;
  int number;

  if (! template_name)
  {
    // use built-in template
    buffer_sect = template_sect;
    return;
  }

  number = MAX(0, atoi(template_name));

  if (number == 0)
    DDF_Error("Bad template reference in sectors.ddf: %s\n", template_name);

  for (i=0; i < num_ddf_sectors; i++)
  {
    // ignore current entry
    if (ddf_sectors[i] == dynamic_sect)
      continue;
    
    if (ddf_sectors[i]->ddf.number == number)
    {
      buffer_sect = ddf_sectors[i][0];
      return;
    }
  }

  I_Error("Template not found for sectors.ddf: %s\n", template_name);
}

static void SectorParseField(const char *field, const char *contents,
    int index, boolean_t is_last)
{
  I_WriteDebug("SECTOR_PARSE: [%s] [%s]\n", field, contents);

  if (! DDF_MainParseField(sect_commands, field, contents))
    DDF_Error("Unknown sectors.ddf command: %s\n", field);
}

static void SectorFinishEntry(void)
{
  ddf_base_t base;
  
  // FIXME: check stuff...
  
  // transfer static entry to dynamic entry
  
  base = dynamic_sect->ddf;
  dynamic_sect[0] = buffer_sect;
  dynamic_sect->ddf = base;

  // compute CRC...
  CRC32_Init(&dynamic_sect->ddf.crc);

  // FIXME: add stuff...
  
  CRC32_Done(&dynamic_sect->ddf.crc);
}


void DDF_ReadSectors(void *data, int size)
{
  readinfo_t sects;

  sects.memfile = data;
  sects.memsize = size;
  sects.tag = "SECTORS";
  sects.entries_per_dot = 1;

  if (sects.memfile)
  {
    sects.message = NULL;
    sects.filename = NULL;
    sects.lumpname = "DDFSECT";
  }
  else
  {
    sects.message = "DDF_InitSectors";
    sects.filename = "sectors.ddf";
    sects.lumpname = NULL;
  }

  sects.start_entry  = SectorStartEntry;
  sects.copy_old     = SectorCopyOld;
  sects.instantiate  = SectorInstantiate;
  sects.parse_field  = SectorParseField;
  sects.finish_entry = SectorFinishEntry;

  DDF_MainReadFile(&sects);
}

void DDF_SectorInit(void)
{
  Z_InitStackArray(&ddf_sectors_a, (void ***)&ddf_sectors,
      sizeof(specialsector_t), 0);

  // clear lookup cache
  memset(sector_lookup_cache, 0, sizeof(sector_lookup_cache));
}

void DDF_SectorCleanUp(void)
{
}

//
// DDF_SectorLookupNum
//
// Returns the special sector properties from given specialtype
//
// -KM-  1998/09/19 Wrote Procedure
// -AJA- 2000/02/09: Reworked.  Hash table is now a cache.

const specialsector_t *DDF_SectorLookupNum(int number)
{
  int i;
  int slot = DDF_SectHashFunc(number);

  // check the lookup cache...
  if (sector_lookup_cache[slot] &&
      sector_lookup_cache[slot]->ddf.number == number)
  {
    return sector_lookup_cache[slot];
  }

  // find sector by number
  // NOTE: we go backwards, so that later entries are found first
  for (i=num_ddf_sectors-1; i >= 0; i--)
  {
    if (ddf_sectors[i]->ddf.number == number)
      break;
  }

  if (i < 0)
  {
    // -AJA- 1999/06/19: Don't crash out if the sector type is unknown,
    // just print a message and ignore it (like for unknown thing types).

    I_Warning("Unknown sector type: %d\n", number);

    return &template_sect;
  }

  // update the cache
  sector_lookup_cache[slot] = ddf_sectors[i];

  return ddf_sectors[i];
}

static specflags_t sector_specials[] =
{
    {"WHOLE REGION", SECSP_WholeRegion, 0},
    {"PROPORTIONAL", SECSP_Proportional, 0},
    {"SYNCHRONISED", SECSP_Synchronised, 0},
    {"PUSH ALL", SECSP_PushAll, 0},
    {"PUSH CONSTANT", SECSP_PushConstant, 0},
    {NULL, 0, 0}
};

//
// DDF_SectGetSpecialFlags
//
// Gets the sector specials.

void DDF_SectGetSpecialFlags(const char *info, void *storage, void *priv)
{
  int flag_value;

  switch (DDF_MainCheckSpecialFlag(info, sector_specials, &flag_value,
      true, false))
  {
    case CHKF_Positive:
      buffer_sect.special_flags |= flag_value;
      break;
    
    case CHKF_Negative:
      buffer_sect.special_flags &= ~flag_value;
      break;

    case CHKF_User:
    case CHKF_Unknown:
      DDF_Error("Unknown sector special: %s", info);
      break;
  }
}

static specflags_t exit_types[] =
{
    {"NONE", EXIT_None, 0},
    {"EXIT", EXIT_Normal, 0},
    {"SECRET", EXIT_Secret, 0},
    {NULL, 0, 0}
};

//
// DDF_MainGetExit
//
// Get the exit type

void DDF_MainGetExit(const char *info, void *storage, void *priv)
{
  int *dest = (int *)storage;
  int flag_value;

  switch (DDF_MainCheckSpecialFlag(info, exit_types, &flag_value,
      false, false))
  {
    case CHKF_Positive:
    case CHKF_Negative:
      (*dest) = flag_value;
      break;

    case CHKF_User:
    case CHKF_Unknown:
      DDF_Error("Unknown Exit type: %s\n", info);
      break;
  }
}

static specflags_t light_types[] =
{
    {"SET", lite_set, 0},
    {"STROBE", lite_strobe, 0},
    {"FLASH", lite_flash, 0},
    {"GLOW", lite_glow, 0},
    {"FLICKER", lite_fireflicker, 0},
    {NULL, 0, 0}
};

//
// DDF_MainGetLighttype
//
// Get the light type

void DDF_MainGetLighttype(const char *info, void *storage, void *priv)
{
  int *dest = (int *)storage;
  int flag_value;

  switch (DDF_MainCheckSpecialFlag(info, light_types, &flag_value,
      false, false))
  {
    case CHKF_Positive:
    case CHKF_Negative:
      (*dest) = flag_value;
      break;

    case CHKF_User:
    case CHKF_Unknown:
      DDF_Error("Unknown light type: %s\n", info);
      break;
  }
}

static specflags_t movement_types[] =
{
    {"MOVE", mov_Once, 0},
    {"MOVEWAITRETURN", mov_MoveWaitReturn, 0},
    {"CONTINUOUS", mov_Continuous, 0},
    {"PLAT", mov_Plat, 0},
    {"BUILDSTAIRS", mov_Stairs, 0},
    {"STOP", mov_Stop, 0},
    {NULL, 0, 0}
};

//
// DDF_MainGetMType
//
// Get movement types: MoveWaitReturn etc

void DDF_MainGetMType(const char *info, void *storage, void *priv)
{
  int *dest = (int *)storage;
  int flag_value;

  switch (DDF_MainCheckSpecialFlag(info, movement_types, &flag_value,
      false, false))
  {
    case CHKF_Positive:
    case CHKF_Negative:
      (*dest) = flag_value;
      break;

    case CHKF_User:
    case CHKF_Unknown:
      DDF_Error("Unknown Movement type: %s\n", info);
      break;
  }
}

static specflags_t reference_types[] =
{
    {"ABSOLUTE", ref_absolute, false},
    {"FLOOR", ref_floorHeight, false},
    {"CEILING", ref_ceilingHeight, false},
    {"LOSURROUNDINGCEILING", ref_LowestSurroundingCeiling, false},
    {"HISURROUNDINGCEILING", ref_HighestSurroundingCeiling, false},
    {"LOSURROUNDINGFLOOR", ref_LowestSurroundingFloor, false},
    {"HISURROUNDINGFLOOR", ref_HighestSurroundingFloor, false},
    {"NEXTHIGHESTFLOOR", ref_NextHighestFloor, false},
    {"NEXTLOWESTFLOOR", ref_NextLowestFloor, false},
    {"NEXTHIGHESTCEILING", ref_NextHighestCeiling, false},
    {"NEXTLOWESTCEILING", ref_NextLowestCeiling, false},
    {"LOWESTBOTTOMTEXTURE", ref_LowestLoTexture, false}
};

//
// DDF_MainGetDestRef
//
// Get surroundingsectorceiling/floorheight etc

void DDF_MainGetDestRef(const char *info, void *storage, void *priv)
{
  int *dest = (int *)storage;
  int flag_value;

  switch (DDF_MainCheckSpecialFlag(info, reference_types, &flag_value,
      false, false))
  {
    case CHKF_Positive:
    case CHKF_Negative:
      (*dest) = flag_value;
      break;

    case CHKF_User:
    case CHKF_Unknown:
      DDF_Error("Unknown Reference Point: %s\n", info);
      break;
  }
}
