//----------------------------------------------------------------------------
//  DOOM Definition Files Code (Sounds)
//----------------------------------------------------------------------------
// 
//  Copyright (c) 1999-2000  The EDGE Team.
// 
//  This program is free software; you can redistribute it and/or
//  modify it under the terms of the GNU General Public License
//  as published by the Free Software Foundation; either version 2
//  of the License, or (at your option) any later version.
//
//  This program is distributed in the hope that it will be useful,
//  but WITHOUT ANY WARRANTY; without even the implied warranty of
//  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//  GNU General Public License for more details.
//
//----------------------------------------------------------------------------
//
// -KM- 1998/09/27 Finished :-)
//

#include "i_defs.h"

#include "dm_defs.h"
#include "dm_type.h"
#include "ddf_locl.h"
#include "m_random.h"
#include "s_sound.h"
#include "z_zone.h"

static sfxinfo_t buffer_sfx;
static sfxinfo_t *dynamic_sfx;

static const sfxinfo_t template_sfx =
{
  DDF_BASE_NIL,  // ddf

  "",     // lump_name
  0,      // sfx_id
  0,      // singularity
  999,    // priority (lower is more important)
  255,    // volume
  false,  // looping
  false,  // precious
  S_CLIPPING_DIST, // max_distance
  8,      // bits
  0,      // stereo
  NULL,   // cache next
  NULL    // cache prev
};

sfxinfo_t ** S_sfx;
int numsfx = 0;

static stack_array_t S_sfx_a;


static const commandlist_t sfx_commands[] =
{
    {"LUMP NAME", DDF_MainGetInlineStr10, &buffer_sfx.lump_name, NULL},
    {"SINGULAR", DDF_MainGetNumeric, &buffer_sfx.singularity, NULL},
    {"PRIORITY", DDF_MainGetNumeric, &buffer_sfx.priority, NULL},
    {"VOLUME", DDF_MainGetPercent, &buffer_sfx.volume, NULL},
    {"LOOP", DDF_MainGetBoolean, &buffer_sfx.looping, NULL},
    {"PRECIOUS", DDF_MainGetBoolean, &buffer_sfx.precious, NULL},
    {"MAX DISTANCE", DDF_MainGetFloat, &buffer_sfx.max_distance, NULL},
    {"BITS", DDF_MainGetNumeric, &buffer_sfx.bits, NULL},
    {"STEREO", DDF_MainGetBoolean, &buffer_sfx.stereo, NULL},
    {NULL, NULL, NULL, NULL}
};

// -KM- 1998/10/29 Done sfx_t first so structure is aligned.
bastard_sfx_t bastard_sfx[] =
{
    {0, "swtchn"},
    {0, "tink"},
    {0, "radio"},
    {0, "oof"},
    {0, "slop"},
    {0, "itmbk"},
    {0, "pstop"},
    {0, "stnmov"},
    {0, "pistol"},
    {0, "swtchx"},
    {0, "noway"},
    {0, "bossit"},
    {0, "bospn"},
    {0, "bosdth"},
    {0, "bospit"},
    {0, "telept"},
    {0, "boscub"},
    {0, "pldeth"},
    {0, "pdiehi"},
    {0, "shotgn"},
    {0, "barexp"},
    {0, "sgcock"},
    {0, "jpmove"},
    {0, "jpidle"},
    {0, "jprise"},
    {0, "jpdown"},
    {0, "jpflow"}
};


//
//  DDF PARSE ROUTINES
//

static boolean_t SoundStartEntry(const char *name)
{
  int i;
  boolean_t replaces = false;

  if (name && name[0])
  {
    for (i=0; i < numsfx; i++)
    {
      if (DDF_CompareName(S_sfx[i]->ddf.name, name) == 0)
      {
        dynamic_sfx = S_sfx[i];
        replaces = true;
        break;
      }
    }
  }

  // not found, create a new one
  if (! replaces)
  {
    Z_SetArraySize(&S_sfx_a, ++numsfx);
    
    dynamic_sfx = S_sfx[numsfx - 1];
    dynamic_sfx->sfx_id = numsfx - 1;
    dynamic_sfx->ddf.name = (name && name[0]) ? Z_StrDup(name) :
        DDF_MainCreateUniqueName("UNNAMED_SOUND", numsfx);
  }

  dynamic_sfx->ddf.number = 0;

  return replaces;
}

static void SoundCopyOld(void)
{
  buffer_sfx = dynamic_sfx[0];
}

static void SoundInstantiate(const char *template_name)
{
  int i;

  if (! template_name)
  {
    // use built-in template
    buffer_sfx = template_sfx;
    return;
  }

  for (i=0; i < numsfx; i++)
  {
    // ignore current entry
    if (S_sfx[i] == dynamic_sfx)
      continue;
    
    if (DDF_CompareName(S_sfx[i]->ddf.name, template_name) == 0)
    {
      buffer_sfx = S_sfx[i][0];
      return;
    }
  }

  I_Error("Template `%s' not found for sounds.ddf\n", template_name);
}

static void SoundParseField(const char *field, const char *contents,
    int index, boolean_t is_last)
{
  I_WriteDebug("SOUND_PARSE: [%s] [%s]\n", field, contents);

  if (! DDF_MainParseField(sfx_commands, field, contents))
    DDF_Error("Unknown sounds.ddf command: %s\n", field);
}

static void SoundFinishEntry(void)
{
  ddf_base_t base;
  int id;
  
  if (!buffer_sfx.lump_name[0])
    DDF_Error("Missing LUMP_NAME for sound.\n");

  // transfer static entry to dynamic entry
  
  base = dynamic_sfx->ddf;
  id = dynamic_sfx->sfx_id;
  dynamic_sfx[0] = buffer_sfx;
  dynamic_sfx->ddf = base;
  dynamic_sfx->sfx_id = id;

  // Compute CRC.  In this case, there is no need, since sounds have
  // zero impact on the game simulation itself.
  dynamic_sfx->ddf.crc = 0;
}


void DDF_ReadSFX(void *data, int size)
{
  readinfo_t sfx_r;

  sfx_r.memfile = data;
  sfx_r.memsize = size;
  sfx_r.tag = "SOUNDS";
  sfx_r.entries_per_dot = 8;

  if (sfx_r.memfile)
  {
    sfx_r.message = NULL;
    sfx_r.filename = NULL;
    sfx_r.lumpname = "DDFSFX";
  }
  else
  {
    sfx_r.message = "DDF_InitSounds";
    sfx_r.filename = "sounds.ddf";
    sfx_r.lumpname = NULL;
  }

  sfx_r.start_entry  = SoundStartEntry;
  sfx_r.copy_old     = SoundCopyOld;
  sfx_r.instantiate  = SoundInstantiate;
  sfx_r.parse_field  = SoundParseField;
  sfx_r.finish_entry = SoundFinishEntry;

  DDF_MainReadFile(&sfx_r);
}

void DDF_SFXInit(void)
{
  Z_InitStackArray(&S_sfx_a, (void ***)&S_sfx, sizeof(sfxinfo_t), 0);

  // create the null sfx
  Z_SetArraySize(&S_sfx_a, numsfx = 1);
  S_sfx[numsfx-1][0] = template_sfx;
}

void DDF_SFXCleanUp(void)
{
  int i;

  for (i = sizeof(bastard_sfx) / sizeof(bastard_sfx_t); i--; )
    bastard_sfx[i].s = DDF_SfxLookupSound(bastard_sfx[i].name);
}

//
// DDF_SfxLookupSound
//
// -ACB- 1999/11/06:
//   Fixed from using zero-length array style malloc - Not ANSI Compliant.
//   Renamed DDF_SfxLookupSound to destingush from the one in DDF_MAIN.C.
//
sfx_t *DDF_SfxLookupSound(const char *name)
{
  int i;
  sfx_t *r;

  // NULL Sound
  if (!name || !name[0])
    return NULL;

  r = Z_New(sfx_t, 1);
  r->num = 0;

  for (i = numsfx; i--;)
  {
    if (!strncasecmpwild(name, S_sfx[i]->ddf.name, 8))
    {
      if (r->num)
        // allocate num+1 elements (one is part of sizeof(sfx_t))
        r = Z_ReMalloc(r, sizeof(sfx_t) + sizeof(int) * r->num);
      r->sounds[r->num++] = i;
    }
  }

  if (!r->num)
    I_Error("Unknown SFX: '%.8s'\n", name);

  return r;
}

//
// DDF_MainLookupSound
//
// Lookup the sound specificed.
//
// -ACB- 1998/07/08 Checked the S_sfx table for sfx names.
// -ACB- 1998/07/18 Removed to the need set *currentcmdlist[commandref].data to -1
// -KM- 1998/09/27 Fixed this func because of sounds.ddf
//
void DDF_MainLookupSound(const char *info, void *storage, void *priv)
{
  sfx_t **dest = (sfx_t **)storage;

  DEV_ASSERT2(info && storage);

  // -KM- 1998/10/29 sfx_t finished
  *dest = DDF_SfxLookupSound(info);
}
