//----------------------------------------------------------------------------
//  DOOM Definition File Codes (Weapons)
//----------------------------------------------------------------------------
// 
//  Copyright (c) 1999-2000  The EDGE Team.
// 
//  This program is free software; you can redistribute it and/or
//  modify it under the terms of the GNU General Public License
//  as published by the Free Software Foundation; either version 2
//  of the License, or (at your option) any later version.
//
//  This program is distributed in the hope that it will be useful,
//  but WITHOUT ANY WARRANTY; without even the implied warranty of
//  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//  GNU General Public License for more details.
//
//----------------------------------------------------------------------------
//
// Player Weapons Setup and Parser Code
//
// -KM- 1998/11/25 File Written
//

#include "i_defs.h"

#include "ddf_locl.h"
#include "ddf_main.h"
#include "dm_state.h"
#include "m_math.h"
#include "m_fixed.h"
#include "p_action.h"
#include "p_mobj.h"
#include "z_zone.h"

static weaponinfo_t buffer_weapon;
static weaponinfo_t *dynamic_weapon;

static const weaponinfo_t template_weapon =
{
  DDF_BASE_NIL,  // ddf

  NULL,        // attack
  AM_NoAmmo,   // ammo;
  0,           // ammopershot;
  1,           // clip;
  false,       // autofire;
  0.0,         // kick;

  NULL,        // sa_attack;
  AM_NoAmmo,   // sa_ammo;
  0,           // sa_ammopershot;
  1,           // sa_clip;
  false,       // sa_autofire;

  0,    // first_state
  0,    // last_state
  
  0,    // up_state;
  0,    // down_state;
  0,    // ready_state;
  0,    // attack_state;
  0,    // reload_state;
  0,    // flash_state;
  0,    // sa_attack_state;
  0,    // sa_reload_state;
  0,    // sa_flash_state;
  0,    // crosshair;

  false,  // autogive;
  false,  // feedback;
  -1,     // replaces;
  0,      // priority;

  NULL,   // eject_attack;
  NULL,   // idle;
  NULL,   // engaged;
  NULL,   // hit;
  NULL,   // start;
  NULL,   // sound1;
  NULL,   // sound2;
  NULL,   // sound3;

  false,  // nothrust;
  -1,     // bind_key;
  0,      // special_flags;
  0,      // zoom_fov
  false   // refire_inacc
};

weaponinfo_t ** weaponinfo = NULL;
int numweapons = 0;

static stack_array_t weaponinfo_a;


// -KM- 1998/11/25 Always 10 weapon keys, 1 - 9 and 0
weaponkey_t weaponkey[10];

static void DDF_WGetAmmo(const char *info, void *storage, void *priv);
static void DDF_WGetUpgrade(const char *info, void *storage, void *priv);
static void DDF_WGetSpecialFlags(const char *info, void *storage, void *priv);

static const commandlist_t weapon_commands[] =
{
    {"AMMOTYPE", DDF_WGetAmmo, &buffer_weapon.ammo, NULL},
    {"AMMOPERSHOT", DDF_MainGetNumeric, &buffer_weapon.ammopershot, NULL},
    {"CLIPSIZE", DDF_MainGetNumeric, &buffer_weapon.clip, NULL},
    {"AUTOMATIC", DDF_MainGetBoolean, &buffer_weapon.autofire, NULL},
    {"SEC AMMOTYPE", DDF_WGetAmmo, &buffer_weapon.sa_ammo, NULL},
    {"SEC AMMOPERSHOT", DDF_MainGetNumeric, &buffer_weapon.sa_ammopershot, NULL},
    {"SEC CLIPSIZE", DDF_MainGetNumeric, &buffer_weapon.sa_clip, NULL},
    {"SEC AUTOMATIC", DDF_MainGetBoolean, &buffer_weapon.sa_autofire, NULL},
    {"ATTACK", DDF_MainRefAttack, &buffer_weapon.attack, NULL},
    {"SECOND ATTACK", DDF_MainRefAttack, &buffer_weapon.sa_attack, NULL},
    {"EJECT ATTACK", DDF_MainRefAttack, &buffer_weapon.eject_attack, NULL},
    {"FREE", DDF_MainGetBoolean, &buffer_weapon.autogive, NULL},
    {"BINDKEY", DDF_MainGetNumeric, &buffer_weapon.bind_key, NULL},
    {"PRIORITY", DDF_MainGetNumeric, &buffer_weapon.priority, NULL},
    {"UPGRADES", DDF_WGetUpgrade, &buffer_weapon.replaces, NULL},
    {"IDLE SOUND", DDF_MainLookupSound, &buffer_weapon.idle, NULL},
    {"ENGAGED SOUND", DDF_MainLookupSound, &buffer_weapon.engaged, NULL},
    {"HIT SOUND", DDF_MainLookupSound, &buffer_weapon.hit, NULL},
    {"START SOUND", DDF_MainLookupSound, &buffer_weapon.start, NULL},
    {"SOUND1", DDF_MainLookupSound, &buffer_weapon.sound1, NULL},
    {"SOUND2", DDF_MainLookupSound, &buffer_weapon.sound2, NULL},
    {"SOUND3", DDF_MainLookupSound, &buffer_weapon.sound3, NULL},
    {"NOTHRUST", DDF_MainGetBoolean, &buffer_weapon.nothrust, NULL},
    {"FEEDBACK", DDF_MainGetBoolean, &buffer_weapon.feedback, NULL},
    {"KICK", DDF_MainGetFloat, &buffer_weapon.kick, NULL},
    {"SPECIAL", DDF_WGetSpecialFlags, &buffer_weapon.special_flags, NULL},
    {"ZOOM FOV", DDF_MainGetAngle, &buffer_weapon.zoom_fov, NULL},
    {"REFIRE INACCURATE", DDF_MainGetBoolean, &buffer_weapon.refire_inacc, NULL},
    {NULL, NULL, NULL, NULL}
};

#define WFIRST  buffer_weapon.first_state

static const state_starter_t weapon_starters[] =
{
  {"STATES(UP)",       "UP",       "UP",    &WFIRST, &buffer_weapon.up_state},
  {"STATES(DOWN)",     "DOWN",     "DOWN",  &WFIRST, &buffer_weapon.down_state},
  {"STATES(READY)",    "READY",    "READY", &WFIRST, &buffer_weapon.ready_state},
  {"STATES(ATTACK)",   "ATTACK",   "READY", &WFIRST, &buffer_weapon.attack_state},
  {"STATES(RELOAD)",   "RELOAD",   "READY", &WFIRST, &buffer_weapon.reload_state},
  {"STATES(FLASH)",    "FLASH",    "REMOVE",&WFIRST, &buffer_weapon.flash_state},
  {"STATES(SECATTACK)","SECATTACK","READY", &WFIRST, &buffer_weapon.sa_attack_state},
  {"STATES(SECRELOAD)","SECRELOAD","READY", &WFIRST, &buffer_weapon.sa_reload_state},
  {"STATES(SECFLASH)", "SECFLASH", "REMOVE",&WFIRST, &buffer_weapon.sa_flash_state},
  {"STATES(CROSSHAIR)","CROSSHAIR","REMOVE",&WFIRST, &buffer_weapon.crosshair},
  {NULL, NULL, NULL, NULL, NULL}
};

#undef WFIRST

static const actioncode_t weapon_actions[] =
{
    {"NOTHING", NULL, NULL},

    {"RANDOMJUMP",       A_RandomJump, NULL},
    {"RAISE",            A_Raise, NULL},
    {"LOWER",            A_Lower, NULL},
    {"READY",            A_WeaponReady, NULL},
    {"SHOOT",            A_WeaponShoot, DDF_StateGetAttack},
    {"EJECT",            A_WeaponEject, DDF_StateGetAttack},
    {"REFIRE",           A_ReFire, NULL},
    {"NOFIRE",           A_NoFire, NULL},
    {"NOFIRE RETURN",    A_NoFireReturn, NULL},
    {"KICK",             A_WeaponKick, DDF_StateGetFloat},
    {"SETCROSS",         A_SetCrosshair, NULL},
    {"TARGET",           A_GotTarget, NULL},
    {"LIGHT0",           A_Light0, NULL},
    {"LIGHT1",           A_Light1, NULL},
    {"LIGHT2",           A_Light2, NULL},
    {"CHECKRELOAD",      A_CheckReload, NULL},
    {"FLASH",            A_GunFlash, NULL},
    {"SOUND1",           A_SFXWeapon1, NULL},
    {"SOUND2",           A_SFXWeapon2, NULL},
    {"SOUND3",           A_SFXWeapon3, NULL},
    {"SEC SHOOT",        A_WeaponShootSA, DDF_StateGetAttack},
    {"SEC REFIRE",       A_ReFireSA, NULL},
    {"SEC NOFIRE",       A_NoFireSA, NULL},
    {"SEC FLASH",        A_GunFlashSA, NULL},
    {"SEC CHECKRELOAD",  A_CheckReloadSA, NULL},

    {NULL, NULL, NULL}
};

const specflags_t ammo_types[] =
{
    {"NOAMMO",  AM_NoAmmo, 0},
    {"BULLETS", AM_Bullet, 0},
    {"SHELLS",  AM_Shell,  0},
    {"ROCKETS", AM_Rocket, 0},
    {"CELLS",   AM_Cell,   0},

    // B.C.
    {"AMMO TYPE1", AM_Bullet, 0},
    {"AMMO TYPE2", AM_Shell,  0},
    {"AMMO TYPE3", AM_Rocket, 0},
    {"AMMO TYPE4", AM_Cell,   0},

    {NULL, 0, 0}
};

static boolean_t WeaponTryParseState(const char *field, 
    const char *contents, int index, boolean_t is_last)
{
  int i;
  const state_starter_t *starter;

  for (i=0; weapon_starters[i].name; i++)
    if (DDF_CompareName(weapon_starters[i].name, field) == 0)
      break;
  
  if (! weapon_starters[i].name)
    return false;
  
  starter = &weapon_starters[i];

  DDF_StateReadState(contents, starter->label,
    &starter->first_and_last[0], &starter->first_and_last[1],
    starter->state_num, is_last ? starter->last_redir : NULL, 
    weapon_actions);

  return true;
}


//
//  DDF PARSE ROUTINES
//

static boolean_t WeaponStartEntry(const char *name)
{
  int i;
  boolean_t replaces = false;

  if (name && name[0])
  {
    for (i=0; i < numweapons; i++)
    {
      if (DDF_CompareName(weaponinfo[i]->ddf.name, name) == 0)
      {
        dynamic_weapon = weaponinfo[i];
        replaces = true;
        break;
      }
    }
  }

  // not found, create a new one
  if (! replaces)
  {
    Z_SetArraySize(&weaponinfo_a, ++numweapons);
 
    dynamic_weapon = weaponinfo[numweapons - 1];
    dynamic_weapon->ddf.name = (name && name[0]) ? Z_StrDup(name) :
        DDF_MainCreateUniqueName("UNNAMED_WEAPON", numweapons);
  }

  dynamic_weapon->ddf.number = 0;

  return replaces;
}

static void WeaponCopyOld(void)
{
  buffer_weapon = dynamic_weapon[0];
}

static void WeaponInstantiate(const char *template_name)
{
  int i;

  if (! template_name)
  {
    // use built-in template
    buffer_weapon = template_weapon;
    return;
  }

  for (i=0; i < numweapons; i++)
  {
    // ignore current entry
    if (weaponinfo[i] == dynamic_weapon)
      continue;
    
    if (DDF_CompareName(weaponinfo[i]->ddf.name, template_name) == 0)
    {
      buffer_weapon = weaponinfo[i][0];
      return;
    }
  }

  I_Error("Template `%s' not found for weapons.ddf\n", template_name);
}

static void WeaponParseField(const char *field, const char *contents,
    int index, boolean_t is_last)
{
  I_WriteDebug("WEAPON_PARSE: [%s] [%s]\n", field, contents);

  if (DDF_MainParseField(weapon_commands, field, contents))
    return;
  
  if (WeaponTryParseState(field, contents, index, is_last))
    return;

  DDF_Error("Unknown weapons.ddf command: %s\n", field);
}

static void WeaponFinishEntry(void)
{
  ddf_base_t base;
  
  if (! buffer_weapon.first_state)
    DDF_Error("Weapon `%s' has missing states.\n",
        dynamic_weapon->ddf.name);

  DDF_StateFinishStates(buffer_weapon.first_state, buffer_weapon.last_state);

  // check stuff...

  if (buffer_weapon.ammopershot < 0)
    DDF_Error("Bad AMMOPERSHOT value for weapon: %d\n",
      buffer_weapon.ammopershot);
  
  // transfer static entry to dynamic entry
  
  base = dynamic_weapon->ddf;
  dynamic_weapon[0] = buffer_weapon;
  dynamic_weapon->ddf = base;

  // compute CRC...
  CRC32_Init(&dynamic_weapon->ddf.crc);

  CRC32_ProcessInt(&dynamic_weapon->ddf.crc, dynamic_weapon->ammo);
  CRC32_ProcessInt(&dynamic_weapon->ddf.crc, dynamic_weapon->ammopershot);

  // FIXME: do more stuff...

  CRC32_Done(&dynamic_weapon->ddf.crc);
}


//
// DDF_ReadWeapons
//
void DDF_ReadWeapons(void *data, int size)
{
  readinfo_t weapons;

  weapons.memfile = data;
  weapons.memsize = size;
  weapons.tag = "WEAPONS";
  weapons.entries_per_dot = 1;

  if (weapons.memfile)
  {
    weapons.message = NULL;
    weapons.filename = NULL;
    weapons.lumpname = "DDFWEAP";
  }
  else
  {
    weapons.message = "DDF_InitWeapons";
    weapons.filename = "weapons.ddf";
    weapons.lumpname = NULL;
  }

  weapons.start_entry  = WeaponStartEntry;
  weapons.copy_old     = WeaponCopyOld;
  weapons.instantiate  = WeaponInstantiate;
  weapons.parse_field  = WeaponParseField;
  weapons.finish_entry = WeaponFinishEntry;

  DDF_MainReadFile(&weapons);
}

void DDF_WeaponInit(void)
{
  Z_InitStackArray(&weaponinfo_a, (void ***)&weaponinfo, sizeof(weaponinfo_t), 0);
}

void DDF_WeaponCleanUp(void)
{
  // compute the weaponkey array

  int key, i;

  memset(&weaponkey, 0, sizeof(weaponkey));

  for (key=0; key < 10; key++)
  {
    weaponkey_t *w = &weaponkey[key];
    
    for (i=0; i < numweapons; i++)
    {
      if (weaponinfo[i]->bind_key != key)
        continue;
      
      Z_Resize(w->choices, int, ++w->numchoices);
      w->choices[w->numchoices-1] = i;
    }
  }
}

static void DDF_WGetAmmo(const char *info, void *storage, void *priv)
{
  int *ammo = (int *)storage;
  int flag_value;

  switch (DDF_MainCheckSpecialFlag(info, ammo_types, &flag_value,
      false, false))
  {
    case CHKF_Positive:
    case CHKF_Negative:
      (*ammo) = flag_value;
      break;
    
    case CHKF_User:
    case CHKF_Unknown:
        DDF_Error("Unknown Ammo type '%s'\n", info);
      break;
  }
}

int DDF_WeaponLookup(const char *name)
{
  int i;

  for (i = 0; i < numweapons; i++)
  {
    if (DDF_CompareName(weaponinfo[i]->ddf.name, name) == 0)
      return i;
  }

  return -1;
}

static void DDF_WGetUpgrade(const char *info, void *storage, void *priv)
{
  int *dest = (int *)storage;

  // this is not ideal, but will do for now
  *dest = DDF_WeaponLookup(info);

  if (*dest < 0)
    DDF_Error("Unknown weapon to upgrade: %s\n", info);
}

static specflags_t weapon_specials[] =
{
    {"SILENT TO MONSTERS", WPSP_SilentToMonsters, 0},
    {NULL, 0, 0}
};

//
// DDF_WGetSpecialFlags
//
static void DDF_WGetSpecialFlags(const char *info, void *storage, void *priv)
{
  int flag_value;

  switch (DDF_MainCheckSpecialFlag(info, weapon_specials, &flag_value,
      true, false))
  {
    case CHKF_Positive:
      buffer_weapon.special_flags |= flag_value;
      break;
    
    case CHKF_Negative:
      buffer_weapon.special_flags &= ~flag_value;
      break;

    case CHKF_User:
    case CHKF_Unknown:
      DDF_Error("DDF_WGetSpecialFlags: Unknown Special: %s", info);
      break;
  }
}
