//----------------------------------------------------------------------------
// EDGE Finale Code on Game Completion
//----------------------------------------------------------------------------
// 
//  Copyright (c) 1999-2000  The EDGE Team.
// 
//  This program is free software; you can redistribute it and/or
//  modify it under the terms of the GNU General Public License
//  as published by the Free Software Foundation; either version 2
//  of the License, or (at your option) any later version.
//
//  This program is distributed in the hope that it will be useful,
//  but WITHOUT ANY WARRANTY; without even the implied warranty of
//  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//  GNU General Public License for more details.
//
//----------------------------------------------------------------------------
//
//  Based on the DOOM source code, released by Id Software under the
//  following copyright:
//
//    Copyright (C) 1993-1996 by id Software, Inc.
//
//----------------------------------------------------------------------------
//
// -KM- 1998/07/21 Clear the background behind those end pics.
// -KM- 1998/09/27 sounds.ddf stuff: seesound -> DDF_LookupSound(seesound)
// -KM- 1998/11/25 Finale generalised.
//

#include "i_defs.h"
#include "f_finale.h"

#include "ddf_main.h"
#include "dm_state.h"
#include "dstrings.h"
#include "hu_stuff.h"
#include "m_random.h"
#include "m_swap.h"
#include "p_action.h"
#include "r_state.h"
#include "s_sound.h"
#include "v_res.h"
#include "z_zone.h"
#include "w_wad.h"

typedef enum
{
  f_text,
  f_pic,
  f_bunny,
  f_cast,
  f_end
}
finalestage_t;

// Stage of animation:
//  0 = text, 1 = art screen, 2 = character cast
static finalestage_t finalestage;

// -ES- 2000/03/11 skip to the next finale stage
static boolean_t skip_finale;
static int finalecount;
static int picnum;

#define TEXTSPEED 3
#define TEXTWAIT  250

#define e1text E1TEXT
#define e2text E2TEXT
#define e3text E3TEXT
#define e4text E4TEXT

#define c1text C1TEXT
#define c2text C2TEXT
#define c3text C3TEXT
#define c4text C4TEXT
#define c5text C5TEXT
#define c6text C6TEXT

#define p1text P1TEXT
#define p2text P2TEXT
#define p3text P3TEXT
#define p4text P4TEXT
#define p5text P5TEXT
#define p6text P6TEXT

#define t1text T1TEXT
#define t2text T2TEXT
#define t3text T3TEXT
#define t4text T4TEXT
#define t5text T5TEXT
#define t6text T6TEXT

static const char *finaletext;

static gameaction_t newgameaction;
static const finale_t *finale;
static void StartCast(void);
static void CastTicker(void);
static boolean_t CastResponder(event_t * ev);

//
// F_StartFinale
//
void F_StartFinale(const finale_t * f, gameaction_t newaction)
{
  player_t *p;

  finalestage = f_text;
  finalecount = 0;
  gameaction = ga_nothing;
  viewactive = false;
  automapactive = false;
  finale = f;
  newgameaction = newaction;
  picnum = 0;
  for (p = players; p; p = p->next)
    p->cmd.buttons = 0;
  F_Ticker();
}

boolean_t F_Responder(event_t * event)
{
  int i;

  if (finalestage == f_cast)
    return CastResponder(event);

  if (event->type != ev_keydown)
    return false;
    
  // -ES- 2000/02/27 Hack: The first parts of the final stage may be
  // accelerated, but not the last one, so we have to check if there are
  // any more to do.
  for (i = f_end-1; i > finalestage; i--)
  {
    if ((i == f_cast && finale->docast) || (i == f_bunny && finale->dobunny)
       || (i == f_pic && finale->pics) || (i == f_text && finale->text))
      break;
  }

  // Skip finale if there either is a next final stage, or if there is a next
  // sub-stage (for the two-stage text acceleration)
  if (i > finalestage ||
      (finalestage == f_text &&
       (unsigned int)finalecount < (TEXTSPEED * strlen(DDF_LanguageLookup(finale->text)))))
  {
    skip_finale = true;
    return true;
  }

  return false;
}

//
// F_Ticker
//
void F_Ticker(void)
{
  int fstage = finalestage;

  if (skip_finale)
  {
    skip_finale = false;
    if (finalestage == f_text && 
        (unsigned int)finalecount < TEXTSPEED * strlen(DDF_LanguageLookup(finale->text)))
    {
      // -ES- 2000/03/08 Two-stage text acceleration. Complete the text the
      // first time, skip to next finale the second time.
      finalecount = TEXTSPEED * strlen(DDF_LanguageLookup(finale->text));
    }
    else
    {
      finalestage++;
      finalecount = 0;
    }
  }

  switch (finalestage)
  {
    case f_text:
      if (finale->text)
      {
        gamestate = GS_FINALE;
        if (!finalecount)
        {
          finaletext = DDF_LanguageLookup(finale->text);
          S_ChangeMusic(finale->music, true);
          wipegamestate = -1;
          break;
        }
        else if ((unsigned int)finalecount > strlen(finaletext) * TEXTSPEED + TEXTWAIT)
        {
          finalecount = 0;
        }
        else
          break;
      }
      finalestage++;
    case f_pic:
      if (finale->pics)
      {
        gamestate = GS_FINALE;
        if ((unsigned int)finalecount > finale->picwait)
        {
          finalecount = 0;
          picnum++;
        }

        if ((unsigned int)picnum >= finale->numpics)
        {
          finalecount = 0;
          picnum = 0;
        }
        else
          break;
      }
      finalestage++;
    case f_bunny:
      if (finale->dobunny)
      {
        gamestate = GS_FINALE;
        if (!finalecount)
          // force a wipe
          wipegamestate = -1;

        break;
      }
      finalestage++;
    case f_cast:
      if (finale->docast)
      {
        gamestate = GS_FINALE;
        if (!finalecount)
          StartCast();
        else
          CastTicker();
        break;
      }
      finalestage++;
    case f_end:
      if (newgameaction != ga_nothing)
        gameaction = newgameaction;
      else
        finalestage = fstage;
      break;
  }

  if (finalestage != (finalestage_t)fstage && finalestage != f_end)
    wipegamestate = -1;

  // advance animation
  finalecount++;

}

//
// TextWrite
//

static void TextWrite(void)
{
  int w;
  int count;
  const char *ch;
  int c;
  int cx;
  int cy;

  // 98-7-10 KM erase the entire screen to a tiled background
  if (finale->text_flat[0])
    V_TextureBackScreen(main_scr, finale->text_flat, 0, 0, SCREENWIDTH, SCREENHEIGHT);
  else
  {
    V_ClearPageBackground(main_scr);
    V_DrawPatchInDirectName(main_scr, 0, 0, finale->text_back);
  }

  V_MarkRect(0, 0, SCREENWIDTH, SCREENHEIGHT);

  // draw some of the text onto the screen
  cx = 10;
  cy = 10;
  ch = finaletext;

  count = (finalecount - 10) / finale->text_speed;
  if (count < 0)
    count = 0;
  for (; count; count--)
  {
    c = *ch++;
    if (!c)
      break;
    if (c == '\n')
    {
      cx = 10;
      cy += 11;
      continue;
    }

    c = toupper(c) - HU_FONTSTART;
    if (c < 0 || c > HU_FONTSIZE)
    {
      cx += 4;
      continue;
    }

    w = SHORT(hu_font[c]->width);
    if (cx + w > SCREENWIDTH)
      break;
    V_DrawPatchInDirect(main_scr, cx, cy, hu_font[c]);
    cx += w;
  }

}

//
// Final DOOM 2 animation
// Casting by id Software.
//   in order of appearance
//

static const mobjinfo_t *castorder;
static int casttics;
static state_t *caststate;
static boolean_t castdeath;
static int castframes;
static int castonmelee;
static boolean_t castattacking;

//
// StartCast
//
static void StartCast(void)
{
  wipegamestate = -1;  // force a screen wipe

  castorder = DDF_MobjLookupCast(2);
  caststate = &states[castorder->chase_state];
  casttics = caststate->tics;
  castdeath = false;
  castframes = 0;
  castonmelee = 0;
  castattacking = false;
  //S_ChangeMusic("d_evil", true);
}

//
// CastTicker
//
// -KM- 1998/10/29 Use sfx_t.
//      Known bug: Chaingun/Spiderdemon's sounds aren't stopped.
//
static int shotsfxchannel = -1;

static void CastTicker(void)
{
  int st;
  sfx_t *sfx = NULL;
  boolean_t wantchannel = false;

  if (--casttics > 0)
    return;  // not time to change state yet

  if (caststate->tics == -1 || caststate->nextstate == S_NULL)
  {
    // switch from deathstate to next monster
    castdeath = false;
    castorder = DDF_MobjLookupCast(castorder->castorder + 1);

    if (castorder->seesound)
      S_StartSound(NULL, castorder->seesound);

    caststate = &states[castorder->chase_state];
    castframes = 0;
  }
  else
  {
    // just advance to next state in animation
    st = caststate->nextstate;
    caststate = &states[st];
    castframes++;
    casttics = caststate->tics;

    // Yuk, handles sounds
    if (caststate->action == P_ActMakeCloseAttemptSound)
    {
      sfx = castorder->closecombat->initsound;
    }
    else if (caststate->action == P_ActMakeRangeAttemptSound)
    {
      sfx = castorder->rangeattack->initsound;
    }
    else if (caststate->action == P_ActRangeAttack)
    {
      if (castorder->rangeattack->attackstyle == ATK_SHOT)
      {
        sfx = castorder->rangeattack->sound;
        wantchannel = true;
        if (sfx && shotsfxchannel >= 0)
          S_StopChannel(shotsfxchannel);
      }
      else if (castorder->rangeattack->attackstyle == ATK_SKULLFLY)
      {
        sfx = castorder->rangeattack->initsound;
      }
      else if (castorder->rangeattack->attackstyle == ATK_SPAWNER)
      {
        sfx = castorder->rangeattack->spawnedobj->rangeattack->initsound;
      }
      else if (castorder->rangeattack->attackstyle == ATK_TRACKER)
      {
        sfx = castorder->rangeattack->initsound;
      }
      else
      {
        sfx = castorder->rangeattack->atk_mobj->seesound;
      }
    }
    else if (caststate->action == P_ActMeleeAttack)
    {
      sfx = castorder->closecombat->sound;
    }
    else if (castorder->activesound && !M_Random() && !castdeath)
    {
      sfx = castorder->activesound;
    }
    else if (caststate->action == P_ActWalkSoundChase)
    {
      sfx = castorder->walksound;
    }
    else if (caststate->action == P_ActComboAttack)
    {
      if (castonmelee)
      {
        sfx = castorder->closecombat->sound;
      }
      else
      {
        if (castorder->rangeattack->attackstyle == ATK_SHOT)
        {
          sfx = castorder->rangeattack->sound;
          wantchannel = true;
          if (sfx && shotsfxchannel >= 0)
            S_StopChannel(shotsfxchannel);
        }
        else if (castorder->rangeattack->attackstyle == ATK_SKULLFLY)
          sfx = castorder->rangeattack->initsound;
        else if (castorder->rangeattack->attackstyle == ATK_SPAWNER)
          sfx = castorder->rangeattack->atk_mobj->rangeattack->initsound;
        else if (castorder->rangeattack->attackstyle == ATK_TRACKER)
          sfx = castorder->rangeattack->initsound;
        else
          sfx = castorder->rangeattack->atk_mobj->seesound;
      }
    }
  }

  if (castframes == 24)
  {
    // go into attack frame
    castattacking = true;
    castonmelee ^= 1;
    if (castonmelee)
      caststate = &states[castorder->melee_state];
    else
      caststate = &states[castorder->missile_state];
    if (caststate == &states[S_NULL])
    {
      castonmelee ^= 1;
      if (castonmelee)
        caststate =
            &states[castorder->melee_state];
      else
        caststate =
            &states[castorder->missile_state];
    }
    sfx = castorder->attacksound;
  }

  if (castattacking)
  {
    if (castframes == 36
        || caststate == &states[castorder->chase_state])
    {
      castattacking = false;
      castframes = 0;
      caststate = &states[castorder->chase_state];
    }
  }

  casttics = caststate->tics;
  if (casttics == -1)
    casttics = 15;

  if (sfx)
  {
    if (wantchannel)
      shotsfxchannel = S_StartSound(NULL, sfx);
    else
      S_StartSound(NULL, sfx);
  }
}

//
// CastResponder
//
static boolean_t CastResponder(event_t * ev)
{
  if (ev->type != ev_keydown)
    return false;

  if (castdeath)
    return true;  // already in dying frames

  // go into death frame
  castdeath = true;
  caststate = &states[castorder->death_state];
  casttics = caststate->tics;
  castframes = 0;
  castattacking = false;
  if (castorder->deathsound)
    S_StartSound(NULL, castorder->deathsound);

  return true;
}

static void CastPrint(char *text)
{
  char *ch;
  int c;
  int cx;
  int w;
  int width;

  // find width
  ch = text;
  width = 0;

  while (ch)
  {
    c = *ch++;
    if (!c)
      break;
    c = toupper(c) - HU_FONTSTART;
    if (c < 0 || c > HU_FONTSIZE)
    {
      width += 4;
      continue;
    }

    w = SHORT(hu_font[c]->width);
    width += w;
  }

  // draw it
  cx = 160 - width / 2;
  ch = text;
  while (ch)
  {
    c = *ch++;
    if (!c)
      break;
    c = toupper(c) - HU_FONTSTART;
    if (c < 0 || c > HU_FONTSIZE)
    {
      cx += 4;
      continue;
    }

    w = SHORT(hu_font[c]->width);
    V_DrawPatchInDirect(main_scr, cx, 180, hu_font[c]);
    cx += w;
  }

}

//
// CastDrawer
//
static void CastDrawer(void)
{
  spritedef_t *sprdef;
  spriteframe_t *sprframe;
  int lump;
  boolean_t flip;
  const patch_t *patch;

  // erase the entire screen to a background
  // -KM- 1998/07/21 Clear around the pic too.
  V_ClearPageBackground(main_scr);
  V_DrawPatchInDirectName(main_scr, 0, 0, "BOSSBACK");

  CastPrint(castorder->ddf.name);

  // draw the current frame in the middle of the screen
  sprdef = &sprites[caststate->sprite];
  sprframe = &sprdef->spriteframes[caststate->frame & FF_FRAMEMASK];
  lump = sprframe->lump[0];
  flip = (boolean_t) sprframe->flip[0];

  patch = W_CacheLumpNum(spritelist[lump]);
  if (flip)
    V_DrawPatchInDirectFlipped(main_scr, 160, 170, patch);
  else
    V_DrawPatchInDirect(main_scr, 160, 170, patch);
  W_DoneWithLump(patch);
}

//
// F_DrawPatchCol
//
// -KM- 1998/07/31 Made our bunny friend take up more of the screen.
static void DrawPatchCol(int x, const patch_t * patch, int col, fixed_t fracstep)
{
// -AJA- 1999/07/05: FIXME. Temporarily disabled, since this should be
  // done by code in v_video1/2.c.  Perhaps there is code which already
  // does this ?  If not, we'll move this code there...
#if 0
  const column_t *column;
  byte *source;
  int count;
  fixed_t frac;

  column = (column_t *) ((byte *) patch + LONG(patch->columnofs[col]));

  if (BPP == 1)
  {
    byte *dest;
    byte *desttop;

    desttop = screens[0] + x + (SCREENWIDTH - SCALEDWIDTH) / 2 +
        SCREENDEPTH * ((SCREENHEIGHT - SCALEDHEIGHT) / 2);
    // step through the posts in a column
    while (column->topdelta != 0xff)
    {
      source = (byte *) column + 3;
      dest = desttop + column->topdelta * SCREENDEPTH;
      count = (column->length << FRACBITS) / fracstep;
      frac = 0;
      while (count--)
      {
        *dest = source[frac >> FRACBITS];
        dest += SCREENDEPTH;
        frac += fracstep;
      }
      column = (column_t *) ((byte *) column + column->length + 4);
    }
  }
  else
  {
    short *dest;
    short *desttop;

    desttop = (short *)(screens[0] + 2 * (x + (SCREENWIDTH - SCALEDWIDTH) / 2 +
            SCREENDEPTH / 2 * ((SCREENHEIGHT - SCALEDHEIGHT) / 2)));
    // step through the posts in a column
    while (column->topdelta != 0xff)
    {
      source = (byte *) column + 3;
      dest = desttop + column->topdelta * SCREENDEPTH / 2;
      count = (column->length << FRACBITS) / fracstep;
      frac = 0;
      while (count--)
      {
        *dest = palette_color[source[frac >> FRACBITS]];
        dest += SCREENDEPTH / 2;
        frac += fracstep;
      }
      column = (column_t *) ((byte *) column + column->length + 4);
    }
  }
#endif
}

//
// BunnyScroll
//
// -KM- 1998/07/31 Made our bunny friend take up more screen space.
// -KM- 1998/12/16 Removed fading routine.
static void BunnyScroll(void)
{
  int scrolled;
  int x;
  const patch_t *p1;
  const patch_t *p2;
  char name[10];
  int stage;
  static int laststage;
  const fixed_t yscale = (200 << FRACBITS) / SCALEDHEIGHT;
  const fixed_t xscale = (320 << FRACBITS) / SCALEDWIDTH;
  fixed_t frac = 0;

  p1 = W_CacheLumpName("PFUB2");
  p2 = W_CacheLumpName("PFUB1");

  V_MarkRect(0, 0, SCREENWIDTH, SCREENHEIGHT);

  scrolled = 320 - (finalecount - 230) / 2;
  if (scrolled > 320)
    scrolled = 320;
  if (scrolled < 0)
    scrolled = 0;
  // 23-6-1998 KM Changed the background colour to black not real dark grey
  // -KM- 1998/07/21  Replaced SCREENWIDTH*BPP with SCREENDEPTH (two places)
  V_ClearPageBackground(main_scr);
  for (x = 0; x < 320; x++)
  {
    do
    {
      if (x + scrolled < 320)
        DrawPatchCol(frac / xscale, p1, x + scrolled, yscale);
      else
        DrawPatchCol(frac / xscale, p2, x + scrolled - 320, yscale);
      frac += xscale;
    }
    while ((frac >> FRACBITS) <= x);
  }

  W_DoneWithLump(p1);
  W_DoneWithLump(p2);

  if (finalecount < 1130)
    return;
  if (finalecount < 1180)
  {
    V_DrawPatchInDirectName(main_scr, (320 - 13 * 8) / 2,
        (200 - 8 * 8) / 2, "END0");
    laststage = 0;
    return;
  }

  stage = (finalecount - 1180) / 5;
  if (stage > 6)
    stage = 6;
  if (stage > laststage)
  {
    S_StartSound(NULL, sfx_pistol);
    laststage = stage;
  }

  sprintf(name, "END%i", stage);
  V_DrawPatchInDirectName(main_scr, (320 - 13 * 8) / 2, (200 - 8 * 8) / 2, name);
}

//
// F_Drawer
//
void F_Drawer(void)
{
  switch (finalestage)
  {
    case f_text:
      TextWrite();
      break;
    case f_pic:
      // -KM- 1998/07/21 Clear around the pic too
      V_ClearPageBackground(main_scr);
      V_DrawPatchInDirectName(main_scr, 0, 0, &finale->pics[picnum * 9]);
      break;
    case f_bunny:
      BunnyScroll();
      break;
    case f_cast:
      CastDrawer();
      break;
    case f_end:
      break;
  }
}
