//----------------------------------------------------------------------------
//  EDGE Linux GGI Video Code
//----------------------------------------------------------------------------
// 
//  Copyright (c) 1999-2000  The EDGE Team.
// 
//  This program is free software; you can redistribute it and/or
//  modify it under the terms of the GNU General Public License
//  as published by the Free Software Foundation; either version 2
//  of the License, or (at your option) any later version.
//
//  This program is distributed in the hope that it will be useful,
//  but WITHOUT ANY WARRANTY; without even the implied warranty of
//  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//  GNU General Public License for more details.
//
//----------------------------------------------------------------------------

#include "i_defs.h"
#include "./i_trans.h"

#include "version.h"
#include "dm_state.h"
#include "dm_defs.h"
#include "dm_type.h"
#include "m_argv.h"
#include "st_stuff.h"
#include "v_video1.h"
#include "v_video2.h"
#include "v_res.h"
#include "v_colour.h"
#include "v_screen.h"

#include "e_main.h"
#include "e_event.h"
#include "z_zone.h"
#include "w_wad.h"

#include "s_sound.h"


#include <stdlib.h>
#include <string.h>
#include <unistd.h>
#include <sys/ipc.h>
#include <sys/shm.h>

#include <stdarg.h>
#include <sys/time.h>
#include <sys/types.h>
#include <sys/socket.h>

#include <netinet/in.h>
#include <signal.h>

#include <ggi/ggi.h>

#ifdef USE_WMH
#include <ggi/wmh.h>
#endif

short hicolortransmask1, hicolortransmask2;

// Dummy screen... 
// mainscreen is a full-size subscreen of this one.
static screen_t dummy_screen;

unsigned char *thepalette;

static int graphics_shutdown = 0;

static ggi_visual_t my_vis;
static ggi_mode my_mode;

static char *my_targetname = NULL;
static char *my_modename = "";

static const ggi_directbuffer *my_dbuf;
static const ggi_pixelformat *my_pixfmt;

static ggi_color my_pal[256];

int leds_always_off = 0;  // Expected by m_misc, not relevant
colourshift_t redshift, greenshift, blueshift;

// Possible Screen Modes
typedef struct
{
  int width;
  int height;
  int bpp;
}
possresmode_t;

static possresmode_t possresmode[] =
{
  { 320, 200, 8},
  { 320, 240, 8},
  { 400, 300, 8},
  { 512, 384, 8},
  { 640, 400, 8},
  { 640, 480, 8},
  { 800, 600, 8},
  {1024, 768, 8},
  {1280,1024, 8},

  { 320, 200, 16},
  { 320, 240, 16},
  { 400, 300, 16},
  { 512, 384, 16},
  { 640, 400, 16},
  { 640, 480, 16},
  { 800, 600, 16},
  {1024, 768, 16},
  {1280,1024, 16},

  {  -1,  -1, -1}
};


// ====================== INTERNALS ======================

//
// MakeCol
//
static inline unsigned long MakeCol(long r, long g, long b)
{
  r = (r >> (8 - redshift.bits))   << redshift.shift;
  g = (g >> (8 - greenshift.bits)) << greenshift.shift;
  b = (b >> (8 - blueshift.bits))  << blueshift.shift;

  return r|g|b;
}

//
// CountLeadingZeros
//
static const inline unsigned long CountLeadingZeros(unsigned long x)
{
  unsigned long r = 0;

  if (x == 0)
    return 0;

  while (!(x & 1))
  {
    x >>= 1;
    r++;
  }

  return r;
}

//
// CountBitsInMask
//
static const inline unsigned long CountBitsInMask(unsigned long x)
{
  unsigned long r = 0;

  while (!(x & 1))
    x >>= 1;

  while (x & 1)
  {
    x >>= 1;
    r++;
  }

  if (x)
    I_Error ("X Version: Mask is not contiguous.\n");

  return r;
}

//
// SetColourShift
//
// Constructs the shifts necessary to set one element of an RGB colour value
//
void SetColourShift(unsigned long mask, colourshift_t * ps)
{
  ps->mask  = mask;
  ps->shift = CountLeadingZeros(mask);
  ps->bits  = CountBitsInMask(mask);
}

//
// BlitToScreen
//
static inline void BlitToScreen(void)
{
  register int y;
  register unsigned char *s1, *d1;
  register unsigned int stride = my_dbuf->buffer.plb.stride;

  s1 = (void *) dummy_screen.data;
  d1 = (void *) my_dbuf->write;

  for (y=0; y < SCREENHEIGHT; y++)
  {
    memcpy(d1, s1, SCREENWIDTH * BPP);
    s1 += SCREENDEPTH;
    d1 += stride;
  }
}

//
// DetectAvailableModes
//
static void DetectAvailableModes(void)
{
  int i;
  int resadded = 0;

  // -ACB- 2000/03/16 Detect Possible Resolutions
  for (i=0; possresmode[i].width != -1; i++)
  {
    ggi_mode template_mode;
    ggi_mode test_mode;

    ggiParseMode("", &template_mode);

    GT_SETSIZE(template_mode.graphtype, possresmode[i].bpp);

    template_mode.visible.x = possresmode[i].width;
    template_mode.visible.y = possresmode[i].height;

    test_mode = template_mode;
    test_mode.virt.x = V_GetDepth(possresmode[i].width, 
        possresmode[i].bpp/8) / (possresmode[i].bpp/8);

    if (ggiCheckMode(my_vis, &test_mode) >= 0)
    {
      V_AddAvailableResolution(possresmode[i].width, possresmode[i].height, 
          possresmode[i].bpp);
      resadded++;
      continue;
    }

    // -AJA- We don't REQUIRE the computed virt.x to be available,
    //       that is just the optimal value to prevent cache thrashing
    //       in the column drawers.  So test again, allowing virt.x to
    //       be decided by the hardware.

    test_mode = template_mode;
    
    if (ggiCheckMode(my_vis, &test_mode) >= 0)
    {
      V_AddAvailableResolution(possresmode[i].width, possresmode[i].height, 
          possresmode[i].bpp);
      resadded++;
      continue;
    }
  }

  if (resadded == 0)
    I_Error("I_StartupGraphics: Find NO supported resolutions !\n");
}

// =================== END OF INTERNALS ===================

//
// I_StartupGraphics
//
void I_StartupGraphics(void)
{
  const char *val_str;

  // check for command-line options
  val_str = M_GetParm("-ggitarget");
  if (val_str)
    my_targetname = strdup(val_str);
  else if (getenv("EDGE_GGI_TARGET"))
    my_targetname = getenv("EDGE_GGI_TARGET");

  val_str = M_GetParm("-ggimode");
  if (val_str)
    my_modename = strdup(val_str);
  else if (getenv("EDGE_GGI_MODE"))
    my_modename = getenv("EDGE_GGI_MODE");

  if (ggiInit() < 0)
    I_Error("Couldn't init LibGGI !\n");

  my_vis = ggiOpen(my_targetname, NULL);

  if (my_vis == NULL)
    I_Error("Couldn't open LibGGI target !\n");

  ggiSetFlags(my_vis, GGIFLAG_ASYNC);

  DetectAvailableModes();

#ifdef USE_WMH
  if (ggiWmhInit() < 0)
    I_Error("Couldn't init LibWMH !\n");
#endif

  //!!! get round LibGGI's signal handlers
  signal(SIGSEGV, SIG_DFL);
}


//
// I_GetTruecolInfo
//
void I_GetTruecolInfo(truecol_info_t * info)
{
  info->red_bits = redshift.bits;
  info->red_shift = redshift.shift;
  info->red_mask = redshift.mask;

  info->green_bits = greenshift.bits;
  info->green_shift = greenshift.shift;
  info->green_mask = greenshift.mask;

  info->blue_bits = blueshift.bits;
  info->blue_shift = blueshift.shift;
  info->blue_mask = blueshift.mask;
}

//
// I_SetScreenSize
//
boolean_t I_SetScreenSize(int width, int height, int bpp)
{
  char title[80];

  ggiParseMode("", &my_mode);

  GT_SETSIZE(my_mode.graphtype, BPP * 8);

  my_mode.visible.x = SCREENWIDTH;
  my_mode.visible.y = SCREENHEIGHT;
  my_mode.virt.x = V_GetDepth(SCREENWIDTH, BPP) / BPP;

  if (ggiSetMode(my_vis, &my_mode) < 0)
  {
    I_Warning("Requested mode not available !\n");

    if (ggiSetMode(my_vis, &my_mode) < 0)
    {
      I_Warning("ARGH ! Mode still not available (bug in LibGGI)\n");

      ggiParseMode("", &my_mode);

      if (ggiSetMode(my_vis, &my_mode) < 0)
        I_Error("Sweet Jesus !! LibGGI fucked beyond all recognition\n");
    }
  }

  BPP = ((my_mode.graphtype & 0x7F) + 7) / 8;  //!!! FIXME
  SCREENWIDTH = my_mode.visible.x;
  SCREENHEIGHT = my_mode.visible.y;
  SCREENDEPTH = V_GetDepth(my_mode.virt.x, BPP);

  my_pixfmt = ggiGetPixelFormat(my_vis);

  // read directbuffer info

  if (! (my_dbuf = ggiDBGetBuffer(my_vis, 0)))
    I_Error("No DirectBuffer available.\n");

  if (! (my_dbuf->type & GGI_DB_SIMPLE_PLB)) 
    I_Error("Error: non-standard display buffer\n");

  if (BPP != 1)
  {
    SetColourShift(my_pixfmt->red_mask,   &redshift);
    SetColourShift(my_pixfmt->green_mask, &greenshift);
    SetColourShift(my_pixfmt->blue_mask,  &blueshift);
  }

  //!!! get round LibGGI's signal handlers
  signal(SIGSEGV, SIG_DFL);

  dummy_screen.width = SCREENWIDTH;
  dummy_screen.height = SCREENHEIGHT;
  dummy_screen.depth = SCREENDEPTH;
  dummy_screen.bpp = BPP;
  dummy_screen.parent = NULL;
  dummy_screen.data = malloc(SCREENDEPTH * SCREENHEIGHT);

  main_scr = V_CreateSubScreen (&dummy_screen, 0, 0, SCREENWIDTH, SCREENHEIGHT);

#ifdef USE_WMH
  if (ggiWmhAttach(my_vis) < 0)
    I_Error("Unable to attach WMH extension to LibGGI visual !\n");

  sprintf(title, "EDGE v%x.%02x", EDGEVER / EDGEVERFIX,
      EDGEVER % EDGEVERFIX);

  ggiWmhSetTitle(my_vis, title);
#endif

  return true;
}

//
// I_StartFrame
//
void I_StartFrame(void)
{
  // Acquire DirectBuffer before we use it.
  if (ggiResourceAcquire(my_dbuf->resource, GGI_ACTYPE_WRITE) != 0)
    I_Error("Error acquiring DirectBuffer\n");
}

//
// I_FinishFrame
//
void I_FinishFrame(void)
{
  BlitToScreen();

  // Release DirectBuffer when done with it.
  ggiResourceRelease(my_dbuf->resource);

  ggiFlush(my_vis);
}

//
// I_SetPalette
//
void I_SetPalette(byte palette[256][3])
{
  int i;

  const byte *gtable = gammatable[usegamma];

  if (BPP != 1)
    return;

  for (i = 0; i < 256; i++)
  {
    int r = gtable[palette[i][0]];
    int g = gtable[palette[i][1]];
    int b = gtable[palette[i][2]];

    my_pal[i].r = (r << 8) + r;
    my_pal[i].g = (g << 8) + g;
    my_pal[i].b = (b << 8) + b;
  }

  ggiSetPalette(my_vis, 0, 256, my_pal);
}

//
// I_Colour2Pixel
//
long I_Colour2Pixel(byte palette[256][3], int col)
{
  if (BPP == 1)
    return col;

  return MakeCol(palette[col][0], palette[col][1], palette[col][2]);
}

//
// I_MakeColourmapRange
//
void I_MakeColourmapRange(void *dest_colmaps, byte palette[256][3], const byte *src_colmaps, int num)
{
  const byte *gtable = gammatable[usegamma];
  unsigned short *colmap16 = dest_colmaps;
  byte tempr[256], tempg[256], tempb[256];
  int i, j;

  if (BPP != 2)
    return;

  stbar_update = true;

  if (interpolate_colmaps && num >= 2)
  {
    for (j = 0; j < 256; j++)
    {
      int c1 = src_colmaps[0 * 256 + j];
      int c2 = src_colmaps[(num - 1) * 256 + j];

      int r1 = palette[c1][0];
      int g1 = palette[c1][1];
      int b1 = palette[c1][2];

      int r2 = palette[c2][0];
      int g2 = palette[c2][1];
      int b2 = palette[c2][2];

      for (i = 0; i < num; i++)
      {
        int r = gtable[r1 + (r2 - r1) * i / (num - 1)];
        int g = gtable[g1 + (g2 - g1) * i / (num - 1)];
        int b = gtable[b1 + (b2 - b1) * i / (num - 1)];

        colmap16[i * 256 + j] = MakeCol (r, g, b);
      }
    }

    return;
  }

  // use the old (but faster) method:

  for (i = 0; i < 256; i++)
  {
    tempr[i] = gtable[palette[i][0]];
    tempg[i] = gtable[palette[i][1]];
    tempb[i] = gtable[palette[i][2]];
  }

  for (i = 0; i < num; i++)
  for (j = 0; j < 256; j++)
  {
    int c = src_colmaps[i * 256 + j];
    colmap16[i * 256 + j] = MakeCol (tempr[c], tempg[c], tempb[c]);
  }
}

//
// I_ShutdownGraphics
//
void I_ShutdownGraphics(void)
{
  if (graphics_shutdown)
    return;

  graphics_shutdown = 1;

  fprintf(stderr, "Shutting down graphics...\n");

#ifdef USE_WMH
  ggiWmhDetach(my_vis);
  ggiWmhExit();
#endif

  ggiClose(my_vis);
  ggiExit();
}

//
//   INPUT CODE 
//
#define GGI_EVENTS  (emKey | emPointer)

//
// Translates a key from GGI -> EDGE
//
// Returns -1 if no suitable translation exists.

static int TranslateGGIKey(gii_event *ev)
{
  int label = ev->key.label;

  if (GII_KTYP(label) == GII_KT_PAD)
    label = ev->key.sym;

  switch (label) 
  {
    case GIIUC_Tab: return KEYD_TAB;
    case GIIUC_Linefeed:
    case GIIUC_Return: return KEYD_ENTER;
    case GIIUC_Escape: return KEYD_ESCAPE;
    case GIIUC_Delete:
    case GIIUC_BackSpace: return KEYD_BACKSPACE;

    case GIIK_Up:    return KEYD_UPARROW;
    case GIIK_Down:  return KEYD_DOWNARROW;
    case GIIK_Left:  return KEYD_LEFTARROW;
    case GIIK_Right: return KEYD_RIGHTARROW;

    case GIIK_Find:   return KEYD_HOME;
    case GIIK_Select: return KEYD_END;
    case GIIK_ScrollBack:
    case GIIK_PageUp:   return KEYD_PGUP;
    case GIIK_ScrollForw:
    case GIIK_PageDown:   return KEYD_PGDN;
    case GIIK_Insert: return KEYD_INSERT;
    case GIIK_Pause: return KEYD_PAUSE;

    case GIIK_F1:  return KEYD_F1;
    case GIIK_F2:  return KEYD_F2;
    case GIIK_F3:  return KEYD_F3;
    case GIIK_F4:  return KEYD_F4;
    case GIIK_F5:  return KEYD_F5;
    case GIIK_F6:  return KEYD_F6;
    case GIIK_F7:  return KEYD_F7;
    case GIIK_F8:  return KEYD_F8;
    case GIIK_F9:  return KEYD_F9;
    case GIIK_F10: return KEYD_F10;
    case GIIK_F11: return KEYD_F11;
    case GIIK_F12: return KEYD_F12;

    case GIIK_PPlus:  return '+';
    case GIIK_PMinus: return '-';
    case GIIK_PDecimal: return '.';
    case GIIK_PEnter: return KEYD_ENTER;
    case GIIK_PSlash: return '/';
    case GIIK_PStar:  return '*';
    case GIIK_PEqual: return '=';

    case GIIK_NumLock:    return KEYD_NUMLOCK;
    case GIIK_ScrollLock: return KEYD_SCRLOCK;
    case GIIK_CapsLock:   return KEYD_CAPSLOCK;

    case GIIK_ShiftL:
    case GIIK_ShiftR: return KEYD_RSHIFT;
    case GIIK_CtrlL:
    case GIIK_CtrlR:  return KEYD_RCTRL;
    case GIIK_MetaL:
    case GIIK_AltL:   return KEYD_LALT;
    case GIIK_MetaR:
    case GIIK_AltR:   return KEYD_RALT;
  }

  if (GII_KTYP(label) == GII_KT_LATIN1)
  {
    label = GII_KVAL(label);

    if (label <= 0x7F)
      return tolower(label);
  }

  return -1;
}

static void HandleKeyEvent(gii_event * ev)
{
  event_t event;

  if (ev->any.type != evKeyPress && ev->any.type != evKeyRelease) 
    return;

  event.type  = (ev->any.type == evKeyPress) ? ev_keydown : ev_keyup;
  event.value.key = TranslateGGIKey(ev);

  if (event.value.key < 0)
    return;
  
  E_PostEvent(&event);
}

static void HandleMouseEvent(gii_event * ev)
{
  event_t event;

  if (ev->any.type == evPtrRelative)
  {
    event.type = ev_analogue;

    if (ev->pmove.x)
    {
      event.value.analogue.axis = mouse_xaxis;
      event.value.analogue.amount = ev->pmove.x * mouseSensitivity;
      E_PostEvent(&event);
    }
    
    if (ev->pmove.y)
    {
      event.value.analogue.axis = mouse_yaxis;
      event.value.analogue.amount = ev->pmove.y * mouseSensitivity;

      if (invertmouse)
        event.value.analogue.amount = -event.value.analogue.amount;

      E_PostEvent(&event);
    }

    // handle the mouse wheel
    if (ev->pmove.wheel == 0)
      return;

    event.type = ev_keydown;
    event.value.key = (ev->pmove.wheel < 0) ? KEYD_MWHEEL_UP : KEYD_MWHEEL_DN;
    
    E_PostEvent(&event);

    event.type = ev_keyup;
    event.value.key = (ev->pmove.wheel < 0) ? KEYD_MWHEEL_UP : KEYD_MWHEEL_DN;
    
    E_PostEvent(&event);
    return;
  }

  if (ev->any.type == evPtrButtonPress) 
    event.type = ev_keydown;
  else if (ev->any.type == evPtrButtonRelease) 
    event.type = ev_keyup;
  else 
    return;

  switch (ev->pbutton.button)
  {
    case GII_PBUTTON_PRIMARY:
      event.value.key = KEYD_MOUSE1; break;
      
    case GII_PBUTTON_SECONDARY:
      event.value.key = KEYD_MOUSE2; break;
      
    case GII_PBUTTON_TERTIARY:
      event.value.key = KEYD_MOUSE3; break;

    default:
      return;
  }

  E_PostEvent(&event);
}

//
// I_ControlGetEvents
//
void I_ControlGetEvents (void)
{
  gii_event ev;

  for (;;) 
  {
    struct timeval tv = {0, 0};

    if (ggiEventPoll(my_vis, GGI_EVENTS, &tv) == 0)
      break;

    ggiEventRead(my_vis, &ev, GGI_EVENTS);

    HandleKeyEvent(&ev);
    HandleMouseEvent(&ev);
  }
}

//
// I_EDGELoop
//
void I_EDGELoop(void)
{
  while (1)
    E_EDGELoopRoutine();
}
