//----------------------------------------------------------------------------
//  EDGE Linux SDL Video Code
//----------------------------------------------------------------------------
// 
//  Copyright (c) 1999-2000  The EDGE Team.
// 
//  This program is free software; you can redistribute it and/or
//  modify it under the terms of the GNU General Public License
//  as published by the Free Software Foundation; either version 2
//  of the License, or (at your option) any later version.
//
//  This program is distributed in the hope that it will be useful,
//  but WITHOUT ANY WARRANTY; without even the implied warranty of
//  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//  GNU General Public License for more details.
//
//----------------------------------------------------------------------------

#include <SDL/SDL.h>

#include "i_defs.h"
#include "./i_trans.h"

#include "version.h"
#include "dm_state.h"
#include "dm_defs.h"
#include "dm_type.h"
#include "m_argv.h"
#include "st_stuff.h"
#include "v_video1.h"
#include "v_video2.h"
#include "v_res.h"
#include "v_colour.h"
#include "v_screen.h"

#include "e_main.h"
#include "e_event.h"
#include "z_zone.h"
#include "w_wad.h"

#include "s_sound.h"


#include <stdlib.h>
#include <unistd.h>
#include <sys/ipc.h>
#include <sys/shm.h>

#include <stdarg.h>
#include <sys/time.h>
#include <sys/types.h>
#include <sys/socket.h>

#include <netinet/in.h>
#include <signal.h>


short hicolortransmask1, hicolortransmask2;

// Dummy screen... 
// mainscreen is a full-size subscreen of this one.
static screen_t dummy_screen;
static SDL_Surface *my_vis;

unsigned char *thepalette;

static int graphics_shutdown = 0;

static SDL_Color my_pal[256];

// Common const strings
static const char lcase_lxdoom[] = {"lxedge"};
static const char ucase_lxdoom[] = {"LXEDGE"};

int leds_always_off = 0;  // Expected by m_misc, not relevant
colourshift_t redshift, greenshift, blueshift;

// Possible Screen Modes
typedef struct
{
  int width;
  int height;
  int bpp;
}
possresmode_t;

possresmode_t possresmode[] =
{
  { 320, 200,  8},
  { 320, 240,  8},
  { 400, 300,  8},
  { 512, 384,  8},
  { 640, 400,  8},
  { 640, 480,  8},
  { 800, 600,  8},
  {1024, 768,  8},
  { 320, 200, 16},
  { 320, 240, 16},
  { 400, 300, 16},
  { 512, 384, 16},
  { 640, 400, 16},
  { 640, 480, 16},
  { 800, 600, 16},
  {1024, 768, 16},
  {  -1,  -1, -1}
};


// ====================== INTERNALS ======================

//
// MakeCol
//
static inline unsigned long MakeCol(long r, long g, long b)
{
  r = (r >> (8 - redshift.bits))   << redshift.shift;
  g = (g >> (8 - greenshift.bits)) << greenshift.shift;
  b = (b >> (8 - blueshift.bits))  << blueshift.shift;

  return r|g|b;
}

//
// CountLeadingZeros
//
static const inline unsigned long CountLeadingZeros(unsigned long x)
{
  unsigned long r = 0;

  if (x == 0)
    return 0;

  while (!(x & 1))
  {
    x >>= 1;
    r++;
  }

  return r;
}

//
// CountBitsInMask
//
static const inline unsigned long CountBitsInMask(unsigned long x)
{
  unsigned long r = 0;

  while (!(x & 1))
    x >>= 1;

  while (x & 1)
  {
    x >>= 1;
    r++;
  }

  if (x)
    I_Error ("X Version: Mask is not contiguous.\n");

  return r;
}

//
// SetColourShift
//
// Constructs the shifts necessary to set one element of an RGB colour value
//
void SetColourShift(unsigned long mask, colourshift_t * ps)
{
  ps->mask  = mask;
  ps->shift = CountLeadingZeros(mask);
  ps->bits  = CountBitsInMask(mask);
}

//
// BlitToScreen
//
static inline void BlitToScreen(void)
{
  register int y;
  register unsigned char *s1, *d1;

  s1 = (void *) dummy_screen.data;
  d1 = (void *) my_vis->pixels;

  for (y=0; y < SCREENHEIGHT; y++)
  {
    memcpy(d1, s1, SCREENWIDTH * BPP);

    s1 += SCREENDEPTH;
    d1 += SCREENWIDTH * BPP;
  }
}

// =================== END OF INTERNALS ===================

//
// I_StartupGraphics
//
void I_StartupGraphics(void)
{
  int i;
  boolean_t resadded;

  if (SDL_Init(SDL_INIT_VIDEO) < 0)
    I_Error("Couldn't init SDL\n");

  my_vis = SDL_SetVideoMode(SCREENWIDTH, SCREENHEIGHT, BPP*8, SDL_SWSURFACE | SDL_HWPALETTE);
  if (my_vis == NULL)
  {
    I_Error("Couldn't set %dx%dx%d video mode: %s\n", SCREENWIDTH,
        SCREENHEIGHT, BPP*8, SDL_GetError());
  }

  // -ACB- 2000/03/16 Detect Possible Resolutions
  resadded = false;
  for (i = 0; possresmode[i].width != -1; i++)
  {
    if (SDL_VideoModeOK (possresmode[i].width, possresmode[i].width,
        possresmode[i].bpp, SDL_SWSURFACE | SDL_HWPALETTE) >= 0)
    {
      V_AddAvailableResolution(possresmode[i].width, possresmode[i].height, possresmode[i].bpp);
      resadded = true;
    }
  }

  if (!resadded)
    I_Error("I_StartupGraphics: Unable to select supported resolution");
}


//
// I_GetTruecolInfo
//
void I_GetTruecolInfo(truecol_info_t * info)
{
  info->red_bits = redshift.bits;
  info->red_shift = redshift.shift;
  info->red_mask = redshift.mask;

  info->green_bits = greenshift.bits;
  info->green_shift = greenshift.shift;
  info->green_mask = greenshift.mask;

  info->blue_bits = blueshift.bits;
  info->blue_shift = blueshift.shift;
  info->blue_mask = blueshift.mask;
}

//
// I_SetScreenSize
//
boolean_t I_SetScreenSize(int width, int height, int bpp)
{
  my_vis = SDL_SetVideoMode(SCREENWIDTH, SCREENHEIGHT, BPP*8, SDL_SWSURFACE | SDL_HWPALETTE);
  if (my_vis == NULL)
  {
    I_Error ("Couldn't set %dx%dx%d video mode: %s\n", SCREENWIDTH,
        SCREENHEIGHT, BPP*8, SDL_GetError());
  }
  
  // -AJA- turn off cursor -- BIG performance increase
  SDL_ShowCursor(0);
  
  SCREENDEPTH = V_GetDepth(SCREENWIDTH,BPP);

  if (BPP != 1)
  {
    SetColourShift(my_vis->format->Rmask,   &redshift);
    SetColourShift(my_vis->format->Gmask, &greenshift);
    SetColourShift(my_vis->format->Bmask,  &blueshift);
  }

  dummy_screen.width = SCREENWIDTH;
  dummy_screen.height = SCREENHEIGHT;
  dummy_screen.depth = SCREENDEPTH;
  dummy_screen.bpp = BPP;
  dummy_screen.parent = NULL;
  dummy_screen.data = malloc(SCREENDEPTH * SCREENHEIGHT);

  main_scr = V_CreateSubScreen (&dummy_screen, 0, 0, SCREENWIDTH, SCREENHEIGHT);

  return true;
}

//
// I_StartFrame
//
void I_StartFrame(void)
{
}

//
// I_FinishFrame
//
void I_FinishFrame(void)
{
  BlitToScreen();

  SDL_UpdateRect(my_vis, 0, 0, SCREENWIDTH, SCREENHEIGHT);
}

//
// I_SetPalette
//
void I_SetPalette(byte palette[256][3])
{
  int i;

  const byte *gtable = gammatable[usegamma];

  if (BPP != 1)
    return;

  for (i = 0; i < 256; i++)
  {
    int r = gtable[palette[i][0]];
    int g = gtable[palette[i][1]];
    int b = gtable[palette[i][2]];

    my_pal[i].r = (r << 8) + r;
    my_pal[i].g = (g << 8) + g;
    my_pal[i].b = (b << 8) + b;
  }

  SDL_SetColors(my_vis, my_pal, 0, 256);
}

//
// I_Colour2Pixel
//
long I_Colour2Pixel(byte palette[256][3], int col)
{
  if (BPP == 1)
    return col;

  return MakeCol(palette[col][0], palette[col][1], palette[col][2]);
}

//
// I_MakeColourmapRange
//
void I_MakeColourmapRange(void *dest_colmaps, byte palette[256][3], const byte *src_colmaps, int num)
{
  const byte *gtable = gammatable[usegamma];
  unsigned short *colmap16 = dest_colmaps;
  byte tempr[256], tempg[256], tempb[256];
  int i, j;

  if (BPP != 2)
    return;

  stbar_update = true;

  if (interpolate_colmaps && num >= 2)
  {
    for (j = 0; j < 256; j++)
    {
      int c1 = src_colmaps[0 * 256 + j];
      int c2 = src_colmaps[(num - 1) * 256 + j];

      int r1 = palette[c1][0];
      int g1 = palette[c1][1];
      int b1 = palette[c1][2];

      int r2 = palette[c2][0];
      int g2 = palette[c2][1];
      int b2 = palette[c2][2];

      for (i = 0; i < num; i++)
      {
        int r = gtable[r1 + (r2 - r1) * i / (num - 1)];
        int g = gtable[g1 + (g2 - g1) * i / (num - 1)];
        int b = gtable[b1 + (b2 - b1) * i / (num - 1)];

        colmap16[i * 256 + j] = MakeCol (r, g, b);
      }
    }

    return;
  }

  // use the old (but faster) method:

  for (i = 0; i < 256; i++)
  {
    tempr[i] = gtable[palette[i][0]];
    tempg[i] = gtable[palette[i][1]];
    tempb[i] = gtable[palette[i][2]];
  }

  for (i = 0; i < num; i++)
  for (j = 0; j < 256; j++)
  {
    int c = src_colmaps[i * 256 + j];
    colmap16[i * 256 + j] = MakeCol (tempr[c], tempg[c], tempb[c]);
  }
}

//
// I_ShutdownGraphics
//
void I_ShutdownGraphics(void)
{
  if (graphics_shutdown)
    return;

  graphics_shutdown = 1;

  fprintf(stderr, "Shutting down graphics...\n");
  SDL_Quit ();
}

//
//   INPUT CODE 
//

//
// Translates a key from SDL -> EDGE
// Returns -1 if no suitable translation exists.

static int TranslateSDLKey(SDL_KeyboardEvent ev)
{
  int label = ev.keysym.sym;

  switch (label) 
  {
    case SDLK_TAB: return KEYD_TAB;
    case SDLK_RETURN: return KEYD_ENTER;
    case SDLK_ESCAPE: return KEYD_ESCAPE;
    case SDLK_DELETE:
    case SDLK_BACKSPACE: return KEYD_BACKSPACE;

    case SDLK_UP:    return KEYD_UPARROW;
    case SDLK_DOWN:  return KEYD_DOWNARROW;
    case SDLK_LEFT:  return KEYD_LEFTARROW;
    case SDLK_RIGHT: return KEYD_RIGHTARROW;

    case SDLK_HOME:   return KEYD_HOME;
    case SDLK_END: return KEYD_END;
    case SDLK_PAGEUP:   return KEYD_PGUP;
    case SDLK_PAGEDOWN:   return KEYD_PGDN;
    case SDLK_INSERT: return KEYD_INSERT;
    case SDLK_PAUSE: return KEYD_PAUSE;

    case SDLK_F1:  return KEYD_F1;
    case SDLK_F2:  return KEYD_F2;
    case SDLK_F3:  return KEYD_F3;
    case SDLK_F4:  return KEYD_F4;
    case SDLK_F5:  return KEYD_F5;
    case SDLK_F6:  return KEYD_F6;
    case SDLK_F7:  return KEYD_F7;
    case SDLK_F8:  return KEYD_F8;
    case SDLK_F9:  return KEYD_F9;
    case SDLK_F10: return KEYD_F10;
    case SDLK_F11: return KEYD_F11;
    case SDLK_F12: return KEYD_F12;

    case SDLK_KP_PLUS:  return '+';
    case SDLK_KP_MINUS: return '-';
    case SDLK_KP_PERIOD: return '.';
    case SDLK_KP_ENTER: return KEYD_ENTER;
    case SDLK_KP_DIVIDE: return '/';
    case SDLK_KP_MULTIPLY:  return '*';
    case SDLK_KP_EQUALS: return '=';

    case SDLK_NUMLOCK:    return KEYD_NUMLOCK;
    case SDLK_SCROLLOCK: return KEYD_SCRLOCK;
    case SDLK_CAPSLOCK:   return KEYD_CAPSLOCK;

    case SDLK_LSHIFT:
    case SDLK_RSHIFT: return KEYD_RSHIFT;
    case SDLK_LCTRL:
    case SDLK_RCTRL:  return KEYD_RCTRL;
    case SDLK_LMETA:
    case SDLK_LALT:   return KEYD_LALT;
    case SDLK_RMETA:
    case SDLK_RALT:   return KEYD_RALT;
  }
  if (label <= 0x7f)
    return tolower(label);

  return -1;
}

static void HandleKeyEvent(SDL_Event* ev)
{
  event_t event;

  if (ev->type != SDL_KEYDOWN && ev->type != SDL_KEYUP) 
    return;

  event.type  = (ev->type == SDL_KEYDOWN) ? ev_keydown : ev_keyup;
  event.value.key = TranslateSDLKey(ev->key);

  if (event.value.key < 0)
    return;
  
  E_PostEvent(&event);
}

static void HandleMouseEvent(SDL_Event * ev)
{
  event_t event;

  if (ev->type == SDL_MOUSEMOTION)
  {
    event.type = ev_analogue;

    if (ev->motion.x)
    {
      event.value.analogue.axis = mouse_xaxis;
      event.value.analogue.amount = ev->motion.xrel * mouseSensitivity;
      E_PostEvent(&event);
    }
    
    if (ev->motion.y)
    {
      event.value.analogue.axis = mouse_yaxis;
      event.value.analogue.amount = ev->motion.yrel * mouseSensitivity;

      if (!invertmouse)
        event.value.analogue.amount = -event.value.analogue.amount;

      E_PostEvent(&event);
    }

    // handle the mouse wheel
/*
    if (ev->pmove.wheel == 0)
      return;

    event.type = ev_keydown;
    event.value.key = (ev->pmove.wheel < 0) ? KEYD_MWHEEL_UP : KEYD_MWHEEL_DN;
    
    E_PostEvent(&event);

    event.type = ev_keyup;
    event.value.key = (ev->pmove.wheel < 0) ? KEYD_MWHEEL_UP : KEYD_MWHEEL_DN;
    
    E_PostEvent(&event);
    return;
*/ 
 }

  if (ev->type == SDL_MOUSEBUTTONDOWN) 
    event.type = ev_keydown;
  else if (ev->type == SDL_MOUSEBUTTONUP) 
    event.type = ev_keyup;
  else 
    return;

  switch (ev->button.button)
  {
    case 1:
      event.value.key = KEYD_MOUSE1; break;
      
    case 2:
      event.value.key = KEYD_MOUSE2; break;
      
    case 3:
      event.value.key = KEYD_MOUSE3; break;

    default:
      return;
  }

  E_PostEvent(&event);
}

//
// I_ControlGetEvents
//
void I_ControlGetEvents (void)
{
  SDL_Event ev;

  for (;;) 
  {
    if (SDL_PollEvent (&ev) == 0)
      break;

    HandleKeyEvent(&ev);
    HandleMouseEvent(&ev);
  }
}

//
// I_EDGELoop
//
void I_EDGELoop(void)
{
  while (1)
    E_EDGELoopRoutine();
}
