//----------------------------------------------------------------------------
//  EDGE Sector Lighting Code
//----------------------------------------------------------------------------
// 
//  Copyright (c) 1999-2000  The EDGE Team.
// 
//  This program is free software; you can redistribute it and/or
//  modify it under the terms of the GNU General Public License
//  as published by the Free Software Foundation; either version 2
//  of the License, or (at your option) any later version.
//
//  This program is distributed in the hope that it will be useful,
//  but WITHOUT ANY WARRANTY; without even the implied warranty of
//  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//  GNU General Public License for more details.
//
//----------------------------------------------------------------------------
//
//  Based on the DOOM source code, released by Id Software under the
//  following copyright:
//
//    Copyright (C) 1993-1996 by id Software, Inc.
//
//----------------------------------------------------------------------------
//
// -KM- 1998/09/27 Lights generalised for ddf
//

#include "i_defs.h"

#include "dm_defs.h"
#include "dm_state.h"
#include "m_random.h"
#include "p_local.h"
#include "r_state.h"
#include "z_zone.h"

static light_t *lights = NULL;

//
// GENERALISED LIGHT
//
static void DoLight(light_t * light)
{
  if (--light->count)
    return;

  // Flashing lights
  switch (light->type)
  {
    case lite_flash:
      {
        int i;

        i = M_Random();
        // Dark
        if (i <= light->probability)
        {
          light->sector->p.lightlevel = light->minlight;
          light->count = light->darktime;
        }
        else
        {
          light->sector->p.lightlevel = light->maxlight;
          light->count = light->brighttime;
        }
        break;
      }

    case lite_strobe:
      if (light->sector->p.lightlevel == light->maxlight)
      {
        // Go dark
        light->sector->p.lightlevel = light->minlight;
        light->count = light->darktime;
      }
      else
      {
        // Go Bright
        light->sector->p.lightlevel = light->maxlight;
        light->count = light->brighttime;
      }
      break;

    case lite_glow:
      if (light->direction == -1)
      {
        // Go dark
        light->sector->p.lightlevel -= GLOWSPEED;
        if (light->sector->p.lightlevel <= light->minlight)
        {
          light->sector->p.lightlevel = light->minlight;
          light->count = light->brighttime;
          light->direction = 1;
        }
        else
        {
          light->count = light->darktime;
        }
      }
      else
      {
        // Go Bright
        light->sector->p.lightlevel += GLOWSPEED;
        if (light->sector->p.lightlevel >= light->maxlight)
        {
          light->sector->p.lightlevel = light->maxlight;
          light->count = light->darktime;
          light->direction = -1;
        }
        else
        {
          light->count = light->brighttime;
        }
      }
      break;

    case lite_fireflicker:
      {
        // -ES- 2000/02/13 Changed this to original DOOM style flicker
        int amount = (M_Random()&3)*16;
    
        if (light->sector->p.lightlevel - amount < light->minlight)
        {
          light->sector->p.lightlevel = light->minlight;
          light->count = light->darktime;
        }
        else
        {
          light->sector->p.lightlevel = light->maxlight - amount;
          light->count = light->brighttime;
        }
      }
    default:
      break;

  }
}

//
// TURN LINE'S TAG LIGHTS ON
//
void EV_LightTurnOn(line_t * line, int bright)
{
  int i;
  int j;
  sector_t *sector;
  sector_t *temp;
  line_t *templine;

  sector = sectors;

  for (i = 0; i < numsectors; i++, sector++)
  {
    if (sector->tag == line->tag)
    {
      // bright = 0 means to search
      // for highest light level
      // surrounding sector
      if (!bright)
      {
        for (j = 0; j < sector->linecount; j++)
        {
          templine = sector->lines[j];
          temp = P_GetNextSector(templine, sector);

          if (!temp)
            continue;

          if (temp->p.lightlevel > bright)
            bright = temp->p.lightlevel;
        }
      }
      sector->p.lightlevel = bright;
    }
  }
}

#if 0 // Currently unused
//
// P_DestroyLight
//
// Unlink and destroy light.
void P_DestroyLight(light_t *light)
{
  if (light->next)
    light->next->prev = light->prev;
  if (light->prev)
    light->prev->next = light->next;
  else
    lights = light->next;
  Z_Free(light);
}
#endif

void P_DestroyAllLights(void)
{
  light_t *l, *next;

  for (l = lights; l; l = next)
  {
    next = l->next;
    Z_Free(l);
  }
  lights = NULL;
}

//
// P_NewLight
//
// Allocate and link in light.
light_t *P_NewLight(void)
{
  light_t *light;

  light = Z_New(light_t, 1);
  light->next = lights;
  light->prev = NULL;
  if (lights)
    lights->prev = light;
  lights = light;

  return light;
}

boolean_t EV_Lights(sector_t * sec, const lighttype_t * type)
{
  light_t *light;

  if (sec->specialdata[LIGHTS])
    return false;

  // check if a light effect already is running on this sector.
  for (light = lights; light && light->sector != sec; light = light->next)
    ;

  if (!light)
    // didn't already exist, create a new one
    light = P_NewLight();

  light->sector = sec;

  light->type = type->type;

  light->minlight = P_FindMinSurroundingLight(sec, sec->p.lightlevel);
  light->maxlight = sec->p.lightlevel;
  light->direction = -1;
  light->darktime = type->darktime;
  light->brighttime = type->brighttime;
  light->probability = type->light;
  light->count = type->sync ? (leveltime % type->sync) + 1 : light->darktime;

  return true;
}

//
// P_RunLights
//
// Executes all light effects of this tic
// Lights are quite simple to handle, since they never destroy
// themselves. Therefore, we do not need to bother about stuff like
// removal queues
void P_RunLights(void)
{
  light_t *l;

  for (l = lights; l; l = l->next)
    DoLight(l);
}
