//----------------------------------------------------------------------------
//  EDGE Moving, Aiming, Shooting & Collision code
//----------------------------------------------------------------------------
// 
//  Copyright (c) 1999-2000  The EDGE Team.
// 
//  This program is free software; you can redistribute it and/or
//  modify it under the terms of the GNU General Public License
//  as published by the Free Software Foundation; either version 2
//  of the License, or (at your option) any later version.
//
//  This program is distributed in the hope that it will be useful,
//  but WITHOUT ANY WARRANTY; without even the implied warranty of
//  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//  GNU General Public License for more details.
//
//----------------------------------------------------------------------------
//
//  Based on the DOOM source code, released by Id Software under the
//  following copyright:
//
//    Copyright (C) 1993-1996 by id Software, Inc.
//
//----------------------------------------------------------------------------
//
// -MH- 1998/07/02 "shootupdown" --> "true3dgameplay"
//
// -AJA- 1999/07/19: Removed P_LineOpening.  Gaps are now stored 
//       in line_t, and updated whenever sector heights change.
//
// -AJA- 1999/07/21: Replaced some non-critical P_Randoms with M_Random.
//
// -AJA- 1999/07/30: Big changes for extra floor handling. Split
//       P_CheckPosition into two new routines (one handling absolute
//       positions, the other handling relative positions). Split the
//       PIT_Check* routines similiarly.
//

#include "i_defs.h"

#include "dm_defs.h"
#include "dm_state.h"
#include "m_bbox.h"
#include "m_random.h"
#include "p_local.h"
#include "r_state.h"
#include "s_sound.h"
#include "z_zone.h"

float_t tmbbox[4];
mobj_t *tmthing;
int tmflags;

static float_t tmx;
static float_t tmy;
static float_t tmz;  // -ACB- 1998/06/15 Z Checking

static float_t tmxmove;
static float_t tmymove;

boolean_t mobj_hit_sky;
line_t *blockline;

// If "floatok" true, move would be ok if at float_destz.
boolean_t floatok;
float_t float_destz;

subsector_t *tm_newsubsec;
static float_t tm_floorz;
static float_t tm_ceilnz;
static float_t tm_dropoff;
mobj_t *tm_support;

// keep track of special lines as they are hit,
// but don't process them until the move is proven valid

stack_array_t spechit_a;
line_t **spechit = NULL;
int numspechit;

mobj_t *linetarget;  // who got hit (or NULL)

mobj_t *shootthing;  // who started the shoot

// startz and stopz on target
float_t attackrange;
static float_t shootz;
static float_t aimslope;
// trace angle -ACB- 1998/09/01
static angle_t shootangle;
static float_t la_damage;
static float_t shoot_prev_z;
static const mobjinfo_t *bulletpuff;

// slopes to top and bottom of target
static float_t targettopslope;
static float_t targetbottomslope;

//
// TELEPORT MOVE
// 

//
// PIT_StompThing
//
static boolean_t PIT_StompThing(mobj_t * thing)
{
  float_t blockdist;

  if (!(thing->flags & MF_SHOOTABLE))
    return true;

  blockdist = thing->radius + tmthing->radius;

  // check to see we hit it
  if (fabs(thing->x - tmx) >= blockdist || fabs(thing->y - tmy) >= blockdist)
    return true;  // no, we did not

  // check we aren't trying to stomp ourselves
  if (thing == tmthing)
    return true;

  // -AJA- 1999/07/30: True 3d gameplay checks.
  if (level_flags.true3dgameplay)
  {
    if (tmz >= thing->z + thing->height)
    {
      // went over
      tm_floorz = MAX(tm_floorz, thing->z + thing->height);
      return true;
    }

    if (tmz + tmthing->height <= thing->z)
    {
      // went under
      tm_ceilnz = MIN(tm_ceilnz, thing->z);
      return true;
    }
  }

  if (!tmthing->player && (currentmap->force_off & MPF_Stomp))
    return false;

  P_DamageMobj(thing, tmthing, tmthing, 10000);

  return true;
}

//
// P_TeleportMove
//
boolean_t P_TeleportMove(mobj_t * thing, float_t x, float_t y, float_t z)
{
  int xl, xh;
  int yl, yh;
  int bx, by;

  // kill anything occupying the position
  tmthing = thing;
  tmflags = thing->flags;

  tmx = x;
  tmy = y;
  tmz = z;

  tmbbox[BOXTOP] = y + tmthing->radius;
  tmbbox[BOXBOTTOM] = y - tmthing->radius;
  tmbbox[BOXRIGHT] = x + tmthing->radius;
  tmbbox[BOXLEFT] = x - tmthing->radius;

  tm_newsubsec = R_PointInSubsector(x, y);

  P_ComputeThingGap(thing, tm_newsubsec->sector, z, &tm_floorz, &tm_ceilnz);

  // The base floor/ceiling is from the subsector that contains the point.
  // Any contacted lines the step closer together will adjust them.
  tm_dropoff = tm_floorz;
  tm_support = NULL;

  validcount++;
  Z_SetArraySize(&spechit_a, numspechit = 0);

  // stomp on any things contacted
  xl = (tmbbox[BOXLEFT] - bmaporgx - MAXRADIUS) / MAPBLOCKUNITS;
  xh = (tmbbox[BOXRIGHT] - bmaporgx + MAXRADIUS) / MAPBLOCKUNITS;
  yl = (tmbbox[BOXBOTTOM] - bmaporgy - MAXRADIUS) / MAPBLOCKUNITS;
  yh = (tmbbox[BOXTOP] - bmaporgy + MAXRADIUS) / MAPBLOCKUNITS;
  for (bx = xl; bx <= xh; bx++)
    for (by = yl; by <= yh; by++)
      P_BlockThingsIterator(bx, by, PIT_StompThing);

  // everything on the spot has been stomped,
  // so link the thing into its new position
  P_UnsetThingPosition(thing);

  thing->floorz = tm_floorz;
  thing->ceilingz = tm_ceilnz;
  thing->x = x;
  thing->y = y;
  thing->z = z;

  P_SetThingPosition(thing);

  return true;
}

//
// ABSOLUTE POSITION CLIPPING
//

//
// PIT_CheckAbsLine
//
static boolean_t PIT_CheckAbsLine(line_t * ld)
{
  int i;

  if (tmbbox[BOXRIGHT] <= ld->bbox[BOXLEFT] ||
      tmbbox[BOXLEFT] >= ld->bbox[BOXRIGHT] ||
      tmbbox[BOXTOP] <= ld->bbox[BOXBOTTOM] ||
      tmbbox[BOXBOTTOM] >= ld->bbox[BOXTOP])
  {
    // no intersection with line
    return true;
  }

  if (P_BoxOnLineSide(tmbbox, ld) != -1)
    return true;

  // The spawning thing's position touches the given line.
  // If this should not be allowed, return false.

  if (!ld->backsector || ld->gap_num == 0)
    return false;  // one sided line

  if (! (tmthing->extendedflags & EF_CROSSLINES))
  {
    // explicitly blocking everything ?
    if (ld->flags & ML_BLOCKING)
      return false;

    // block monsters only ?
    if ((tmthing->extendedflags & EF_MONSTER) &&
        (ld->flags & ML_BLOCKMONSTERS))
    {
      return false;
    }
  }

  // does the thing fit in one of the line gaps ?
  for (i = 0; i < ld->gap_num; i++)
  {
    if (ld->gaps[i].f <= tmz &&
        tmz + tmthing->height <= ld->gaps[i].c)
      return true;
  }

  return false;
}

//
// PIT_CheckAbsThing
//
static boolean_t PIT_CheckAbsThing(mobj_t * thing)
{
  float_t blockdist;
  boolean_t solid;

  if (thing == tmthing)
    return true;

  if (!(thing->flags & (MF_SOLID | MF_SHOOTABLE)))
    return true;

  blockdist = thing->radius + tmthing->radius;

  // Check that we didn't hit it
  if (fabs(thing->x - tmx) >= blockdist || fabs(thing->y - tmy) >= blockdist)
    return true;  // no we missed this thing

  // -KM- 1998/9/19 True 3d gameplay checks.
  if ((tmflags & MF_MISSILE) || level_flags.true3dgameplay)
  {
    // overhead ?
    if (tmz >= thing->z + thing->height)
      return true;

    // underneath ?
    if (tmz + tmthing->height <= thing->z)
      return true;
  }

  solid = (thing->flags & MF_SOLID);

  // check for missiles making contact
  // -ACB- 1998/08/04 Procedure for missile contact

  if (tmflags & MF_MISSILE)
  {
    // thing isn't shootable, return depending on if the thing is solid.
    if (!(thing->flags & MF_SHOOTABLE))
      return !solid;

    if (tmthing->source && tmthing->source == thing)
      return true;

    P_ActMissileContact(tmthing, thing);
    return false;
  }

  // -AJA- 2000/06/09: Follow MBF semantics: allow the non-solid
  // moving things to pass through solid things.
  return !solid || (thing->flags & MF_NOCLIP) || !(tmflags & MF_SOLID);
}

//
// P_CheckAbsPosition
//
// Check whether the thing can be placed at the absolute position
// (x,y,z).  Makes no assumptions about the thing's current position.
// 
// This is purely informative, nothing is modified, nothing is picked
// up, no special lines are recorded, no special things are touched, and
// no information (apart from true/false) is returned.
// 
// Only used for checking if an object can be spawned at a
// particular location.

boolean_t P_CheckAbsPosition(mobj_t * thing, float_t x, float_t y, float_t z)
{
  int xl, xh;
  int yl, yh;
  int bx, by;

  // can go anywhere
  if (tmflags & MF_NOCLIP)
    return true;

  tmthing = thing;
  tmflags = thing->flags;

  tmx = x;
  tmy = y;
  tmz = z;

  tmbbox[BOXTOP] = y + tmthing->radius;
  tmbbox[BOXBOTTOM] = y - tmthing->radius;
  tmbbox[BOXRIGHT] = x + tmthing->radius;
  tmbbox[BOXLEFT] = x - tmthing->radius;

  tm_newsubsec = R_PointInSubsector(x, y);

  validcount++;

  // Check things first.
  // The bounding box is extended by MAXRADIUS
  // because mobj_ts are grouped into mapblocks
  // based on their origin point, and can overlap
  // into adjacent blocks by up to MAXRADIUS units.

  xl = (tmbbox[BOXLEFT] - bmaporgx - MAXRADIUS) / MAPBLOCKUNITS;
  xh = (tmbbox[BOXRIGHT] - bmaporgx + MAXRADIUS) / MAPBLOCKUNITS;
  yl = (tmbbox[BOXBOTTOM] - bmaporgy - MAXRADIUS) / MAPBLOCKUNITS;
  yh = (tmbbox[BOXTOP] - bmaporgy + MAXRADIUS) / MAPBLOCKUNITS;

  for (bx = xl; bx <= xh; bx++)
    for (by = yl; by <= yh; by++)
      if (!P_BlockThingsIterator(bx, by, PIT_CheckAbsThing))
        return false;

  // check lines

  xl = (tmbbox[BOXLEFT] - bmaporgx) / MAPBLOCKUNITS;
  xh = (tmbbox[BOXRIGHT] - bmaporgx) / MAPBLOCKUNITS;
  yl = (tmbbox[BOXBOTTOM] - bmaporgy) / MAPBLOCKUNITS;
  yh = (tmbbox[BOXTOP] - bmaporgy) / MAPBLOCKUNITS;

  for (bx = xl; bx <= xh; bx++)
    for (by = yl; by <= yh; by++)
      if (!P_BlockLinesIterator(bx, by, PIT_CheckAbsLine))
        return false;

  return true;
}

//
// RELATIVE MOVEMENT CLIPPING
//

//
// PIT_CheckRelLine
//
// Adjusts tm_floorz & tm_ceilnz as lines are contacted

static boolean_t PIT_CheckRelLine(line_t * ld)
{
  int i;

  if (tmbbox[BOXRIGHT] <= ld->bbox[BOXLEFT] ||
      tmbbox[BOXLEFT] >= ld->bbox[BOXRIGHT] ||
      tmbbox[BOXTOP] <= ld->bbox[BOXBOTTOM] ||
      tmbbox[BOXBOTTOM] >= ld->bbox[BOXTOP])
  {
    // no intersection with line
    return true;
  }

  if (P_BoxOnLineSide(tmbbox, ld) != -1)
    return true;

  // A line has been hit

  // The moving thing's destination position will cross the given line.
  // If this should not be allowed, return false.
  // If the line is special, keep track of it
  // to process later if the move is proven ok.
  // NOTE: specials are NOT sorted by order,
  // so two special lines that are only 8 pixels apart
  // could be crossed in either order.

  if (!ld->backsector)
  {
    blockline = ld;

    // one sided line
    return false;
  }

  if (! (tmthing->extendedflags & EF_CROSSLINES))
  {
    // explicitly blocking everything ?
    // or just blocking monsters ?

    if (ld->flags & ML_BLOCKING ||
        (ld->flags & ML_BLOCKMONSTERS &&
         tmthing->extendedflags & EF_MONSTER))
    {
      blockline = ld;
      return false;
    }
  }

  // check for hitting a sky-hack line
  {
    float_t f1, c1;
    float_t f2, c2;

    f1 = ld->frontsector->floor.h;
    c1 = ld->frontsector->ceil.h;
    f2 = ld->backsector->floor.h;
    c2 = ld->backsector->ceil.h;
    
    if (c1 != c2 && ld->frontsector->ceil.pic == skyflatnum &&
        ld->backsector->ceil.pic == skyflatnum &&
        tmz > MIN(c1, c2))
    {
      mobj_hit_sky = true;
    }

    if (f1 != f2 && ld->frontsector->floor.pic == skyflatnum &&
        ld->backsector->floor.pic == skyflatnum &&
        tmz+tmthing->height < MAX(f1, f2))
    {
      mobj_hit_sky = true;
    }
  }
  
  // CHOOSE GAP
  //
  // If this line borders a sector with multiple floors, then there will
  // be multiple gaps and we must choose one here, based on the thing's
  // current position (esp. Z).

  i = P_FindThingGap(ld->gaps, ld->gap_num, tmz, tmz + tmthing->height);

  // gap has been chosen. apply it.

  if (i >= 0)
  {
    if (ld->gaps[i].f >= tm_floorz)
    {
      tm_floorz = ld->gaps[i].f;
      tm_support = NULL;
    }

    if (ld->gaps[i].c < tm_ceilnz)
      tm_ceilnz = ld->gaps[i].c;

    if (ld->gaps[i].f < tm_dropoff)
      tm_dropoff = ld->gaps[i].f;
  }
  else
  {
    tm_ceilnz = tm_floorz;
  }

  if (tm_ceilnz < tm_floorz + tmthing->height)
    blockline = ld;

  // if contacted a special line, add it to the list
  if (ld->special)
  {
    Z_SetArraySize(&spechit_a, ++numspechit);
    spechit[numspechit - 1] = ld;
  }

  return true;
}

//
// PIT_CheckRelThing
//
static boolean_t PIT_CheckRelThing(mobj_t * thing)
{
  float_t blockdist;
  boolean_t solid;

  if (thing == tmthing)
    return true;

  if (!(thing->flags & (MF_SOLID | MF_SPECIAL | MF_SHOOTABLE | MF_TOUCHY)))
    return true;

  blockdist = thing->radius + tmthing->radius;

  // Check that we didn't hit it
  if (fabs(thing->x - tmx) >= blockdist || fabs(thing->y - tmy) >= blockdist)
    return true;  // no we missed this thing

  // -KM- 1998/9/19 True 3d gameplay checks.
  if (level_flags.true3dgameplay && !(thing->flags & MF_SPECIAL))
  {
    float_t top_z = thing->z + thing->height;

    // see if we went over
    if (tmz >= top_z)
    {
      if (top_z > tm_floorz && !(thing->flags & MF_MISSILE))
      {
        tm_floorz = top_z;
        tm_support = thing;
      }
      return true;
    }

    // see if we went underneath
    if (tmz + tmthing->height <= thing->z)
    {
      if (thing->z < tm_ceilnz && !(thing->flags & MF_MISSILE))
      {
        tm_ceilnz = thing->z;
      }
      return true;
    }

    // -AJA- 1999/07/21: allow climbing on top of things.

    if ((thing->extendedflags & EF_CLIMBABLE) &&
        (tmthing->player || (tmthing->extendedflags & EF_MONSTER)) &&
        ((tmflags & MF_DROPOFF) ||
         (tmthing->extendedflags & EF_EDGEWALKER)) &&
        (tmz + tmthing->info->step_size >= top_z) &&
        (top_z > tm_floorz))
    {
      tm_floorz = top_z;
      tm_support = thing;
      return true;
    }
  }

  // check for skulls slamming into things
  // -ACB- 1998/08/04 Use procedure
  // -KM- 1998/09/01 After I noticed Skulls slamming into boxes of rockets...

  solid = (thing->flags & MF_SOLID);

  if ((tmflags & MF_SKULLFLY) && solid)
  {
    P_ActSlammedIntoObject(tmthing, thing);

    // stop moving
    return false;
  }

  // check for missiles making contact
  // -ACB- 1998/08/04 Procedure for missile contact

  if (tmflags & MF_MISSILE)
  {
    // see if it went over / under
    if (tmz > thing->z + thing->height)
      return true;  // overhead

    if (tmz + tmthing->height < thing->z)
      return true;  // underneath

    // thing isn't shootable, return depending on if the thing is solid.
    if (!(thing->flags & MF_SHOOTABLE))
      return !solid;

    // don't hurt the missile's shooter:
    if (tmthing->source && tmthing->source == thing)
      return true;

    P_ActMissileContact(tmthing, thing);
    return false;
  }

  // check for special pickup
  if ((tmflags & MF_PICKUP) && (thing->flags & MF_SPECIAL))
  {
    // can remove thing
    P_TouchSpecialThing(thing, tmthing);
  }

  // -AJA- 1999/08/21: check for touchy objects.
  if ((thing->flags & MF_TOUCHY) && (tmflags & MF_SOLID) &&
      !(thing->extendedflags & EF_USABLE))
  {
    P_ActTouchyContact(thing, tmthing);
    return !solid;
  }

  // -AJA- 2000/06/09: Follow MBF semantics: allow the non-solid
  // moving things to pass through solid things.
  return !solid || (thing->flags & MF_NOCLIP) || !(tmflags & MF_SOLID);
}

//
// P_CheckRelPosition
//
// Checks whether the thing can be moved to the position (x,y), which is
// assumed to be relative to the thing's current position.
//
// This is purely informative, nothing is modified
// (except things picked up).
//
// Only used by P_TryMove and P_ThingHeightClip.
// 
// in:
//  a mobj_t (can be valid or invalid)
//  a position to be checked
//
// during:
//  special things are touched if MF_PICKUP
//  early out on solid lines?
//
// out:
//  tm_newsubsec
//  tm_floorz
//  tm_ceilnz
//  tm_dropoff
//  tm_support
//  speciallines[]
//  numspeciallines

static boolean_t P_CheckRelPosition(mobj_t * thing, float_t x, float_t y)
{
  int xl, xh;
  int yl, yh;
  int bx, by;

  mobj_hit_sky = false;
  blockline = NULL;

  tmthing = thing;
  tmflags = thing->flags;

  tmx = x;
  tmy = y;
  tmz = thing->z;

  tmbbox[BOXTOP] = y + tmthing->radius;
  tmbbox[BOXBOTTOM] = y - tmthing->radius;
  tmbbox[BOXRIGHT] = x + tmthing->radius;
  tmbbox[BOXLEFT] = x - tmthing->radius;

  tm_newsubsec = R_PointInSubsector(x, y);

  // The base floor / ceiling is from the sector that contains the
  // point.  Any contacted lines the step closer together will adjust them.
  // -AJA- 1999/07/19: Extra floor support.
  P_ComputeThingGap(thing, tm_newsubsec->sector, tmz, &tm_floorz, &tm_ceilnz);

  tm_dropoff = tm_floorz;
  tm_support = NULL;

  // can go anywhere
  if (tmflags & MF_NOCLIP)
    return true;

  validcount++;
  Z_SetArraySize(&spechit_a, numspechit = 0);

  // -KM- 1998/11/25 Corpses aren't supposed to hang in the air...
  if (!(tmflags & (MF_NOCLIP | MF_CORPSE)))
  {
    // Check things first, possibly picking things up.
    // The bounding box is extended by MAXRADIUS
    // because mobj_ts are grouped into mapblocks
    // based on their origin point, and can overlap
    // into adjacent blocks by up to MAXRADIUS units.

    xl = (tmbbox[BOXLEFT] - bmaporgx - MAXRADIUS) / MAPBLOCKUNITS;
    xh = (tmbbox[BOXRIGHT] - bmaporgx + MAXRADIUS) / MAPBLOCKUNITS;
    yl = (tmbbox[BOXBOTTOM] - bmaporgy - MAXRADIUS) / MAPBLOCKUNITS;
    yh = (tmbbox[BOXTOP] - bmaporgy + MAXRADIUS) / MAPBLOCKUNITS;

    for (bx = xl; bx <= xh; bx++)
      for (by = yl; by <= yh; by++)
        if (!P_BlockThingsIterator(bx, by, PIT_CheckRelThing))
          return false;
  }

  // check lines

  xl = (tmbbox[BOXLEFT] - bmaporgx) / MAPBLOCKUNITS;
  xh = (tmbbox[BOXRIGHT] - bmaporgx) / MAPBLOCKUNITS;
  yl = (tmbbox[BOXBOTTOM] - bmaporgy) / MAPBLOCKUNITS;
  yh = (tmbbox[BOXTOP] - bmaporgy) / MAPBLOCKUNITS;;

  for (bx = xl; bx <= xh; bx++)
    for (by = yl; by <= yh; by++)
      if (!P_BlockLinesIterator(bx, by, PIT_CheckRelLine))
        return false;

  return true;
}

//
// P_TryMove
//
// Attempt to move to a new position,
// crossing special lines unless MF_TELEPORT is set.

boolean_t P_TryMove(mobj_t * thing, float_t x, float_t y)
{
  float_t oldx;
  float_t oldy;
  int side;
  int oldside;
  line_t *ld;

  floatok = false;

  if (!P_CheckRelPosition(thing, x, y))
    // solid wall or thing
    return false;

  if (!(thing->flags & MF_NOCLIP))
  {
    if (thing->height > tm_ceilnz - tm_floorz)
    {
      // doesn't fit
      return false;
    }

    floatok = true;
    float_destz = tm_floorz;

    if (!(thing->flags & MF_TELEPORT) &&
        (thing->z + thing->height > tm_ceilnz))
    {
      // mobj must lower itself to fit.
      return false;
    }

    if (!(thing->flags & MF_TELEPORT) &&
        (thing->z + thing->info->step_size) < tm_floorz)
    {
      // too big a step up.
      return false;
    }

    if (!(thing->flags & (MF_TELEPORT | MF_DROPOFF | MF_FLOAT)) &&
        (thing->z - thing->info->step_size) > tm_floorz &&
        thing->z <= thing->floorz + 1.0)
    {
      // too big a step down.
      return false;
    }

    if (!((thing->flags & (MF_DROPOFF | MF_FLOAT)) ||
            (thing->extendedflags & (EF_EDGEWALKER | EF_WATERWALKER))) &&
        (tm_floorz - tm_dropoff > thing->info->step_size))
    {
      // don't stand over a dropoff.
      return false;
    }
  }

  // the move is ok, so link the thing into its new position
  P_UnsetThingPosition(thing);

  oldx = thing->x;
  oldy = thing->y;
  thing->floorz = tm_floorz;
  thing->ceilingz = tm_ceilnz;
  P_MobjSetRideEm(thing, (thing->flags & MF_FLOAT) ? NULL : tm_support);
  thing->x = x;
  thing->y = y;

  // -AJA- 1999/08/02: Improved MF_TELEPORT handling.
  if (thing->flags & (MF_TELEPORT | MF_NOCLIP))
  {
    if (thing->z <= thing->floorz)
      thing->z = thing->floorz;
    else if (thing->z + thing->height > thing->ceilingz)
      thing->z = thing->ceilingz - thing->height;
  }

  P_SetThingPosition(thing);

  // if any special lines were hit, do the effect
  if (!(thing->flags & (MF_TELEPORT | MF_NOCLIP)))
  {
    line_t **hits;
    int i;
    
    // -ES- 2000/02/05 spechit could be changed inside the loop
    hits = I_TmpMalloc(numspechit * sizeof(line_t *));
    Z_MoveData(hits, spechit, line_t *, numspechit);
    i = numspechit;
    Z_SetArraySize(&spechit_a, numspechit = 0);
    
    while (i)
    {
      // see if the line was crossed
      ld = hits[--i];
      side = P_PointOnLineSide(thing->x, thing->y, ld);
      oldside = P_PointOnLineSide(oldx, oldy, ld);

      if (side != oldside && ld->special)
      {
        if (thing->flags & MF_MISSILE)
          P_ShootSpecialLine(thing, ld);
        else
          P_CrossSpecialLine(ld, oldside, thing);
      }
    }
    I_TmpFree(hits);
  }

  return true;
}

//
// P_ThingHeightClip
//
// Takes a valid thing and adjusts the thing->floorz, thing->ceilingz,
// and possibly thing->z.
//
// This is called for all nearby things whenever a sector changes height.
//
// If the thing doesn't fit, the z will be set to the lowest value
// and false will be returned.

static boolean_t P_ThingHeightClip(mobj_t * thing)
{
  boolean_t onfloor = (thing->z == thing->floorz);

  P_CheckRelPosition(thing, thing->x, thing->y);

  thing->floorz = tm_floorz;
  thing->ceilingz = tm_ceilnz;

  P_MobjSetRideEm(thing, (thing->flags & MF_FLOAT) ? NULL : tm_support);

  if (onfloor)
  {
    // walking monsters rise and fall with the floor
    thing->z = thing->floorz;
  }
  else
  {
    // don't adjust a floating monster unless forced to
    if (thing->z + thing->height > thing->ceilingz)
      thing->z = thing->ceilingz - thing->height;
  }

  if (thing->ceilingz - thing->floorz < thing->height)
    return false;

  return true;
}

//
// SLIDE MOVE
// Allows the player to slide along any angled walls.
//
static float_t bestslidefrac;
static float_t secondslidefrac;

static line_t *bestslideline;
static line_t *secondslideline;

mobj_t *slidemo;

//
// P_HitSlideLine
// Adjusts the xmove / ymove
// so that the next move will slide along the wall.
//
static void HitSlideLine(line_t * ld)
{
  int side;

  angle_t lineangle;
  angle_t moveangle;
  angle_t deltaangle;

  float_t movelen;
  float_t newlen;

  if (ld->slopetype == ST_HORIZONTAL)
  {
    tmymove = 0;
    return;
  }

  if (ld->slopetype == ST_VERTICAL)
  {
    tmxmove = 0;
    return;
  }

  side = P_PointOnLineSide(slidemo->x, slidemo->y, ld);

  lineangle = R_PointToAngle(0, 0, ld->dx, ld->dy);

  if (side == 1)
    lineangle += ANG180;

  moveangle = R_PointToAngle(0, 0, tmxmove, tmymove);
  deltaangle = moveangle - lineangle;

  if (deltaangle > ANG180)
    deltaangle += ANG180;
  // I_Error ("SlideLine: ang>ANG180");

  movelen = P_ApproxDistance(tmxmove, tmymove);
  newlen = movelen * M_Cos(deltaangle);

  tmxmove = newlen * M_Cos(lineangle);
  tmymove = newlen * M_Sin(lineangle);
}

//
// PTR_SlideTraverse
//
static boolean_t PTR_SlideTraverse(intercept_t * in)
{
  line_t *li;
  int i;

  if (!in->isaline)
    I_Error("PTR_SlideTraverse: not a line?");

  li = in->d.line;

  if (!(li->flags & ML_TWOSIDED))
  {
    // hit the back side ?
    if (P_PointOnLineSide(slidemo->x, slidemo->y, li))
      return true;
  }

  // -AJA- 1999/07/19: Gaps are now stored in line_t.

  for (i = 0; i < li->gap_num; i++)
  {
    // check if it can fit in the space
    if (slidemo->height > li->gaps[i].c - li->gaps[i].f)
      continue;

    // check slide mobj is not too high
    if (slidemo->z + slidemo->height > li->gaps[i].c)
      continue;

    // check slide mobj can step over
    if (slidemo->z + slidemo->info->step_size < li->gaps[i].f)
      continue;

    return true;
  }

  // the line does block movement,
  // see if it is closer than best so far
  if (in->frac < bestslidefrac)
  {
    secondslidefrac = bestslidefrac;
    secondslideline = bestslideline;
    bestslidefrac = in->frac;
    bestslideline = li;
  }

  // stop
  return false;
}

//
// P_SlideMove
// The momx / momy move is bad, so try to slide along a wall.
//
// Find the first line hit, move flush to it, and slide along it
//
// -ACB- 1998/07/28 This is NO LONGER a kludgy mess; removed goto rubbish.
//
void P_SlideMove(mobj_t * mo)
{
  float_t leadx;
  float_t leady;
  float_t trailx;
  float_t traily;
  float_t newx;
  float_t newy;
  int hitcount;
  boolean_t retry;

  slidemo = mo;
  hitcount = 0;
  retry = true;

  while (retry)
  {
    if (++hitcount == 3)
    {
      if (!P_TryMove(mo, mo->x, mo->y + mo->momy))
        P_TryMove(mo, mo->x + mo->momx, mo->y);
      return;
    }

    // trace along the three leading corners
    if (mo->momx > 0)
    {
      leadx = mo->x + mo->radius;
      trailx = mo->x - mo->radius;
    }
    else
    {
      leadx = mo->x - mo->radius;
      trailx = mo->x + mo->radius;
    }

    if (mo->momy > 0)
    {
      leady = mo->y + mo->radius;
      traily = mo->y - mo->radius;
    }
    else
    {
      leady = mo->y - mo->radius;
      traily = mo->y + mo->radius;
    }

    bestslidefrac = 1.0001;

    P_PathTraverse(leadx, leady, leadx + mo->momx, leady + mo->momy,
        PT_ADDLINES, PTR_SlideTraverse);
    P_PathTraverse(trailx, leady, trailx + mo->momx, leady + mo->momy,
        PT_ADDLINES, PTR_SlideTraverse);
    P_PathTraverse(leadx, traily, leadx + mo->momx, traily + mo->momy,
        PT_ADDLINES, PTR_SlideTraverse);

    // move up to the wall
    if (bestslidefrac == 1.0001)
    {
      // the move must have hit the middle, so stairstep
      if (!P_TryMove(mo, mo->x, mo->y + mo->momy))
        P_TryMove(mo, mo->x + mo->momx, mo->y);
      return;
    }

    // fudge a bit to make sure it doesn't hit
    bestslidefrac -= 0.01;
    if (bestslidefrac > 0)
    {
      newx = mo->momx * bestslidefrac;
      newy = mo->momy * bestslidefrac;

      if (!P_TryMove(mo, mo->x + newx, mo->y + newy))
      {
        if (!P_TryMove(mo, mo->x, mo->y + mo->momy))
          P_TryMove(mo, mo->x + mo->momx, mo->y);
        return;
      }
    }

    // Now continue along the wall.
    // First calculate remainder.
    bestslidefrac = 1.0 - (bestslidefrac + 0.01);

    if (bestslidefrac > 1.0)
      bestslidefrac = 1.0;

    if (bestslidefrac <= 0)
      return;

    tmxmove = mo->momx * bestslidefrac;
    tmymove = mo->momy * bestslidefrac;

    HitSlideLine(bestslideline);  // clip the moves

    mo->momx = tmxmove;
    mo->momy = tmymove;

    if (P_TryMove(mo, mo->x + tmxmove, mo->y + tmymove))
      retry = false;
  }
}

//
// PTR_AimTraverse
//
// Sets linetarget and aimslope when a target is aimed at.

static boolean_t PTR_AimTraverse(intercept_t * in)
{
  line_t *li;
  mobj_t *th;
  float_t slope;
  float_t thingtopslope;
  float_t thingbottomslope;
  float_t dist;

  dist = attackrange * in->frac;

  if (dist < 0.01)
    return true;

  if (in->isaline)
  {
    li = in->d.line;

    if (!(li->flags & ML_TWOSIDED) || li->gap_num == 0)
      return false;  // stop

    // Crosses a two sided line.
    // A two sided line will restrict
    // the possible target ranges.
    //
    // -AJA- 1999/07/19: Gaps are now kept in line_t.

    if (li->frontsector->floor.h != li->backsector->floor.h)
    {
      float_t maxfloor = MAX(li->frontsector->floor.h, li->backsector->floor.h);

      slope = (maxfloor - shootz) / dist;

      if (slope > targetbottomslope)
        targetbottomslope = slope;
    }

    if (li->frontsector->ceil.h != li->backsector->ceil.h)
    {
      float_t minceil = MIN(li->frontsector->ceil.h, li->backsector->ceil.h);

      slope = (minceil - shootz) / dist;
      
      if (slope < targettopslope)
        targettopslope = slope;
    }

    if (targettopslope <= targetbottomslope)
      return false;  // stop

    // shot continues
    return true;
  }

  // shoot a thing
  th = in->d.thing;

  if (th == shootthing)
    return true;  // can't shoot self

  if (!(th->flags & MF_SHOOTABLE))
    return true;  // has to be able to be shot

  // check angles to see if the thing can be aimed at
  thingtopslope = (th->z + th->height - shootz) / dist;

  if (thingtopslope < targetbottomslope)
    return true;  // shot over the thing

  thingbottomslope = (th->z - shootz) / dist;

  if (thingbottomslope > targettopslope)
    return true;  // shot under the thing

  // this thing can be hit!
  if (thingtopslope > targettopslope)
    thingtopslope = targettopslope;

  if (thingbottomslope < targetbottomslope)
    thingbottomslope = targetbottomslope;

  aimslope = (thingtopslope + thingbottomslope) / 2;
  linetarget = th;

  return false;  // don't go any farther
}

//
// PTR_ShootTraverse
//
// -ACB- 1998/07/28 Cleaned up.

static boolean_t PTR_ShootTraverse(intercept_t * in)
{
  float_t x, y, z;
  float_t frac;
  float_t dist;
  float_t thingtopslope;
  float_t thingbottomslope;
  int i, sidenum;

  line_t *li;
  mobj_t *th;
  side_t *side;

  dist = attackrange * in->frac;

  if (dist < 0.1)
    dist = 0.1;

  // Intercept is a line?
  if (in->isaline)
  {
    li = in->d.line;

    // Line is a special, Cause action....
    if (li->special)
      P_ShootSpecialLine(shootthing, li);

    // determine coordinates of intersect
    frac = in->frac;
    z = shootz + frac * aimslope * attackrange;
    x = trace.x + trace.dx * frac;
    y = trace.y + trace.dy * frac;

    sidenum = P_PointOnLineSide(trace.x, trace.y, li);
    side = li->side[sidenum];

    // check if shot has hit a floor or ceiling...
    if (side)
    {
      vert_region_t *reg;

      for (reg=side->sector->regions; reg; reg=reg->sec_next)
      {
        if (shoot_prev_z >= reg->floor->h && z <= reg->floor->h)
        {
          // -- shot hit the region's floor --

          // don't shoot the sky!
          if (!reg->sec_prev && reg->floor->pic == skyflatnum)
            return false;
          
          if (reg->sec_prev && reg->sec_prev->extrafloor &&
              (reg->sec_prev->extrafloor->type & EXFL_ShootThrough))
            break;
          
          z = reg->floor->h;
          // -ES- 2000/04/30 Avoid div by zero.
          if (aimslope != 0)
            frac = (z - shootz) / (aimslope * attackrange);
          else
            frac = in->frac;
          
          x = trace.x + trace.dx * frac;
          y = trace.y + trace.dy * frac;

          // Spawn bullet puffs.
          if (bulletpuff)
            P_SpawnPuff(x, y, z + 2.0, bulletpuff);

          // don't go any farther
          return false;
        }
        else if (shoot_prev_z <= reg->ceil->h && z >= reg->ceil->h)
        {
          // -- shot hit the region's ceiling --

          // don't shoot the sky!
          if (!reg->sec_next && reg->ceil->pic == skyflatnum)
            return false;
          
          if (reg->extrafloor &&
              (reg->extrafloor->type & EXFL_ShootThrough))
            break;
          
          z = reg->ceil->h;
          if (aimslope != 0)
            frac = (z - shootz) / (aimslope * attackrange);
          else
            frac = in->frac;
            
          x = trace.x + trace.dx * frac;
          y = trace.y + trace.dy * frac;

          // Spawn bullet puffs.
          if (bulletpuff)
            P_SpawnPuff(x, y, z - 2.0, bulletpuff);

          // don't go any farther
          return false;
        }
      }
    }

    // shoot doesn't go through a one-sided line, since one sided lines
    // do not have a sector the other side.

    if ((li->flags & ML_TWOSIDED) && li->gap_num > 0)
    {
      DEV_ASSERT2(li->backsector);

      // check all line gaps
      for (i=0; i < li->gap_num; i++)
      {
        if (li->gaps[i].f <= z && z <= li->gaps[i].c)
        {
          shoot_prev_z = z;
          return true;
        }
      }
    }
    
    // check if bullet hit a sky hack line...
    if (li->frontsector && li->backsector)
    {
      if (li->frontsector->ceil.pic == skyflatnum &&
          li->backsector->ceil.pic == skyflatnum)
      {
        float_t c1 = li->frontsector->ceil.h;
        float_t c2 = li->backsector->ceil.h;

        if (MIN(c1,c2) <= z && z <= MAX(c1,c2))
          return false;
      }

      if (li->frontsector->floor.pic == skyflatnum &&
          li->backsector->floor.pic == skyflatnum)
      {
        float_t f1 = li->frontsector->floor.h;
        float_t f2 = li->backsector->floor.h;

        if (MIN(f1,f2) <= z && z <= MAX(f1,f2))
          return false;
      }
    }

    // position puff off the wall

    x -= trace.dx * 6.0 / attackrange;
    y -= trace.dy * 6.0 / attackrange;

    // Spawn bullet puffs.
    if (bulletpuff)
      P_SpawnPuff(x, y, z, bulletpuff);

    // don't go any farther
    return false;
  }

  // shoot a thing
  th = in->d.thing;

  if (th == shootthing)
    return true;  // don't shoot self

  if (!(th->flags & MF_SHOOTABLE))
    return true;  // got to able to shoot it

  // check angles to see if the thing can be aimed at
  thingtopslope = (th->z + th->height - shootz) / dist;

  if (thingtopslope < aimslope)
    return true;  // shot over the thing

  thingbottomslope = (th->z - shootz) / dist;

  if (thingbottomslope > aimslope)
    return true;  // shot under the thing

  // hit thing
  // position a bit closer
  frac = in->frac - 10.0 / attackrange;

  x = trace.x + trace.dx * frac;
  y = trace.y + trace.dy * frac;
  z = shootz + frac * aimslope * attackrange;

  // Spawn bullet puffs or blood spots,
  // depending on target type.
  if (in->d.thing->flags & MF_NOBLOOD)
  {
    if (bulletpuff)
      P_SpawnPuff(x, y, z, bulletpuff);
  }
  else
    P_SpawnBlood(x, y, z, la_damage, shootangle, in->d.thing->info->blood);

  P_ActBulletContact(shootthing, th, la_damage);

  // don't go any farther
  return false;
}

//
// P_AimLineAttack
//
float_t P_AimLineAttack(mobj_t * t1, angle_t angle, float_t distance)
{
  float_t x2;
  float_t y2;

  shootthing = t1;

  x2 = t1->x + distance * M_Cos(angle);
  y2 = t1->y + distance * M_Sin(angle);
  shootz = t1->z + t1->height / 2 + 8;

  if (t1->player)
  {
    targettopslope = (t1->vertangle * 256 + 100.0) / 160.0;
    targetbottomslope = (t1->vertangle * 256 - 100.0) / 160.0;
  }
  else
  {
    targettopslope = 100.0 / 160.0;
    targetbottomslope = -100.0 / 160.0;
  }

  attackrange = distance;
  linetarget = NULL;

  P_PathTraverse(t1->x, t1->y, x2, y2, PT_ADDLINES | PT_ADDTHINGS, PTR_AimTraverse);

  if (linetarget)
    return aimslope;

  return 0;
}

//
// P_LineAttack
//
// If damage == 0, it is just a test trace
// that will leave linetarget set.

void P_LineAttack(mobj_t * t1, angle_t angle, float_t distance, 
    float_t slope, float_t damage, const mobjinfo_t *puff)
{
  float_t x2;
  float_t y2;

  shootangle = angle;
  shootthing = t1;
  la_damage = damage;
  x2 = t1->x + distance * M_Cos(angle);
  y2 = t1->y + distance * M_Sin(angle);
  shootz = t1->z + (t1->height / 2) + 8;  // -AJA- FIXME: DDF-itize
  shoot_prev_z = shootz;
  attackrange = distance;
  aimslope = slope;
  bulletpuff = puff;

  P_PathTraverse(t1->x, t1->y, x2, y2, PT_ADDLINES | PT_ADDTHINGS, 
      PTR_ShootTraverse);
}

//
// P_MapTargetTheory
//
// Returns a dummy moving object for a target, used for mobjs
// that don't have a specific target; but need to launch some sort
// of projectile in the direction based upon the angle & vertical
// angle of the mobj.
//
// -ACB- 1998/09/01
//
mobj_t *P_MapTargetTheory(mobj_t * source)
{
  angle_t angle;
  float_t slope;
  float_t distance;
  static mobj_t theorytarget;

  angle = source->angle;
  distance = MISSILERANGE;

  // not slope = source->vertangle; must allow for player weapon
  slope = source->vertangle * 1.25;

  Z_Clear(&theorytarget, mobj_t, 1);

  theorytarget.x = distance * M_Cos(angle) + source->x;
  theorytarget.y = distance * M_Sin(angle) + source->y;
  theorytarget.z = distance * slope + source->z;

  theorytarget.extendedflags |= EF_DUMMYMOBJ;

  return &theorytarget;
}

//
// P_MapTargetAutoAim
//
// Returns a moving object for a target; will search for a mobj to
// lock onto, however a dummy target returned if no object cannot be
// locked onto.
//
// -ACB- 1998/09/01
// -AJA- 1999/08/08: Added `force_aim' to fix chainsaw.
//

mobj_t *P_MapTargetAutoAim(mobj_t * source, angle_t angle, float_t distance, boolean_t force_aim)
{
  float_t x2;
  float_t y2;

  shootthing = source;

  x2 = source->x + distance * M_Cos(angle);
  y2 = source->y + distance * M_Sin(angle);
  shootz = source->z + source->height / 2 + 8;

  if (source->player)
  {
    targettopslope = (100 + source->vertangle * 256) / 160.0;
    targetbottomslope = (-100 + source->vertangle * 256) / 160.0;
  }
  else
  {
    targettopslope = 100.0 / 160.0;
    targetbottomslope = -100.0 / 160.0;
  }

  attackrange = distance;
  linetarget = NULL;

  // -KM- 1999/01/31 Autoaim is an option.
  if (!source->player || level_flags.autoaim || force_aim)
  {
    P_PathTraverse(source->x, source->y, x2, y2,
        PT_ADDLINES | PT_ADDTHINGS, PTR_AimTraverse);
  }

  if (linetarget)
  {
    // -KM- 1999/01/31 Look at the thing you aimed at.  Is sometimes
    //  useful, sometimes annoying :-)
    if (source->player && level_flags.autoaim == AA_MLOOK)
    {
      source->vertangle = (linetarget->z - source->z) /
          P_ApproxDistance(source->x - linetarget->x, source->y - linetarget->y);
      if (source->vertangle > LOOKUPLIMIT)
        source->vertangle = LOOKUPLIMIT;
      if (source->vertangle < LOOKDOWNLIMIT)
        source->vertangle = LOOKDOWNLIMIT;
    }
    return linetarget;
  }
  else
    return P_MapTargetTheory(source);
}

//
// USE LINES
//
static mobj_t *usething;
static float use_lower, use_upper;

static boolean_t PTR_UseTraverse(intercept_t * in)
{
  int sidenum;
  line_t *line;
  side_t *side;
  mobj_t *th;

  // intercept is a thing ?
  if (! in->isaline)
  {
    th = in->d.thing;

    // not a usable thing ?
    if (! (th->extendedflags & EF_USABLE) || ! th->info->touch_state)
      return true;
    
    if (! P_ActUseThing(usething, th, use_lower, use_upper))
      return true;

    // don't go any farther (thing was usable)
    return false;
  }

  line = in->d.line;

  sidenum = P_PointOnLineSide(usething->x, usething->y, line);
  sidenum = (sidenum == 1) ? 1 : 0;

  side = line->side[sidenum];

  // update open vertical range (extrafloors are NOT checked)
  if (side)
  {
    use_lower = MAX(use_lower, side->sector->floor.h);
    use_upper = MIN(use_upper, side->sector->ceil.h);
  }

  if (!line->special)
  {
    if (line->gap_num == 0 || use_upper <= use_lower)
    {
      // can't use through a wall
      S_StartSound(usething, sfx_noway);
      return false;
    }

    // not a special line, but keep checking
    return true;
  }

  P_UseSpecialLine(usething, line, sidenum, use_lower, use_upper);

  // can't use more than one special line in a row
  // -AJA- 1999/09/25: ...unless the line has the ML_PASSTHRU flag
  //       (Boom compatibility).

  return (line->flags & ML_PASSTHRU) ? true : false;
}

//
// P_UseLines
// Looks for special lines in front of the player to activate.
//
void P_UseLines(player_t * player)
{
  int angle;
  float_t x1;
  float_t y1;
  float_t x2;
  float_t y2;

  usething = player->mo;
  use_lower = INT_MIN;
  use_upper = INT_MAX;

  angle = player->mo->angle;

  x1 = player->mo->x;
  y1 = player->mo->y;
  x2 = x1 + USERANGE * M_Cos(angle);
  y2 = y1 + USERANGE * M_Sin(angle);

  P_PathTraverse(x1, y1, x2, y2, PT_ADDLINES | PT_ADDTHINGS, PTR_UseTraverse);
}

//
// RADIUS ATTACK
//
mobj_t *bombsource;
mobj_t *bombspot;
static float_t bombdamage;
boolean_t bombthrust;

//
// PIT_RadiusAttack
// "bombsource" is the creature
// that caused the explosion at "bombspot".
//
static boolean_t PIT_RadiusAttack(mobj_t * thing)
{
  float_t dx;
  float_t dy;
  float_t dist;

  // ignore the bomb spot itself
  if (thing == bombspot)
    return true;

  if (!(thing->flags & MF_SHOOTABLE))
    return true;

  // Boss types take no damage from concussion.
  // -ACB- 1998/06/14 Changed enum reference to extended flag check.
  if (thing->info->extendedflags & EF_BOSSMAN)
    return true;

  dx = fabs(thing->x - bombspot->x);
  dy = fabs(thing->y - bombspot->y);

  dist = dx > dy ? dx : dy;
  dist = (dist - thing->radius);

  if (dist < 0)
    dist = 0;

  if (dist >= bombdamage)
    return true;  // out of range

  if (P_CheckSight(thing, bombspot))
  {
    // must be in direct path
    if (bombthrust)
    {
      if (bombspot)
        P_ThrustMobj(thing, bombspot, bombdamage - dist);
      else if (bombsource)
        P_ThrustMobj(thing, bombsource, bombdamage - dist);
    }
    else
      P_DamageMobj(thing, bombspot, bombsource, bombdamage - dist);
  }

  return true;
}

//
// PIT_SphereAttack
// "bombsource" is the creature
// that caused the explosion at "bombspot".
//
// -ACB- 1998/07/15 New procedure that differs for RadiusAttack - it checks
//                  Height, therefore it is a sphere attack.
//
// -KM- 1998/11/25 Fixed.  Added z movement for rocket jumping.
//
static boolean_t PIT_SphereAttack(mobj_t * thing)
{
  float_t dx;
  float_t dy;
  float_t dz;
  float_t dist;
  float_t hdist;

  // ignore the bomb spot itself
  if (thing == bombspot)
    return true;

  if (!(thing->flags & MF_SHOOTABLE))
    return true;

  // Boss types take no damage from concussion.
  // -ACB- 1998/06/14 Changed enum reference to extended flag check.
  if (thing->info->extendedflags & EF_BOSSMAN)
    return true;

  dx = fabs(thing->x - bombspot->x) - thing->radius;
  dy = fabs(thing->y - bombspot->y) - thing->radius;
  // -KM- 1999/01/31 Use thing->height/2
  dz = fabs(thing->z + thing->height / 2 - bombspot->z) - thing->height / 2;

  hdist = P_ApproxDistance(dx > 0 ? dx : 0, dy > 0 ? dy : 0);
  dist = P_ApproxDistance(hdist, dz > 0 ? dz : 0);

  if (dist < 0)
    dist = 0;

  if (dist >= bombdamage)
    return true;  // out of range

  if (P_CheckSight(thing, bombspot))
  {
    float_t momz = (thing->z + thing->height / 2 - bombspot->z);
    int updown = momz >= 0 ? 1 : -1;

    if (bombthrust)
    {
      if (bombspot)
        P_ThrustMobj(thing, bombspot, bombdamage - dist);
      else if (bombsource)
        P_ThrustMobj(thing, bombsource, bombdamage - dist);
    }
    else
      P_DamageMobj(thing, bombspot, bombsource, bombdamage - dist);

    // -AJA- FIXME: do this in P_ThrustMobj.
    if (fabs(momz) < bombdamage)
    {
      momz = (bombdamage - fabs(momz)) / bombdamage;

      // -ACB- 2000/03/11 Div-by-zero check...
      CHECKVAL(thing->info->mass);

      momz *= (bombdamage - dist) / thing->info->mass;

      thing->momz += updown * momz;
    }
  }
  return true;
}

//
// P_RadiusAttack
// Source is the creature that caused the explosion at spot.
//
void P_RadiusAttack(mobj_t * spot, mobj_t * source, float_t damage)
{
  int x;
  int y;

  int xl;
  int xh;
  int yl;
  int yh;

  float_t dist;

  dist = damage;
  yh = (spot->y + dist - bmaporgy) / MAPBLOCKUNITS;
  yl = (spot->y - dist - bmaporgy) / MAPBLOCKUNITS;
  xh = (spot->x + dist - bmaporgx) / MAPBLOCKUNITS;
  xl = (spot->x - dist - bmaporgx) / MAPBLOCKUNITS;

  bombspot = spot;
  bombsource = source;
  bombdamage = damage;
  bombthrust = false;

  // -ACB- 1998/07/15 This normally does damage to everything within
  //                  a radius regards of height, however true 3D uses
  //                  a sphere attack, which checks height.
  if (level_flags.true3dgameplay)
  {
    for (y = yl; y <= yh; y++)
      for (x = xl; x <= xh; x++)
        P_BlockThingsIterator(x, y, PIT_SphereAttack);
  }
  else
  {
    for (y = yl; y <= yh; y++)
      for (x = xl; x <= xh; x++)
        P_BlockThingsIterator(x, y, PIT_RadiusAttack);
  }
}

//
// P_RadiusThrust
//
// Source is the creature that caused the thrust at Spot.
//
// -AJA- 1999/11/06: written.

void P_RadiusThrust(mobj_t * spot, mobj_t * source, float_t thrust)
{
  int x;
  int y;

  int xl;
  int xh;
  int yl;
  int yh;

  float_t dist;

  dist = thrust;
  yh = (spot->y + dist - bmaporgy) / MAPBLOCKUNITS;
  yl = (spot->y - dist - bmaporgy) / MAPBLOCKUNITS;
  xh = (spot->x + dist - bmaporgx) / MAPBLOCKUNITS;
  xl = (spot->x - dist - bmaporgx) / MAPBLOCKUNITS;

  bombspot = spot;
  bombsource = source;
  bombdamage = thrust;
  bombthrust = true;

  // -ACB- 1998/07/15 This normally does damage to everything within
  //                  a radius regards of height, however true 3D uses
  //                  a sphere attack, which checks height.
  if (level_flags.true3dgameplay)
  {
    for (y = yl; y <= yh; y++)
      for (x = xl; x <= xh; x++)
        P_BlockThingsIterator(x, y, PIT_SphereAttack);
  }
  else
  {
    for (y = yl; y <= yh; y++)
      for (x = xl; x <= xh; x++)
        P_BlockThingsIterator(x, y, PIT_RadiusAttack);
  }
}

//
// SECTOR HEIGHT CHANGING
// After modifying a sectors floor or ceiling height,
// call this routine to adjust the positions
// of all things that touch the sector.
//
// If anything doesn't fit anymore, true will be returned.
//
// If crunch is true, they will take damage as they are being crushed.
//
// If crunch is false, you should set the sector height back the way
//   it was and call P_ChangeSector again to undo the changes.

static boolean_t crushchange;
static boolean_t nofit;
static int crush_time;
static int crush_damage;

//
// P_UpdateMultipleFloors
//
// Call this whenever any heights in a sector (including any extra floor
// heights) are altered.
// 
// -AJA- 1999/07/20: Externalised this.

void P_UpdateMultipleFloors(sector_t * sector)
{
  int i;
  vert_region_t *reg;

  // Firstly, recheck all the regions, making sure they are all valid.
  // When they are invalid, we must recompute the regions.
  
  for (reg=sector->regions; reg; reg=reg->sec_next)
  {
    if (reg->ceil->h < reg->floor->h ||
        (reg->sec_next && reg->ceil->h  >= reg->sec_next->floor->h) ||
        (!reg->sec_next && reg->ceil->h >= sector->ceil.h) ||
        (reg->sec_prev && reg->floor->h <= reg->sec_prev->ceil->h) ||
        (!reg->sec_prev && reg->floor->h <= sector->floor.h))
    {
      // nope, not valid
      P_ComputeSectorRegions(sector);
      break;
    }
  }

  // secondly, recompute the line gaps
  for (i = 0; i < sector->linecount; i++)
  {
    P_ComputeGaps(sector->lines[i]);
  }
}

//
// PIT_ChangeSector
//
static boolean_t PIT_ChangeSector(mobj_t * thing)
{
  mobj_t *mo;

  if (P_ThingHeightClip(thing))
  {
    // keep checking
    return true;
  }

  // crunch bodies to giblets
  if (thing->health <= 0)
  {
    // -AJA- FIXME: don't use states not belonging to us
    if (thing->info->gib->spawn_state)
      P_SetMobjState(thing, thing->info->gib->spawn_state);

    thing->flags &= ~MF_SOLID;  // just been crushed, isn't solid.

    thing->height = 0;
    thing->radius = 0;
    return true;
  }

  // dropped items get removed by a falling ceiling
  if (thing->flags & MF_DROPPED)
  {
    P_RemoveMobj(thing);
    return true;
  }

  // if thing is not shootable, can't be crushed
  if (!(thing->flags & MF_SHOOTABLE) ||
      (thing->flags & MF_NOCLIP))
    return true;

  nofit = true;

  if (crushchange && !(leveltime % crush_time))
  {
    P_DamageMobj(thing, NULL, NULL, crush_damage);

    // spray blood in a random direction
    mo = P_MobjCreateObject(thing->x, thing->y,
        thing->z + thing->height / 2,
        thing->info->blood);

    mo->momx = (M_Random() - 128) / 4.0;
    mo->momy = (M_Random() - 128) / 4.0;
  }

  // keep checking (crush other things) 
  return true;
}

//
// ChangeSector2
//
static void DoChangeSector(sector_t * sector)
{
  int x;
  int y;

  // -AJA- 1999/07/19: Recompute all the line gaps.
  // -AJA- 1999/11/23: If this sector has multiple floors, handle
  //       possible overlapping or invalid combinations.
  P_UpdateMultipleFloors(sector);
  
  // -AJA- 2000/01/02: DDF-itisation of crushers.
  crush_time   = 4;
  crush_damage = 10;

  if (sector->p.special)
  {
    crush_time   = sector->p.special->crush_time;
    crush_damage = sector->p.special->crush_damage;
  }

  // re-check heights for all things near the moving sector
  for (x = sector->blockbox[BOXLEFT]; x <= sector->blockbox[BOXRIGHT]; x++)
    for (y = sector->blockbox[BOXBOTTOM]; y <= sector->blockbox[BOXTOP]; y++)
      P_BlockThingsIterator(x, y, PIT_ChangeSector);
}

boolean_t P_ChangeSector(sector_t * sector, boolean_t crunch)
{
  nofit = false;
  crushchange = crunch;

  DoChangeSector(sector);

  // -AJA- 1999/07/23: If this sector is a control sector, then
  // updated all the things & gaps in the controlled sectors.

  if (sector->controller)
  {
    int i;
    sector_t *slave;

    for (i=0; i < sector->linecount; i++)
    {
      if (!sector->lines[i]->special || !sector->lines[i]->tag ||
          !sector->lines[i]->special->ef.type)
        continue;

      for (slave = P_FindSectorFromTag(sector->lines[i]->tag); 
           slave != NULL; slave = slave->tag_next)
      {
        DoChangeSector(slave);
      }
    }
  }
    
  return nofit;
}

//
// PIT_CorpseCheck
//
// Detect a corpse that could be raised.
//
// Based upon PIT_VileCheck: checks for any corpse within thing's radius.
//
// -ACB- 1998/08/22

static mobj_t *corpsehit;
static mobj_t *raiserobj;
static float_t raisertryx;
static float_t raisertryy;

static boolean_t PIT_CorpseCheck(mobj_t * thing)
{
  float_t maxdist;
  float_t oldradius;
  float_t oldheight;
  boolean_t check;

  if (!(thing->flags & MF_CORPSE))
    return true;  // not a corpse

  if (thing->tics != -1)
    return true;  // not lying still yet

  if (thing->info->raise_state == S_NULL)
    return true;  // monster doesn't have a raise state

  // -KM- 1998/12/21 Monster can't be resurrected.
  if (thing->info->extendedflags & EF_NORESURRECT)
    return true;

  // -ACB- 1998/08/06 Use raiserobj for radius info
  maxdist = thing->info->radius + raiserobj->info->radius;

  if (fabs(thing->x - raisertryx) > maxdist || fabs(thing->y - raisertryy) > maxdist)
    return true;  // not actually touching

  corpsehit = thing;
  corpsehit->momx = corpsehit->momy = 0;

  oldradius = corpsehit->radius;
  oldheight = corpsehit->height;

  // -ACB- 1998/08/22 Check making sure with have the correct radius & height.
  corpsehit->radius = corpsehit->info->radius;
  corpsehit->height = corpsehit->info->height;

  check = P_CheckAbsPosition(corpsehit, corpsehit->x, corpsehit->y, corpsehit->z);

  // -ACB- 1998/08/22 Restore radius & height: we are only checking.
  corpsehit->radius = oldradius;
  corpsehit->height = oldheight;

  if (!check)
    return true;  // doesn't fit here

  // got one, so stop checking
  return false;
}

//
// P_MapFindCorpse
//
// Used to detect corpses that have a raise state and therefore can be
// raised. Arch-Viles (Raisers in-general) use this procedure to pick
// their corpse. NULL is returned if no corpse is found, if one is
// found it is returned.
//
// -ACB- 1998/08/22

mobj_t *P_MapFindCorpse(mobj_t * thing)
{
  int xlow, xhigh, xcount, ylow, yhigh, ycount;

  if (thing->movedir != DI_NODIR)
  {
    raiserobj = thing;

    // check for corpses to raise
    raisertryx = thing->x + thing->speed * xspeed[thing->movedir];
    raisertryy = thing->y + thing->speed * yspeed[thing->movedir];

    xlow = (raisertryx - bmaporgx - MAXRADIUS * 2) / MAPBLOCKUNITS;
    xhigh = (raisertryx - bmaporgx + MAXRADIUS * 2) / MAPBLOCKUNITS;
    ylow = (raisertryy - bmaporgy - MAXRADIUS * 2) / MAPBLOCKUNITS;
    yhigh = (raisertryy - bmaporgy + MAXRADIUS * 2) / MAPBLOCKUNITS;

    for (xcount = xlow; xcount <= xhigh; xcount++)
      for (ycount = ylow; ycount <= yhigh; ycount++)
        if (!P_BlockThingsIterator(xcount, ycount, PIT_CorpseCheck))
          return corpsehit;  // got one - return it
  }

  return NULL;
}

//
// PIT_CheckBlockingLine
//
// Used for checking that any movement between one set of coordinates does not cross
// blocking lines. If the line is twosided and has no restrictions, the move is
// allowed; the next check is to check the respective bounding boxes, see if any
// contact is made and the check is made to see if the objects are on different
// sides of the line.
//
// -ACB- 1998/08/23
//
// -AJA- 1999/09/30: Updated for extra floors.
//
static boolean_t crosser;

// Moving Object x,y cordinates
// for object one and object two.

static float_t mx1;
static float_t my1;
static float_t mx2;
static float_t my2;

// spawn object base
static float_t mb2;

// spawn object top
static float_t mt2;

static boolean_t PIT_CheckBlockingLine(line_t * line)
{
  int i;

  // -KM- 1999/01/31 Changed &&s to ||s.  This condition actually does something
  //  now.
  if (tmbbox[BOXRIGHT] <= line->bbox[BOXLEFT] ||
      tmbbox[BOXLEFT] >= line->bbox[BOXRIGHT] ||
      tmbbox[BOXTOP] <= line->bbox[BOXBOTTOM] ||
      tmbbox[BOXBOTTOM] >= line->bbox[BOXTOP])
  {
    // no possible contact made between the respective bounding boxes.
    return true;
  }

  // if the result is the same, we haven't crossed the line.
  if (P_PointOnLineSide(mx1, my1, line) == P_PointOnLineSide(mx2, my2, line))
    return true;

  // -KM- 1999/01/31 Save ceilingline for bounce.
  if (!crosser && (line->flags & (ML_BLOCKING | ML_BLOCKMONSTERS)))
  {
    blockline = line;
    return false;
  }

  if (!(line->flags & ML_TWOSIDED) || line->gap_num == 0)
  {
    blockline = line;
    return false;
  }

  for (i = 0; i < line->gap_num; i++)
  {
    // gap with no restriction ?
    if (line->gaps[i].f <= mb2 && mt2 <= line->gaps[i].c)
      return true;
  }

  // stop checking, objects are on different sides of a blocking line
  blockline = line;
  return false;
}

//
// P_MapCheckBlockingLine
//
// Checks for a blocking line between thing and the spawnthing coordinates
// given. Return true if there is a line; blockable indicates whether or not
// whether the ML_BLOCKING & ML_BLOCKMONSTERS should be ignored or not.
//
// -ACB- 1998/08/23

boolean_t P_MapCheckBlockingLine(mobj_t * thing, mobj_t * spawnthing)
{
  int xlow, xhigh, xcount, ylow, yhigh, ycount;

  mx1 = thing->x;
  my1 = thing->y;
  mx2 = spawnthing->x;
  my2 = spawnthing->y;
  mb2 = spawnthing->z;
  mt2 = spawnthing->z + spawnthing->height;

  crosser = (spawnthing->extendedflags & EF_CROSSLINES);

  tmbbox[BOXLEFT] = mx1 < mx2 ? mx1 : mx2;
  tmbbox[BOXRIGHT] = mx1 > mx2 ? mx1 : mx2;
  tmbbox[BOXBOTTOM] = my1 < my2 ? my1 : my2;
  tmbbox[BOXTOP] = my1 > my2 ? my1 : my2;

  xlow = (tmbbox[BOXLEFT] - bmaporgx) / MAPBLOCKUNITS;
  xhigh = (tmbbox[BOXRIGHT] - bmaporgx) / MAPBLOCKUNITS;
  ylow = (tmbbox[BOXBOTTOM] - bmaporgy) / MAPBLOCKUNITS;
  yhigh = (tmbbox[BOXTOP] - bmaporgy) / MAPBLOCKUNITS;

  validcount++;

  mobj_hit_sky = false;
  blockline = NULL;

  for (xcount = xlow; xcount <= xhigh; xcount++)
    for (ycount = ylow; ycount <= yhigh; ycount++)
      if (!P_BlockLinesIterator(xcount, ycount, PIT_CheckBlockingLine))
        return true;

  return false;
}

void P_MapInit(void)
{
  Z_InitStackArray(&spechit_a, (void***)&spechit, sizeof(line_t *), -8);
}

