//----------------------------------------------------------------------------
//  EDGE Movement, Collision & Blockmap utility functions
//----------------------------------------------------------------------------
// 
//  Copyright (c) 1999-2000  The EDGE Team.
// 
//  This program is free software; you can redistribute it and/or
//  modify it under the terms of the GNU General Public License
//  as published by the Free Software Foundation; either version 2
//  of the License, or (at your option) any later version.
//
//  This program is distributed in the hope that it will be useful,
//  but WITHOUT ANY WARRANTY; without even the implied warranty of
//  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//  GNU General Public License for more details.
//
//----------------------------------------------------------------------------
//
//  Based on the DOOM source code, released by Id Software under the
//  following copyright:
//
//    Copyright (C) 1993-1996 by id Software, Inc.
//
//----------------------------------------------------------------------------
//
// DESCRIPTION:
//   Movement/collision utility functions,
//   as used by function in p_map.c. 
//   BLOCKMAP Iterator functions,
//   and some PIT_* functions to use for iteration.
//   Gap/extrafloor utility functions.
//

#include "i_defs.h"

#include "m_bbox.h"
#include "dm_defs.h"
#include "p_local.h"

#include "z_zone.h"

// State.
#include "r_state.h"

//
// P_ApproxDistance
// Gives an estimation of distance (not exact)
//

float_t P_ApproxDistance(float_t dx, float_t dy)
{
  dx = fabs(dx);

  dy = fabs(dy);

  if (dx < dy)
    return dx / 2 + dy;

  return dx + dy / 2;
}

//
// P_ApproxSlope
// Gives an estimation of slope (not exact)
//
// -AJA- 1999/09/11: written.

float_t P_ApproxSlope(float_t dx, float_t dy, float_t dz)
{
  float_t dist = P_ApproxDistance(dx, dy);

  // kludge to prevent overflow or division by zero.
  if (dist < 1.0 / 32)
    dist = 1.0 / 32;

  return dz / dist;
}

//
// P_PointOnLineSide
// Returns 0 or 1
//
int P_PointOnLineSide(float_t x, float_t y, line_t * line)
{
  float_t dx;
  float_t dy;
  float_t left;
  float_t right;

  if (line->dx == 0)
  {
    if (x <= line->v1->x)
      return line->dy > 0;

    return line->dy < 0;
  }
  if (line->dy == 0)
  {
    if (y <= line->v1->y)
      return line->dx < 0;

    return line->dx > 0;
  }

  dx = x - line->v1->x;
  dy = y - line->v1->y;

  left = line->dy * dx;
  right = dy * line->dx;

  if (right < left)
    return 0;  // front side

  return 1;  // back side

}

//
// P_BoxOnLineSide
// Considers the line to be infinite
// Returns side 0 or 1, -1 if box crosses the line.
//
int P_BoxOnLineSide(float_t * tmbox, line_t * ld)
{
  int p1 = 0;
  int p2 = 0;

  switch (ld->slopetype)
  {
    case ST_HORIZONTAL:
      p1 = tmbox[BOXTOP] > ld->v1->y;
      p2 = tmbox[BOXBOTTOM] > ld->v1->y;
      if (ld->dx < 0)
      {
        p1 ^= 1;
        p2 ^= 1;
      }
      break;

    case ST_VERTICAL:
      p1 = tmbox[BOXRIGHT] < ld->v1->x;
      p2 = tmbox[BOXLEFT] < ld->v1->x;
      if (ld->dy < 0)
      {
        p1 ^= 1;
        p2 ^= 1;
      }
      break;

    case ST_POSITIVE:
      p1 = P_PointOnLineSide(tmbox[BOXLEFT], tmbox[BOXTOP], ld);
      p2 = P_PointOnLineSide(tmbox[BOXRIGHT], tmbox[BOXBOTTOM], ld);
      break;

    case ST_NEGATIVE:
      p1 = P_PointOnLineSide(tmbox[BOXRIGHT], tmbox[BOXTOP], ld);
      p2 = P_PointOnLineSide(tmbox[BOXLEFT], tmbox[BOXBOTTOM], ld);
      break;
  }

  if (p1 == p2)
    return p1;
  return -1;
}

//
// P_PointOnDivlineSide
// Returns 0 or 1.
//
int P_PointOnDivlineSide(float_t x, float_t y, divline_t * line)
{
  float_t dx;
  float_t dy;
  float_t left;
  float_t right;

  if (line->dx == 0)
  {
    if (x <= line->x)
      return line->dy > 0;

    return line->dy < 0;
  }
  if (line->dy == 0)
  {
    if (y <= line->y)
      return line->dx < 0;

    return line->dx > 0;
  }

  dx = x - line->x;
  dy = y - line->y;

  // try to quickly decide by looking at sign bits
  if ((line->dy < 0) ^ (line->dx < 0) ^ (dx < 0) ^ (dy < 0))
  {
    if ((line->dy < 0) ^ (dx < 0))
      return 1;  // (left is negative)

    return 0;
  }

  left = line->dy * dx;
  right = dy * line->dx;

  if (right < left)
    return 0;  // front side

  return 1;  // back side

}

//
// P_MakeDivline
//
void P_MakeDivline(line_t * li, divline_t * dl)
{
  dl->x = li->v1->x;
  dl->y = li->v1->y;
  dl->dx = li->dx;
  dl->dy = li->dy;
}

//
// P_InterceptVector
//
// Returns the fractional intercept point along the first divline.
// This is only called by the addthings and addlines traversers.

float_t P_InterceptVector(divline_t * v2, divline_t * v1)
{
  float_t frac;
  float_t num;
  float_t den;
  float_t v1x;
  float_t v1y;
  float_t v1dx;
  float_t v1dy;
  float_t v2x;
  float_t v2y;
  float_t v2dx;
  float_t v2dy;

  v1x = v1->x;
  v1y = v1->y;
  v1dx = v1->dx;
  v1dy = v1->dy;
  v2x = v2->x;
  v2y = v2->y;
  v2dx = v2->dx;
  v2dy = v2->dy;

  den = v1dy * v2dx - v1dx * v2dy;

  if (den == 0)
    return 0;  // parallel

  num = (v1x - v2x) * v1dy + (v2y - v1y) * v1dx;
  frac = num / den;

  return frac;
}

//
//  GAP UTILITY FUNCTIONS
//

static int GAP_RemoveSolid(vgap_t * dest, int d_num, float_t z1, float_t z2)
{
  int d;
  int new_num = 0;
  vgap_t new_gaps[MAXOPENGAPS * 2];

#ifdef DEVELOPERS
  if (z1 > z2)
    I_Error("RemoveSolid: z1 > z2");
#endif

  for (d = 0; d < d_num; d++)
  {
    if (dest[d].c <= dest[d].f)
      continue;  // ignore empty gaps.

    if (z1 <= dest[d].f && z2 >= dest[d].c)
      continue;  // completely blocks it.

    if (z1 >= dest[d].c || z2 <= dest[d].f)
    {
      // no intersection.

      new_gaps[new_num].f = dest[d].f;
      new_gaps[new_num].c = dest[d].c;
      new_num++;
      continue;
    }

    // partial intersections.

    if (z1 > dest[d].f)
    {
      new_gaps[new_num].f = dest[d].f;
      new_gaps[new_num].c = z1;
      new_num++;
    }

    if (z2 < dest[d].c)
    {
      new_gaps[new_num].f = z2;
      new_gaps[new_num].c = dest[d].c;
      new_num++;
    }
  }

  Z_MoveData(dest, new_gaps, vgap_t, new_num);

  return new_num;
}

int GAP_Construct(vgap_t * gaps, sector_t * sec, mobj_t * thing)
{
  vert_region_t * reg;

  int num = 1;

  // early out for FUBAR sectors
  if (sec->floor.h >= sec->ceil.h)
    return 0;

  gaps[0].f = sec->floor.h;
  gaps[0].c = sec->ceil.h;

  for (reg = sec->regions; reg && reg->sec_next; reg = reg->sec_next)
  {
     boolean_t fall_thru = (reg->extrafloor && 
         (reg->extrafloor->type & EXFL_Liquid));
 
     boolean_t walk_water = (thing != NULL) &&
         (thing->extendedflags & EF_WATERWALKER) &&
         reg->extrafloor && (reg->extrafloor->type & EXFL_Water);

     // ignore FUBAR regions
     if (reg->floor->h > reg->ceil->h)
       continue;
     
     if (!fall_thru || walk_water)
     {
       num = GAP_RemoveSolid(gaps, num, reg->ceil->h,
           reg->sec_next->floor->h);
     }
  }
  
  return num;
}

int GAP_Constrict(vgap_t * dest, int d_num, vgap_t * src, int s_num)
{
  int d, s;
  float_t f, c;

  int new_num = 0;
  vgap_t new_gaps[32];

  for (s = 0; s < s_num; s++)
  {
    // ignore empty gaps.
    if (src[s].c <= src[s].f)
      continue;

    for (d = 0; d < d_num; d++)
    {
      // ignore empty gaps.
      if (dest[d].c <= dest[d].f)
        continue;

      f = MAX(src[s].f, dest[d].f);
      c = MIN(src[s].c, dest[d].c);

      if (f < c)
      {
        new_gaps[new_num].c = c;
        new_gaps[new_num].f = f;
        new_num++;
      }
    }
  }

  Z_MoveData(dest, new_gaps, vgap_t, new_num);

  return new_num;
}

//
// P_FindThingGap
//
// Find the best gap that the thing could fit in, given a certain Z
// position (z1 is foot, z2 is head).  Assuming at least two gaps exist,
// the best gap is chosen as follows:
//
// 1. if the thing fits in one of the gaps without moving vertically,
//    then choose that gap.
//
// 2. if there is only *one* gap which the thing could fit in, then
//    choose that gap.
//
// 3. if there is multiple gaps which the thing could fit in, choose
//    the gap whose floor is closest to the thing's current Z.
//
// 4. if there is no gaps which the thing could fit in, do the same.
//
// Returns the gap number, or -1 if there are no gaps at all.

int P_FindThingGap(vgap_t * gaps, int gap_num, float_t z1, float_t z2)
{
  int i;
  float_t dist;

  int fit_num = 0;
  int fit_last = -1;

  int fit_closest = -1;
  float_t fit_mindist = INT_MAX;

  int nofit_closest = -1;
  float_t nofit_mindist = INT_MAX;

  // check for trivial gaps...

  if (gap_num == 0)
  {
    return -1;
  }
  else if (gap_num == 1)
  {
    return 0;
  }

  // There are 2 or more gaps. Now it gets interesting :-)

  for (i = 0; i < gap_num; i++)
  {
    if (z1 >= gaps[i].f && z2 <= gaps[i].c)
    { // [1]
      return i;
    }

    dist = ABS(z1 - gaps[i].f);

    if (z2 - z1 <= gaps[i].c - gaps[i].f)
    { // [2]
      fit_num++;

      fit_last = i;
      if (dist < fit_mindist)
      { // [3]
        fit_mindist = dist;
        fit_closest = i;
      }
    }
    else
    {
      if (dist < nofit_mindist)
      { // [4]
        nofit_mindist = dist;
        nofit_closest = i;
      }
    }
  }

  if (fit_num == 1)
    return fit_last;
  else if (fit_num > 1)
    return fit_closest;
  else
    return nofit_closest;
}

//
// P_ComputeThingGap
//
// Determine the initial floorz and ceilingz that a thing placed at a
// particular Z would have.  Returns the nominal Z height.  Some special
// values of Z are recognised: ONFLOORZ & ONCEILINGZ.

float_t P_ComputeThingGap(mobj_t * thing, sector_t * sec, float_t z,
    float_t * f, float_t * c)
{
  vgap_t temp_gaps[MAXOPENGAPS];
  int temp_num;

  temp_num = GAP_Construct(temp_gaps, sec, thing);

  if (z == ONFLOORZ)
    z = sec->floor.h;

  if (z == ONCEILINGZ)
    z = sec->ceil.h - thing->height;

  temp_num = P_FindThingGap(temp_gaps, temp_num, z, z + thing->height);

  if (temp_num < 0)
  {
    // thing is stuck in a closed door.
    *f = *c = sec->floor.h;
    return *f;
  }

  *f = temp_gaps[temp_num].f;
  *c = temp_gaps[temp_num].c;

  return z;
}

//
// P_ComputeGaps
//
// Determine the gaps between the front & back sectors of the line,
// taking into account any extra floors.
//
// -AJA- 1999/07/19: This replaces P_LineOpening.

void P_ComputeGaps(line_t * ld)
{
  sector_t *front = ld->frontsector;
  sector_t *back  = ld->backsector;

  int temp_num;
  vgap_t temp_gaps[MAXOPENGAPS];

  ld->gap_num = 0;

  if (!front || !back)
  {
    // single sided line
    return;
  }

  if (front->ceil.h <= front->floor.h ||
      back->ceil.h <= back->floor.h ||
      front->ceil.h <= back->floor.h ||
      back->ceil.h <= front->floor.h)
  {
    // closed door.
    return;
  }
 
  ld->gap_num = GAP_Construct(ld->gaps, front, NULL);
  temp_num = GAP_Construct(temp_gaps, back, NULL);

  ld->gap_num = GAP_Constrict(ld->gaps, ld->gap_num, temp_gaps, temp_num);
}

//
// P_ExtraFloorFits
//
// Check if a solid extrafloor fits, error out if it doesn't.

static void ExtraFloorFits(sector_t *sec, float_t z1, float_t z2)
{
  vert_region_t *reg;

  if (z2 > sec->ceil.h)
  {
    I_Error("Extrafloor with z range of %1.0f / %1.0f is stuck "
        "in sector #%d's ceiling.\n",
        z1, z2, (int)(sec - sectors));
  }

  if (z1 < sec->floor.h)
  {
    I_Error("Extrafloor with z range of %1.0f / %1.0f is stuck "
        "in sector #%d's floor.\n",
        z1, z2, (int)(sec - sectors));
  }

  for (reg=sec->regions; reg && reg->sec_next; reg=reg->sec_next)
  {
    float_t bottom = reg->ceil->h;
    float_t top = reg->sec_next->floor->h;

    DEV_ASSERT2(top >= bottom);

    if (!reg->extrafloor || (reg->extrafloor->type & EXFL_Liquid))
      continue;
    
    // here is another solid extrafloor, check for overlap
    
    if (z2 <= bottom || z1 >= top)
      continue;
    
    I_Error("Extrafloor with z range of %1.0f / %1.0f is stuck "
        "in sector #%d in another extrafloor.\n",
        z1, z2, (int)(sec - sectors));
  }
}


//
// P_ComputeSectorRegions
//
// Use the extrafloor records of the sector to recompute the regions
// of the sector.  Apart from initial setup, this function only needs
// to be called whenever a region in the sector becomes invalid (e.g.
// water sinking below the floor).
//
// -AJA- 1999/11/23: written.

#define GET_NEW_REGION()  \
  if (spare)  \
  {  \
    new_reg = spare;  \
    spare = spare->sec_next;  \
  }  \
  else  \
  {  \
    new_reg = Z_New(vert_region_t, 1);  \
  }  \
  Z_Clear(new_reg, vert_region_t, 1);

void P_ComputeSectorRegions(sector_t *sector)
{
  vert_region_t *spare = sector->regions;
  vert_region_t *cur, *new_reg, *next;
  extrafloor_record_t *rec;

  plane_info_t *top;
  plane_info_t *bottom;

  sector->regions = NULL;

  // handle plain sector
  GET_NEW_REGION();

  new_reg->floor = &sector->floor;
  new_reg->ceil  = &sector->ceil;
  new_reg->p     = &sector->p;

  sector->regions = new_reg;

  // early out if sector is FUBAR, e.g. has floor >= ceiling
  if (sector->floor.h < sector->ceil.h)
  {
    for (rec=sector->extrafloor_records; rec; rec=rec->next)
    {
      // get control sector
      sector_t *ctrl = rec->line->frontsector;
      int liquid = (rec->info->type & EXFL_Liquid);

      bottom = &ctrl->floor;
      top = (rec->info->type & EXFL_Thick) ? &ctrl->ceil : bottom;

      // ignore control sectors that are FUBAR, i.e. floor > ceiling
      if (top->h < bottom->h)
        continue;
      
      // check for overlap
      if (!liquid)
        ExtraFloorFits(sector, bottom->h, top->h);

      // compare this extrafloor against all current regions, updating
      // them accordingly.

      for (cur=sector->regions; cur; cur=next)
      {
        float_t f1 = cur->floor->h;
        float_t c1 = cur->ceil->h;
        
        next = cur->sec_next;

        // region totally above liquid ?
        if (f1 > top->h)
          break;

        // region totally below liquid ?
        if (c1 < bottom->h)
        {
          // for flooders, apply properties
          // FIXME: don't flood already flooded areas
          if (rec->info->type & EXFL_Flooder)
          {
            cur->p = &ctrl->p;
          }
          continue;
        }

        // OK, we need to split the current region into two halves.  The
        // new region will be the one _above_ the current one.

        GET_NEW_REGION();

        new_reg->floor = top;
        new_reg->ceil  = cur->ceil;
        new_reg->p     = cur->p;
        new_reg->extrafloor = cur->extrafloor;
        new_reg->extraside  = cur->extraside;
        
        // cur->floor remains the same
        cur->ceil = bottom;
        cur->p = &ctrl->p;
        cur->extrafloor = rec->info;
        cur->extraside  = &rec->line->side[0]->middle;

#ifdef DEVELOPERS
        // more sanity checks (shouldn't happen)
        if (cur->ceil->h < cur->floor->h)
          I_Error("INTERNAL ERROR: bad cur floor/ceil for sector region.\n");
        if (new_reg->ceil->h < new_reg->floor->h)
          I_Error("INTERNAL ERROR: bad new_reg floor/ceil for sector region.\n");
#endif 
        // link in the newbie

        new_reg->sec_next = cur->sec_next;
        new_reg->sec_prev = cur;

        if (cur->sec_next)
          cur->sec_next->sec_prev = new_reg;
        
        cur->sec_next = new_reg;
      }
    }
  }

  // free any left-over vert_regions
  while (spare)
  {
    cur = spare;
    spare = spare->sec_next;

    Z_Free(cur);
  }
}

//
//  THING POSITION SETTING
//

//
// P_UnsetThingPosition
//
// Unlinks a thing from block map and subsector.
// On each position change, BLOCKMAP and other
// lookups maintaining lists ot things inside
// these structures need to be updated.
//
// -ES- 1999/12/04 Better error checking: Clear prev/next fields.
// This catches errors which can occur if the position is unset twice.

void P_UnsetThingPosition(mobj_t * thing)
{
  int blockx;
  int blocky;
  int bnum;

//  I_WriteDebug("Attempt to Unlink thing %d known as %s to sector: %d\n",thing,thing->info->name,thing->subsector);

  if (!(thing->flags & MF_NOSECTOR))
  {
    // unlink from subsector
    // (inert things don't need to be in subsector list)

    if (thing->snext)
    {
#ifdef DEVELOPERS
      if (thing->snext->sprev != thing)
        I_Error("INTERNAL ERROR: Bad subsector NEXT link in thing.\n");
#endif
      thing->snext->sprev = thing->sprev;
    }

    if (thing->sprev)
    {
#ifdef DEVELOPERS
      if (thing->sprev->snext != thing)
        I_Error("INTERNAL ERROR: Bad subsector PREV link in thing.\n");
#endif
      thing->sprev->snext = thing->snext;
    }
    else
    {
#ifdef DEVELOPERS
      if (thing->subsector->thinglist != thing)
        I_Error("INTERNAL ERROR: Bad subsector link (HEAD) in thing, possibly double P_UnsetThingPosition call.\n");
#endif
      thing->subsector->thinglist = thing->snext;
    }

    thing->snext = NULL;
    thing->sprev = NULL;
  }

  if (!(thing->flags & MF_NOBLOCKMAP))
  {
    // inert things don't need to be in blockmap
    // unlink from block map
    if (thing->bnext)
    {
#ifdef DEVELOPERS
      if (thing->bnext->bprev != thing)
        I_Error("INTERNAL ERROR: Bad block NEXT link in thing.\n");
#endif
      thing->bnext->bprev = thing->bprev;
    }

    if (thing->bprev)
    {
#ifdef DEVELOPERS
      if (thing->bprev->bnext != thing)
        I_Error("INTERNAL ERROR: Bad block PREV link in thing.\n");
#endif
      thing->bprev->bnext = thing->bnext;
    }
    else
    {
      blockx = (thing->x - bmaporgx) / MAPBLOCKUNITS;
      blocky = (thing->y - bmaporgy) / MAPBLOCKUNITS;

      if (blockx >= 0 && blockx < bmapwidth &&
          blocky >= 0 && blocky < bmapheight)
      {
        bnum = blocky * bmapwidth + blockx;
#ifdef DEVELOPERS
        if (blocklinks[bnum] != thing)
          I_Error("INTERNAL ERROR: Bad block link (HEAD) in thing.\n");
#endif
        blocklinks[bnum] = thing->bnext;
      }
    }

    thing->bprev = NULL;
    thing->bnext = NULL;
  }
}

//
// P_SetThingPosition
//
// Links a thing into both a block and a subsector
// based on it's x y.
//
// Sets thing->subsector properly

void P_SetThingPosition(mobj_t * thing)
{
  subsector_t *ss;
  int blockx;
  int blocky;
  int bnum;

  // -ES- 1999/12/04 The position must be unset before it's set again.
#ifdef DEVELOPERS
  if (thing->snext || thing->sprev || thing->bnext || thing->bprev)
    I_Error("INTERNAL ERROR: Double P_SetThingPosition call.");
#endif

  // link into subsector
  ss = R_PointInSubsector(thing->x, thing->y);
  thing->subsector = ss;

  if (! (thing->flags & MF_NOSECTOR))
  {
    thing->snext = ss->thinglist;
    thing->sprev = NULL;

    if (ss->thinglist)
      ss->thinglist->sprev = thing;

    ss->thinglist = thing;
  }

  // link into blockmap
  if (!(thing->flags & MF_NOBLOCKMAP))
  {
    blockx = (thing->x - bmaporgx) / MAPBLOCKUNITS;
    blocky = (thing->y - bmaporgy) / MAPBLOCKUNITS;

    if (blockx >= 0 && blockx < bmapwidth &&
        blocky >= 0 && blocky < bmapheight)
    {
      bnum = blocky * bmapwidth + blockx;

      thing->bprev = NULL;
      thing->bnext = blocklinks[bnum];

      if (blocklinks[bnum])
        (blocklinks[bnum])->bprev = thing;

      blocklinks[bnum] = thing;
    }
    else
    {
      // thing is off the map
      thing->bnext = thing->bprev = NULL;
    }
  }

  //I_WriteDebug("Linked thing %d known as %s to sector: %d\n",thing,thing->info->name,ss);
}

//
// BLOCK MAP ITERATORS
// For each line/thing in the given mapblock,
// call the passed PIT_* function.
// If the function returns false,
// exit with false without checking anything else.
//

//
// P_BlockLinesIterator
// The validcount flags are used to avoid checking lines
// that are marked in multiple mapblocks,
// so increment validcount before the first call
// to P_BlockLinesIterator, then make one or more calls
// to it.
//
// 23-6-98 KM Changed to reflect blockmap is now int* not short*

boolean_t P_BlockLinesIterator (int x, int y,
    boolean_t(*func) (line_t *))
{
  int offset;
  int *list;
  line_t *ld;

  if (x < 0 || y < 0 || x >= bmapwidth || y >= bmapheight)
    return true;

  offset = blockmap[y * bmapwidth + x];

  for (list = blockmaplump + offset; *list != -1; list++)
  {
    ld = &lines[*list];

    // has line already been checked ?
    if (ld->validcount == validcount)
      continue;

    ld->validcount = validcount;

    if (!func(ld))
      return false;
  }

  // everything was checked
  return true;
}

//
// P_BlockThingsIterator
//
boolean_t P_BlockThingsIterator(int x, int y, boolean_t(*func) (mobj_t *))
{
  mobj_t *mobj;

  if (x < 0 || y < 0 || x >= bmapwidth || y >= bmapheight)
  {
    return true;
  }

  for (mobj = blocklinks[y * bmapwidth + x]; mobj; mobj = mobj->bnext)
  {
    if (!func(mobj))
      return false;
  }

  return true;
}

//
// INTERCEPT ROUTINES
//
stack_array_t intercept_a;
intercept_t **intercepts = NULL;
int intercept_p;

divline_t trace;
boolean_t earlyout;
int ptflags;

//
// PIT_AddLineIntercepts.
//
// Looks for lines in the given block
// that intercept the given trace
// to add to the intercepts list.
//
// A line is crossed if its endpoints
// are on opposite sides of the trace.
// Returns true if earlyout and a solid line hit.

static boolean_t PIT_AddLineIntercepts(line_t * ld)
{
  int s1;
  int s2;
  float_t frac;
  divline_t dl;

  // avoid precision problems with two routines
  if (trace.dx > 16 || trace.dy > 16 || trace.dx < -16 || trace.dy < -16)
  {
    s1 = P_PointOnDivlineSide(ld->v1->x, ld->v1->y, &trace);
    s2 = P_PointOnDivlineSide(ld->v2->x, ld->v2->y, &trace);
  }
  else
  {
    s1 = P_PointOnLineSide(trace.x, trace.y, ld);
    s2 = P_PointOnLineSide(trace.x + trace.dx, trace.y + trace.dy, ld);
  }

  if (s1 == s2)
    return true;  // line isn't crossed

  // hit the line
  P_MakeDivline(ld, &dl);
  frac = P_InterceptVector(&trace, &dl);

  if (frac < 0)
    return true;  // behind source

  // try to early out the check
  if (earlyout && frac < 1.0 && !ld->backsector)
    return false;  // stop checking

  Z_SetArraySize(&intercept_a, intercept_p + 1);

  intercepts[intercept_p]->frac = frac;
  intercepts[intercept_p]->isaline = true;
  intercepts[intercept_p]->d.line = ld;
  intercept_p++;

  return true;  // continue

}

//
// PIT_AddThingIntercepts
//
static boolean_t PIT_AddThingIntercepts(mobj_t * thing)
{
  float_t x1;
  float_t y1;
  float_t x2;
  float_t y2;

  int s1;
  int s2;

  boolean_t tracepositive;

  divline_t dl;

  float_t frac;

  tracepositive = (trace.dx >= 0) == (trace.dy >= 0);

  // check a corner to corner crossection for hit
  if (tracepositive)
  {
    x1 = thing->x - thing->radius;
    y1 = thing->y + thing->radius;

    x2 = thing->x + thing->radius;
    y2 = thing->y - thing->radius;
  }
  else
  {
    x1 = thing->x - thing->radius;
    y1 = thing->y - thing->radius;

    x2 = thing->x + thing->radius;
    y2 = thing->y + thing->radius;
  }

  s1 = P_PointOnDivlineSide(x1, y1, &trace);
  s2 = P_PointOnDivlineSide(x2, y2, &trace);

  // line isn't crossed ?
  if (s1 == s2)
    return true;

  dl.x = x1;
  dl.y = y1;
  dl.dx = x2 - x1;
  dl.dy = y2 - y1;

  frac = P_InterceptVector(&trace, &dl);

  // behind source ?
  if (frac < 0)
    return true;

  Z_SetArraySize(&intercept_a, intercept_p + 1);

  intercepts[intercept_p]->frac = frac;
  intercepts[intercept_p]->isaline = false;
  intercepts[intercept_p]->d.thing = thing;
  intercept_p++;

  // keep going
  return true;
}

//
// P_TraverseIntercepts
//
// Returns true if the traverser function returns true
// for all lines.

static boolean_t TraverseIntercepts(traverser_t func, float_t maxfrac)
{
  int count;
  float_t dist;
  int scan;
  intercept_t *in = NULL;

  count = intercept_p;

  while (count--)
  {
    dist = INT_MAX;
    for (scan = 0; scan < intercept_p; scan++)
    {
      if (intercepts[scan]->frac < dist)
      {
        dist = intercepts[scan]->frac;
        in = intercepts[scan];
      }
    }

    if (dist > maxfrac)
    {
      // checked everything in range  
      return true;
    }
#if 0  // UNUSED
    {
      // don't check these yet, there may be others inserted
      in = scan = intercepts;
      for (scan = intercepts; scan < &intercepts[intercept_p]; scan++)
        if (scan->frac > maxfrac)
          *in++ = *scan;
      intercept_p = in;
      return false;
    }
#endif

    if (!func(in))
    {
      // don't bother going farther
      return false;
    }

    in->frac = INT_MAX;
  }

  // everything was traversed
  return true;
}

//
// P_PathTraverse
//
// Traces a line from x1,y1 to x2,y2,
// calling the traverser function for each.
// Returns true if the traverser function returns true
// for all lines.

boolean_t P_PathTraverse(float_t x1, float_t y1, float_t x2, float_t y2, 
    int flags, traverser_t trav)
{
  static boolean_t firsttime = true;

  int xt1;
  int yt1;
  int xt2;
  int yt2;

  float_t xstep;
  float_t ystep;

  float_t partial;

  float_t xintercept;
  float_t yintercept;

  double tmp;

  int mapx;
  int mapy;

  int mapxstep;
  int mapystep;

  int count;

  earlyout = flags & PT_EARLYOUT;

  validcount++;
  
  if (firsttime)
  {
    Z_InitStackArray(&intercept_a, (void ***)&intercepts, sizeof(intercept_t), 0);
    firsttime = false;
  }

  intercept_p = 0;

  if (fmod(x1 - bmaporgx, MAPBLOCKUNITS) == 0)
  {
    // don't side exactly on a line
    x1 += 1.0;
  }

  if (fmod(y1 - bmaporgy, MAPBLOCKUNITS) == 0)
  {
    // don't side exactly on a line
    y1 += 1.0;
  }

  trace.x = x1;
  trace.y = y1;
  trace.dx = x2 - x1;
  trace.dy = y2 - y1;

  x1 -= bmaporgx;
  y1 -= bmaporgy;
  x2 -= bmaporgx;
  y2 -= bmaporgy;

  // -AJA- use floor() here, for more precise behaviour.
  // FIXME: is this correct though ?
 
  xt1 = floor(x1 / MAPBLOCKUNITS);
  yt1 = floor(y1 / MAPBLOCKUNITS);
  xt2 = floor(x2 / MAPBLOCKUNITS);
  yt2 = floor(y2 / MAPBLOCKUNITS);

  if (xt2 > xt1)
  {
    mapxstep = 1;
    partial = 1.0 - modf(x1 / MAPBLOCKUNITS, &tmp);

    // -ACB- 2000/03/11 Div-by-zero check...
    CHECKVAL(x2-x1);

    ystep = (y2 - y1) / fabs(x2 - x1);
  }
  else if (xt2 < xt1)
  {
    mapxstep = -1;
    partial = modf(x1 / MAPBLOCKUNITS, &tmp);

    // -ACB- 2000/03/11 Div-by-zero check...
    CHECKVAL(x2-x1);

    ystep = (y2 - y1) / fabs(x2 - x1);
  }
  else
  {
    mapxstep = 0;
    partial = 1.0;
    ystep = 256.0;
  }

  yintercept = y1 / MAPBLOCKUNITS + partial * ystep;

  if (yt2 > yt1)
  {
    mapystep = 1;
    partial = 1.0 - modf(y1 / MAPBLOCKUNITS, &tmp);

    // -ACB- 2000/03/11 Div-by-zero check...
    CHECKVAL(y2-y1);

    xstep = (x2 - x1) / fabs(y2 - y1);
  }
  else if (yt2 < yt1)
  {
    mapystep = -1;
    partial = modf(y1 / MAPBLOCKUNITS, &tmp);

    // -ACB- 2000/03/11 Div-by-zero check...
    CHECKVAL(y2-y1);

    xstep = (x2 - x1) / fabs(y2 - y1);
  }
  else
  {
    mapystep = 0;
    partial = 1.0;
    xstep = 256.0;
  }
  xintercept = x1 / MAPBLOCKUNITS + partial * xstep;

  // Step through map blocks.
  // Count is present to prevent a round off error
  // from skipping the break.
  mapx = xt1;
  mapy = yt1;

  for (count = 0; count < 64; count++)
  {
    if (flags & PT_ADDLINES)
    {
      if (!P_BlockLinesIterator(mapx, mapy, PIT_AddLineIntercepts))
        return false;
    }

    if (flags & PT_ADDTHINGS)
    {
      if (!P_BlockThingsIterator(mapx, mapy, PIT_AddThingIntercepts))
        return false;
    }

    if (mapx == xt2 && mapy == yt2)
      break;

    if (mapy == (int)yintercept)
    {
      yintercept += ystep;
      mapx += mapxstep;
    }
    else if (mapx == (int)xintercept)
    {
      xintercept += xstep;
      mapy += mapystep;
    }
  }

  // go through the sorted list
  return TraverseIntercepts(trav, 1.0);
}
