//----------------------------------------------------------------------------
//  EDGE Moving Object Handling Code
//----------------------------------------------------------------------------
// 
//  Copyright (c) 1999-2000  The EDGE Team.
// 
//  This program is free software; you can redistribute it and/or
//  modify it under the terms of the GNU General Public License
//  as published by the Free Software Foundation; either version 2
//  of the License, or (at your option) any later version.
//
//  This program is distributed in the hope that it will be useful,
//  but WITHOUT ANY WARRANTY; without even the implied warranty of
//  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//  GNU General Public License for more details.
//
//----------------------------------------------------------------------------
//
//  Based on the DOOM source code, released by Id Software under the
//  following copyright:
//
//    Copyright (C) 1993-1996 by id Software, Inc.
//
//----------------------------------------------------------------------------
//
// -MH- 1998/07/02  "shootupdown" --> "true3dgameplay"
//
// -ACB- 1998/07/30 Took an axe to the item respawn code: now uses a
//                  double-linked list to store to individual items;
//                  limit removed; P_MobjItemRespawn replaces P_RespawnSpecials
//                  as the procedure that handles respawning of items.
//
//                  P_NightmareRespawnOld -> P_TeleportRespawn
//                  P_NightmareRespawnNew -> P_ResurrectRespawn
//
// -ACB- 1998/07/31 Use new procedure to handle flying missiles that hammer
//                  into sky-hack walls & ceilings. Also don't explode the
//                  missile if it hits sky-hack ceiling or floor.
//
// -ACB- 1998/08/06 Implemented limitless mobjinfo list, altered/removed all
//                  mobjinfo[] references.
//
// -AJA- 1999/07/21: Replaced some non-critical P_Randoms with M_Random.
//
// -AJA- 1999/07/30: Removed redundant code from P_SpawnMobj (it was
//                   virtually identical to P_MobjCreateObject).
//
// -AJA- 1999/09/15: Removed P_SpawnMobj itself :-).
//

#include "i_defs.h"
#include "p_mobj.h"

#include "con_main.h"
#include "con_cvar.h"
#include "dm_defs.h"
#include "dm_state.h"
#include "g_game.h"
#include "hu_stuff.h"
#include "m_argv.h"
#include "m_inline.h"
#include "m_random.h"
#include "p_local.h"
#include "r_main.h"
#include "s_sound.h"
#include "st_stuff.h"
#include "z_zone.h"

// Holds the players real Z location, so the missile when teleported to
// a different height sector stay relative to it.
float_t realplayerz = 0;  //-jc-

// List of all objects in map.
mobj_t *mobjlisthead;

// Where objects go to die...
iteminque_t *itemquehead;

//
// P_SetMobjState
// Returns true if the mobj is still present.
//
boolean_t P_SetMobjState(mobj_t * mobj, statenum_t state)
{
  state_t *st;

  do
  {
    if (state == S_NULL)
    {
      mobj->state = (state_t *) S_NULL;
      P_RemoveMobj(mobj);
      return false;
    }

    // -AJA- Forget any deferred state.
    mobj->deferred_state = -1;

    st = &states[state];
    mobj->state = st;
    mobj->tics = st->tics;
    mobj->sprite = st->sprite;
    mobj->frame = st->frame;

    if (st->action)
      (* st->action)(mobj);

    state = st->nextstate;
  }
  while (mobj->tics == 0);

  return true;
}

//
// P_SetMobjStateDeferred
//
// Similiar to P_SetMobjState, but no actions are performed yet.
// The new state will entered when the P_MobjThinker code reaches it,
// which may happen in the current tick, or at worst the next tick.
//
// Prevents re-entrancy into code like P_CheckRelPosition which is
// inherently non re-entrant.
//
// -AJA- 1999/09/12: written.

boolean_t P_SetMobjStateDeferred(mobj_t * mo, statenum_t state, int tic_skip)
{
  if (state == S_NULL)
  {
    mo->state = (state_t *) S_NULL;
    P_RemoveMobj(mo);
    return false;
  }

  mo->deferred_state = state;
  mo->defer_tic_skip = tic_skip;
  mo->tics = 1;

  return true;
}

//
// P_SetMobjDirAndSpeed
//
// -AJA- 1999/09/11: written.

void P_SetMobjDirAndSpeed(mobj_t * mo, angle_t angle, float_t slope, float_t speed)
{
  mo->angle = angle;
  mo->vertangle = slope;

  mo->momx = M_Cos(angle) * speed;
  mo->momy = M_Sin(angle) * speed;
  mo->momz = slope * speed;
}

//
// P_MobjExplodeMissile  
//
// -AJA- 1999/09/12: Now uses P_SetMobjStateDeferred, since this
//       routine can be called by TryMove/PIT_CheckRelThing.
//
void P_MobjExplodeMissile(mobj_t * mo)
{
  mo->momx = mo->momy = mo->momz = 0;

  mo->flags &= ~(MF_MISSILE | MF_TOUCHY);
  mo->extendedflags &= ~(EF_BOUNCE | EF_USABLE);

  if (mo->info->deathsound)
    S_StartSound(mo, mo->info->deathsound);

  // mobjinfo used -ACB- 1998/08/06
  P_SetMobjStateDeferred(mo, mo->info->death_state, P_Random() & 3);
}

//
// EnterBounceStates
//
// -AJA- 1999/10/18: written.

static void EnterBounceStates(mobj_t * mo)
{
  if (! mo->info->bounce_state)
    return;

  // ignore if disarmed
  if (mo->extendedflags & EF_JUSTBOUNCED)
    return;

  // give deferred states a higher priority
  if (mo->deferred_state >= 0)
    return;

  mo->extendedflags |= EF_JUSTBOUNCED;
  P_SetMobjState(mo, mo->info->bounce_state);
}

//
// BounceOffWall
//
// -AJA- 1999/08/22: written.

static void BounceOffWall(mobj_t * mo, line_t * wall)
{
  angle_t angle;
  angle_t wall_angle;
  angle_t diff;

  float_t dest_x, dest_y;

  angle = R_PointToAngle(0, 0, mo->momx, mo->momy);
  wall_angle = R_PointToAngle(0, 0, wall->dx, wall->dy);

  diff = wall_angle - angle;

  if (diff > ANG90 && diff < ANG270)
    diff -= ANG180;

  // -AJA- Prevent getting stuck at some walls...

  dest_x = mo->x + M_Cos(angle) * (mo->speed + mo->info->radius) * 4.0f;
  dest_y = mo->y + M_Sin(angle) * (mo->speed + mo->info->radius) * 4.0f;

  if (P_PointOnLineSide(mo->x, mo->y, wall) ==
      P_PointOnLineSide(dest_x, dest_y, wall))
  {
    // Result is the same, thus we haven't crossed the line.  Choose a
    // random angle to bounce away.  And don't attenuate the speed (so
    // we can get far enough away).

    angle = P_Random() << (ANGLEBITS - 9);
  }
  else
  {
    angle += diff << 1;
  }

  // calculate new momentum

  mo->speed *= mo->info->bounce_speed;

  mo->momx = M_Cos(angle) * mo->speed;
  mo->momy = M_Sin(angle) * mo->speed;
  mo->angle = angle;

  EnterBounceStates(mo);
}

//
// BounceOffPlane
//
// -AJA- 1999/10/18: written.

static void BounceOffPlane(mobj_t * mo, float_t dir)
{
  // calculate new momentum

  mo->speed *= mo->info->bounce_speed;

  mo->momx = (float_t)(M_Cos(mo->angle) * mo->speed);
  mo->momy = (float_t)(M_Sin(mo->angle) * mo->speed);
  mo->momz = (float_t)(dir * mo->speed / 2.0f);

  EnterBounceStates(mo);
}

//
// MissileHitSpecialLine
//
// -AJA- 1999/08/22: written.

static void MissileHitSpecialLine(mobj_t * mo)
{
  line_t **hits;
  int i;
    
  // -ES- 2000/02/05 spechit could be changed inside the loop
  hits = I_TmpMalloc(numspechit * sizeof(line_t *));
  Z_MoveData(hits, spechit, line_t *, numspechit);
  i = numspechit;
  Z_SetArraySize(&spechit_a, numspechit = 0);
  
  while (i)
  {
    line_t *ld = hits[--i];

    P_ShootSpecialLine(mo, ld);
  }
  
  I_TmpFree(hits);
}

//
// CorpseShouldSlide
//
// -AJA- 1999/09/25: written.

static boolean_t CorpseShouldSlide(mobj_t * mo)
{
  float_t floor, ceil;
  
  if (-0.25 < mo->momx && mo->momx < 0.25 &&
      -0.25 < mo->momy && mo->momy < 0.25)
  {
    return false;
  }
 
  P_ComputeThingGap(mo, mo->subsector->sector, mo->z, &floor, &ceil);

  return (mo->floorz != floor);
}

//
// P_XYMovement  
//
static void P_XYMovement(mobj_t * mo, region_properties_t *props)
{
  player_t *player;

  float_t ptryx;
  float_t ptryy;
  float_t xmove;
  float_t ymove;
  float_t xstep;
  float_t ystep;
  float_t absx,absy;
  float_t maxstep;

  int did_move;

  if (mo->momx == 0 && mo->momy == 0 && (mo->flags & MF_SKULLFLY))
  {
    // the skull slammed into something
    mo->flags &= ~MF_SKULLFLY;
    mo->momx = mo->momy = mo->momz = 0;

    P_SetMobjState(mo, mo->info->idle_state);
    return;
  }

#if 0  // -AJA- doesn't work... will fix soon.
  // -AJA- 1999/07/31: Ride that rawhide :->
  if (mo->ride_em && (mo->z == mo->floorz))
  {
    mo->momx += mo->ride_em->momx;
    mo->momy += mo->ride_em->momy;
  }
#endif

  player = mo->player;

  if (mo->momx > MAXMOVE)
    mo->momx = MAXMOVE;
  else if (mo->momx < -MAXMOVE)
    mo->momx = -MAXMOVE;

  if (mo->momy > MAXMOVE)
    mo->momy = MAXMOVE;
  else if (mo->momy < -MAXMOVE)
    mo->momy = -MAXMOVE;

  xmove = mo->momx;
  ymove = mo->momy;

  // -AJA- 1999/10/09: Reworked viscosity.
  xmove *= 1.0f - props->viscosity;
  ymove *= 1.0f - props->viscosity;

  // -ES- 1999/10/16 For fast mobjs, break down
  //  the move into steps of max half radius for collision purposes.

  // Use half radius as max step, if not exceptionally small.
  if (mo->radius > STEPMOVE)
    maxstep = mo->radius / 2;
  else
    maxstep = STEPMOVE / 2;

  // precalculate these two, they are used frequently
  absx = (float_t)fabs(xmove);
  absy = (float_t)fabs(ymove);

  if (absx > maxstep || absy > maxstep)
  {
    // Do it in the most number of steps.
    if (absx > absy)
    {
      xstep = (xmove > 0) ? maxstep : -maxstep;

      // almost orthogonal movements are rounded to orthogonal, to prevent
      // an infinite loop in some extreme cases.
      if (absy * 256 < absx)
        ystep = ymove = 0;
      else
        ystep = ymove * xstep / xmove;
    }
    else
    {
      ystep = (ymove > 0) ? maxstep : -maxstep;

      if (absx * 256 < absy)
        xstep = xmove = 0;
      else
        xstep = xmove * ystep / ymove;
    }
  }
  else
  {
    // Step is less than half radius, so one iteration is enough.
    xstep = xmove;
    ystep = ymove;
  }

  // Keep attempting moves until object has lost all momentum.
  do
  {
    // if movement is more than half that of the maximum, attempt the move
    // in two halves or move.
    if (fabs(xmove) > fabs(xstep))
    {
      ptryx = mo->x + xstep;
      xmove -= xstep;
    }
    else
    {
      ptryx = mo->x + xmove;
      xmove = 0;
    }

    if (fabs(ymove) > fabs(ystep))
    {
      ptryy = mo->y + ystep;
      ymove -= ystep;
    }
    else
    {
      ptryy = mo->y + ymove;
      ymove = 0;
    }

    did_move = P_TryMove(mo, ptryx, ptryy);

    // unable to complete desired move ?
    if (!did_move)
    { 
      if (numspechit > 0 && (mo->flags & MF_MISSILE))
        MissileHitSpecialLine(mo);

      if (mo->info->flags & MF_SLIDE)
        P_SlideMove(mo);

      // -KM- 1999/01/31 Bouncy objects (grenades)
      // -AJA- 1999/07/30: Moved up here.
      else if (mo->extendedflags & EF_BOUNCE)
      {
        if (! blockline)
        {
          if (mobj_hit_sky)
            P_MobjRemoveMissile(mo);
          else
            P_MobjExplodeMissile(mo);
          
          return;
        }

        BounceOffWall(mo, blockline);
        xmove = ymove = 0;
      }
      else if (mo->flags & MF_MISSILE)
      {
        if (mobj_hit_sky)
          P_MobjRemoveMissile(mo);  // New Procedure -ACB- 1998/07/30
        else
          P_MobjExplodeMissile(mo);

        return;
      }
      else
      {
        xmove = ymove = 0;
        mo->momx = mo->momy = 0;
      }
    }
  }
  while (xmove || ymove);

  // slow down
  if (player && player->cheats & CF_NOMOMENTUM)
  {
    // debug option for no sliding at all
    mo->momx = mo->momy = 0;
    return;
  }

  if ((mo->extendedflags & EF_NOFRICTION) || (mo->flags & MF_SKULLFLY))
  {
    return;
  }

  // -MH- 1998/08/18 - make mid-air movement normal when using the jetpack
  //                    When in mid-air there's no friction so you slide
  //                    about uncontrollably. This is realistic but makes
  //                    the game difficult to control to the extent that
  //                    for normal people, it's not worth playing - a bit
  //                    like having auto-aim permanently off (as most
  //                    real people are not crack-shots!)
  if ((mo->z > mo->floorz) &&
      !(level_flags.true3dgameplay && mo->player && 
        mo->player->powers[PW_Jetpack]))
  {
    // no friction when airborne unless player has jetpack
    return;
  }

  if (mo->flags & MF_CORPSE)
  {
    // do not stop sliding if halfway off a step with some momentum
    if (CorpseShouldSlide(mo))
      return;
  }

  mo->momx *= props->friction;
  mo->momy *= props->friction;
}

//
// P_ZMovement
//
static void P_ZMovement(mobj_t * mo, region_properties_t *props)
{
  float_t dist;
  float_t delta;
  float_t zmove;

  // -KM- 1998/11/25 Gravity is now not precalculated so that
  //  menu changes affect instantly.
  float_t gravity = props->gravity / 8.0 * 
      (float_t)level_flags.menu_grav / (float_t)MENU_GRAV_NORMAL;
  
  // check for smooth step up
  if (mo->player && mo->z < mo->floorz)
  {
    mo->player->viewheight -= mo->floorz - mo->z;
    mo->player->deltaviewheight = (VIEWHEIGHT - mo->player->viewheight) / 8.0f;
  }

  zmove = mo->momz * (1.0f - props->viscosity);

  // adjust height
  mo->z += zmove;

  if (mo->flags & MF_FLOAT && mo->target)
  {
    // float down towards target if too close
    if (!(mo->flags & MF_SKULLFLY) && !(mo->flags & MF_INFLOAT))
    {
      dist = P_ApproxDistance(mo->x - mo->target->x, mo->y - mo->target->y);
      delta = mo->target->z + (mo->height / 2) - mo->z;

      if (delta < 0 && dist < -(delta * 3))
        mo->z -= mo->info->float_speed;
      else if (delta > 0 && dist < (delta * 3))
        mo->z += mo->info->float_speed;
    }
  }
  
  //
  //  HIT FLOOR ?
  //

  if (mo->z <= mo->floorz)
  {
    if (mo->flags & MF_SKULLFLY)
      mo->momz = -mo->momz;

    if (mo->momz < 0)
    {
      float_t hurt_momz = gravity * mo->info->maxfall;

      if (mo->player && gravity > 0 && -zmove > OOF_SPEED)
      {
        // Squat down. Decrease viewheight for a moment after hitting the
        // ground (hard), and utter appropriate sound.
        mo->player->deltaviewheight = zmove / 8.0;
        S_StartSound(mo, sfx_oof);
      }
      // -KM- 1998/12/16 If bigger than max fall, take damage.
      if (mo->info->maxfall && gravity > 0 && -mo->momz > hurt_momz)
      {
        P_DamageMobj(mo, NULL, NULL, (-mo->momz - hurt_momz));
      }

      // -KM- 1999/01/31 Bouncy bouncy...
      if (mo->extendedflags & EF_BOUNCE)
      {
        BounceOffPlane(mo, +1.0);

        // don't bounce forever on the floor
        if (! (mo->flags & MF_NOGRAVITY) &&
            fabs(mo->momz) < STOPSPEED + fabs(gravity))
        {
          mo->momx = mo->momy = mo->momz = 0;
        }
      }
      else
        mo->momz = 0;
    }

    mo->z = mo->floorz;

    if ((mo->flags & MF_MISSILE) && !(mo->flags & MF_NOCLIP))
    {
      // if the floor is sky, don't explode missile -ACB- 1998/07/31
      if (mo->subsector->sector->floor.pic == skyflatnum &&
          mo->subsector->sector->floor.h >= mo->floorz)
      {
        P_MobjRemoveMissile(mo);
      }
      else
      {
        if (! (mo->extendedflags & EF_BOUNCE))
          P_MobjExplodeMissile(mo);
      }
      return;
    }
  }
  else if (gravity > 0.0)
  {
    // thing is above the ground, therefore apply gravity

    // -MH- 1998/08/18 - Disable gravity while player has jetpack
    //                   (nearly forgot this one:-)

    if (!(mo->flags & MF_NOGRAVITY) &&
        !(level_flags.true3dgameplay && mo->player && 
         mo->player->powers[PW_Jetpack]))
    {
      mo->momz -= gravity;
    }
  }

  //
  //  HIT CEILING ?
  //

  if (mo->z + mo->height > mo->ceilingz)
  {
    if (mo->flags & MF_SKULLFLY)
      mo->momz = -mo->momz;  // the skull slammed into something

    // hit the ceiling
    if (mo->momz > 0)
    {
      float_t hurt_momz = gravity * mo->info->maxfall;

      if (mo->player && gravity < 0 && zmove > OOF_SPEED)
      {
        mo->player->deltaviewheight = zmove / 8.0;
        S_StartSound(mo, sfx_oof);
      }
      if (mo->info->maxfall && gravity < 0 && mo->momz > hurt_momz)
      {
        P_DamageMobj(mo, NULL, NULL, (mo->momz - hurt_momz));
      }

      // -KM- 1999/01/31 More bouncing.
      if (mo->extendedflags & EF_BOUNCE)
      {
        BounceOffPlane(mo, -1.0);
        
        // don't bounce forever on the ceiling
        if (! (mo->flags & MF_NOGRAVITY) &&
            fabs(mo->momz) < STOPSPEED + fabs(gravity))
        {
          mo->momx = mo->momy = mo->momz = 0;
        }
      }
      else
        mo->momz = 0;
    }

    mo->z = mo->ceilingz - mo->height;

    if ((mo->flags & MF_MISSILE) && !(mo->flags & MF_NOCLIP))
    {
      // if the ceiling is sky, don't explode missile -ACB- 1998/07/31
      if (mo->subsector->sector->ceil.pic == skyflatnum &&
          mo->subsector->sector->ceil.h <= mo->ceilingz)
      {
        P_MobjRemoveMissile(mo);
      }
      else
      {
        if (! (mo->extendedflags & EF_BOUNCE))
          P_MobjExplodeMissile(mo);
      }
      return;
    }
  }
  else if (gravity < 0.0)
  {
    // thing is below ceiling, therefore apply any negative gravity

    // -MH- 1998/08/18 - Disable gravity while player has jetpack
    //                   (nearly forgot this one:-)

    if (!(mo->flags & MF_NOGRAVITY) &&
        !(level_flags.true3dgameplay && mo->player && 
         mo->player->powers[PW_Jetpack]))
    {
      mo->momz += -gravity;
    }
  }

  if (level_flags.true3dgameplay && mo->player)
  {
    if (fabs(mo->momz) < STOPSPEED &&
        mo->player->cmd.upwardmove == 0)
    {
      mo->momz = 0;
    }

    // -AJA- 1999/10/09: removed old code here which applied
    //       _friction_ to vertical momentum (?).
  }
}

//
// P_TeleportRespawn
//
static void TeleportRespawn(mobj_t * mobj)
{
  float_t x;
  float_t y;
  float_t z;
  mobj_t *mo;
  mapthing_t *mthing;

  x = mobj->spawnpoint.x;
  y = mobj->spawnpoint.y;

  if (mobj->info->flags & MF_SPAWNCEILING)
    z = ONCEILINGZ;
  else
    z = ONFLOORZ;

  // something is occupying it's position?
  if (!P_CheckAbsPosition(mobj, x, y, z))
    return;  // no respawn

  // spawn a teleport fog at old spot
  // because of removal of the body?

  // temp fix for teleport flash...
  P_MobjCreateObject(mobj->x, mobj->y, ONFLOORZ, mobj->info->respawneffect);

  // spawn a teleport fog at the new spot
  // temp fix for teleport flash...
  P_MobjCreateObject(x, y, ONFLOORZ, mobj->info->respawneffect);

  // spawn the new monster
  mthing = &mobj->spawnpoint;

  // spawn it
  // inherit attributes from deceased one
  // -ACB- 1998/08/06 Create Object
  mo = P_MobjCreateObject(x, y, z, mobj->info);

  mo->spawnpoint = mobj->spawnpoint;
  mo->angle = FLOAT_2_ANG((float) mthing->angle);

  if (mthing->options & MTF_AMBUSH)
    mo->flags |= MF_AMBUSH;

  mo->reactiontime = 18;

  // remove the old monster.
  P_RemoveMobj(mobj);
}

//
// P_ResurrectRespawn
//
// -ACB- 1998/07/29 Prevented respawning of ghosts
//                  Make monster deaf, if originally deaf
//                  Given a reaction time, delays monster starting up immediately.
//                  Given Invisibility level.
//                  Doesn't try to raise an object with no raisestate

static void ResurrectRespawn(mobj_t * object)
{
  float_t x;
  float_t y;
  float_t z;
  const mobjinfo_t *info;
  mapthing_t *mthing;

  x = object->x;
  y = object->y;
  z = object->z;

  info = object->info;

  if (!object->info->raise_state)
    return;  // cannot raise the unraisable

  // something is occupying it's position?
  // restore original height is position occupied
  if (!P_CheckAbsPosition(object, x, y, z))
    return;

  if (object->state == &states[object->info->gib->spawn_state])
    return;  // don't respawn gibs

  // restore objects original height
  object->radius = info->radius;
  object->height = info->height;

  // Resurrect monster
   
  if (info->overkill_sound)
    S_StartSound(object, info->overkill_sound);

  P_SetMobjState(object, info->raise_state);

  object->flags = info->flags;
  object->extendedflags = object->extendedflags;
  object->health = info->spawnhealth;
  P_MobjSetTarget(object, NULL);
  object->visibility = info->visibility;
  object->movecount = 0;  // -ACB- 1998/08/03 Don't head off in any direction

  mthing = &object->spawnpoint;

  if (mthing->options & MTF_AMBUSH)
    object->flags |= MF_AMBUSH;

  object->reactiontime = 18;
}

//
// P_MobjThinker
//
#define MAX_DEFER_LOOP  4

static void P_MobjThinker(mobj_t * mobj)
{
  int defer_count = 0;

  region_properties_t *props;

  DEV_ASSERT(-1 != (int)mobj->next, ("P_MobjThinker INTERNAL ERROR: mobj has been Z_Freed"));

  // ignore removed objects
  if (! mobj->state)
    return;

  mobj->visibility = ((TICRATE - 1) * mobj->visibility + mobj->delta_vis) / 35;

  if (mobj->fuse >= 0)
  {
    if (!--mobj->fuse)
      P_MobjExplodeMissile(mobj);
  }

  // determine properties.  Later on, things may be linked into
  // vertical regions, making it feasible to do get the correct
  // properties for all objects.  For now, we only determine the
  // correct props for players.

  if (mobj->player)
    props = R_PointInVertRegion(mobj->subsector, mobj->z + 1.0)->p;
  else
    props = &mobj->subsector->sector->p;

  // handle pushing sectors
  if (props->special && (props->special->push_speed > 0 || 
       props->special->push_zspeed > 0) &&
      ((props->special->special_flags & SECSP_PushAll) ||
       !(mobj->flags & (MF_NOGRAVITY))) &&
      !(mobj->flags & MF_NOCLIP) &&
      ((props->special->special_flags & SECSP_WholeRegion) ||
       mobj->z <= mobj->floorz + 1.0))
  {
    if (props->special->special_flags & SECSP_PushConstant)
    {
      float_t mul = props->special->push_speed / 100.0;

      mobj->momx += M_Cos(props->special->push_angle) * mul;
      mobj->momy += M_Sin(props->special->push_angle) * mul;
      mobj->momz += props->special->push_zspeed / 100.0;
    }
    else if (mobj->info->mass > 0)
    {
      float_t mul = props->special->push_speed / mobj->info->mass;

      mobj->momx += M_Cos(props->special->push_angle) * mul;
      mobj->momy += M_Sin(props->special->push_angle) * mul;
      mobj->momz += props->special->push_zspeed / mobj->info->mass;
    }
  }
 
  // momentum movement
  if (mobj->momx != 0 || mobj->momy != 0 || 
      (mobj->flags & MF_SKULLFLY) || mobj->ride_em)
  {
    P_XYMovement(mobj, props);
  }

  if ((mobj->z != mobj->floorz) || mobj->momz != 0 || mobj->ride_em)
  {
    P_ZMovement(mobj, props);
  }

  if (mobj->tics < 0)
  {
    // check for nightmare respawn
    if (!(mobj->extendedflags & EF_MONSTER))
      return;

    // replaced respawnmonsters & newnmrespawn with respawnsetting
    // -ACB- 1998/07/30
    if (!level_flags.respawn)
      return;

    mobj->movecount++;

    //
    // Uses movecount as a timer, when movecount hits 12*TICRATE the
    // object will try to respawn. So after 12 seconds the object will
    // try to respawn.
    //
    if (mobj->movecount < mobj->info->respawntime)
      return;

    // if the first 5 bits of leveltime are on, don't respawn now...ok?
    if (leveltime & 31)
      return;

    // give a limited "random" chance that respawn don't respawn now
    if (P_Random() > 4)
      return;

    // replaced respawnmonsters & newnmrespawn with respawnsetting
    // -ACB- 1998/07/30
    if (level_flags.res_respawn)
      ResurrectRespawn(mobj);
    else
      TeleportRespawn(mobj);

    return;
  }

  // this can only happen if one of the above movement routines caused
  // the mobj to be removed (e.g. exploded missile).
  if (! mobj->state)
    return;

  // Cycle through states, calling action functions at transitions.
  // -AJA- 1999/09/12: reworked for deferred states.
  do
  {
    statenum_t next;
    int skip;

    mobj->tics--;

    if (mobj->tics >= 1)
      return;

    if (mobj->deferred_state >= 0)
    {
      defer_count++;
      next = mobj->deferred_state;
      skip = mobj->defer_tic_skip;
      mobj->deferred_state = -1;
    }
    else
    {
      next = mobj->state->nextstate;
      skip = 0;
    }

    // You can cycle through multiple states in a tic.
    // NOTE: returns false if object freed itself.
    if (!P_SetMobjState(mobj, next))
      return;

    // thing died ?
    if (mobj->tics < 0)
      return;

    mobj->tics -= skip;

    if (mobj->tics < 1)
      mobj->tics = 1;

    // keep going if there was a deferred state.
  }
  while (mobj->deferred_state >= 0 && defer_count < MAX_DEFER_LOOP);
}

//
// P_RunMobjThinkers
//
// Cycle through all mobjs and let them think.
void P_RunMobjThinkers(void)
{
  mobj_t *mo;

  for (mo = mobjlisthead; mo; mo = mo->next)
  {
    P_MobjThinker(mo);
  }

  P_RemoveQueuedMobjs();
}

//
// DoRemoveMobj
//
// Remove the map object.  This routine common to P_RemoveMobj and
// P_MobjRemoveMissile.
//
// -AJA- 1999/09/15: written.

static void DoRemoveMobj(mobj_t * mo)
{
  // -------------------------------------------------------------------
  // -ACB- 1998/08/27 Mobj Linked-List Removal
  //
  // A useful way of cycling through the current things without
  // having to deref and search everything using thinkers.
  //

  // unlink from sector and block lists
  P_UnsetThingPosition(mo);

  if (mo->prev == NULL)  // no previous, must be first item
  {
    DEV_ASSERT2(mobjlisthead == mo);
    mobjlisthead = mo->next;

    if (mobjlisthead != NULL)
    {
      DEV_ASSERT2(mobjlisthead->prev == mo);
      mobjlisthead->prev = NULL;
    }
  }
  else
  {
    DEV_ASSERT2(mo->prev->next == mo);
    mo->prev->next = mo->next;

    if (mo->next != NULL)
    {
      DEV_ASSERT2(mo->next->prev == mo);
      mo->next->prev = mo->prev;
    }
  }

  // Clear all references to other mobjs
  P_MobjSetTracer(mo, NULL);
  P_MobjSetSource(mo, NULL);
  P_MobjSetTarget(mo, NULL);
  P_MobjSetSupportObj(mo, NULL);
  P_MobjSetRideEm(mo, NULL);

  // clear all references FROM other mobjs. A bit more tricky: We have to
  // search through the reference list for links to this mobj. Luckily
  // we have the refcount that tells us how many occurances we have to find.
  if (mo->refcount)
  {
    mobj_t *ref;

#define CHECK_REF(field)\
if (ref->field == mo)\
{\
  *(mobj_t**)&ref->field = NULL;\
  mo->refcount--;\
}
    for (ref = mobjlisthead; ref && mo->refcount; ref = ref->next)
    {
      CHECK_REF(tracer);
      CHECK_REF(source);
      CHECK_REF(target);
      CHECK_REF(supportobj);
      CHECK_REF(ride_em);
    }
#undef CHECK_REF
#ifdef DEVELOPERS
    if (mo->refcount != 0)
      // someone forgot to clear a reference after use
      I_Error("INTERNAL ERROR: Reference count %d", mo->refcount);
#endif
  }

  // free shoot-to-spot info
  if (mo->spot_info)
  {
    Z_Free(mo->spot_info->targets);
    Z_Free(mo->spot_info);
  }

  // Sound might still be playing, so use indirect Z_Free.
  S_AddToFreeQueue(mo, mo);
}

static mobj_t **removeque = NULL;
static int removeque_size = 0;

//
// AddMobjToRemoveQue
//
// Queues a mobj to the removal queue. It will be removed at the end of the
// frame.
//
// -ES- 1999/10/24 Written.
static void AddMobjToRemoveQue(mobj_t *mo)
{
  int i;

  // we don't have to add the mobj if it's already there.
  for (i = 0; i < removeque_size; i++)
    if (removeque[i] == mo)
      return;

  Z_Resize(removeque, mobj_t *, ++removeque_size);
  removeque[removeque_size-1] = mo;
}

//
// void P_RemoveQueuedMobjs(void)
//
// Removes all the mobjs in the removeque list.
// -ES- 1999/10/24 Written.
void P_RemoveQueuedMobjs(void)
{
  int i;

  for (i = 0; i < removeque_size; i++)
    DoRemoveMobj(removeque[i]);

  removeque_size = 0;
}

//
// P_RemoveMobj
//
// Removes the object from the play simulation: no longer thinks, if
// the mobj is MF_SPECIAL: i.e. item can be picked up, it is added to
// the item-respawn-que, so it gets respawned if needed; The respawning
// only happens if itemrespawn is set or the deathmatch mode is
// version 2.0: altdeath.

void P_RemoveMobj(mobj_t *mo)
{
  if ((mo->info->flags & MF_SPECIAL) && !(mo->extendedflags & EF_NORESPAWN) &&
      !(mo->flags & MF_DROPPED))
  {
    iteminque_t *newiteminque;

    // itemquehead is NULL, allocate space for new item-in-respawn-que
    if (itemquehead == NULL)
    {
      itemquehead = Z_New(iteminque_t, 1);
      newiteminque = itemquehead;
      newiteminque->prev = NULL;
      newiteminque->next = NULL;
    }
    else  // item-respawn-que exists, add new item to que
    {
      iteminque_t *currentitem = itemquehead;

      while (currentitem->next != NULL)
        currentitem = currentitem->next;

      newiteminque = Z_New(iteminque_t, 1);
      currentitem->next = newiteminque;
      newiteminque->prev = currentitem;
      newiteminque->next = NULL;
    }

    newiteminque->info = mo->spawnpoint;
    newiteminque->time = mo->info->respawntime;
  }

  AddMobjToRemoveQue(mo);
}

//
// P_SpawnPlayer
//
// Called when a player is spawned on the level.
// Most of the player structure stays unchanged between levels.
//
// -KM- 1998/12/21 Cleaned this up a bit.
// -KM- 1999/01/31 Removed all those nasty cases for doomednum (1/4001)

void P_SpawnPlayer(player_t *p, const mapthing_t * mthing)
{
  float_t x;
  float_t y;
  float_t z;

  mobj_t *mobj;
  const mobjinfo_t *objtype;

  int i;

  // -ES- FIXME: Move these checks higher up.
  if (!p)
    return;
  // -jc- not playing?
  if (!p->in_game)
    return;

  // -KM- 1998/11/25 This is in preparation for skins.  The creatures.ddf
  //   will hold player start objects, sprite will be taken for skin.
  objtype = DDF_MobjLookupPlayer(p->pnum+1);

  if (p->playerstate == PST_REBORN)
  {
    G_PlayerReborn(p);
  }

  x = mthing->x;
  y = mthing->y;
  z = ONFLOORZ;

  mobj = P_MobjCreateObject(x, y, z, objtype);

  mobj->angle = FLOAT_2_ANG((float) mthing->angle);
  mobj->player = p;
  mobj->health = p->health;

  p->mo = mobj;
  p->playerstate = PST_LIVE;
  p->refire = 0;
  p->damagecount = 0;
  p->bonuscount = 0;
  p->extralight = 0;
  p->effect_colourmap = NULL;
  p->viewheight = VIEWHEIGHT;
  p->jumpwait = 0;

  // setup gun psprite
  P_SetupPsprites(p);

  // give all cards in death match mode
  if (deathmatch)
    for (i = 0; i < NUMCARDS; i++)
      p->cards[i] = true;

  // -AJA- FIXME: surely this belongs elsewhere.
  if (p == consoleplayer)
  {
    char buffer[16];

    // wake up the status bar
    ST_Start();

    // wake up the heads up text
    HU_Start();

    CON_DeleteCVar("health");
    CON_CreateCVarReal("health", cf_read | cf_delete, &mobj->health);
    CON_DeleteCVar("armour");
    CON_CreateCVarReal("armour", cf_read | cf_delete, &p->armourpoints);
    CON_DeleteCVar("a_type");
    CON_CreateCVarInt("a_type", cf_read | cf_delete, (int*)&p->armourtype);
    CON_DeleteCVar("frags");
    CON_CreateCVarInt("frags", cf_read | cf_delete, &p->frags);
    CON_DeleteCVar("totalfrags");
    CON_CreateCVarInt("totalfrags", cf_read | cf_delete, &p->totalfrags);

    for (i = 0; i < NUMAMMO; i++)
    {
      sprintf(buffer, "ammo%d", i);
      CON_DeleteCVar(buffer);
      CON_CreateCVarInt(buffer, cf_read | cf_delete, &p->ammo[i].num);

      sprintf(buffer, "maxammo%d", i);
      CON_DeleteCVar(buffer);
      CON_CreateCVarInt(buffer, cf_read | cf_delete, &p->ammo[i].max);
    }

    for (i = 0; i < numweapons; i++)
    {
      sprintf(buffer, "weapon%d", i);
      CON_DeleteCVar(buffer);
      CON_CreateCVarBool(buffer, cf_read | cf_delete, &p->weapons[i].owned);
    }

    for (i = 0; i < NUMCARDS; i++)
    {
      sprintf(buffer, "key%d", i);
      CON_DeleteCVar(buffer);
      CON_CreateCVarBool(buffer, cf_read | cf_delete, &p->cards[i]);
    }

    for (i = 0; i < NUMPOWERS; i++)
    {
      sprintf(buffer, "power%d", i);
      CON_DeleteCVar(buffer);
      CON_CreateCVarReal(buffer, cf_read | cf_delete, &p->powers[i]);
    }
  }

  // Heh, make a drone player invisible and no clip
  if (doomcom->drone & (1 << p->pnum))
  {
    mobj->delta_vis = mobj->visibility = INVISIBLE;
    mobj->flags |= MF_NOCLIP;
    mobj->flags &= ~(MF_SHOOTABLE | MF_SOLID);
  }

  // Don't get stuck spawned in things: telefrag them.
  if (deathmatch >= 3)
    P_TeleportMove(mobj, mobj->x, mobj->y, mobj->z);
}

//
// P_SpawnMapThing
//
// The fields of the mapthing should
// already be in host byte order.
//
void P_SpawnMapThing(const mapthing_t * mthing)
{
  int bit;
  mobj_t *mobj;
  const mobjinfo_t *objtype;
  float_t x;
  float_t y;
  float_t z;

  objtype = DDF_MobjLookupNum(mthing->type);

  // MOBJTYPE not found, don't crash out: JDS Compliance.
  // -ACB- 1998/07/21
  if (objtype == NULL)
  {
    I_WriteDebug("Unknown thing type %i at (%i, %i)\n", mthing->type, 
        mthing->x, mthing->y);
    return;
  }

  // -KM- 1999/01/31 Use playernum property.
  // count deathmatch start positions
  if (objtype->playernum < 0)
  {
    if (deathmatch_p == &deathmatchstarts[max_deathmatch_starts])
    {
      int length = deathmatch_p - deathmatchstarts;

      Z_Resize(deathmatchstarts, mapthing_t, ++max_deathmatch_starts);

      deathmatch_p = deathmatchstarts + length;
    }
    *deathmatch_p = *mthing;
    deathmatch_p++;
    return;
  }

  // check for players specially
  //-jc-
  if (objtype->playernum > 0)
  {
    // save spots for respawning in network games
    playerstarts[objtype->playernum - 1] = *mthing;

    // -AJA- FIXME: belongs higher up (g_game).
    if (!deathmatch)
      P_SpawnPlayer(playerlookup[objtype->playernum - 1], mthing);

    return;
  }

  // check for apropriate skill level
  // -ES- 1999/04/13 Implemented Kester's Bugfix.
  // -AJA- 1999/10/21: Reworked again.
  if (!netgame && !deathmatch && (mthing->options & MTF_NOT_SINGLE))
    return;

  // -AJA- 1999/09/22: Boom compatibility flags.
  if (netgame && !deathmatch && (mthing->options & MTF_NOT_COOP))
    return;

  if (deathmatch && (mthing->options & MTF_NOT_DM))
    return;

  if (gameskill == sk_baby)
    bit = 1;
  else if (gameskill == sk_nightmare)
    bit = 4;
  else
    bit = 1 << (gameskill - 1);

  if (!(mthing->options & bit))
    return;

  // don't spawn keycards and players in deathmatch
  if (deathmatch && objtype->flags & MF_NOTDMATCH)
    return;

  // don't spawn any monsters if -nomonsters
  if (level_flags.nomonsters && (objtype->extendedflags & EF_MONSTER))
    return;

  // -AJA- 1999/10/07: don't spawn extra things if -noextra.
  if (!level_flags.have_extra && (objtype->extendedflags & EF_EXTRA))
    return;

  // spawn it
  x = mthing->x;
  y = mthing->y;

  if (objtype->flags & MF_SPAWNCEILING)
    z = ONCEILINGZ;
  else
    z = ONFLOORZ;

  // Use MobjCreateObject -ACB- 1998/08/06
  mobj = P_MobjCreateObject(x, y, z, objtype);

  mobj->spawnpoint = *mthing;

  if (mobj->tics > 0)
    mobj->tics = 1 + (P_Random() % mobj->tics);

  mobj->angle = FLOAT_2_ANG((float) mthing->angle);

  if (mthing->options & MTF_AMBUSH)
    mobj->flags |= MF_AMBUSH;
}

//
// GAME SPAWN FUNCTIONS
//

//
// P_SpawnPuff
//
void P_SpawnPuff(float_t x, float_t y, float_t z, const mobjinfo_t * puff)
{
  mobj_t *th;

  z += (float_t) P_RandomNegPos() / 127.0f;

  // -ACB- 1998/08/06 Specials table for non-negotiables....
  th = P_MobjCreateObject(x, y, z, puff);

  // -AJA- 1999/07/14: DDF-itised.
  th->momz = puff->float_speed;

  th->tics -= P_Random() & 3;

  if (th->tics < 1)
    th->tics = 1;
}

//
// P_SpawnBlood
//
// -KM- 1998/11/25 Made more violent. :-)
// -KM- 1999/01/31 Different blood objects for different mobjs.

void P_SpawnBlood(float_t x, float_t y, float_t z, float_t damage,
    angle_t angle, const mobjinfo_t * blood)
{
  int num;
  mobj_t *th;

  angle += ANG180;

  num = (int) (!level_flags.more_blood ? 1.0f : (M_Random() % 7) + 
      (float_t)((MAX(damage / 4.0f, 7.0f))));

  while (num--)
  {
    z += (float_t)(P_RandomNegPos() / 64.0f);

    angle += (angle_t) (P_RandomNegPos() * (int)(ANG1 / 2));

    th = P_MobjCreateObject(x, y, z, blood);

    P_SetMobjDirAndSpeed(th, angle, ((float_t)num + 12.0f) / 6.0f, 
        (float_t)num / 4.0f);

    th->tics -= P_Random() & 3;

    if (th->tics < 1)
      th->tics = 1;

    if (damage <= 12 && th->state)
      P_SetMobjState(th, th->state->nextstate);
    if (damage < 9 && th->state)
      P_SetMobjState(th, th->state->nextstate);
  }
}

//
// P_MobjItemRespawn
//
// Replacement procedure for P_RespawnSpecials, uses a linked list to go through
// the item-respawn-que. The time until respawn (in tics) is decremented every tic,
// when the item-in-the-que has a time of zero is it respawned.
//
// -ACB- 1998/07/30 Procedure written.
// -KM- 1999/01/31 Custom respawn fog.
//
void P_MobjItemRespawn(void)
{
  float_t x;
  float_t y;
  float_t z;
  mobj_t *mo;
  const mobjinfo_t *objtype;

  iteminque_t *currentitem;
  iteminque_t *respawneditem;

  // No item-respawn-que exists, so nothing to process.
  if (itemquehead == NULL)
    return;

  // only respawn items in deathmatch or if itemrespawn
  if ((deathmatch <= 1) && (!level_flags.itemrespawn))
    return;

  currentitem = itemquehead;  // lets start from the beginning....

  while (currentitem != NULL)
  {
    currentitem->time--;

    if (!currentitem->time)  // no time left? if so respawn object
    {
      x = currentitem->info.x;
      y = currentitem->info.y;

      // spawn a teleport fog at the new spot

//      S_StartSound (mo, sfx_itmbk);

      // -ACB- 1998/08/06 Cycle through the linked mobjinfo list....
      objtype = DDF_MobjLookupNum(currentitem->info.type);

      if (objtype == NULL)
      {
        I_Error("P_MobjItemRespawn: No such item type!");
        return;  // shouldn't happen.
      }

      P_MobjCreateObject(x, y, ONFLOORZ, objtype->respawneffect);

      if (objtype->flags & MF_SPAWNCEILING)
        z = ONCEILINGZ;
      else
        z = ONFLOORZ;

      // -ACB- 1998/08/06 Use MobjCreateObject
      mo = P_MobjCreateObject(x, y, z, objtype);

      mo->spawnpoint = currentitem->info;
      mo->angle = FLOAT_2_ANG((float) currentitem->info.angle);

      // Taking this item-in-que out of the que, remove
      // any references by the previous and next items to
      // the current one.....
      if (currentitem->prev == NULL)
      {
        // no previous, must be first item
        itemquehead = currentitem->next;

        if (itemquehead != NULL)
          itemquehead->prev = NULL;
      }
      else
      {
        currentitem->prev->next = currentitem->next;

        if (currentitem->next != NULL)
          currentitem->next->prev = currentitem->prev;
      }

      // unlink & free
      respawneditem = currentitem;
      currentitem = currentitem->next;
      Z_Free(respawneditem);
    }
    else
    {
      currentitem = currentitem->next;
    }
  }
}

//
// P_MobjRemoveMissile
//
// This procedure only is used when a flying missile is removed because
// it "hit" a wall or ceiling that in the simulation acts as a sky. The
// only major differences with P_RemoveMobj are that now item respawn check
// is not done (not needed) and any sound will continue playing despite
// the fact the missile has been removed: This is only done due to the
// fact that a missile in reality would continue flying through a sky and
// you should still be able to hear it.
//
// -ACB- 1998/07/31 Procedure written.
// -AJA- 1999/09/15: Functionality subsumed by DoRemoveMobj.
// -ES- 1999/10/24 Removal Queue.

void P_MobjRemoveMissile(mobj_t * missile)
{
  missile->momx = missile->momy = missile->momz = 0;

  missile->flags &= ~(MF_MISSILE | MF_TOUCHY);
  missile->extendedflags &= ~(EF_BOUNCE);

  AddMobjToRemoveQue(missile);
}

// Use these functions to set mobj entries. Never modify the entries directly.
#define CREATE_FUNCTION(name, field)\
void name(mobj_t *mo, mobj_t *target)\
{\
  if (mo->field)\
    mo->field->refcount--;\
  mo->field = target;\
  if (target)\
    target->refcount++;\
}

CREATE_FUNCTION(P_MobjSetTracer, tracer)
CREATE_FUNCTION(P_MobjSetSource, source)
CREATE_FUNCTION(P_MobjSetTarget, target)
CREATE_FUNCTION(P_MobjSetSupportObj, supportobj)
CREATE_FUNCTION(P_MobjSetRideEm, ride_em)

#undef CREATE_FUNCTION

//
// P_MobjCreateObject
//
// Creates a Map Object (MOBJ) at the specified location, with the
// specified type (given by DDF).  The special z values ONFLOORZ and
// ONCEILINGZ are recognised and handled appropriately.
//
// -ACB- 1998/08/02 Procedure written.
//
mobj_t *P_MobjCreateObject(float_t x, float_t y, float_t z, const mobjinfo_t *type)
{
  mobj_t *mobj;
  state_t *st;
  sector_t *sec;

  mobj = Z_ClearNew(mobj_t, 1);

  mobj->info = type;
  mobj->x = x;
  mobj->y = y;
  mobj->radius = type->radius;
  mobj->height = type->height;
  mobj->flags = type->flags;
  mobj->health = type->spawnhealth;
  mobj->speed = type->speed;
  mobj->fuse = type->fuse;
  mobj->side = type->side;

  if (level_flags.fastparm)
    mobj->speed *= type->fast;

  // -ACB- 1998/06/25 new mobj Stuff (1998/07/11 - invisibility added)
  mobj->extendedflags = type->extendedflags;
  mobj->delta_vis = mobj->visibility = type->visibility;
  mobj->currentattack = NULL;

  if (gameskill != sk_nightmare)
    mobj->reactiontime = type->reactiontime;

  mobj->lastlook = P_Random() % maxplayers;

  //
  // Do not set the state with P_SetMobjState,
  // because action routines can not be called yet
  //
  // if we have a spawnstate use that; else try the meanderstate
  // -ACB- 1998/09/06
  //
  if (type->spawn_state)
    st = &states[type->spawn_state];
  else if (type->meander_state)
    st = &states[type->meander_state];
  else
    st = &states[type->idle_state];
 
  mobj->state = st;
  mobj->tics = st->tics;
  mobj->fasttics = 0;
  mobj->sprite = st->sprite;
  mobj->frame = st->frame;
  mobj->deferred_state = -1;

  // enable usable items
  if (mobj->extendedflags & EF_USABLE)
    mobj->flags |= MF_TOUCHY;
  
  // set subsector and/or block links
  P_SetThingPosition(mobj);

  // -AJA- 1999/07/30: Updated for extra floors.

  sec = mobj->subsector->sector;

  mobj->z = P_ComputeThingGap(mobj, sec, z, &mobj->floorz, &mobj->ceilingz);

  // Find the real players height (TELEPORT WEAPONS).
  mobj->origheight = z - realplayerz;  //-jc-

  // update totals for countable items.  Doing it here means that
  // things spawned dynamically can be counted as well.  Whilst this
  // has its dangers, at least it is consistent (more than can be said
  // when RTS comes into play -- trying to second guess which
  // spawnthings should not be counted just doesn't work).

  if (mobj->flags & MF_COUNTKILL)
    totalkills++;
  if (mobj->flags & MF_COUNTITEM)
    totalitems++;

  //
  // -ACB- 1998/08/27 Mobj Linked-List Addition
  //
  // A useful way of cycling through the current things without
  // having to deref everything using thinkers.
  //
  // -AJA- 1999/09/15: now adds to _head_ of list (for speed).
  //
  mobj->prev = NULL;
  mobj->next = mobjlisthead;

  if (mobjlisthead != NULL)
    mobjlisthead->prev = mobj;

  mobjlisthead = mobj;

  return mobj;
}
