//----------------------------------------------------------------------------
//  EDGE SaveGame Handling Code
//----------------------------------------------------------------------------
// 
//  Copyright (c) 1999-2000  The EDGE Team.
// 
//  This program is free software; you can redistribute it and/or
//  modify it under the terms of the GNU General Public License
//  as published by the Free Software Foundation; either version 2
//  of the License, or (at your option) any later version.
//
//  This program is distributed in the hope that it will be useful,
//  but WITHOUT ANY WARRANTY; without even the implied warranty of
//  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//  GNU General Public License for more details.
//
//----------------------------------------------------------------------------
//
//  Based on the DOOM source code, released by Id Software under the
//  following copyright:
//
//    Copyright (C) 1993-1996 by id Software, Inc.
//
//----------------------------------------------------------------------------
//
// -ACB- 1998/09/15 DEVELOPERS support added.
// -KM- 1998/11/25 Invisibility is fixed_t.
// -KM- 1998/12/16 Fixed savegames.
//

#include "i_defs.h"
#include "p_saveg.h"

#include "dm_state.h"
#include "m_inline.h"
#include "p_local.h"
#include "p_spec.h"
#include "r_state.h"
#include "w_wad.h"
#include "wi_stuff.h"
#include "z_zone.h"

byte *savebuffer;
int savegame_size;
byte *save_p;

//
// Save Game Map Object definition.
//
// If anything new is added to the mobj_t and is information that
// would need to be recorded here also, anything that is a pointer
// (that is hopefully in a linked list..) should be record as an int
// with the _id suffix. All _id types represent the number in the
// linked list that the pointer pointed to in the game so that it
// can restored when loaded, you must also add (de)referencing code
// in P_ArchiveThinkers, UnArchiveSingleThinker & P_UnArchiveThinker.
//
// -ACB- 1998/09/14 Written for DDF Items.
//
typedef struct savegmobj_s savegmobj_t;
struct savegmobj_s
{
  fixed_t x;
  fixed_t y;
  fixed_t z;

  angle_t angle;
  char sprite[4];
  int frame;

  fixed_t floorz;
  fixed_t ceilingz;

  fixed_t radius;
  fixed_t height;

  fixed_t momx;
  fixed_t momy;
  fixed_t momz;

  fixed_t speed;
  fixed_t fasttics;

  int tics;
  int fuse;
  int state;
  int flags;
  fixed_t health;

  int movedir;
  int movecount;

  int reactiontime;
  int threshold;

  int lastlook;

  mapthing_t spawnpoint;

  int source;
  int target;
  int tracer;
  int supportobj;
  int ride_em;
  int side;

  fixed_t origheight;

  fixed_t visibility;
  fixed_t delta_vis;

  int extendedflags;

  fixed_t vertangle;
  int spreadcount;

  int player;

  char *type;
  char *currentattack;

  // next & prev in the saved_mobjs list
  savegmobj_t *next, *prev;
};

// Pads save_p to a 4-byte boundary so that the load/save works on SGI&Gecko.
#define PADSAVEP() save_p += (4 - ((int) save_p & 3)) & 3

static void P_GrowSGBuffer(int numBytes)
{
  int length;

  if ((savegame_size - (save_p - savebuffer)) < numBytes)
  {
    savegame_size += numBytes;
    length = save_p - savebuffer;
    Z_Resize(savebuffer, byte, savegame_size);
    save_p = savebuffer + length;
  }
}

//
// P_ArchivePlayers
//
void P_ArchivePlayers(void)
{
  int i;
  player_t *dest;
  int *s;
  player_t *p;

  P_GrowSGBuffer(maxplayers * sizeof(player_t) +
      2 * sizeof(int) +
      numweapons * sizeof(boolean_t) +
      2 * NUMAMMO * sizeof(int));

  for (p = players; p; p = p->next)
  {
    if (!p->in_game)
      continue;

#ifdef DEVELOPERS
    I_WriteDebug("P_ArchivePlayers: Current position: %d\n",
      (int)(save_p - savebuffer));
#endif

    PADSAVEP();

    s = (int *)save_p;
    s[0] = numweapons;
    s[1] = NUMAMMO;
    save_p += 2 * sizeof(int);

    dest = (player_t *) save_p;
    Z_MoveData(dest, p, player_t, 1);
    save_p += sizeof(player_t);

//!!!    memcpy(save_p, players[i].ammo, NUMAMMO * sizeof(int));
//!!!    save_p += NUMAMMO * sizeof(int);
//!!!    memcpy(save_p, players[i].maxammo, NUMAMMO * sizeof(int));
//!!!    save_p += NUMAMMO * sizeof(int);

    // -AJA- 1999/08/11: Updated for new playerweapon_t field.
    for (i = 0; i < numweapons; i++)
    {
      s = (int *)save_p;
      *s++ = p->weapons[i].owned ? 1 : 0;
      *s++ = p->weapons[i].fire_count;
      save_p = (void *)s;
    }

    for (i = 0; i < NUMPSPRITES; i++)
    {
      if (dest->psprites[i].state)
        dest->psprites[i].state = (state_t *) (dest->psprites[i].state - states);
    }
  }

#ifdef DEVELOPERS
  I_WriteDebug("P_ArchivePlayers: Current position: %d\n",
      (int)(save_p - savebuffer));
#endif

}

//
// P_UnArchivePlayers
//
void P_UnArchivePlayers(void)
{
  int i;
  int p_numammo;
  int p_numweapons;
  int *s;
  playerweapon_t *weaps;
  playerammo_t *ammos;
  player_t *p, *next, *prev;

#ifdef DEVELOPERS
  I_WriteDebug("P_UnArchivePlayers: Current position: %d\n",
      (int)(save_p - savebuffer));
#endif

  for (p = players; p; p = p->next)
  {
#ifdef DEVELOPERS
    I_WriteDebug("P_UnArchivePlayers: Current position: %d\n", 
        (int)(save_p - savebuffer));
#endif

    if (!p->in_game)
      continue;

    PADSAVEP();

    s = (int *)save_p;
    p_numweapons = s[0];
    p_numammo = s[1];
    save_p += sizeof(int) * 2;

    weaps = p->weapons;
    ammos = p->ammo;
    next = p->next;
    prev = p->prev;

    *p = *(player_t *)save_p;
    save_p += sizeof(player_t);

    p->weapons = weaps;
    p->ammo = ammos;
    p->next = next;
    p->prev = prev;

    if (p_numweapons >= numweapons)
    {
      // -AJA- 1999/08/11: Updated for new playerweapon_t field.
      for (i = 0; i < p_numweapons; i++)
      {
        s = (int *)save_p;
        p->weapons[i].owned = (*s++) ? true : false;
        p->weapons[i].fire_count = (*s++);
        save_p = (void *)s;
      }
#ifdef DEVELOPERS
      if (p_numweapons != numweapons)
        I_WriteDebug("P_UnArchivePlayers: Savegame numweapons > current numweapons!");
#endif
    }
    else
    {

#if 0  // -AJA- FIXME
      Z_MoveData(weaponowned, save_p, boolean_t, p_numweapons);
      Z_Clear(&weaponowned[p_numweapons], boolean_t, numweapons - p_numweapons);
      save_p += p_numweapons * sizeof(boolean_t);
      I_WriteDebug("P_UnArchivePlayers: Savegame numweapons < current numweapons!");
#endif
    }

//!!!    if (p_numammo >= NUMAMMO)
//!!!    {
//!!!      memcpy(ammo, save_p, NUMAMMO * sizeof(int));
//!!!      save_p += p_numammo * sizeof(int);
//!!!      memcpy(maxammo, save_p, NUMAMMO * sizeof(int));
//!!!      save_p += p_numammo * sizeof(int);
//!!!
//!!!#ifdef DEVELOPERS
//!!!      if (p_numammo != NUMAMMO)
//!!!        I_WriteDebug("P_UnArchivePlayers: Savegame numammo > current numammo!");
//!!!#endif
//!!!    }
//!!!    else
//!!!    {
//!!!      memcpy(ammo, save_p, p_numammo * sizeof(int));
//!!!      memset(&ammo[p_numammo], 0, sizeof(int) * (NUMAMMO - p_numammo));
//!!!      save_p += p_numammo * sizeof(int);
//!!!      memcpy(maxammo, save_p, p_numammo * sizeof(int));
//!!!      memset(&maxammo[p_numammo], 0, sizeof(int) * (NUMAMMO - p_numammo));
//!!!      save_p += p_numammo * sizeof(int);
//!!!
//!!!      I_WriteDebug("P_UnArchivePlayers: Savegame numweapons < current numweapons!");
//!!!    }

    // will be set when we unarc thinker
    p->mo = NULL;
    p->attacker = NULL;

    for (i = 0; i < NUMPSPRITES; i++)
    {
      if (p->psprites[i].state)
        p->psprites[i].state = &states[(int)p->psprites[i].state];
    }
  }

#ifdef DEVELOPERS
  I_WriteDebug("P_UnArchivePlayers: Current position: %d\n", 
      (int)(save_p - savebuffer));
#endif

}

typedef struct
{
  short textureoffset;
  short rowoffset;
  short toptexture;
  short bottomtexture;
  short midtexture;
}
saveside_t;

typedef struct
{
  short flags;
  short special;
  short tag;
  saveside_t front;
  saveside_t back;
}
saveline_t;

typedef struct
{
  short floorheight;
  short ceilingheight;
  char floorpic[8];
  char ceilingpic[8];
  short lightlevel;
  short special;
  short tag;
  char colourmapname[32];
  fixed_t gravity;
  fixed_t friction;
  fixed_t viscosity;
}
savesector_t;

//
// P_ArchiveWorld
//
// Archives sector and linedef states.
//
// -ACB- 1998/09/15 Saving of sectors and lines now use set structures, Originally
//                  written as an attempted bugfix from a savegame de-sync, but it
//                  has been kept as it is much more stable then the old method.
//
void P_ArchiveWorld(void)
{
  int i;
  line_t *li;
  sector_t *sec;
  savesector_t buffsec;
  saveline_t buffline;

  // do sectors
  // -ACB- 1998/09/14 Altered to store a defined savesector type....
  // -AJA- 1999/07/09: Small change for colmap.ddf.
  for (i = 0, sec = sectors; i < numsectors; i++, sec++)
  {
    buffsec.floorheight = sec->floor.h;
    buffsec.ceilingheight = sec->ceil.h;
    Z_MoveData(buffsec.floorpic, lumpinfo[flatlist[sec->floor.pic]].name, char, 8);
    Z_MoveData(buffsec.ceilingpic, lumpinfo[flatlist[sec->ceil.pic]].name, char, 8);
    strcpy(buffsec.colourmapname, sec->p.colourmap->ddf.name);
    buffsec.lightlevel = sec->p.lightlevel;
    buffsec.special = sec->p.special ? sec->p.special->ddf.number : 0;
    buffsec.gravity = M_FloatToFixed(sec->p.gravity);
    buffsec.friction = M_FloatToFixed(sec->p.friction);
    buffsec.viscosity = M_FloatToFixed(sec->p.viscosity);
    buffsec.tag = sec->tag;

    P_GrowSGBuffer(sizeof(savesector_t));
    PADSAVEP();
    Z_MoveData((savesector_t *) save_p, &buffsec, savesector_t, 1);
    save_p += sizeof(savesector_t);
  }

  // do lines
  // -ACB- 1998/09/14 Altered to store a defined saveline type....
  for (i = 0, li = lines; i < numlines; i++, li++)
  {
#ifdef DEVELOPERS
    I_WriteDebug("P_ArchiveWorld: Line: %d Tag: %d Special: %p\n", i,
      li->tag, li->special);
#endif

    Z_Clear(&buffline, saveline_t, 1);

    buffline.flags = li->flags;
    buffline.special = li->special ? li->special->ddf.number : 0;
    buffline.tag = li->tag;

    if (li->side[0])
    {
      buffline.front.textureoffset = li->side[0]->top.x_offset;
      buffline.front.rowoffset = li->side[0]->top.y_offset;
      buffline.front.toptexture = li->side[0]->top.tex;
      buffline.front.midtexture = li->side[0]->middle.tex;
      buffline.front.bottomtexture = li->side[0]->bottom.tex;
    }

    if (li->side[1])
    {
      buffline.back.textureoffset = li->side[1]->top.x_offset;
      buffline.back.rowoffset = li->side[1]->top.y_offset;
      buffline.back.toptexture = li->side[1]->top.tex;
      buffline.back.midtexture = li->side[1]->middle.tex;
      buffline.back.bottomtexture = li->side[1]->bottom.tex;
    }

    P_GrowSGBuffer(sizeof(saveline_t));
    PADSAVEP();
    Z_MoveData((saveline_t *) save_p, &buffline, saveline_t, 1);
    save_p += sizeof(saveline_t);
  }

  // -KM- 1998/12/17 Save the intermission
  if (worldmap.ddf.name)
  {
    P_GrowSGBuffer(strlen(worldmap.ddf.name) + 1);
    sprintf(save_p, "%s", worldmap.ddf.name);
    save_p += strlen(worldmap.ddf.name) + 1;

    P_GrowSGBuffer(sizeof(boolean_t) * worldmap.nummaps + sizeof(int));

    *(int *)save_p = worldmap.nummaps;
    save_p += sizeof(int);

    Z_MoveData((boolean_t *)save_p, worldmap.mapdone, boolean_t, worldmap.nummaps);
    save_p += sizeof(boolean_t) * worldmap.nummaps;
  }
  else
  {
    P_GrowSGBuffer(1);
    *save_p++ = 0;
  }

#ifdef DEVELOPERS
  I_WriteDebug("P_ArchiveWorld: Current Position: %d\n", 
      (int)(save_p - savebuffer));
#endif
}

//
// P_UnArchiveWorld
//
// Un-Archives sector and linedef states.
//
// -ACB- 1998/09/15 Saving of sectors and lines now use set structures, Originally
//                  written as an attempted bugfix from a savegame de-sync, but it
//                  has been kept as it is much more stable then the old method.
//
void P_UnArchiveWorld(void)
{
  int i;
  line_t *li;
  sector_t *sec;
  saveline_t buffline;
  savesector_t buffsec;

  // do sectors
  // -AJA- 1999/07/09: Small change for colmap.ddf.
  for (i = 0, sec = sectors; i < numsectors; i++, sec++)
  {
    PADSAVEP();
    Z_MoveData(&buffsec, (savesector_t *) save_p, savesector_t, 1);

    sec->floor.h = buffsec.floorheight;
    sec->floor.pic = R_FlatNumForName(buffsec.floorpic);
    sec->floor.x_offset = sec->floor.y_offset = 0;
    sec->floor.translucency = 1.0;

    sec->ceil = sec->floor;
    sec->ceil.h = buffsec.ceilingheight;
    sec->ceil.pic = R_FlatNumForName(buffsec.ceilingpic);

    sec->p.colourmap = DDF_ColmapLookup(buffsec.colourmapname);
    sec->p.friction = M_FixedToFloat(buffsec.friction);
    sec->p.viscosity = M_FixedToFloat(buffsec.viscosity);
    sec->p.lightlevel = M_FixedToFloat(buffsec.lightlevel);
    sec->p.special = (buffsec.special == 0) ? NULL :
      DDF_SectorLookupNum(buffsec.special);

    sec->tag = buffsec.tag;
    sec->specialdata[0] = 0;
    sec->specialdata[1] = 0;
    sec->specialdata[2] = 0;
    sec->soundtarget = 0;

    save_p += sizeof(savesector_t);
  }

  // do lines
  for (i = 0, li = lines; i < numlines; i++, li++)
  {
    PADSAVEP();
    buffline = *(saveline_t *)save_p;
    save_p += sizeof(saveline_t);

    li->flags = buffline.flags;
    li->special = (buffline.special == 0) ? NULL :
      DDF_LineLookupNum(buffline.special);
    li->tag = buffline.tag;

    // we have a frontside?
    if (li->side[0])
    {
      li->side[0]->top.tex = buffline.front.toptexture;
      li->side[0]->top.x_offset = buffline.front.textureoffset;
      li->side[0]->top.y_offset = buffline.front.rowoffset;
      li->side[0]->top.translucency = 1.0;

      li->side[0]->middle = li->side[0]->top;
      li->side[0]->middle.tex = buffline.front.midtexture;

      li->side[0]->bottom = li->side[0]->top;
      li->side[0]->bottom.tex = buffline.front.bottomtexture;
    }

    // we have a backside?
    if (li->side[1])
    {
      li->side[1]->top.tex = buffline.back.toptexture;
      li->side[1]->top.x_offset = buffline.back.textureoffset;
      li->side[1]->top.y_offset = buffline.back.rowoffset;
      li->side[1]->top.translucency = 1.0;
      
      li->side[1]->middle = li->side[1]->top;
      li->side[1]->middle.tex = buffline.back.midtexture;

      li->side[1]->bottom = li->side[1]->top;
      li->side[1]->bottom.tex = buffline.back.bottomtexture;
    }

#ifdef DEVELOPERS
    I_WriteDebug("P_UnArchiveWorld: Line: %d Tag: %d Special: %p\n",
      i, li->tag, li->special);
#endif
  }

  // -KM- 1998/12/17 Reload the intermission.
  if (*save_p)
  {
    int nummaps;

    WI_MapInit(DDF_GameLookup(save_p));
    save_p += strlen(save_p) + 1;

    nummaps = *(int *)save_p;
    save_p += sizeof(int);

    if (nummaps >= worldmap.nummaps)
    {
      Z_MoveData(worldmap.mapdone, (boolean_t *)save_p, boolean_t, worldmap.nummaps);
      save_p += sizeof(boolean_t) * nummaps;
    }
    else
    {
      Z_MoveData(worldmap.mapdone, (boolean_t *)save_p, boolean_t, nummaps);
      Z_Clear(&worldmap.mapdone[nummaps], boolean_t, worldmap.nummaps - nummaps);
      save_p += sizeof(boolean_t) * nummaps;
    }
  }
  else
    save_p++;
}

//
// Thinkers
//
typedef enum
{
  tc_end,
  tc_mobj
}
thinkerclass_e;

//
// P_ArchiveThinkers
//
// Records the list of objects in the current world. The list
// is taken from the top to the end. The copy if not direct, since
// the mobj contains pointers that must be translated into id numbers;
// The id numbers are the position in the linked list of that type,
// these numbers will always be the same if the same DDF procedures
// and files are used; the mobj list is loaded the same because it
// will be stored in the order is currently exists.
//
// The Save-game object structure (savegmobj_t) is used to store the
// objects: it is the mobj_t with all the pointers replaced with id
// refs.
//
// -ACB- 1998/09/14 Procedure rewritten for DDF Objects...
//
void P_ArchiveThinkers(void)
{
#if 0  //!!!!! FIXME
  attacktype_t *searchattack;
#endif
  mobj_t *currmobj;
  mobj_t *searchmobj;

  savegmobj_t savemobj;
  int count;

#ifdef DEVELOPERS
  I_WriteDebug("P_ArchiveThinkers: Current Position: %d\n", 
      (int)(save_p - savebuffer));
#endif

  for (currmobj = mobjlisthead; currmobj != NULL; currmobj = currmobj->next)
  {
#ifdef DEVELOPERS
    I_WriteDebug("P_ArchiveThinkers: Thing: %s %f %f %f\n",
    currmobj->info->ddf.name, currmobj->x, currmobj->y, currmobj->z);
#endif
    // Lets go through the straight number storage bit...
    savemobj.type = currmobj->info->ddf.name;

    savemobj.x = M_FloatToFixed(currmobj->x);
    savemobj.y = M_FloatToFixed(currmobj->y);
    savemobj.z = M_FloatToFixed(currmobj->z);

    savemobj.angle = currmobj->angle;
    Z_MoveData(savemobj.sprite, sprnames[currmobj->sprite], char, 4);
    savemobj.frame = currmobj->frame;

    savemobj.floorz = M_FloatToFixed(currmobj->floorz);
    savemobj.ceilingz = M_FloatToFixed(currmobj->ceilingz);

    savemobj.radius = M_FloatToFixed(currmobj->radius);
    savemobj.height = M_FloatToFixed(currmobj->height);

    savemobj.momx = M_FloatToFixed(currmobj->momx);
    savemobj.momy = M_FloatToFixed(currmobj->momy);
    savemobj.momz = M_FloatToFixed(currmobj->momz);

    savemobj.speed = M_FloatToFixed(currmobj->speed);

    savemobj.tics = currmobj->tics;
    savemobj.fasttics = M_FloatToFixed(currmobj->fasttics);
    savemobj.fuse = currmobj->fuse;
    savemobj.state = currmobj->state - states;
    savemobj.flags = currmobj->flags;
    savemobj.health = M_FloatToFixed(currmobj->health);

    savemobj.side = currmobj->side;

    savemobj.movedir = currmobj->movedir;
    savemobj.movecount = currmobj->movecount;

    savemobj.reactiontime = currmobj->reactiontime;
    savemobj.threshold = currmobj->threshold;

    savemobj.lastlook = currmobj->lastlook;

    savemobj.spawnpoint = currmobj->spawnpoint;

    savemobj.origheight = M_FloatToFixed(currmobj->origheight);

    savemobj.visibility = M_FloatToFixed(currmobj->visibility);
    savemobj.delta_vis = M_FloatToFixed(currmobj->delta_vis);

    savemobj.extendedflags = currmobj->extendedflags;

    savemobj.spreadcount = currmobj->spreadcount;
    savemobj.vertangle = M_FloatToFixed(currmobj->vertangle);

#if 0  //!!!!! FIXME
    // Now store the current attack type id number
    searchattack = attackhead;

    while (searchattack != currmobj->currentattack && searchattack)
      searchattack = searchattack->next;

    if (searchattack)
      savemobj.currentattack = searchattack->ddf.name;
    else
#endif
      savemobj.currentattack = "";

    // Moving Object References....
    //
    // Searches through the list of objects to see if any match the
    // following objects: if so the object is stored.
    //
    savemobj.source = -1;
    savemobj.target = -1;
    savemobj.tracer = -1;
    savemobj.supportobj = -1;
    savemobj.ride_em = -1;

    count = 0;

    for (searchmobj = mobjlisthead; searchmobj != NULL; searchmobj = searchmobj->next)
    {
      if (currmobj->source == searchmobj)
        savemobj.source = count;

      if (currmobj->target == searchmobj)
        savemobj.target = count;

      if (currmobj->tracer == searchmobj)
        savemobj.tracer = count;

      if (currmobj->supportobj == searchmobj)
        savemobj.supportobj = count;

      if (currmobj->ride_em == searchmobj)
        savemobj.ride_em = count;

      count++;
    }

    // lets store the ref for what the object is (mobjinfo)
    count = 0;

    if (currmobj->player)
      savemobj.player = currmobj->player->pnum;
    else
      savemobj.player = -1;

    P_GrowSGBuffer(sizeof(savegmobj_t) + 3 - sizeof(char *) + strlen(savemobj.currentattack) +
        strlen(savemobj.type));

    *save_p++ = tc_mobj;
    PADSAVEP();
    *(savegmobj_t *)save_p = savemobj;
    save_p += sizeof(savegmobj_t) - (2 * sizeof(char *));

    save_p += sprintf(save_p, "%s", savemobj.type) + 1;
    save_p += sprintf(save_p, "%s", savemobj.currentattack) + 1;
  }

  // add a terminating marker
  P_GrowSGBuffer(1);
  *save_p++ = tc_end;

#ifdef DEVELOPERS
  I_WriteDebug("P_ArchiveThinkers: Current Position: %d\n", 
      (int)(save_p - savebuffer));
#endif

}

//
// UnArchiveSingleThinker
//
// Converts the saved save-game object into the world object. Its
// de-references the id numbers to pointers in the object, this
// does NOT happen with pointers to other objects here. Its is
// attempted when the objects list has been fully reconstructed (for
// obvious reasons).
//
// -ACB- 1998/09/14 Procedure Written.
//
static void UnArchiveSingleThinker(savegmobj_t * savedmobj)
{
#if 0  //!!!!! FIXME
  attacktype_t *searchattack;
#endif
  mobj_t *mobj;
  mobj_t *listcurrmobj;
  const mobjinfo_t *objecttype;
  int i;

  mobj = Z_New(mobj_t, 1);

  mobj->x = M_FixedToFloat(savedmobj->x);
  mobj->y = M_FixedToFloat(savedmobj->y);
  mobj->z = M_FixedToFloat(savedmobj->z);

  mobj->angle = savedmobj->angle;
  for (i = 0; i < num_sprnames; i++)
    if (!memcmp(savedmobj->sprite, sprnames[i], 4))
    {
      mobj->sprite = i;
      break;
    }

  if (i == num_sprnames)
    I_Error("UnArchiveSingleThinker: Unknown sprite name '%.4s'\n", savedmobj->sprite);

  mobj->frame = savedmobj->frame;

  mobj->floorz = M_FixedToFloat(savedmobj->floorz);
  mobj->ceilingz = M_FixedToFloat(savedmobj->ceilingz);

  mobj->radius = M_FixedToFloat(savedmobj->radius);
  mobj->height = M_FixedToFloat(savedmobj->height);

  mobj->momx = M_FixedToFloat(savedmobj->momx);
  mobj->momy = M_FixedToFloat(savedmobj->momy);
  mobj->momz = M_FixedToFloat(savedmobj->momz);

  mobj->speed = M_FixedToFloat(savedmobj->speed);

  savedmobj->type = save_p;
  save_p += strlen(savedmobj->type) + 1;
  objecttype = DDF_MobjLookup(savedmobj->type);

  if (!objecttype)
    I_Error("UnArchiveSingleThinker: Unknown Mobjinfo type encountered\n");

  mobj->info = objecttype;
  mobj->tics = savedmobj->tics;
  mobj->fasttics = M_FixedToFloat(savedmobj->fasttics);
  mobj->fuse = savedmobj->fuse;
  mobj->state = &states[savedmobj->state];
  mobj->flags = savedmobj->flags;
  mobj->health = M_FixedToFloat(savedmobj->health);
  mobj->deferred_state = -1;  // -AJA- FIXME.

  mobj->side = savedmobj->side;

  mobj->movedir = savedmobj->movedir;
  mobj->movecount = savedmobj->movecount;

  mobj->reactiontime = savedmobj->reactiontime;
  mobj->threshold = savedmobj->threshold;

  mobj->lastlook = savedmobj->lastlook;

  mobj->spawnpoint = savedmobj->spawnpoint;

  mobj->origheight = M_FixedToFloat(savedmobj->origheight);

  mobj->visibility = M_FixedToFloat(savedmobj->visibility);
  mobj->delta_vis = M_FixedToFloat(savedmobj->delta_vis);

  mobj->extendedflags = savedmobj->extendedflags;

  mobj->vertangle = M_FixedToFloat(savedmobj->vertangle);
  mobj->spreadcount = savedmobj->spreadcount;

  mobj->currentattack = NULL;

  if (savedmobj->player != -1)
  {
    mobj->player = playerlookup[savedmobj->player];
    mobj->player->mo = mobj;
  }
  else
  {
    mobj->player = NULL;
  }

  savedmobj->currentattack = save_p;
  save_p += strlen(savedmobj->currentattack) + 1;

#if 0  // !!!!! FIXME
  if (savedmobj->currentattack[0])
  {
    searchattack = attackhead;
    while (searchattack && !strcmp(searchattack->ddf.name, savedmobj->currentattack))
      searchattack = searchattack->next;

    if (!searchattack)
      I_Error("UnArchiveSingleThinker: Unknown Attack Type id found\n");

    mobj->currentattack = searchattack;
  }
  else
#endif
  {
    mobj->currentattack = NULL;
  }

  P_SetThingPosition(mobj);

  //
  // -ACB- 1998/08/27 Mobj Linked-List Addition
  //
  // A useful way of cycling through the current things without
  // having to deref everything using thinkers.
  //
  if (mobjlisthead == NULL)
  {
    mobjlisthead = mobj;
    mobj->prev = NULL;
    mobj->next = NULL;
  }
  else
  {
    listcurrmobj = mobjlisthead;

    while (listcurrmobj->next != NULL)
      listcurrmobj = listcurrmobj->next;

    listcurrmobj->next = mobj;
    mobj->prev = listcurrmobj;
    mobj->next = NULL;
  }
}

//
// P_UnArchiveThinkers
//
// Loaded the objects from the same games, the conversion from
// save-game object to world-object (mobj_t) takes place. Once
// all of the objects have been loaded from the file, the final
// action of this procedure is to translate the id numbers of a
// save-game object into pointers for the world-objects to other
// world-objects. Confused? You will be...
//
// -ACB- 1998/09/14 Rewritten for DDF Items.
//
void P_UnArchiveThinkers(void)
{
  boolean_t endfound;
  byte tclass;
  mobj_t *mobj;
  savegmobj_t *savedmobj;

  // List of all the saved mobjs. Circular list with savedmobjs as a dummy
  // between the head and tail.
  savegmobj_t savedmobjs;

  int nummobjs = 0;
  mobj_t **mobj_list;

  int i;

  endfound = false;

#ifdef DEVELOPERS
  I_WriteDebug("P_UnArchiveThinkers: Current Position: %d\n", 
      (int)(save_p - savebuffer));
#endif

  // Initialise the savedmobjs list.
  savedmobjs.next = savedmobjs.prev = &savedmobjs;

  while (!endfound)
  {
    tclass = *save_p++;

    // so what we reading exactly? tclass should be either tc_end or tc_mobj.
    switch (tclass)
    {
      case tc_end:
        {
          endfound = true;  // end of list

          break;
        }

      case tc_mobj:
        {
          PADSAVEP();

          savedmobj = Z_New(savegmobj_t, 1);
          *savedmobj = *(savegmobj_t *)save_p;
          savedmobj->type = savedmobj->currentattack = NULL;
          save_p += sizeof(savegmobj_t) - (2 * sizeof(char *));

          savedmobjs.prev->next = savedmobj;
          savedmobj->prev = savedmobjs.prev;
          savedmobjs.prev = savedmobj;
          savedmobj->next = &savedmobjs;

          UnArchiveSingleThinker(savedmobj);
          nummobjs++;

          break;
        }
      default:
        {
          I_Error("Unknown tclass %i in savegame", tclass);
          // lets be pedantic :)
          break;
        }
    }
  }

  //
  // The list of objects has now been reconstructed in the order from the
  // savegame file (and in the same order that it existed when the game was
  // saved), we will now reference the id numbers to pointers to the
  // correct object, any failure indicates a corrupt savegame file.
  //
  mobj = mobjlisthead;
  mobj_list = Z_New(mobj_t *, nummobjs);

  for (mobj = mobjlisthead, i = 0; mobj; i++, mobj = mobj->next)
    mobj_list[i] = mobj;

  savedmobj = savedmobjs.next;
  for (mobj = mobjlisthead; mobj; mobj = mobj->next)
  {
    P_MobjSetTarget(mobj, savedmobj->target == -1 ? NULL : mobj_list[savedmobj->target]);
    P_MobjSetSource(mobj, savedmobj->source == -1 ? NULL : mobj_list[savedmobj->source]);
    P_MobjSetTracer(mobj, savedmobj->tracer == -1 ? NULL : mobj_list[savedmobj->tracer]);
    P_MobjSetSupportObj(mobj, savedmobj->supportobj == -1 ? NULL : mobj_list[savedmobj->supportobj]);
    P_MobjSetRideEm(mobj, savedmobj->ride_em == -1 ? NULL : mobj_list[savedmobj->ride_em]);

    // we don't need the saved mobj any more, so throw it away.
    savedmobj = savedmobj->next;
    Z_Free(savedmobj->prev);
  }

  Z_Free(mobj_list);

#ifdef DEVELOPERS
  I_WriteDebug("P_UnArchiveThinkers: Current Position: %d\n", 
      (int)(save_p - savebuffer));
#endif

}

typedef enum
{
  tq_end,
  tq_item
}
queclass_t;

//
// P_ArchiveItemRespawnQue
//
// Archives the Item-Respawn-Que.
//
// -ACB- 1998/09/15
//
void P_ArchiveItemRespawnQue(void)
{
  iteminque_t *curritem;

  for (curritem = itemquehead; curritem != NULL; curritem = curritem->next)
  {
    P_GrowSGBuffer(sizeof(mapthing_t) + 1 + sizeof(int));

    *save_p++ = tq_item;
    PADSAVEP();
    *(mapthing_t *)save_p = curritem->info;
    save_p += sizeof(mapthing_t);
    *(int *)save_p = curritem->time;
    save_p += sizeof(int);
  }

  P_GrowSGBuffer(1);
  *save_p++ = tq_end;
}

//
// P_ArchiveItemRespawnQue
//
// Archives the Item-Respawn-Que.
//
// -ACB- 1998/09/15
//
void P_UnArchiveItemRespawnQue(void)
{
  iteminque_t *newitem;
  iteminque_t *searchitem;

  newitem = Z_New(iteminque_t, 1);

  while (*save_p++ != tq_end)
  {
    PADSAVEP();
    newitem->info = *(mapthing_t *) save_p;
    save_p += sizeof(mapthing_t);
    newitem->time = *(int *)save_p;
    save_p += sizeof(int);

    searchitem = itemquehead;

    if (itemquehead)
    {
      while (searchitem->next != NULL)
        searchitem = searchitem->next;

      searchitem->next = newitem;
      newitem->prev = searchitem;
    }
    else
    {
      itemquehead = newitem;
      newitem->prev = NULL;
    }

    newitem->next = NULL;
  }

}

//
// P_ArchiveSpecials
//
typedef enum
{
  tc_sector,
  tc_light,
  tc_endspecials
}
specials_e;

//
// P_ArchiveSpecials
//
void P_ArchiveSpecials(void)
{
#ifdef DEVELOPERS
  I_WriteDebug("P_ArchiveSpecials: Current Position: %d\n", 
      (int)(save_p - savebuffer));
#endif

  // -ES- Fixme: Re-add stuff here.
  // We currently need to save light_t, sec_move_t and sectorsfx_t
  
  // add a terminating marker
  P_GrowSGBuffer(1);
  *save_p++ = tc_endspecials;

#ifdef DEVELOPERS
  I_WriteDebug("P_ArchiveSpecials: Current Position: %d\n", 
      (int)(save_p - savebuffer));
#endif

}

typedef struct alloc_mp_s
{
  struct alloc_mp_s *next;
  moving_plane_t mp;
} alloc_mp_t;

static alloc_mp_t *alloc_mps;

void P_DestroyMovingPlanes(void)
{
  alloc_mp_t *next;

  while (alloc_mps)
  {
    next = alloc_mps->next;
    Z_Free(alloc_mps);
    alloc_mps = next;
  }
}

static moving_plane_t *NewMovingPlane(void)
{
  alloc_mp_t *mp = Z_New(alloc_mp_t, 1);

  mp->next = alloc_mps;
  alloc_mps = mp;
  return &mp->mp;
}

//
// P_UnArchiveSpecials
//
// -KM- 1998/09/01 Remove doors/plats (DDF)
//
void P_UnArchiveSpecials(void)
{
  byte tclass;

#ifdef DEVELOPERS
  I_WriteDebug("P_UnArchiveSpecials: Current Position: %d\n", 
      (int)(save_p - savebuffer));
#endif

  // read in saved thinkers
  while (1)
  {
    tclass = *save_p++;
    switch (tclass)
    {
      case tc_endspecials:
#ifdef DEVELOPERS
        I_WriteDebug("P_UnArchiveSpecials: Current Position: %d\n", 
            (int)(save_p - savebuffer));
#endif
        return;  // end of list

      // -ES- Fixme: Re-add stuff here.

      default:
        I_Error("P_UnarchiveSpecials: Unknown tclass %i in savegame", tclass);
    }
  }
}
