//----------------------------------------------------------------------------
//  EDGE Level Loading/Setup Code
//----------------------------------------------------------------------------
// 
//  Copyright (c) 1999-2000  The EDGE Team.
// 
//  This program is free software; you can redistribute it and/or
//  modify it under the terms of the GNU General Public License
//  as published by the Free Software Foundation; either version 2
//  of the License, or (at your option) any later version.
//
//  This program is distributed in the hope that it will be useful,
//  but WITHOUT ANY WARRANTY; without even the implied warranty of
//  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//  GNU General Public License for more details.
//
//----------------------------------------------------------------------------
//
//  Based on the DOOM source code, released by Id Software under the
//  following copyright:
//
//    Copyright (C) 1993-1996 by id Software, Inc.
//
//----------------------------------------------------------------------------

#include "i_defs.h"
#include "p_setup.h"

#include "ddf_main.h"
#include "dm_defs.h"
#include "dm_state.h"
#include "g_game.h"
#include "m_swap.h"
#include "m_bbox.h"
#include "m_fixed.h"
#include "p_local.h"
#include "p_bot.h"
#include "r2_defs.h"
#include "s_sound.h"
#include "w_textur.h"
#include "w_wad.h"
#include "z_zone.h"


// debugging aide:
#define FORCE_LOCATION  0
#define FORCE_LOC_X     1120
#define FORCE_LOC_Y     644
#define FORCE_LOC_ANG   90


//
// MAP related Lookup tables.
// Store VERTEXES, LINEDEFS, SIDEDEFS, etc.
//
int numvertexes;
vertex_t *vertexes;

int num_gl_vertexes;
vertex_t *gl_vertexes;

int numsegs;
seg_t *segs;

int numsectors;
sector_t *sectors;

int numsubsectors;
subsector_t *subsectors;

int numnodes;
node_t *nodes;

int numlines;
line_t *lines;

int numsides;
side_t *sides;

// BLOCKMAP
//
// Created from axis aligned bounding box
// of the map, a rectangular array of
// blocks of size ...
// Used to speed up collision detection
// by spatial subdivision in 2D.
//
// Blockmap size.
// 23-6-98 KM Promotion of short * to int *
int bmapwidth;
int bmapheight;  // size in mapblocks

int *blockmap = NULL;  // int for larger maps

// offsets in blockmap are from here
int *blockmaplump = NULL;

// origin of block map
float_t bmaporgx;
float_t bmaporgy;

// for thing chains
mobj_t **blocklinks;

// REJECT
//
// For fast sight rejection.
// Speeds up enemy AI by skipping detailed
//  LineOf Sight calculation.
// Without special effect, this could be
//  used as a PVS lookup as well.
//
const byte *rejectmatrix;

// Maintain single and multi player starting spots.
int max_deathmatch_starts = 10;
mapthing_t *deathmatchstarts;
mapthing_t *deathmatch_p;

mapthing_t *playerstarts;

//
// LoadVertexes
//
static void LoadVertexes(int lump)
{
  const void *data;
  int i;
  const mapvertex_t *ml;
  vertex_t *li;

  if (! W_VerifyLumpName(lump, "VERTEXES"))
    I_Error("Bad WAD: level %s missing VERTEXES.\n", currentmap->lump);

  // Determine number of lumps:
  //  total lump length / vertex record length.
  numvertexes = W_LumpLength(lump) / sizeof(mapvertex_t);

  if (numvertexes == 0)
    I_Error("Bad WAD: level %s contains 0 vertexes.\n", currentmap->lump);

  // Allocate zone memory for buffer.
  vertexes = Z_New(vertex_t, numvertexes);

  // Load data into cache.
  data = W_CacheLumpNum(lump);

  ml = (const mapvertex_t *) data;
  li = vertexes;

  // Copy and convert vertex coordinates,
  // internal representation as fixed.
  for (i = 0; i < numvertexes; i++, li++, ml++)
  {
    li->x = SHORT(ml->x);
    li->y = SHORT(ml->y);
  }

  // Free buffer memory.
  W_DoneWithLump(data);
}

//
// LoadGLVertexes
//
static void LoadGLVertexes(int lump)
{
  const void *data;
  int i;
  const mapvertex_t *ml;
  vertex_t *vert;

  if (!W_VerifyLumpName(lump, "GL_VERT"))
    I_Error("Bad WAD: level %s missing GL_VERT.\n", currentmap->lump);

  // Determine number of lumps:
  //  total lump length / vertex record length.
  num_gl_vertexes = W_LumpLength(lump) / sizeof(mapvertex_t);

  gl_vertexes = Z_New(vertex_t, num_gl_vertexes);

  // Load data into cache.
  data = W_CacheLumpNum(lump);

  ml = (const mapvertex_t *) data;
  vert = gl_vertexes;

  // Copy and convert vertex coordinates,
  // internal representation as fixed.
  for (i = 0; i < num_gl_vertexes; i++, vert++, ml++)
  {
    vert->x = SHORT(ml->x);
    vert->y = SHORT(ml->y);
  }

  // Free buffer memory.
  W_DoneWithLump(data);
}

//
// LoadSegs
//
static void LoadSegs(int lump)
{
  const void *data;
  int i;
  const mapseg_t *ml;
  seg_t *seg;
  int linedef;
  int side;

  if (! W_VerifyLumpName(lump, "SEGS"))
    I_Error("Bad WAD: level %s missing SEGS.\n", currentmap->lump);

  numsegs = W_LumpLength(lump) / sizeof(mapseg_t);

  if (numsegs == 0)
    I_Error("Bad WAD: level %s contains 0 segs.\n", currentmap->lump);

  segs = Z_ClearNew(seg_t, numsegs);
  data = W_CacheLumpNum(lump);

  ml = (const mapseg_t *) data;
  seg = segs;
  
  for (i = 0; i < numsegs; i++, seg++, ml++)
  {
    vertex_t *v1 = &vertexes[SHORT(ml->v1)];
    vertex_t *v2 = &vertexes[SHORT(ml->v2)];

    seg->x1 = v1->x;
    seg->y1 = v1->y;
    seg->x2 = v2->x;
    seg->y2 = v2->y;

    seg->angle = (SHORT(ml->angle)) << 16;
    seg->offset = SHORT(ml->offset);

    // compute length
    seg->length = R_PointToDist(seg->x1, seg->y1, seg->x2, seg->y2);
    
    linedef = SHORT(ml->linedef);
    side = SHORT(ml->side);

    seg->linedef = &lines[linedef];
    seg->sidedef = seg->linedef->side[side];
    seg->frontsector = seg->sidedef->sector;
    seg->backsector = NULL;

    if (seg->linedef->flags & ML_TWOSIDED)
      seg->backsector = seg->linedef->side[side^1]->sector;

    // The following fields are filled out elsewhere:
    //     sub_next, front_sub, back_sub, frontsector, backsector.

    seg->miniseg = false;
    seg->sub_next = (void *) -3;
    seg->front_sub = seg->back_sub = (void *) -3;
  }

  W_DoneWithLump(data);
}

//
// LoadGLSegs
//
static void LoadGLSegs(int lump)
{
  const void *data;
  int i;
  const map_glseg_t *ml;
  seg_t *seg;
  int linedef;
  int side;
  int partner;

  if (! W_VerifyLumpName(lump, "GL_SEGS"))
    I_Error("Bad WAD: level %s missing GL_SEGS.\n", currentmap->lump);

  numsegs = W_LumpLength(lump) / sizeof(map_glseg_t);

  if (numsegs == 0)
    I_Error("Bad WAD: level %s contains 0 gl-segs.\n", currentmap->lump);

  segs = Z_ClearNew(seg_t, numsegs);
  data = W_CacheLumpNum(lump);

  ml = (const map_glseg_t *) data;
  seg = segs;
  
  for (i = 0; i < numsegs; i++, seg++, ml++)
  {
    int v1num = SHORT(ml->v1);
    int v2num = SHORT(ml->v2);

    vertex_t *v1, *v2;

    if (v1num & SF_GL_VERTEX)
      v1 = &gl_vertexes[v1num & ~SF_GL_VERTEX];
    else
      v1 = &vertexes[v1num];

    if (v2num & SF_GL_VERTEX)
      v2 = &gl_vertexes[v2num & ~SF_GL_VERTEX];
    else
      v2 = &vertexes[v2num];

    seg->x1 = v1->x;
    seg->y1 = v1->y;
    seg->x2 = v2->x;
    seg->y2 = v2->y;

    seg->angle  = R_PointToAngle(seg->x1, seg->y1, seg->x2, seg->y2);
    seg->length = R_PointToDist (seg->x1, seg->y1, seg->x2, seg->y2);

    linedef = SHORT(ml->linedef);
    side = SHORT(ml->side);

    seg->frontsector = seg->backsector = NULL;

    if (linedef < 0)
      seg->miniseg = 1;
    else
    {
      float_t sx, sy;

      seg->miniseg = 0;
      seg->linedef = &lines[linedef];

      sx = side ? seg->linedef->v2->x : seg->linedef->v1->x;
      sy = side ? seg->linedef->v2->y : seg->linedef->v1->y;

      seg->offset = R_PointToDist(sx, sy, seg->x1, seg->y1);

      seg->sidedef = seg->linedef->side[side];
      seg->frontsector = seg->sidedef->sector;

      if (seg->linedef->flags & ML_TWOSIDED)
        seg->backsector = seg->linedef->side[side^1]->sector;
    }

    partner = SHORT(ml->partner);

    if (partner < 0)
      seg->partner = NULL;
    else
      seg->partner = &segs[partner];

    // The following fields are filled out elsewhere:
    //     sub_next, front_sub, back_sub, frontsector, backsector.

    seg->sub_next = (void *) -3;
    seg->front_sub = seg->back_sub = (void *) -3;
  }

  W_DoneWithLump(data);
}

//
// LoadSubsectors
//
static void LoadSubsectors(int lump, const char *name)
{
  int i, j;
  const void *data;
  const mapsubsector_t *ms;
  subsector_t *ss;

  if (! W_VerifyLumpName(lump, name))
    I_Error("Bad WAD: level %s missing %s.\n", currentmap->lump, name);

  numsubsectors = W_LumpLength(lump) / sizeof(mapsubsector_t);

  if (numsubsectors == 0)
    I_Error("Bad WAD: level %s contains 0 ssectors.\n", currentmap->lump);

  subsectors = Z_ClearNew(subsector_t, numsubsectors);

  data = W_CacheLumpNum(lump);
  ms = (const mapsubsector_t *) data;
  ss = subsectors;

  for (i = 0; i < numsubsectors; i++, ss++, ms++)
  {
    int countsegs = SHORT(ms->numsegs);
    int firstseg = SHORT(ms->firstseg);

    // -AJA- 1999/09/23: New linked list for the segs of a subsector
    //       (part of true bsp rendering).
    seg_t **prevptr = & ss->segs;

    if (countsegs <= 0 || firstseg < 0 || firstseg+countsegs > numsegs)
      I_Error("Bad WAD: level %s has invalid SSECTORS.\n", currentmap->lump);
        
    ss->sector = NULL;
    ss->thinglist = NULL;

    for (j = 0; j < countsegs; j++)
    {
      seg_t *cur = &segs[firstseg + j];

      *prevptr = cur;
       prevptr = &cur->sub_next;

      cur->front_sub = ss;
      cur->back_sub = NULL;

      if (!ss->sector && !cur->miniseg)
        ss->sector = cur->sidedef->sector;
    }

    if (ss->sector == NULL)
      I_Error("Bad WAD: level %s has crazy SSECTORS.\n", currentmap->lump);

    *prevptr = NULL;

    // link subsector into parent sector's list.
    // order is not important, so add it to the head of the list.
    
    ss->sec_next = ss->sector->subsectors;
    ss->sector->subsectors = ss;
  }

  W_DoneWithLump(data);
}

//
// GroupSectorTags
//
// Called during P_LoadSectors to set the tag_next & tag_prev fields of
// each sector_t, which keep all sectors with the same tag in a linked
// list for faster handling.
//
// -AJA- 1999/07/29: written.

static void GroupSectorTags(sector_t * dest, sector_t * seclist, int numsecs)
{
  // NOTE: `numsecs' does not include the current sector.

  dest->tag_next = dest->tag_prev = NULL;

  for (; numsecs > 0; numsecs--)
  {
    sector_t *src = &seclist[numsecs - 1];

    if (src->tag == dest->tag)
    {
      src->tag_next = dest;
      dest->tag_prev = src;
      return;
    }
  }
}

//
// LoadSectors
//
static void LoadSectors(int lump)
{
  const void *data;
  int i;
  const mapsector_t *ms;
  sector_t *ss;

  if (! W_VerifyLumpName(lump, "SECTORS"))
    I_Error("Bad WAD: level %s missing SECTORS.\n", currentmap->lump);

  numsectors = W_LumpLength(lump) / sizeof(mapsector_t);

  if (numsectors == 0)
    I_Error("Bad WAD: level %s contains 0 sectors.\n", currentmap->lump);
    
  sectors = Z_ClearNew(sector_t, numsectors);

  data = W_CacheLumpNum(lump);
  ms = (const mapsector_t *) data;
  ss = sectors;
  for (i = 0; i < numsectors; i++, ss++, ms++)
  {
    ss->floor.h = SHORT(ms->floorheight);
    ss->floor.pic = R_FlatNumForName(ms->floorpic);
    ss->floor.x_offset = ss->floor.y_offset = 0;
    ss->floor.translucency = 1.0;

    ss->ceil = ss->floor;
    ss->ceil.h = SHORT(ms->ceilingheight);
    ss->ceil.pic = R_FlatNumForName(ms->ceilingpic);

    ss->p.lightlevel = SHORT(ms->lightlevel);
    ss->p.special = (SHORT(ms->special) <= 0) ? NULL :
      DDF_SectorLookupNum(SHORT(ms->special));

    ss->tag = SHORT(ms->tag);

    // -AJA- 1999/07/29: Keep sectors with same tag in a list.
    GroupSectorTags(ss, sectors, i);

    // -AJA- 1999/10/09: Create initial region.
    P_ComputeSectorRegions(ss);
  }

  W_DoneWithLump(data);
}

//
// LoadNodes
//
static void LoadNodes(int lump, char *name)
{
  const void *data;
  int i;
  int j;
  int k;
  const mapnode_t *mn;
  node_t *no;

  if (! W_VerifyLumpName(lump, name))
    I_Error("Bad WAD: level %s missing %s.\n", currentmap->lump, name);

  // Note: zero numnodes is valid.
  numnodes = W_LumpLength(lump) / sizeof(mapnode_t);
  nodes = Z_New(node_t, numnodes);
  data = W_CacheLumpNum(lump);

  mn = (const mapnode_t *) data;
  no = nodes;

  for (i = 0; i < numnodes; i++, no++, mn++)
  {
    no->line.x = SHORT(mn->x);
    no->line.y = SHORT(mn->y);
    no->line.dx = SHORT(mn->dx);
    no->line.dy = SHORT(mn->dy);

    for (j = 0; j < 2; j++)
    {
      no->children[j] = SHORT(mn->children[j]);

      for (k = 0; k < 4; k++)
        no->bbox[j][k] = (float_t) SHORT(mn->bbox[j][k]);
    }
  }

  W_DoneWithLump(data);
}

//
// LoadThings
//
static void LoadThings(int lump)
{
  const void *data;
  int i;
  const mapthing_t *mt;
  mapthing_t t;
  int numthings;

  if (!W_VerifyLumpName(lump, "THINGS"))
    I_Error("Bad WAD: level %s missing THINGS.\n", currentmap->lump);

  data = W_CacheLumpNum(lump);
  numthings = W_LumpLength(lump) / sizeof(mapthing_t);

  if (numthings == 0)
    I_Error("Bad WAD: level %s contains 0 things.\n", currentmap->lump);

  mt = (const mapthing_t *) data;
  for (i = 0; i < numthings; i++, mt++)
  {
    // Do spawn all other stuff. 
    t.x = SHORT(mt->x);
    t.y = SHORT(mt->y);
    t.angle = SHORT(mt->angle);
    t.type = SHORT(mt->type);
    t.options = SHORT(mt->options);

#if FORCE_LOCATION
    if (t.type == 1)
    {
      t.x = FORCE_LOC_X;
      t.y = FORCE_LOC_Y;
      t.angle = FORCE_LOC_ANG;
    }
#endif

    P_SpawnMapThing(&t);
  }

  W_DoneWithLump(data);
}

//
// LoadLineDefs
// Also counts secret lines for intermissions.
//
static void LoadLineDefs(int lump)
{
  const void *data;
  int i;
  const maplinedef_t *mld;
  line_t *ld;
  vertex_t *v1;
  vertex_t *v2;
  int side0, side1;

  if (! W_VerifyLumpName(lump, "LINEDEFS"))
    I_Error("Bad WAD: level %s missing LINEDEFS.\n", currentmap->lump);

  numlines = W_LumpLength(lump) / sizeof(maplinedef_t);

  if (numlines == 0)
    I_Error("Bad WAD: level %s contains 0 linedefs.\n", currentmap->lump);
    
  lines = Z_ClearNew(line_t, numlines);
  data = W_CacheLumpNum(lump);

  mld = (const maplinedef_t *) data;
  ld = lines;
  for (i = 0; i < numlines; i++, mld++, ld++)
  {
    ld->flags = SHORT(mld->flags);
    ld->tag = SHORT(mld->tag);
    v1 = ld->v1 = &vertexes[SHORT(mld->v1)];
    v2 = ld->v2 = &vertexes[SHORT(mld->v2)];
    ld->dx = v2->x - v1->x;
    ld->dy = v2->y - v1->y;

    ld->special = (SHORT(mld->special) <= 0) ? NULL :
        DDF_LineLookupNum(SHORT(mld->special));
    
    if (ld->dx == 0)
      ld->slopetype = ST_VERTICAL;
    else if (ld->dy == 0)
      ld->slopetype = ST_HORIZONTAL;
    else
    {
      if (ld->dy / ld->dx > 0)
        ld->slopetype = ST_POSITIVE;
      else
        ld->slopetype = ST_NEGATIVE;
    }

    if (v1->x < v2->x)
    {
      ld->bbox[BOXLEFT] = v1->x;
      ld->bbox[BOXRIGHT] = v2->x;
    }
    else
    {
      ld->bbox[BOXLEFT] = v2->x;
      ld->bbox[BOXRIGHT] = v1->x;
    }

    if (v1->y < v2->y)
    {
      ld->bbox[BOXBOTTOM] = v1->y;
      ld->bbox[BOXTOP] = v2->y;
    }
    else
    {
      ld->bbox[BOXBOTTOM] = v2->y;
      ld->bbox[BOXTOP] = v1->y;
    }

    side0 = SHORT(mld->sidenum[0]);
    side1 = SHORT(mld->sidenum[1]);

    ld->side[0] = ld->side[1] = NULL;
    ld->frontsector = ld->backsector = NULL;

    if (side0 != -1)
    {
      ld->side[0] = &sides[side0];
      ld->frontsector = ld->side[0]->sector;
    }

    if (side1 != -1)
    {
      ld->side[1] = &sides[side1];
      ld->backsector = ld->side[1]->sector;
    }

    if ((ld->flags & ML_TWOSIDED) && ((side0 == -1) || (side1 == -1)))
      I_Error("Bad WAD: level %s has linedef #%d marked TWOSIDED, "
          "but it has only one side.\n", currentmap->lump, i);

    P_ComputeGaps(ld);
  }

  W_DoneWithLump(data);
}

//
// LoadSideDefs
//
static void LoadSideDefs(int lump)
{
  const void *data;
  int i;
  const mapsidedef_t *msd;
  side_t *sd;

  if (! W_VerifyLumpName(lump, "SIDEDEFS"))
    I_Error("Bad WAD: level %s missing SIDEDEFS.\n", currentmap->lump);

  numsides = W_LumpLength(lump) / sizeof(mapsidedef_t);

  if (numsides == 0)
    I_Error("Bad WAD: level %s contains 0 sidedefs.\n", currentmap->lump);
    
  sides = Z_ClearNew(side_t, numsides);
  data = W_CacheLumpNum(lump);

  msd = (const mapsidedef_t *) data;
  sd = sides;

  for (i = 0; i < numsides; i++, msd++, sd++)
  {
    sd->top.tex = W_TextureNumForName(msd->toptexture);
    sd->top.x_offset = SHORT(msd->textureoffset);
    sd->top.y_offset = SHORT(msd->rowoffset);
    sd->top.translucency = 1.0;

    sd->middle = sd->top;
    sd->middle.tex = W_TextureNumForName(msd->midtexture);

    sd->bottom = sd->top;
    sd->bottom.tex = W_TextureNumForName(msd->bottomtexture);

    sd->sector = &sectors[SHORT(msd->sector)];
  }

  W_DoneWithLump(data);
}

//
// LoadBlockMap
//

// 23-6-98 KM Brand Spanking New Blockmap code:
// I think there is a bug somewhere in P_GenerateBlockMap, so it
// is commented out, but since P_LoadBlockMap should be able
// to generate any level that doom can take, no problemo.

static void LoadBlockMap(int lump)
{
  int i;
  int count;
  const short *bmap;

  if (! W_VerifyLumpName(lump, "BLOCKMAP"))
    I_Error("Bad WAD: level %s missing BLOCKMAP.\n", currentmap->lump);

  bmap = W_CacheLumpNum(lump);
  count = W_LumpLength(lump) / sizeof(short);
  blockmaplump = Z_New(int, count);

  blockmap = blockmaplump + 4;

  for (i = 0; i < 4; i++)
    blockmaplump[i] = (int)(signed short)SHORT(bmap[i]);

  bmaporgx = (float_t)blockmaplump[0];
  bmaporgy = (float_t)blockmaplump[1];
  bmapwidth = blockmaplump[2];
  bmapheight = blockmaplump[3];

  for (; i < (bmapwidth * bmapheight); i++)
    blockmaplump[i] = (int)(unsigned short)SHORT(bmap[i]);

  for (; i < count; i++)
    blockmaplump[i] = (int)(signed short)SHORT(bmap[i]);

  W_DoneWithLump(bmap);
}

static void GenerateBlockMap(int min_x, int min_y, int max_x, int max_y)
{
  int count;
  int **bmap;
  int *bmaplen;
  int *maxbmaplen;
  int i, j;
  // number of items in blockmaplump
  int size;

  int x0, x1, y0, y1;
  int blocknum;

  int blockstep;
  boolean_t column, row;

  int x_sign, y_sign;
  float_t slope, x_jump, next_x, this_x;
  int lastblock;

  line_t *line;

  int pointer_p;
  int blockmap_p;

  bmaporgx = min_x - 8;
  bmaporgy = min_y - 8;
  bmapwidth = (max_x - (min_x - 8) + 127) / 128;
  bmapheight = (max_y - (min_y - 8) + 127) / 128;

  count = bmapwidth * bmapheight;

  bmap = Z_New(int *, count);
  bmaplen = Z_New(int, count);
  maxbmaplen = Z_New(int, count);
  // for the -1 terminations of each bmap
  size = count;

  for (i = count; i--;)
  {
    bmap[i] = Z_New(int, 16);

    bmap[i][0] = 0;
    bmaplen[i] = 1;
    maxbmaplen[i] = 16;
    size++;
  }

  for (i = 0; i < numlines; i++)
  {
    line = &lines[i];
    x0 = line->v1->x - bmaporgx;
    x1 = line->v2->x - bmaporgx;
    y0 = line->v1->y - bmaporgy;
    y1 = line->v2->y - bmaporgy;

    blocknum = (y0 / 128) * bmapwidth + (x0 / 128);
    if (bmaplen[blocknum] == maxbmaplen[blocknum])
    {
      maxbmaplen[blocknum] += 16;
      Z_Resize(bmap[blocknum], int, maxbmaplen[blocknum]);
    }
    size++;

    bmap[blocknum][bmaplen[blocknum]] = i;
    bmaplen[blocknum]++;

    // Check if this line spans multiple blocks.
    column = ((x0 / 128) == (x1 / 128));
    row = ((y0 / 128) == (y1 / 128));
    switch (column * 2 + row)
    {
        // Entirely in one block
      case 3:
        break;
      case 2:
        blockstep = (y1 >= y0) ? bmapwidth : -bmapwidth;
        for (j = 0; j < abs((y1 / 128) - (y0 / 128)); j++)
        {
          blocknum += blockstep;
          if (bmaplen[blocknum] == maxbmaplen[blocknum])
          {
            maxbmaplen[blocknum] += 16;
            Z_Resize(bmap[blocknum], int, maxbmaplen[blocknum]);
          }
          size++;

          bmap[blocknum][bmaplen[blocknum]] = i;
          bmaplen[blocknum]++;
        }
        break;
      case 1:
        blockstep = (x1 >= x0) ? 1 : -1;
        for (j = 0; j < abs((x1 / 128) - (x0 / 128)); j++)
        {
          blocknum += blockstep;
          if (bmaplen[blocknum] == maxbmaplen[blocknum])
          {
            maxbmaplen[blocknum] += 16;
            Z_Resize(bmap[blocknum], int, maxbmaplen[blocknum]);
          }
          size++;

          bmap[blocknum][bmaplen[blocknum]] = i;
          bmaplen[blocknum]++;
        }
        break;
      default:
        y_sign = (y1 >= y0) ? 1 : -1;
        x_sign = (x1 >= x0) ? 1 : -1;

        slope = ((float_t)(y1 - y0)) / (x1 - x0);
        x_jump = y_sign * 128.0 / slope;
        if (y_sign < 0)
          next_x = x0 + ((y0 & ~127) - 1 - y0) / slope;
        else
          next_x = x0 + ((y0 & ~127) + 128 - y0) / slope;

        lastblock = (int)(next_x / 128) - (x0 / 128) + blocknum;
        if (lastblock > blocknum)
        {
          for (j = blocknum + x_sign; j <= lastblock; j += x_sign)
          {
            if (bmaplen[j] == maxbmaplen[j])
            {
              maxbmaplen[j] += 16;
              Z_Resize(bmap[j], int, maxbmaplen[j]);
            }
            size++;

            bmap[j][bmaplen[j]] = i;
            bmaplen[j]++;
          }
        }

        do
        {
          this_x = next_x;
          next_x = this_x + x_jump;
          if ((x_sign * next_x) > (x_sign * x1))
          {
            j = lastblock + y_sign * bmapwidth;
            lastblock = j + (int)(next_x / 128) - (int)(this_x / 128);
            for (; j <= lastblock; j += x_sign)
            {
              if (bmaplen[j] == maxbmaplen[j])
              {
                maxbmaplen[j] += 16;
                Z_Resize(bmap[j], int, maxbmaplen[j]);
              }
              size++;

              bmap[j][bmaplen[j]] = i;
              bmaplen[j]++;
            }
          }
        }
        while ((int)next_x != x1);
        break;
    }
  }

  // 4 for the header, count for the pointers
  size += 4 + count + count;

  blockmaplump = Z_New(int, size);

  blockmaplump[0] = min_x - 8;
  blockmaplump[1] = min_y - 8;
  blockmaplump[2] = bmapwidth;
  blockmaplump[3] = bmapheight;

  blockmap = blockmaplump + 4;
  pointer_p = 4;
  blockmap_p = 4 + count;

  for (i = 0; i < count; i++)
  {
    blockmaplump[pointer_p] = blockmap_p;
    for (j = 0; j < bmaplen[i]; j++)
      blockmaplump[blockmap_p + j] = bmap[i][j];
    blockmaplump[blockmap_p + j] = -1;
    blockmap_p += bmaplen[i] + 1;
    pointer_p++;
    Z_Free(bmap[i]);
  }
  Z_Free(bmap);
  Z_Free(bmaplen);
  Z_Free(maxbmaplen);
}

static void DoBlockMap(int lump)
{
  vertex_t *v;
  int i, x, y;
  int min_x = vertexes[0].x, min_y = vertexes[0].y;
  int max_x = vertexes[0].x, max_y = vertexes[0].y;
  int map_width, map_height;
  int count;

  for (v = vertexes, i = numvertexes; i--; v++)
  {
    x = v->x;
    y = v->y;
    if (x < min_x)
      min_x = x;
    else if (x > max_x)
      max_x = x;
    if (y < min_y)
      min_y = y;
    else if (y > max_y)
      max_y = y;
  }
  map_width = max_x - min_x;
  map_height = max_y - min_y;
//    if ((map_width < 32768) && (map_height < 32768))
  LoadBlockMap(lump);
//    else
//      GenerateBlockMap(min_x, min_y, max_x, max_y);

  // clear out mobj chains
  count = bmapwidth * bmapheight;
  blocklinks = Z_ClearNew(mobj_t *, count);
}

// all sectors' line tables
static line_t **linebuffer;

//
// GroupLines
//
// Builds sector line lists and subsector sector numbers.
// Finds block bounding boxes for sectors.
//
static void GroupLines(void)
{
  int i;
  int j;
  int total;
  line_t *li;
  sector_t *sector;
  seg_t *seg;
  float_t bbox[4];
  int block;
  line_t **line_p;

  // setup remaining seg information
  for (i=0, seg=segs; i < numsegs; i++, seg++)
  {
    if (seg->partner)
      seg->back_sub = seg->partner->front_sub;
    
    if (!seg->frontsector)
      seg->frontsector = seg->front_sub->sector;

    if (!seg->backsector && seg->back_sub)
      seg->backsector = seg->back_sub->sector;
  }
  
  // count number of lines in each sector
  li = lines;
  total = 0;
  for (i = 0; i < numlines; i++, li++)
  {
    total++;
    li->frontsector->linecount++;

    if (li->backsector && li->backsector != li->frontsector)
    {
      li->backsector->linecount++;
      total++;
    }
  }

  // build line tables for each sector 
  line_p = linebuffer = Z_New(line_t *, total);
  sector = sectors;
  for (i = 0; i < numsectors; i++, sector++)
  {
    M_ClearBox(bbox);
    sector->lines = line_p;
    li = lines;
    for (j = 0; j < numlines; j++, li++)
    {
      if (li->frontsector == sector || li->backsector == sector)
      {
        *line_p++ = li;
        M_AddToBox(bbox, li->v1->x, li->v1->y);
        M_AddToBox(bbox, li->v2->x, li->v2->y);
      }
    }
    if (line_p - sector->lines != sector->linecount)
      I_Error("GroupLines: miscounted");

    // set the degenmobj_t to the middle of the bounding box
    sector->soundorg.x = (bbox[BOXRIGHT] + bbox[BOXLEFT]) / 2;
    sector->soundorg.y = (bbox[BOXTOP] + bbox[BOXBOTTOM]) / 2;
    sector->soundorg.z = (sector->floor.h + sector->ceil.h) / 2;

    // adjust bounding box to map blocks
    block = (bbox[BOXTOP] - bmaporgy + MAXRADIUS) / MAPBLOCKUNITS;
    block = block >= bmapheight ? bmapheight - 1 : block;
    sector->blockbox[BOXTOP] = block;

    block = (bbox[BOXBOTTOM] - bmaporgy - MAXRADIUS) / MAPBLOCKUNITS;
    block = block < 0 ? 0 : block;
    sector->blockbox[BOXBOTTOM] = block;

    block = (bbox[BOXRIGHT] - bmaporgx + MAXRADIUS) / MAPBLOCKUNITS;
    block = block >= bmapwidth ? bmapwidth - 1 : block;
    sector->blockbox[BOXRIGHT] = block;

    block = (bbox[BOXLEFT] - bmaporgx - MAXRADIUS) / MAPBLOCKUNITS;
    block = block < 0 ? 0 : block;
    sector->blockbox[BOXLEFT] = block;
  }
}

//
// LoadReject
//
static void LoadReject(int lump)
{
  if (! W_VerifyLumpName(lump, "REJECT"))
    I_Error("Bad WAD: level %s missing REJECT.\n", currentmap->lump);

  rejectmatrix = W_CacheLumpNum(lump);
}

static void RemoveMobjs(void)
{
  mobj_t *mo;
  iteminque_t *tmp;

  for (mo = mobjlisthead; mo; mo = mo->next)
    P_RemoveMobj(mo);
  P_RemoveQueuedMobjs();
  P_RemoveBots();

  while (itemquehead)
  {
    tmp = itemquehead;
    itemquehead = itemquehead->next;
    Z_Free(tmp);
  }
}

static void RemoveSectors(void)
{
  int i;
  vert_region_t *curreg, *nextreg;
  extrafloor_record_t *currec, *nextrec;

  for (i = 0; i < numsectors; i++)
  {
    // Free regions.
    for (curreg = sectors[i].regions; curreg; curreg = nextreg)
    {
      nextreg = curreg->sec_next;
      Z_Free(curreg);
    }
    // Free extrafloor records.
    for (currec = sectors[i].extrafloor_records; currec; currec = nextrec)
    {
      nextrec = currec->next;
      Z_Free(currec);
    }
    // Might still be playing a sound.
    S_StopSound((mobj_t *)&sectors[i].soundorg);
  }
  Z_Free(sectors);
}

static boolean_t level_active = false;

//
// P_ShutdownLevel
//
// Destroys everything on the level.
static void ShutdownLevel(void)
{
#ifdef DEVELOPERS
  if (!level_active)
    I_Error("ShutdownLevel: no level to shut down!");
#endif

  RemoveMobjs();

  if (gl_vertexes)
  {
    Z_Free(gl_vertexes);
    gl_vertexes = NULL;
  }
  Z_Free(segs);
  R2_FreeAllocatedSegs();
  Z_Free(nodes);
  Z_Free(vertexes);
  Z_Free(sides);
  Z_Free(lines);
  Z_Free(subsectors);
  P_DestroyMovingPlanes();
  RemoveSectors();
  Z_Free(linebuffer);
  Z_Free(blocklinks);
  Z_Free(blockmaplump);
  P_DestroyAllLights();
  P_RemoveAllActiveSectors();
  P_DestroyAllSectorSFX();
  blockmaplump = blockmap = NULL;
  W_DoneWithLump(rejectmatrix);
  
  level_active = false;
}

//
// P_SetupLevel
//
// Sets up the current level using the skill passed and the
// information in currentmap.
//
// -ACB- 1998/08/09 Use currentmap to ref lump and par time
// -KM- 1998/11/25 Added autotag.
//
void P_SetupLevel(skill_t skill, int autotag)
{
  int lumpnum;
  int gl_lumpnum;
  player_t *p;
  char gl_lumpname[16];

  if (level_active)
    ShutdownLevel();

  totalkills = totalitems = totalsecret = wminfo.maxfrags = 0;

  wminfo.partime = currentmap->partime;

  for (p = players; p; p = p->next)
    p->killcount = p->secretcount = p->itemcount = 0;

  // Initial height of PointOfView
  // will be set by player think.
  consoleplayer->viewz = M_PI;

  // if working with a development map, reload it
  W_Reload();

  lumpnum = W_GetNumForName(currentmap->lump);

  leveltime = 0;

  // -AJA- 1999/12/20: Support for "GL-Friendly Nodes".
  sprintf(gl_lumpname, "GL_%s", currentmap->lump);
  gl_lumpnum = W_CheckNumForName(gl_lumpname);

  // ignore GL info if the level marker occurs _before_ the normal
  // level marker.
  if (gl_lumpnum >= 0 && gl_lumpnum < lumpnum)
    gl_lumpnum = -1;
  
  // -ACB- 1998/08/27 NULL the head pointers for the linked lists....
  itemquehead = NULL;
  mobjlisthead = NULL;

  // note: most of this ordering is important
  // 23-6-98 KM, eg, Sectors must be loaded before sidedefs,
  // Sidedefs + Vertexes must be loaded before LineDefs,
  // LineDefs + Vertexes must be loaded before BlockMap,
  // Sectors must be loaded before Segs

  LoadVertexes(lumpnum + ML_VERTEXES);
  LoadSectors(lumpnum + ML_SECTORS);
  LoadSideDefs(lumpnum + ML_SIDEDEFS);
  LoadLineDefs(lumpnum + ML_LINEDEFS);

  DoBlockMap(lumpnum + ML_BLOCKMAP);

  if (gl_lumpnum >= 0)
  {
    LoadGLVertexes(gl_lumpnum + ML_GL_VERT);
    LoadGLSegs(gl_lumpnum + ML_GL_SEGS);
    LoadSubsectors(gl_lumpnum + ML_GL_SSECT, "GL_SSECT");
    LoadNodes(gl_lumpnum + ML_GL_NODES, "GL_NODES");

    if (! force_classic)
      use_true_bsp = true;
  }
  else
  {
    LoadSegs(lumpnum + ML_SEGS);
    LoadSubsectors(lumpnum + ML_SSECTORS, "SSECTORS");
    LoadNodes(lumpnum + ML_NODES, "NODES");
  }
  LoadReject(lumpnum + ML_REJECT);

  GroupLines();

  // do we need to polygonise ?
  if (gl_lumpnum < 0 && use_true_bsp)
    R2_Polygonise();

  bodyqueslot = 0;
  deathmatch_p = deathmatchstarts;

  // -AJA- 1999/10/21: Clear out playerstarts.
  Z_Clear(playerstarts, mapthing_t, maxplayers);

  LoadThings(lumpnum + ML_THINGS);

  // if deathmatch, randomly spawn the active players
  if (deathmatch)
  {
    for (p = players; p; p = p->next)
    {
      if (p->in_game)
      {
        p->mo = NULL;
        G_DeathMatchSpawnPlayer(p);
        if (p->level_init)
          p->level_init(p, p->data);
      }
    }
  }

  // -AJA- 1999/10/21: if not netgame/deathmatch, then check for
  //       missing player start.  NOTE: temp fix, player handling
  //       desperately needs a massive overhaul.
  if (!(netgame || deathmatch) && consoleplayer->mo == NULL)
    I_Error("Missing player start !\n");

  // set up world state
  P_SpawnSpecials(autotag);

  // preload graphics
  if (precache)
    R_PrecacheLevel();

  S_SoundLevelInit(); // Clear out the playing sounds
  S_ChangeMusic(currentmap->music, true); // start level music

  level_active = true;
}

//
// P_Init
//
boolean_t P_Init(void)
{
  static boolean_t firsttime = true;

  if (firsttime)
  {
    P_MapInit();
    R_InitSprites(sprnames);
    firsttime = false;
  }
  deathmatchstarts = Z_New(mapthing_t, max_deathmatch_starts);
  playerstarts = Z_New(mapthing_t, maxplayers);

  P_InitSwitchList();
  P_InitPicAnims();

  return true;
}
