//----------------------------------------------------------------------------
//  EDGE Sight Code
//----------------------------------------------------------------------------
// 
//  Copyright (c) 1999-2000  The EDGE Team.
// 
//  This program is free software; you can redistribute it and/or
//  modify it under the terms of the GNU General Public License
//  as published by the Free Software Foundation; either version 2
//  of the License, or (at your option) any later version.
//
//  This program is distributed in the hope that it will be useful,
//  but WITHOUT ANY WARRANTY; without even the implied warranty of
//  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//  GNU General Public License for more details.
//
//----------------------------------------------------------------------------
//
//  Based on the DOOM source code, released by Id Software under the
//  following copyright:
//
//    Copyright (C) 1993-1996 by id Software, Inc.
//
//----------------------------------------------------------------------------

#include "i_defs.h"

#include "dm_defs.h"
#include "p_local.h"
#include "r_state.h"

//
// P_CheckSight
//

// eye z of looker
static float_t sightzstart;
static float_t sight_topslope;
static float_t sight_bottomslope;  // slopes to top and bottom of target

static divline_t strace;  // from t1 to t2

static float_t t2x;
static float_t t2y;

static int sightcounts[2];

//
// DivlineSide
// Returns side 0 (front), 1 (back), or 2 (on).
//
static int DivlineSide(float_t x, float_t y, divline_t * node)
{
  float_t dx;
  float_t dy;
  float_t left;
  float_t right;

  if (0 == node->dx)
  {
    if (x == node->x)
      return 2;

    if (x <= node->x)
      return node->dy > 0;

    return node->dy < 0;
  }

  if (0 == node->dy)
  {
    if (x == node->y)
      return 2;

    if (y <= node->y)
      return node->dx < 0;

    return node->dx > 0;
  }

  dx = x - node->x;
  dy = y - node->y;

  left = node->dy * dx;
  right = dy * node->dx;

  if (right < left)
    return 0;  // front side

  if (left == right)
    return 2;
  return 1;  // back side

}

//
// InterceptVector2
// Returns the fractional intercept point
// along the first divline.
//
static float_t InterceptVector2(divline_t * v2, divline_t * v1)
{
  float_t num;
  float_t den;

  den = v1->dy * v2->dx - v1->dx * v2->dy;

  // parallel ?
  if (den == 0)
    return 0;

  num = (v1->x - v2->x) * v1->dy +
      (v2->y - v1->y) * v1->dx;

  return num / den;
}

//
// P_CrossSubsector
//
// Returns true if strace crosses the given subsector successfully.
//
// -AJA- 1999/07/31: FIXME: doesn't handle extra floors

static boolean_t CrossSubsector(int num)
{
  seg_t *seg;
  line_t *line;
  int s1;
  int s2;
  subsector_t *sub;
  sector_t *front;
  sector_t *back;
  float_t opentop;
  float_t openbottom;
  divline_t divl;
  vertex_t *v1;
  vertex_t *v2;
  float_t frac;
  float_t slope;
  float_t dist;

#ifdef DEVELOPERS
  if (num >= numsubsectors)
    I_Error("CrossSubsector: ss %i with numss = %i",
        num,
        numsubsectors);
#endif

  sub = &subsectors[num];

  // check lines
  for (seg = sub->segs; seg != NULL; seg = seg->sub_next)
  {
    if (seg->miniseg)
      continue;

    line = seg->linedef;

    // allready checked other side?
    if (line->validcount == validcount)
      continue;

    line->validcount = validcount;

    v1 = line->v1;
    v2 = line->v2;
    s1 = DivlineSide(v1->x, v1->y, &strace);
    s2 = DivlineSide(v2->x, v2->y, &strace);

    // line isn't crossed?
    if (s1 == s2)
      continue;

    divl.x = v1->x;
    divl.y = v1->y;
    divl.dx = v2->x - v1->x;
    divl.dy = v2->y - v1->y;
    s1 = DivlineSide(strace.x, strace.y, &divl);
    s2 = DivlineSide(t2x, t2y, &divl);

    // line isn't crossed?
    if (s1 == s2)
      continue;

    // stop because it is not two sided anyway
    // -AJA- 1999/07/28: moved up here, and uses the line gaps.
    if (!(line->flags & ML_TWOSIDED) || line->gap_num == 0)
      return false;

    // crosses a two sided line
    front = seg->frontsector;
    back = seg->backsector;

    // no wall to block sight with?
    if (front->floor.h == back->floor.h &&
        front->ceil.h == back->ceil.h)
      continue;

    // possible occluder because of ceiling height differences
    if (front->ceil.h < back->ceil.h)
      opentop = front->ceil.h;
    else
      opentop = back->ceil.h;

    // possible occluder because of floor height differences
    if (front->floor.h > back->floor.h)
      openbottom = front->floor.h;
    else
      openbottom = back->floor.h;

    // -AJA- OPTIMISE...
    frac = InterceptVector2(&strace, &divl);
    dist = P_ApproxDistance(strace.dx * frac, strace.dy * frac);

    // -ACB- 2000/03/11 Div-by-zero
    if (!dist)
      dist=1;

    if (front->floor.h != back->floor.h)
    {
      slope = (openbottom - sightzstart) / dist;
      if (slope > sight_bottomslope)
        sight_bottomslope = slope;
    }

    if (front->ceil.h != back->ceil.h)
    {
      slope = (opentop - sightzstart) / dist;
      if (slope < sight_topslope)
        sight_topslope = slope;
    }

    if (sight_topslope <= sight_bottomslope)
      return false;  // stop    

  }
  // passed the subsector ok
  return true;
}

//
// CrossBSPNode
//
// Returns true if strace crosses the given node successfully.

static boolean_t CrossBSPNode(int bspnum)
{
  node_t *bsp;
  int side;

  if (bspnum & NF_SUBSECTOR)
  {
    if (bspnum == -1)
      return CrossSubsector(0);
    else
      return CrossSubsector(bspnum & (~NF_SUBSECTOR));
  }

  bsp = &nodes[bspnum];

  // decide which side the start point is on
  side = DivlineSide(strace.x, strace.y, &bsp->line);
  if (side == 2)
    side = 0;  // an "on" should cross both sides

  // cross the starting side
  if (!CrossBSPNode(bsp->children[side]))
    return false;

  // the partition plane is crossed here
  if (side == DivlineSide(t2x, t2y, &bsp->line))
  {
    // the line doesn't touch the other side
    return true;
  }

  // cross the ending side  
  return CrossBSPNode(bsp->children[side ^ 1]);
}

//
// P_CheckSight
//
// Returns true if a straight line between t1 and t2 is unobstructed.
// Uses the REJECT info.

boolean_t P_CheckSight(mobj_t * t1, mobj_t * t2)
{
  int s1;
  int s2;
  int pnum;
  int bytenum;
  int bitnum;

  // First check for trivial rejection.

  // -ACB- 1998/07/20 t2 is Invisible, t1 cannot possibly see it.
  if (t2->visibility == INVISIBLE)
    return false;

  // Determine subsector entries in REJECT table.
  s1 = (t1->subsector->sector - sectors);
  s2 = (t2->subsector->sector - sectors);
  pnum = s1 * numsectors + s2;
  bytenum = pnum >> 3;
  bitnum = 1 << (pnum & 7);

  // Check in REJECT table.
  if (rejectmatrix[bytenum] & bitnum)
  {
    sightcounts[0]++;

    // can't possibly be connected
    return false;
  }

  // An unobstructed LOS is possible.
  // Now look from eyes of t1 to any part of t2.
  sightcounts[1]++;

  validcount++;

  // The "eyes" of a thing is 75% of its height.
  sightzstart = t1->z + t1->height - t1->height / 4;

  // -AJA- 1999/10/24: DDF control over slope.
  sight_topslope = t1->info->sight_slope;
  sight_bottomslope = -t1->info->sight_slope;

  strace.x = t1->x;
  strace.y = t1->y;
  t2x = t2->x;
  t2y = t2->y;
  strace.dx = t2->x - t1->x;
  strace.dy = t2->y - t1->y;

  // the head node is the last node output
  return CrossBSPNode(numnodes - 1);
}
