//----------------------------------------------------------------------------
//  EDGE Specials Lines & Floor Code
//----------------------------------------------------------------------------
// 
//  Copyright (c) 1999-2000  The EDGE Team.
// 
//  This program is free software; you can redistribute it and/or
//  modify it under the terms of the GNU General Public License
//  as published by the Free Software Foundation; either version 2
//  of the License, or (at your option) any later version.
//
//  This program is distributed in the hope that it will be useful,
//  but WITHOUT ANY WARRANTY; without even the implied warranty of
//  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//  GNU General Public License for more details.
//
//----------------------------------------------------------------------------
//
//  Based on the DOOM source code, released by Id Software under the
//  following copyright:
//
//    Copyright (C) 1993-1996 by id Software, Inc.
//
//----------------------------------------------------------------------------
//
// -KM- 1998/09/01 Lines.ddf
//

#include "i_defs.h"
#include "p_spec.h"

#include "con_main.h"
#include "dm_defs.h"
#include "dm_state.h"
#include "g_game.h"
#include "m_argv.h"
#include "m_inline.h"
#include "m_random.h"
#include "p_local.h"
#include "r_local.h"
#include "r_state.h"
#include "rad_trig.h"
#include "s_sound.h"
#include "w_textur.h"
#include "w_wad.h"
#include "z_zone.h"

#include "ddf_main.h"  // -KM- 98/07/31 Need animation definitions

//
// Animating textures and planes
// There is another anim_t used in wi_stuff, unrelated.
//
typedef struct
{
  boolean_t istexture;
  int picnum;
  int basepic;
  int numpics;
  int speed;
}
anim_t;

// -KM- 98/07/31 Moved init to ddf_anim.h
// -ES- 2000/02/04 Switched to stack_array
static stack_array_t anim_a;
static anim_t **anims;
static int numanims;

//
// Animating line specials
//
static stack_array_t linespecial_a;
static int numlinespecials;
static line_t ** linespeciallist;

//
// Animating sector specials
//
static stack_array_t sect_special_a;
static int num_sect_specials;
static sector_t ** sect_speciallist;

boolean_t P_DoSectorsFromTag(int tag, const void *p1, void *p2,
    boolean_t(*func) (sector_t *, const void *, void *));

static boolean_t DoSector_wrapper(sector_t *s, const void *p1, void *p2)
{
  return EV_DoSector(s, p1, p2);
}
static boolean_t DoLights_wrapper(sector_t *s, const void *p1, void *p2)
{
  return EV_Lights(s, p1);
}
static boolean_t DoDonut_wrapper(sector_t *s, const void *p1, void *p2)
{
  return EV_DoDonut(s, p2);
}

//
// P_InitPicAnims
//
// -ACB- 1999/09/25 modified for new boolean_t type
//
void P_InitPicAnims(void)
{
  int i, picnum, basepic;
  anim_t *addanim;

  if (!anim_a.ptr)
    Z_InitStackArray(&anim_a, (void ***)&anims, sizeof(anim_t), 0);
    
  numanims = 0;
  Z_SetArraySize(&anim_a, 0);

  // loop through animdefs, and add relevant anims.
  for (i = 0; i < numanimdefs; i++)
  {
    animdef_t *A = animdefs[i];

    if (A->istexture == true)
    {
      basepic = W_CheckTextureNumForName(A->startname);
      if (basepic == -1)
        continue;

      picnum = W_TextureNumForName(A->endname);
    }
    else
    {
      if (W_CheckNumForName(A->startname) == -1)
        continue;

      picnum = R_FlatNumForName(A->endname);
      basepic = R_FlatNumForName(A->startname);
    }

    numanims++;
    Z_SetArraySize(&anim_a, numanims);
    addanim = anims[numanims-1];

    addanim->picnum = picnum;
    addanim->basepic = basepic;
    addanim->istexture = A->istexture;

    if (addanim->istexture == true)
      addanim->numpics = addanim->picnum - addanim->basepic + 1;
    else
      addanim->numpics = flatlist[addanim->picnum] - flatlist[addanim->basepic] + 1;

    if (addanim->numpics < 2)
    {
      I_Error("P_InitPicAnims: bad cycle from %s to %s",
          A->startname, A->endname);
    }

    addanim->speed = A->speed;
  }
}

void P_ReInitPicAnims(int pos)
{
  int i;
  anim_t *addanim;

  // Init animation
  for (i = 0; i < numanims; i++)
  {
    animdef_t *A = animdefs[i];

    addanim = anims[i];

    if (!A->istexture)
    {
      if (W_CheckNumForName(A->startname) == -1)
        continue;

      if (addanim->picnum >= pos)
        addanim->picnum++;
      if (addanim->basepic >= pos)
        addanim->basepic++;

      addanim->numpics = flatlist[addanim->picnum] - flatlist[addanim->basepic] + 1;
    }
  }
}

typedef struct sectorsfx_s
{
  sector_t *sector;
  sfx_t *sfx;
  boolean_t sfxstarted;
  int count;
  struct sectorsfx_s *prev, *next;
}
sectorsfx_t;

static sectorsfx_t *sectorsfx_list;

//
// Creates a sector sfx and adds it to the list.
//
static sectorsfx_t *NewSectorSFX(void)
{
  sectorsfx_t *sfx;

  sfx = Z_New(sectorsfx_t, 1);
  sfx->next = sectorsfx_list;
  sfx->prev = NULL;
  sectorsfx_list = sfx;
  
  return sfx;
}

void P_DestroyAllSectorSFX(void)
{
  sectorsfx_t *sfx, *next;

  for (sfx = sectorsfx_list; sfx; sfx = next)
  {
    next = sfx->next;
    Z_Free(sfx);
  }
  sectorsfx_list = NULL;
}

//
// DoSectorSFX
//
// -KM- 1998/09/27
//
// This function is called every so often to keep a sector's ambient
// sound going.  The sound should be looped.
//
static void DoSectorSFX(sectorsfx_t *sec)
{
  if (--sec->count)
    return;

  sec->count = 7;

  if (!sec->sfxstarted)
  {
    if (P_ApproxDistance(displayplayer->mo->x - sec->sector->soundorg.x,
            displayplayer->mo->y - sec->sector->soundorg.y)
        < S_CLIPPING_DIST)
    {
      int channel = S_StartSound((mobj_t *) & sec->sector->soundorg,
          sec->sfx);

      if (channel >= 0)
        sec->sfxstarted = true;
    }
  }
  else
  {
    if (P_ApproxDistance(displayplayer->mo->x - sec->sector->soundorg.x,
            displayplayer->mo->y - sec->sector->soundorg.y)
        > S_CLIPPING_DIST)
    {
      S_StopSound((mobj_t *) &sec->sector->soundorg);
      sec->sfxstarted = false;
    }
  }
}

void P_RunSectorSFX(void)
{
  sectorsfx_t *sfx;

  for (sfx = sectorsfx_list; sfx; sfx = sfx->next)
  {
    DoSectorSFX(sfx);
  }
}

//
// UTILITIES
//

//
// P_GetSide
//
// Will return a side_t * given the number of the current sector,
// the line number, and the side (0/1) that you want.
//
side_t *P_GetSide(int currentSector, int line, int side)
{
  line_t *linedef = sectors[currentSector].lines[line];

  return linedef->side[side];
}

//
// P_GetSector
// Will return a sector_t*
//  given the number of the current sector,
//  the line number and the side (0/1) that you want.
//
sector_t *P_GetSector(int currentSector, int line, int side)
{
  line_t *linedef = sectors[currentSector].lines[line];

  return side ? linedef->backsector : linedef->frontsector;
}

//
// P_TwoSided
//
// Given the sector number and the line number, it will tell you whether the
// line is two-sided or not.
//
int P_TwoSided(int sector, int line)
{
  return (sectors[sector].lines[line])->flags & ML_TWOSIDED;
}

//
// P_GetNextSector
//
// Return sector_t * of sector next to current; NULL if not two-sided line
//
sector_t *P_GetNextSector(line_t * line, sector_t * sec)
{
  if (!(line->flags & ML_TWOSIDED))
    return NULL;

  if (line->frontsector == sec)
    return line->backsector;

  return line->frontsector;
}

//
// P_FindLowestFloorSurrounding()
// FIND LOWEST FLOOR HEIGHT IN SURROUNDING SECTORS
//
float_t P_FindLowestFloorSurrounding(sector_t * sec)
{
  int i;
  line_t *check;
  sector_t *other;
  float_t floor = sec->floor.h;

  for (i = 0; i < sec->linecount; i++)
  {
    check = sec->lines[i];
    other = P_GetNextSector(check, sec);

    if (!other)
      continue;

    if (other->floor.h < floor)
      floor = other->floor.h;
  }

  return floor;
}

//
// P_FindRaiseToTexture()
//
// FIND THE SHORTEST BOTTOM TEXTURE SURROUNDING sec
// AND RETURN IT'S TOP HEIGHT
//
// -KM- 1998/09/01 Lines.ddf; used to be inlined in p_floors
//
float_t P_FindRaiseToTexture(sector_t * sec)
{
  int i;
  side_t *side;
  float_t minsize = INT_MAX;
  int secnum = sec - sectors;

  for (i = 0; i < sec->linecount; i++)
  {
    if (P_TwoSided(secnum, i))
    {
      side = P_GetSide(secnum, i, 0);

      if (side->bottom.tex >= 0)
      {
        if (textureheight[side->bottom.tex] < minsize)
          minsize = textureheight[side->bottom.tex];
      }

      side = P_GetSide(secnum, i, 1);

      if (side->bottom.tex >= 0)
      {
        if (textureheight[side->bottom.tex] < minsize)
          minsize = textureheight[side->bottom.tex];
      }
    }
  }

  return sec->floor.h + minsize;
}

//
// P_FindHighestFloorSurrounding()
//
// FIND HIGHEST FLOOR HEIGHT IN SURROUNDING SECTORS
//
float_t P_FindHighestFloorSurrounding(sector_t * sec)
{
  int i;
  line_t *check;
  sector_t *other;
  float_t floor = INT_MIN;

  for (i = 0; i < sec->linecount; i++)
  {
    check = sec->lines[i];
    other = P_GetNextSector(check, sec);

    if (!other)
      continue;

    if (other->floor.h > floor)
      floor = other->floor.h;
  }

  return floor;
}

//
// P_FindNextHighestFloor
//
// FIND NEXT HIGHEST FLOOR IN SURROUNDING SECTORS
//
// 23-6-98 KM, done without a fixed array, based on
// the other find next thingys, fixed by ACB.
//
// -ACB- 1998/07/21 if no sectors of greater height found, set min to currentheight.
//
float_t P_FindNextHighestFloor(sector_t * sec, float_t height)
{
  int i, count = 0;
  float_t min = INT_MAX;
  sector_t *other;

  for (i = 0; i < sec->linecount; i++)
  {
    other = P_GetNextSector(sec->lines[i], sec);

    if (!other)
      continue;

    if ((other->floor.h > height) && (other->floor.h < min))
    {
      min = other->floor.h;
      count++;
    }
  }

  // -ACB- 1998/07/21 if No sectors of great height found, set min to currentheight.
  if (count == 0)
    return height;

  return min;
}

//
// P_FindNextLowestFloor
//
// -AJA- 1999/10/22: written.

float_t P_FindNextLowestFloor(sector_t * sec, float_t height)
{
  int i, count = 0;
  float_t max = INT_MIN;
  sector_t *other;

  for (i = 0; i < sec->linecount; i++)
  {
    other = P_GetNextSector(sec->lines[i], sec);

    if (!other)
      continue;

    if ((other->floor.h < height) && (other->floor.h > max))
    {
      max = other->floor.h;
      count++;
    } 
  }

  if (count == 0)
    return height;

  return max;
}

//
// P_FindNextHighestCeiling
//
// -AJA- 1999/10/22: written.

float_t P_FindNextHighestCeiling(sector_t * sec, float_t height)
{
  int i, count = 0;
  float_t min = INT_MAX;
  sector_t *other;

  for (i = 0; i < sec->linecount; i++)
  {
    other = P_GetNextSector(sec->lines[i], sec);

    if (!other)
      continue;

    if ((other->ceil.h > height) && (other->ceil.h < min))
    {
      min = other->ceil.h;
      count++;
    }
  }

  if (count == 0)
    return height;

  return min;
}

//
// P_FindNextLowestCeiling
//
// -AJA- 1999/10/22: written.

float_t P_FindNextLowestCeiling(sector_t * sec, float_t height)
{
  int i, count = 0;
  float_t max = INT_MIN;
  sector_t *other;

  for (i = 0; i < sec->linecount; i++)
  {
    other = P_GetNextSector(sec->lines[i], sec);

    if (!other)
      continue;

    if ((other->ceil.h < height) && (other->ceil.h > max))
    {
      max = other->ceil.h;
      count++;
    } 
  }

  if (count == 0)
    return height;

  return max;
}

//
// FIND LOWEST CEILING IN THE SURROUNDING SECTORS
//
float_t P_FindLowestCeilingSurrounding(sector_t * sec)
{
  int i;
  line_t *check;
  sector_t *other;
  float_t height = INT_MAX;

  for (i = 0; i < sec->linecount; i++)
  {
    check = sec->lines[i];
    other = P_GetNextSector(check, sec);

    if (!other)
      continue;

    if (other->ceil.h < height)
      height = other->ceil.h;
  }
  return height;
}

//
// FIND HIGHEST CEILING IN THE SURROUNDING SECTORS
//
float_t P_FindHighestCeilingSurrounding(sector_t * sec)
{
  int i;
  line_t *check;
  sector_t *other;
  float_t height = 0;

  for (i = 0; i < sec->linecount; i++)
  {
    check = sec->lines[i];
    other = P_GetNextSector(check, sec);

    if (!other)
      continue;

    if (other->ceil.h > height)
      height = other->ceil.h;
  }
  return height;
}

//
// P_FindSectorFromTag
//
// Returns the FIRST sector that tag refers to.
//
// -KM- 1998/09/27 Doesn't need a line.
// -AJA- 1999/09/29: Now returns a sector_t, and has no start.

sector_t *P_FindSectorFromTag(int tag)
{
  int i;

  for (i = 0; i < numsectors; i++)
  {
    if (sectors[i].tag == tag)
      return sectors + i;
  }

  return NULL;
}

//
// Find minimum light from an adjacent sector
//
int P_FindMinSurroundingLight(sector_t * sector, int max)
{
  int i;
  int min;
  line_t *line;
  sector_t *check;

  min = max;
  for (i = 0; i < sector->linecount; i++)
  {
    line = sector->lines[i];
    check = P_GetNextSector(line, sector);

    if (!check)
      continue;

    if (check->p.lightlevel < min)
      min = check->p.lightlevel;
  }
  return min;
}

//
// P_AddExtraFloor
//
static void P_AddExtraFloor(sector_t *sec, line_t *line, 
    const extrafloor_t *ef)
{
  extrafloor_record_t *rec, **succ_p;

  // create new record
  rec = Z_New(extrafloor_record_t, 1);

  rec->line = line;
  rec->info = ef;

  // determine priority
  rec->priority = 0;

  if (ef->type & EXFL_Liquid)
    rec->priority += 32;
  
  if (! (ef->type & EXFL_Thick))
    rec->priority += 16;
  
  if (ef->type & EXFL_Flooder)
    rec->priority += 8;

  // scan list, looking for the right position to add it such that the
  // extrafloor records are kept in priority order.

  succ_p = &sec->extrafloor_records;
  
  while (*succ_p)
  {
    if ((*succ_p)->priority <= rec->priority)
    {
      succ_p = & ((*succ_p)->next);
      continue;
    }

    // the current extrafloor record has a higher priority than the
    // new record, thus we need to insert the new record here.

    break;
  }
  
  rec->next = (*succ_p);
  (*succ_p) = rec;
}

//
// P_EFTransferTrans
//
static void P_EFTransferTrans(sector_t *ctrl, sector_t *sec, line_t *line, 
    const extrafloor_t *ef)
{
  int i;

  // floor and ceiling

  if (ctrl->floor.translucency > ef->translucency)
    ctrl->floor.translucency = ef->translucency;

  if (ctrl->ceil.translucency > ef->translucency)
    ctrl->ceil.translucency = ef->translucency;

  // sides

  if (! (ef->type & EXFL_Thick))
    return;

  if (ef->type & (EXFL_SideUpper | EXFL_SideLower))
  {
    for (i=0; i < sec->linecount; i++)
    {
      line_t *L = sec->lines[i];
      side_t *S;

      if (L->frontsector == sec)
        S = L->side[1];
      else if (L->backsector == sec)
        S = L->side[0];
      else
        // WTF ?
        continue;

      if (! S)
        continue;

      if (ef->type & EXFL_SideUpper)
        S->top.translucency = ef->translucency;
      else  // EXFL_SideLower
        S->bottom.translucency = ef->translucency;
    }

    return;
  }

  line->side[0]->middle.translucency = ef->translucency;
}

//
// EVENTS
//
// Events are operations triggered by using, crossing,
// or shooting special lines, or by timed thinkers.
//

//
// P_ActivateSpecialLine
//
// Called when a special line is activated.
//
// line is the line to be activated, side is the side activated from,
// (as lines can only be activated from the right), thing is the thing
// activating, to check for player/monster only lines trig is how it
// was activated, ie shot/crossed/pushed.  `line' can be NULL for
// non-line activations.
//
// -KM- 1998/09/01 Procedure Written.  
// -ACB- 1998/09/11 Return Success or Failure.
// -AJA- 1999/09/29: Updated for new tagged sector links.
// -AJA- 1999/10/21: Allow non-line activation (line == NULL), and
//       added `typenum' and `tag' parameter.
// -AJA- 2000/01/02: New trigger method `line_Any'.

static boolean_t P_ActivateSpecialLine(line_t * line,
    const linedeftype_t * special, int tag, int side, mobj_t * thing,
    trigger_e trig, int can_reach)
{
  boolean_t texSwitch = false;
  boolean_t failedsecurity = true;  // -ACB- 1998/09/11 Security pass/fail check
  sfx_t *sfx[4];
  sector_t *tsec;

#ifdef DEVELOPERS
  if (!special)
  {
    if (line == NULL)
      I_Error("P_ActivateSpecialLine: Special type is 0\n");
    else
      I_Error("P_ActivateSpecialLine: Line %d is not Special\n", 
          (int)(line - lines));
  }
#endif

  if (! G_CheckWhenAppear(special->appear))
  {
    if (line)
      line->special = NULL;
    return true;
  }

  if (trig != line_Any && special->type != trig)
    return false;

  // Check for use once.
  if (line && !line->count)
    return false;

  // Single sided line
  if (trig != line_Any && special->singlesided && side == 1)
    return false;

  // -AJA- 1999/12/07: Height checking.
  if (line && thing && thing->player &&
      (special->special_flags & LINSP_MustReach) && !can_reach)
  {
    S_StartSound(thing, sfx_noway);
    return false;
  }
  
  if (thing)
  {
    // Check this type of thing can trigger
    if (!thing->player)  // Missile / Monster
    {
      // Monsters/Missiles don't trigger secrets
      if (line && (line->flags & ML_SECRET))
        return false;

      // -ACB- 1998/09/07 Remove missileteleport check, not needed.
      if (thing->flags & MF_MISSILE)
      {
        // Missiles can only trigger linedefs with the trig_projectile flag set
        if (!(special->obj & trig_projectile))
          return false;

      }
      else
      {
        // Monsters can only trigger if the trig_monster flag is set
        if (!(special->obj & trig_monster))
          return false;
      }
    }
    else
    {
      // Players can only trigger if the trig_player is set
      if (!(special->obj & trig_player))
        return false;
    }

    // Check for keys
    // -ACB- 1998/09/11 Key possibilites extended
    if (special->keys != KF_NONE)
    {
      // Monsters/Missiles have no keys
      if (!thing->player)
        return false;

      //
      // New Security Checks, allows for any combination of keys in
      // an AND or OR function. Therefore it extends the possibilities
      // of security above 3 possible combinations..
      //
      // Could be better, but cheap and effective.
      if ((special->keys & KF_BLUECARD) && thing->player->cards[KEY_BlueCard])
        failedsecurity = false;

      if ((special->keys & KF_YELLOWCARD) && thing->player->cards[KEY_YellowCard])
        failedsecurity = false;

      if ((special->keys & KF_REDCARD) && thing->player->cards[KEY_RedCard])
        failedsecurity = false;

      if ((special->keys & KF_BLUESKULL) && thing->player->cards[KEY_BlueSkull])
        failedsecurity = false;

      if ((special->keys & KF_YELLOWSKULL) && thing->player->cards[KEY_YellowSkull])
        failedsecurity = false;

      if ((special->keys & KF_REDSKULL) && thing->player->cards[KEY_RedSkull])
        failedsecurity = false;

      if ((special->keys & (KF_BLUECARD << 8)) && !thing->player->cards[KEY_BlueCard])
        failedsecurity = true;

      if ((special->keys & (KF_YELLOWCARD << 8)) && !thing->player->cards[KEY_YellowCard])
        failedsecurity = true;

      if ((special->keys & (KF_REDCARD << 8)) && !thing->player->cards[KEY_RedCard])
        failedsecurity = true;

      if ((special->keys & (KF_BLUESKULL << 8)) && !thing->player->cards[KEY_BlueSkull])
        failedsecurity = true;

      if ((special->keys & (KF_YELLOWSKULL << 8)) && !thing->player->cards[KEY_YellowSkull])
        failedsecurity = true;

      if ((special->keys & (KF_REDSKULL << 8)) && !thing->player->cards[KEY_RedSkull])
        failedsecurity = true;

      if (failedsecurity)
      {
        if (special->failedmessage)
          CON_PlayerMessageLDF(thing->player, special->failedmessage);

        return false;
      }
    }
  }

  // Check if button already pressed
  if (line && P_ButtonCheckPressed(line))
    return false;
  
  // Do lights
  // -KM- 1998/09/27 Generalised light types.
  switch (special->l.type)
  {
    case lite_set:
      if (line)
        EV_LightTurnOn(line, special->l.light);
      texSwitch = true;
      break;

    case lite_none:
      break;

    default:
      texSwitch = P_DoSectorsFromTag(tag, &special->l, NULL, DoLights_wrapper);
      break;
  }

  // -ACB- 1998/09/13 Use teleport define..
  if (special->t.teleport)
  {
    texSwitch = EV_Teleport(line, tag, 1, thing, special->t.delay, 
        special->t.special, special->t.inspawnobj, special->t.outspawnobj);
  }

  if (special->e_exit == EXIT_Normal)
  {
    G_ExitLevel(5);
    texSwitch = true;
  }
  else if (special->e_exit == EXIT_Secret)
  {
    G_SecretExitLevel(5);
    texSwitch = true;
  }

  if (special->d.dodonut)
  {
    // Proper ANSI C++ Init
    sfx[0] = special->d.d_sfxout;
    sfx[1] = special->d.d_sfxoutstop;
    sfx[2] = special->d.d_sfxin;
    sfx[3] = special->d.d_sfxinstop;

    texSwitch = P_DoSectorsFromTag(tag, NULL, sfx, DoDonut_wrapper);
  }

  // - Plats/Floors -
  if (special->f.type != mov_undefined)
  {
    if (!tag)
    {
      if (line)
        texSwitch = EV_Manual(line, thing, &special->f);
    }
    else
    {
      texSwitch = P_DoSectorsFromTag(tag, &special->f, 
          line ? line->frontsector : NULL, DoSector_wrapper);
    }
  }

  // - Doors/Ceilings -
  if (special->c.type != mov_undefined)
  {
    if (!tag)
    {
      if (line)
        texSwitch = EV_Manual(line, thing, &special->c);
    }
    else
    {
      texSwitch = P_DoSectorsFromTag(tag, &special->c,
          line ? line->frontsector : NULL, DoSector_wrapper);
    }
  }

  if (special->use_colourmap)
  {
    for (tsec = P_FindSectorFromTag(tag); tsec; tsec = tsec->tag_next)
    {
      tsec->p.colourmap = special->use_colourmap;
      texSwitch = true;
    }
  }

  // Note: M_PI means "not changed" for gravity, friction and
  // viscosity.

  if (special->gravity != (float_t)M_PI)
  {
    for (tsec = P_FindSectorFromTag(tag); tsec; tsec = tsec->tag_next)
    {
      tsec->p.gravity = special->gravity;
      texSwitch = true;
    }
  }

  if (special->friction != (float_t)M_PI)
  {
    for (tsec = P_FindSectorFromTag(tag); tsec; tsec = tsec->tag_next)
    {
      tsec->p.friction = special->friction;
      texSwitch = true;
    }
  }

  if (special->viscosity != (float_t)M_PI)
  {
    for (tsec = P_FindSectorFromTag(tag); tsec; tsec = tsec->tag_next)
    {
      tsec->p.viscosity = special->viscosity;
      texSwitch = true;
    }
  }

  if (line && special->ef.type)
  {
    sector_t *ctrl = line->frontsector;

    for (tsec = P_FindSectorFromTag(tag); tsec; tsec = tsec->tag_next)
    {
      P_AddExtraFloor(tsec, line, &special->ef);

      // transfer any translucency
      if (special->ef.translucency <= 0.99)
        P_EFTransferTrans(ctrl, tsec, line, &special->ef);

      ctrl->controller = true;

      // update the regions
      P_ComputeSectorRegions(tsec);

      // update the line gaps & things:
      P_ChangeSector(tsec, false);

      texSwitch = true;
    }
  }

  if (special->trigger_effect && tag > 0)
  {
    RAD_EnableByTag(tag, special->trigger_effect < 0);
    texSwitch = true;
  }

  if (special->sfx)
  {
    for (tsec = P_FindSectorFromTag(tag); tsec; tsec = tsec->tag_next)
    {
      sectorsfx_t *sfx = NewSectorSFX();

      sfx->count = 7;
      sfx->sector = tsec;
      sfx->sfx = special->sfx;
      sfx->sfxstarted = false;

      texSwitch = true;
    }
  }

  if (special->music)
  {
    S_ChangeMusic(special->music, true);
    texSwitch = true;
  }

  // reduce count & clear special if necessary
  if (line && texSwitch)
  {
    if (line->count != -1)
    {
      line->count--;

      if (!line->count)
        line->special = NULL;
    }
    // -KM- 1998/09/27 Reversable linedefs.
    if (line->special && special->newtrignum)
      line->special = (special->newtrignum <= 0) ? NULL :
        DDF_LineLookupNum(special->newtrignum);

    P_ChangeSwitchTexture(line, line->special && (special->newtrignum == 0),
        special->special_flags);
  }

  return true;
}

//
// P_CrossSpecialLine - TRIGGER
//
// Called every time a thing origin is about
// to cross a line with a non 0 special.
//
// -KM- 1998/09/01 Now much simpler
// -ACB- 1998/09/12 Return success/failure
//
boolean_t P_CrossSpecialLine(line_t *ld, int side, mobj_t * thing)
{
  return P_ActivateSpecialLine(ld, ld->special,
      ld->tag, side, thing, line_walkable, 1);
}

//
// P_ShootSpecialLine - IMPACT SPECIALS
// Called when a thing shoots a special line.
//
void P_ShootSpecialLine(mobj_t * thing, line_t * line)
{
  P_ActivateSpecialLine(line, line->special, line->tag, 0, 
      thing, line_shootable, 1);
}

//
// P_UseSpecialLine
//
// Called when a thing uses a special line.
// Only the front sides of lines are usable.
//
// -KM- 1998/09/01 Uses new lines.ddf code in p_spec.c
//
// -ACB- 1998/09/07 Uses the return value to discern if a move if possible.
//
// -AJA- 1999/12/07: New parameters `open_bottom' and `open_top',
//       which give a vertical range through which the linedef is
//       accessible.  Could be used for smarter switches, like one on
//       a lower wall-part which is out of reach (e.g. MAP02).
//
boolean_t P_UseSpecialLine(mobj_t * thing, line_t * line, int side,
    float_t open_bottom, float_t open_top)
{
  int can_reach = (thing->z < open_top) &&
      (thing->z + thing->height + USE_Z_RANGE >= open_bottom);

  return P_ActivateSpecialLine(line, line->special, line->tag, side,
      thing, line_pushable, can_reach);
}

//
// P_RemoteActivation
//
// Called by the RTS `ACTIVATE_LINETYPE' primitive, and also the code
// pointer in things.ddf of the same name.
//
// -AJA- 1999/10/21: written.

void P_RemoteActivation(mobj_t * thing, int typenum, int tag, 
    int side, trigger_e method)
{
  const linedeftype_t *spec = DDF_LineLookupNum(typenum);

  P_ActivateSpecialLine(NULL, spec, tag, side, thing, method, 1);
}

//
// P_PlayerInSpecialSector
//
// Called every tic frame that the player origin is in a special sector
//
// -KM- 1998/09/27 Generalised for sectors.ddf
// -AJA- 1999/10/09: Updated for new sector handling.

void P_PlayerInSpecialSector(player_t * player, sector_t * sec)
{
  vert_region_t *reg;
  const specialsector_t *special;
  float_t damage;

  for (reg=sec->regions; reg; reg=reg->sec_next)
  {
    if (! reg->p->special)
      continue;

    special = reg->p->special;

    if (! G_CheckWhenAppear(special->appear))
    {
      reg->p->special = NULL;
      continue;
    }

    damage = special->damage;

    if (special->special_flags & SECSP_WholeRegion)
    {
      float_t bz = player->mo->z;
      float_t tz = player->mo->z + player->mo->height;

      if (special->special_flags & SECSP_Proportional)
      {
        // only partially in region -- mitigate damage

        if (tz > reg->ceil->h)
          damage -= damage * (tz - reg->ceil->h) / (tz-bz);
      
        if (bz < reg->floor->h)
          damage -= damage * (reg->floor->h - bz) / (tz-bz);
      }
      else
      {
        if (bz > reg->ceil->h || tz < reg->floor->h)
          damage = 0;
      }
    }
    else
    {
      // Not touching the floor ?
      if (player->mo->z > reg->floor->h + 4.0)
        continue;
    }

    if (damage > 0)
      if (!player->powers[PW_AcidSuit])
        if ((leveltime % special->damagetime) == 0)
          P_DamageMobj(player->mo, NULL, NULL, damage);

    if (special->secret)
    {
      player->secretcount++;
      reg->p->special = NULL;
    }

    if (special->e_exit == EXIT_Normal)
    {
      player->cheats &= ~CF_GODMODE;
      if (player->health <= special->damage)
      {
        S_StartSound(player->mo, player->mo->info->deathsound);
        // -KM- 1998/12/16 We don't want to alter the special type,
        //   modify the sector's attributes instead.
        reg->p->special = NULL;
        G_ExitLevel(1);
        return;
      }
    }
    else if (special->e_exit == EXIT_Secret)
    {
      player->cheats &= ~CF_GODMODE;
      if (player->health <= special->damage)
      {
        S_StartSound(player->mo, player->mo->info->deathsound);
        reg->p->special = NULL;
        G_SecretExitLevel(1);
        return;
      }
    }
  }
}

//
// P_UpdateSpecials
//
// Animate planes, scroll walls, etc.

boolean_t levelTimer;
int levelTimeCount;

void P_UpdateSpecials(void)
{
  int i;
  int n;
  anim_t *anim;
  line_t *line;
  const linedeftype_t *special;

  // LEVEL TIMER
  if (levelTimer == true)
  {
    levelTimeCount--;
    if (!levelTimeCount)
      G_ExitLevel(1);
  }

  // ANIMATE FLATS AND TEXTURES GLOBALLY
  for (n = 0; n < numanims; n++)
  {
    anim = anims[n];
    for (i = anim->basepic; i < anim->basepic + anim->numpics; i++)
    {
      if (anim->istexture)
        texturetranslation[i] = anim->basepic + ((leveltime / anim->speed + i) % anim->numpics);
      else
        flattranslation[i] = wad2flatlist[flatlist[anim->basepic] + ((leveltime / anim->speed + i) % anim->numpics)];
    }
  }

  // ANIMATE LINE SPECIALS
  // -KM- 1998/09/01 Lines.ddf
  for (i = 0; i < numlinespecials; i++)
  {
    line = linespeciallist[i];
    
    // lines can change their special, so we check...
    if (! line->special)
      continue;

    special = line->special;

    // -KM- 1999/01/31 Use new method.
    // -AJA- 1999/07/01: Handle both sidedefs.
    if (line->side[0])
    {
      if (special->scroll_parts & SCPT_RightUpper)
      {
        line->side[0]->top.x_offset += special->s_xspeed;
        line->side[0]->top.y_offset += special->s_yspeed;
      }
      if (special->scroll_parts & SCPT_RightMiddle)
      {
        line->side[0]->middle.x_offset += special->s_xspeed;
        line->side[0]->middle.y_offset += special->s_yspeed;
      }
      if (special->scroll_parts & SCPT_RightLower)
      {
        line->side[0]->bottom.x_offset += special->s_xspeed;
        line->side[0]->bottom.y_offset += special->s_yspeed;
      }
    }

    if (line->side[1])
    {
      int xmul = (special->scroll_parts & SCPT_LeftRevX) ? -1 : 1;
      int ymul = (special->scroll_parts & SCPT_LeftRevY) ? -1 : 1;
      
      if (special->scroll_parts & SCPT_LeftUpper)
      {
        line->side[1]->top.x_offset += special->s_xspeed * xmul;
        line->side[1]->top.y_offset += special->s_yspeed * ymul;
      }
      if (special->scroll_parts & SCPT_LeftMiddle)
      {
        line->side[1]->middle.x_offset += special->s_xspeed * xmul;
        line->side[1]->middle.y_offset += special->s_yspeed * ymul;
      }
      if (special->scroll_parts & SCPT_LeftLower)
      {
        line->side[1]->bottom.x_offset += special->s_xspeed * xmul;
        line->side[1]->bottom.y_offset += special->s_yspeed * ymul;
      }
    }
  }

  // ANIMATE SECTOR SPECIALS
  for (i = 0; i < num_sect_specials; i++)
  {
    sector_t *sec = sect_speciallist[i];
    float_t dx, dy;
    
    if (! sec->p.special)
      continue;

    if (sec->p.special->f.scroll_speed > 0)
    {
      dx = -M_Cos(sec->p.special->f.scroll_angle);
      dy =  M_Sin(sec->p.special->f.scroll_angle);
      
      sec->floor.x_offset += dx * sec->p.special->f.scroll_speed / 32.0;
      sec->floor.y_offset += dy * sec->p.special->f.scroll_speed / 32.0;
    }

    if (sec->p.special->c.scroll_speed > 0)
    {
      dx = -M_Cos(sec->p.special->c.scroll_angle);
      dy =  M_Sin(sec->p.special->c.scroll_angle);
      
      sec->ceil.x_offset += dx * sec->p.special->c.scroll_speed / 32.0;
      sec->ceil.y_offset += dy * sec->p.special->c.scroll_speed / 32.0;
    }
  }

  // DO BUTTONS
  for (i = 0; i < maxbuttons; i++)
  {
    if (buttonlist[i].btimer == 0)
      continue;

    buttonlist[i].btimer--;

    if (buttonlist[i].btimer != 0)
      continue;

    switch (buttonlist[i].where)
    {
      case BWH_Top:
        buttonlist[i].line->side[0]->top.tex = buttonlist[i].btexture;
        break;

      case BWH_Middle:
        buttonlist[i].line->side[0]->middle.tex = buttonlist[i].btexture;
        break;

      case BWH_Bottom:
        buttonlist[i].line->side[0]->bottom.tex = buttonlist[i].btexture;
        break;
    }

    S_StartSound((mobj_t *) & buttonlist[i].soundorg, sfx_swtchn);
    Z_Clear(&buttonlist[i], button_t, 1);
  }
}

//
//  SPECIAL SPAWNING
//

//
// P_SpawnSpecials
//
// This function is called at the start of every level.  It parses command line
// parameters for level timer, spawns passive special sectors, (ie sectors that
// act even when a player is not in them, and counts total secrets) spawns
// passive lines, (ie scrollers) and resets floor/ceiling movement.
//
// (-ACB- 1998/09/06 Its parses command line parameters? really? cool. I going to
//                   to get my C book out and see how this works [:)]. )
// -KM- 1998/09/27 Generalised for sectors.ddf
// -KM- 1998/11/25 Lines with auto tag are automatically triggered.

void P_SpawnSpecials(int autotag)
{
  sector_t *sector;
  const specialsector_t *secSpecial;
  int i;
  const linedeftype_t *special;
  const char *s;

  // See if -TIMER needs to be used.
  levelTimer = false;

  i = M_CheckParm("-avg");
  if (i && deathmatch)
  {
    levelTimer = true;
    levelTimeCount = 20 * 60 * TICRATE;
  }

  s = M_GetParm("-timer");
  if (s && deathmatch)
  {
    int time;

    time = atoi(s) * 60 * TICRATE;
    levelTimer = true;
    levelTimeCount = time;
  }

  for (i = 0; i < maxbuttons; i++)
    Z_Clear(&buttonlist[i], button_t, 1);

  //
  // Init special SECTORs.
  //
  if (!sect_special_a.ptr)
    Z_InitStackArray(&sect_special_a, (void ***)&sect_speciallist, 
        sizeof(sector_t *), -16);
  Z_SetArraySize(&sect_special_a, num_sect_specials = 0);

  sector = sectors;
  for (i = 0; i < numsectors; i++, sector++)
  {
    sector->p.gravity = GRAVITY;
    sector->p.friction = FRICTION;
    sector->p.viscosity = VISCOSITY;
    
    // -AJA- 1999/07/10: Updated for colmap.ddf.
    sector->p.colourmap = DDF_ColmapLookup("NORMAL");

    if (!sector->p.special)
      continue;

    secSpecial = sector->p.special;

    if (! G_CheckWhenAppear(secSpecial->appear))
    {
      sector->p.special = NULL;
      continue;
    }

    if (secSpecial->l.type != lite_none)
      EV_Lights(sector, &secSpecial->l);

    if (secSpecial->secret)
      totalsecret++;

    if (secSpecial->use_colourmap)
      sector->p.colourmap = secSpecial->use_colourmap;

    if (secSpecial->sfx)
    {
      sectorsfx_t *sfx = NewSectorSFX();

      sfx->count = 7;
      sfx->sector = sector;
      sfx->sfx = secSpecial->sfx;
      sfx->sfxstarted = false;
    }

    // - Plats/Floors -
    if (secSpecial->f.type != mov_undefined)
      EV_DoSector(sector, &secSpecial->f, sector);

    // - Doors/Ceilings -
    if (secSpecial->c.type != mov_undefined)
      EV_DoSector(sector, &secSpecial->c, sector);

    sector->p.gravity   = secSpecial->gravity;
    sector->p.friction  = secSpecial->friction;
    sector->p.viscosity = secSpecial->viscosity;

    // Scrollers
    if (secSpecial->f.scroll_speed || secSpecial->c.scroll_speed)
    {
      Z_SetArraySize(&sect_special_a, ++num_sect_specials);
      sect_speciallist[num_sect_specials - 1] = sector;
    }
  }

  //
  // Init line EFFECTs
  //
  // -ACB- & -JC- 1998/06/10 Implemented additional scroll effects
  //
  // -ACB- Added the code
  // -JC-  Designed and contributed code
  // -KM-  Removed Limit
  // -KM- 1998/09/01 Added lines.ddf support
  //
  if (!linespecial_a.ptr)
    Z_InitStackArray(&linespecial_a, (void ***)&linespeciallist, 
        sizeof(line_t *), -16);
  Z_SetArraySize(&linespecial_a, numlinespecials = 0);

  for (i = 0; i < numlines; i++)
  {
    if (lines[i].special)
    {
      special = lines[i].special;

      // -AJA- 1999/10/23: weed out non-appearing lines.
      if (! G_CheckWhenAppear(special->appear))
      {
        lines[i].special = NULL;
        continue;
      }

      lines[i].count = special->count;

      if (special->s_xspeed || special->s_yspeed)
      {
        Z_SetArraySize(&linespecial_a, ++numlinespecials);
        linespeciallist[numlinespecials - 1] = &lines[i];
      }

      // -AJA- 1999/06/30: Translucency effect.
      if (lines[i].side[0])
        lines[i].side[0]->middle.translucency = special->translucency;

      if (lines[i].side[1])
        lines[i].side[1]->middle.translucency = special->translucency;

      if (special->autoline)
        P_ActivateSpecialLine(&lines[i], lines[i].special,
          lines[i].tag, 0, NULL, line_Any, 1);

      // -KM- 1998/11/25 This line should be pushed automatically
      if (autotag && lines[i].tag == autotag)
        P_ActivateSpecialLine(&lines[i], lines[i].special,
          lines[i].tag, 0, NULL, line_pushable, 1);
    }
    else
    {
      lines[i].count = 0;
    }
  }

  // UNUSED: no horizonal sliders.
  // P_InitSlidingDoorFrames();
}

// -KM- 1998/09/27 This helper function is used to do stuff to all the sectors
// with the specified line's tag.
// -AJA- 1999/09/29: Updated for new tagged sector links.

boolean_t P_DoSectorsFromTag(int tag, const void *p1, void *p2,
    boolean_t(*func) (sector_t *, const void *, void *))
{
  sector_t *tsec;
  boolean_t rtn = false;

  for (tsec = P_FindSectorFromTag(tag); tsec; tsec = tsec->tag_next)
  {
    if ((*func) (tsec, p1, p2))
      rtn = true;
  }
  return rtn;
}
