//----------------------------------------------------------------------------
//  EDGE Player User Code
//----------------------------------------------------------------------------
// 
//  Copyright (c) 1999-2000  The EDGE Team.
// 
//  This program is free software; you can redistribute it and/or
//  modify it under the terms of the GNU General Public License
//  as published by the Free Software Foundation; either version 2
//  of the License, or (at your option) any later version.
//
//  This program is distributed in the hope that it will be useful,
//  but WITHOUT ANY WARRANTY; without even the implied warranty of
//  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//  GNU General Public License for more details.
//
//----------------------------------------------------------------------------
//
//  Based on the DOOM source code, released by Id Software under the
//  following copyright:
//
//    Copyright (C) 1993-1996 by id Software, Inc.
//
//----------------------------------------------------------------------------

#include "i_defs.h"

#include "e_event.h"
#include "dm_defs.h"
#include "dm_state.h"
#include "g_game.h"
#include "m_inline.h"
#include "p_inter.h"
#include "p_local.h"
#include "rad_trig.h"
#include "s_sound.h"
#include "st_stuff.h"
#include "w_wad.h"

// 16 pixels of bob
#define MAXBOB	      16
#define JUMPBREATHER  1*TICRATE  // -ACB- 1998/08/09 Gap before next jump

//
// ThrustPlayer
//
// Moves the given origin along a given angle.

static void ThrustPlayer(player_t * player, angle_t angle, float_t move)
{
  player->mo->momx += move * M_Cos(angle);
  player->mo->momy += move * M_Sin(angle);
}

//
// CalcHeight
//
// Calculate the walking / running height adjustment

static void CalcHeight(player_t * player)
{
  angle_t angle;
  float_t bob;
  boolean_t onground = player->mo->z <= player->mo->floorz;
  boolean_t hasjetpack = player->powers[PW_Jetpack] > 0;
  
  // Check for cheat.
  if ((player->cheats & CF_NOMOMENTUM) || !onground)
  {
    // -ES- FIXME: What is this?????
    player->viewz = player->mo->z + VIEWHEIGHT;
  
    if (player->viewz > player->mo->ceilingz - 4)
      player->viewz = player->mo->ceilingz - 4;
  
    player->viewz = player->mo->z + player->viewheight;
    return;
  }

  //
  // Sorted this, bob and viewheight are only calculated when
  // player is alive. -ACB- 1998/07/27
  //
  if (player->playerstate == PST_LIVE && !hasjetpack)
  {
    // ----CALCULATE BOB EFFECT----
    angle = ANG90 / 5 * leveltime;
  
    // Regular movement bobbing (needs to be calculated for gun swing even
    // if not on ground) OPTIMISE: tablify angle; Note: a LUT allows for
    // effects like a ramp with low health.
    player->bob = (player->mo->momx * player->mo->momx
        + player->mo->momy * player->mo->momy) / 4;

    if (player->bob > MAXBOB)
      player->bob = MAXBOB;

    bob = player->bob * M_Sin(angle) / 2;

    // ----CALCULATE VIEWHEIGHT----
    player->viewheight += player->deltaviewheight;

    if (player->viewheight > VIEWHEIGHT)
    {
      player->viewheight = VIEWHEIGHT;
      player->deltaviewheight = 0;
    }
    else if (player->viewheight < VIEWHEIGHT / 2)
    {
      player->viewheight = VIEWHEIGHT / 2;
      player->deltaviewheight = 0;
    }
    else if (player->viewheight < VIEWHEIGHT)
    {
       player->deltaviewheight += 0.2;
    }

    player->viewz = player->mo->z + player->viewheight + bob;

  }
  else
  {
    // If one is dead, one looks from the floor with no effects.
    player->viewz = player->mo->z + player->viewheight;
  }

  // No heads above the ceiling
  if (player->viewz > player->mo->ceilingz - 4)
    player->viewz = player->mo->ceilingz - 4;
}

//
// MovePlayer
//
static void MovePlayer(player_t * player)
{
  ticcmd_t *cmd;

  boolean_t onground = player->mo->z <= player->mo->floorz;
  boolean_t hasjetpack = player->powers[PW_Jetpack] > 0;

  cmd = &player->cmd;

  player->mo->angle += M_FloatToFixed(cmd->angleturn * player->mo->speed);

  // Do not let the player control movement if not onground.
  // -MH- 1998/06/18  unless he has the JetPack!
  
  if (hasjetpack)
  {
    if (player->powers[PW_Jetpack] <= (5 * TICRATE))
    {
      if (!(leveltime & 10))
        S_StartSound(player->mo, sfx_jpflow);  // fuel low
    }
    else
    {
      if (cmd->upwardmove > 0)
        S_StartSound(player->mo, sfx_jprise);
      else
      {
        if (cmd->upwardmove < 0)
          S_StartSound(player->mo, sfx_jpdown);
        else
        {
          if (cmd->forwardmove || cmd->sidemove)
            S_StartSound(player->mo, (onground ? sfx_jpidle : sfx_jpmove));
          else
            S_StartSound(player->mo, sfx_jpidle);
        }
      }
    }
  }

  // -MH- 1998/08/18  do vertical move
  if (cmd->upwardmove)
  {
    float_t zmove = player->mo->speed * cmd->upwardmove / 256.0;

    if (hasjetpack)
      zmove *= 16;
    
    // -AJA- Viscosity applied in P_ZMovement
    player->mo->momz += zmove;
  }

  // -MH- 1998/06/18  ...or has the JetPack
  if (cmd->forwardmove)
  {
    float_t fmove = player->mo->speed * cmd->forwardmove / 512.0;

    if (onground || hasjetpack)
      fmove *= 16;

    // -AJA- Viscosity applied in P_XYMovement
    ThrustPlayer(player, player->mo->angle, fmove);
  }

  // -MH- 1998/06/18  ...or has the JetPack
  if (cmd->sidemove)
  {
    float_t smove = player->mo->speed * cmd->sidemove / 512.0;

    if (onground || hasjetpack)
      smove *= 16;

    // -AJA- Viscosity applied in P_XYMovement
    ThrustPlayer(player, player->mo->angle - ANG90, smove);
  }

  if ((cmd->forwardmove || cmd->sidemove)
      && player->mo->state == &states[player->mo->info->spawn_state])
  {
    if (player->mo->info->chase_state)
      P_SetMobjState(player->mo, player->mo->info->chase_state);
  }
}

//
// P_DeathThink
//
// Fall on your face when dying.
// Decrease POV height to floor height.
//
#define ANG5   	(ANG90/18)

static void DeathThink(player_t * player)
{
  angle_t angle;
  angle_t delta;

  // -AJA- 1999/12/07: don't die mid-air.
  player->powers[PW_Jetpack] = 0;

  P_MovePsprites(player);

  // fall to the ground
  if (player->viewheight > DEATHVIEWHEIGHT)
    player->viewheight -= 1.0;
  else if (player->viewheight < DEATHVIEWHEIGHT)
    player->viewheight = DEATHVIEWHEIGHT;
  
  player->deltaviewheight = 0;
  player->kick_offset = 0;

  CalcHeight(player);

  if (player->attacker && player->attacker != player->mo)
  {
    angle = R_PointToAngle(player->mo->x, player->mo->y,
        player->attacker->x, player->attacker->y);

    delta = angle - player->mo->angle;

    if (delta <= ANG5 || delta >= (angle_t)(0 - ANG5))
    {
      // Looking at killer,
      //  so fade damage flash down.
      player->mo->angle = angle;

      if (player->damagecount)
        player->damagecount--;
    }
    else if (delta < ANG180)
      player->mo->angle += ANG5;
    else
      player->mo->angle -= ANG5;
  }
  else if (player->damagecount)
    player->damagecount--;

  // -AJA- 1999/08/07: Fade out armor points too.
  if (player->bonuscount)
    player->bonuscount--;

  // lose the zoom when dead
  if (viewiszoomed)
  {
    R_SetFOV(normalfov);
    viewiszoomed = false;
  }

  if (deathmatch >= 3 && player->mo->movecount > player->mo->info->respawntime)
    return;

  if (player->cmd.buttons & BT_USE)
    player->playerstate = PST_REBORN;
}

//
// P_ConsolePlayerThinker
//
// Does the thinking of the console player, i.e. read from input
void P_ConsolePlayerThinker(const player_t *p, void *data, ticcmd_t *dest)
{
  G_BuildTiccmd(dest);
}

//
// P_PlayerThink
//
void P_PlayerThink(player_t * player)
{
  ticcmd_t *cmd;
  int key;

  // fixme: do this in the cheat code
  if (player->cheats & CF_NOCLIP)
    player->mo->flags |= MF_NOCLIP;
  else
    player->mo->flags &= ~MF_NOCLIP;

  // chain saw run forward
  cmd = &player->cmd;
  if (player->mo->flags & MF_JUSTATTACKED)
  {
    cmd->angleturn = 0;
    cmd->forwardmove = 64;
    cmd->sidemove = 0;
    player->mo->flags &= ~MF_JUSTATTACKED;
  }

  if (player->playerstate == PST_DEAD)
  {
    DeathThink(player);
    return;
  }

  // Move around.
  // Reactiontime is used to prevent movement
  //  for a bit after a teleport.
  if (player->mo->reactiontime)
    player->mo->reactiontime--;
  else
    MovePlayer(player);

  CalcHeight(player);

  if (player->mo->subsector->sector->p.special ||
      player->mo->subsector->sector->regions->sec_next)
    P_PlayerInSpecialSector(player, player->mo->subsector->sector);

  // -AJA- FIXME: make RTS execution independent from players
  RAD_DoRadiTrigger(player);

  // Check for weapon change.

  // A special event has no other buttons.
  if (cmd->buttons & BT_SPECIAL)
    cmd->buttons = 0;

  if (cmd->buttons & BT_CHANGE)
  {
    int i, j;
    weaponkey_t *wk;

    // The actual changing of the weapon is done when the weapon
    // psprite can do it (read: not in the middle of an attack).

    key = (cmd->buttons & BT_WEAPONMASK) >> BT_WEAPONSHIFT;
    wk = &weaponkey[key];

    for (i=j=player->key_choices[key]; i < (j + wk->numchoices); i++)
    {
      int choice = wk->choices[i % wk->numchoices];

      if (! player->weapons[choice].owned)
        continue;
      
      if (choice == player->ready_wp)
        continue;
       
      if (! P_CheckWeaponSprite(weaponinfo[choice]))
        continue;
      
      player->pending_wp = choice;
      player->key_choices[key] = choice;
      break;
    }
  }

  // check for use
  if (cmd->buttons & BT_USE)
  {
    if (!player->usedown)
    {
      P_UseLines(player);
      player->usedown = true;
    }
  }
  else
    player->usedown = false;

  // EDGE Feature: Jump Code
  //
  // -ACB- 1998/08/09 Check that jumping is allowed in the currentmap
  //                  Make player pause before jumping again
  
  if (cmd->extbuttons & EBT_JUMP)
  {
    if (!(currentmap->force_off & MPF_Jumping) && !player->jumpwait)
    {
      if (player->mo->z <= player->mo->floorz)
      {
        player->mo->momz += player->mo->info->jumpheight / 1.4;
        player->jumpwait = JUMPBREATHER;

        // -AJA- 1999/09/11: New JUMP_SOUND for ddf.
        if (player->mo->info->jump_sound)
          S_StartSound(player->mo, player->mo->info->jump_sound);
      }
    }
  }

  // EDGE Feature: Zooming
  //
  if (cmd->extbuttons & EBT_ZOOM)
  {
    P_Zoom(player);
  }

  // EDGE Feature: Vertical Look (Mlook)
  //
  // -ACB- 1998/07/02 New Code used, rerouted via Ticcmd
  // -ACB- 1998/07/27 Used defines for look limits.
  //
  if (level_flags.mlook && (cmd->extbuttons & EBT_MLOOK))
  {
    player->mo->vertangle += cmd->vertangle / 256.0;

    if (player->mo->vertangle > LOOKUPLIMIT)
      player->mo->vertangle = LOOKUPLIMIT;

    if (player->mo->vertangle < LOOKDOWNLIMIT)
      player->mo->vertangle = LOOKDOWNLIMIT;
  }

  // EDGE Feature: Vertical Look (Mlook)
  //
  // -ACB- 1998/07/02 Re-routed via Ticcmd
  //
  if (cmd->extbuttons & EBT_CENTER)
    player->mo->vertangle = 0;

  // decrement jumpwait counter
  if (player->jumpwait)
    player->jumpwait--;

  // cycle psprites
  P_MovePsprites(player);

  // Counters, time dependend power ups.

  if (player->powers[PW_Berserk] > 0)
    player->powers[PW_Berserk]--;

  // -MH- 1998/06/18  jetpack "fuel" counter
  if (player->powers[PW_Jetpack] > 0)
    player->powers[PW_Jetpack]--;

  // -ACB- 1998/07/16  nightvision counter decrementation
  if (player->powers[PW_NightVision] > 0)
    player->powers[PW_NightVision]--;

  if (player->powers[PW_Invulnerable] > 0)
    player->powers[PW_Invulnerable]--;

  if (player->powers[PW_PartInvis] > 0)
  {
    player->powers[PW_PartInvis]--;

    if (player->powers[PW_PartInvis])
      player->mo->flags |=  MF_SHADOW;
    else
      player->mo->flags &= ~MF_SHADOW;
  }

  if (player->powers[PW_Infrared] > 0)
    player->powers[PW_Infrared]--;

  if (player->powers[PW_AcidSuit] > 0)
    player->powers[PW_AcidSuit]--;

  if (player->damagecount)
    player->damagecount--;

  if (player->bonuscount)
    player->bonuscount--;

  if (player->grin_count)
    player->grin_count--;

  if (player->attackdown || player->secondatk_down)
    player->attackdown_count++;
  else
    player->attackdown_count = 0;
   
  player->kick_offset /= 1.6;

  // Handling colourmaps.
  //
  // -AJA- 1999/07/10: Updated for colmap.ddf. Ideally, the exact
  // colourmap & palette effects to use would be specifiable somewhere
  // in DDF -- for now it is hardcoded.

  player->effect_colourmap = NULL;

  if (player->powers[PW_Invulnerable] > 0)
  {
    float_t s = player->powers[PW_Invulnerable];

    player->effect_colourmap = DDF_ColmapLookup("ALLWHITE");
    player->effect_strength = (float_t)((s >= 128) ? 1.0 : s / 128.0);
  }
  else if (player->powers[PW_Infrared] > 0)
  {
    float_t s = player->powers[PW_Infrared];

    player->effect_colourmap = DDF_ColmapLookup("ALLBRIGHT");
    player->effect_strength = (float_t)((s >= 128) ? 1.0 : s / 128.0);
  }
  // -ACB- 1998/07/15 NightVision Code
  else if (player->powers[PW_NightVision] > 0)
  {
    float_t s = player->powers[PW_NightVision];

    player->effect_colourmap = DDF_ColmapLookup("ALLGREEN");
    player->effect_strength = (float_t)((s >= 128) ? 1.0 : s / 128.0);
  }
}

//
// P_AddPlayer
//
// Creates a new player.
player_t *P_AddPlayer(int pnum)
{
  int i;
  player_t *p;
  player_t *prev;

  p = Z_ClearNew(player_t, 1);

  p->weapons = Z_ClearNew(playerweapon_t, numweapons);
  p->ammo = Z_ClearNew(playerammo_t, NUMAMMO);

  // Link it in. The list is sorted by pnum.
  if (players == NULL)
  {
    playerlookup = Z_ClearNew(player_t *, pnum + 1);
    playerlookup[pnum] = players = p;
  }
  else
  {
    for (prev = players; prev->next && prev->next->pnum < pnum; prev = prev->next)
      ;
    p->next = prev->next;
    p->prev = prev;
    prev->next = p;
    if (p->next)
    {
      // not at the end of the list, so fix prev link.
      DEV_ASSERT(pnum != p->next->pnum, ("P_AddPlayer: Attempt to create a second player %d", pnum));
      p->next->prev = p;
    }
    else
    {
      // Insert at the end of the list. Must grow playerlookup to fit more
      // elements.
      Z_Resize(playerlookup, player_t *, pnum + 1);
      for (i = prev->pnum + 1; i < pnum; i++)
        playerlookup[i] = NULL;
    }
    playerlookup[pnum] = p;
  }

  p->pnum = pnum;

  return p;
}

//
// P_GiveInitialBenefits
//
// Give the player the initial benefits when they start a game (or
// restart after dying).  Sets up: ammo, ammo-limits, health, armour,
// and weapons.
//
// -AJA- 2000/03/02: written.

void P_GiveInitialBenefits(player_t *p, const mobjinfo_t *info)
{
  int priority = -100;
  int i;
  
  p->ready_wp   = WPSEL_None;
  p->pending_wp = WPSEL_NoChange;

  for (i = 0; i < numweapons; i++)
  {
    p->weapons[i].owned = p->weapons[i].info->autogive;
    p->weapons[i].fire_count = 0;
    p->weapons[i].sa_fire_count = 0;

    // choose highest priority FREE weapon as the default
    if (p->weapons[i].owned)
    {
      if (p->weapons[i].info->priority > priority)
      {
        priority = p->weapons[i].info->priority;
        p->pending_wp = i;
        p->ready_wp = i;
      }
    }
  }

  // clear out ammo & ammo-limits
  for (i = 0; i < NUMAMMO; i++)
  {
    p->ammo[i].num = p->ammo[i].max = 0;
  }

  // set health and armour
  p->health = NORMHEALTH;
  p->armourtype = ARMOUR_None;
  p->armourpoints = 0;

  for (i = 0; i < NUMCARDS; i++)
    p->cards[i] = false;

  // give all initial benefits
  P_GiveBenefitList(p, NULL, info->initial_benefits);

  // refresh to remove all stuff from status bar
  P_UpdateAvailWeapons(p);
  stbar_update = true;
}

