//----------------------------------------------------------------------------
//  EDGE True BSP Rendering (Drawing code)
//----------------------------------------------------------------------------
// 
//  Copyright (c) 1999-2000  The EDGE Team.
// 
//  This program is free software; you can redistribute it and/or
//  modify it under the terms of the GNU General Public License
//  as published by the Free Software Foundation; either version 2
//  of the License, or (at your option) any later version.
//
//  This program is distributed in the hope that it will be useful,
//  but WITHOUT ANY WARRANTY; without even the implied warranty of
//  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//  GNU General Public License for more details.
//
//----------------------------------------------------------------------------
//
//  Based on the DOOM source code, released by Id Software under the
//  following copyright:
//
//    Copyright (C) 1993-1996 by id Software, Inc.
//
//----------------------------------------------------------------------------
//
// -AJA- 1999/08/31: Wrote this file.
//

#include "i_defs.h"

#include "dm_state.h"
#include "m_inline.h"
#include "m_swap.h"
#include "e_search.h"
#include "r_data.h"
#include "r_defs.h"
#include "r_main.h"
#include "r_plane.h"
#include "r_sky.h"
#include "r_state.h"
#include "v_colour.h"
#include "v_res.h"
#include "w_wad.h"
#include "w_textur.h"
#include "z_zone.h"

#include "r2_defs.h"


#define SHOW_WALLS   2
#define SHOW_PLANES  2
#define SHOW_THINGS  2

#define SHOW_OVERDRAW  0
#define SHOW_FIRST_IN  0


// FIXME:
extern float_t basexscale;
extern float_t baseyscale;

#if (SHOW_WALLS < 2 || SHOW_PLANES < 2 || SHOW_THINGS < 2)

// Temporary routine
static void BOGUS_Column16(screen_t * scr, int x, int y, int h, int col)
{
  short *dest;

  x += viewwindowx;
  y += viewwindowy;

  dest = (short *)(scr->data + y * scr->depth) + x;

  for (; h > 0; h--)
  {
      dest[0] = col;
      dest += scr->depth / 2;
  }
}

// Temporary routine
static void BOGUS_Column(screen_t * scr, int x, int y, int h, int col)
{
  byte *dest;

  if (BPP == 2)
  {
    BOGUS_Column16(scr, x, y, h, pixel_values[col & 0xFF]);
    return;
  }
  else if (BPP != 1)
    return;

  x += viewwindowx;
  y += viewwindowy;

  dest = scr->data + y * scr->depth + x;

  for (; h > 0; h--, dest += scr->depth)
  {
#if SHOW_OVERDRAW
    if (col > 5 && *dest != 1) 
    {
      if (SHOW_FIRST_IN)
        *dest = col;
      else if (*dest >= 192)
        *dest += 2;
      else
        *dest = 194;
    }
    else 
#endif
      *dest = col;
  }
}

// Temporary routine
static void BOGUS_DrawArea(screenline_t *area, int col)
{
  int x;

  float_t yt  = area->yt;
  float_t yb  = area->yb;

  for (x=area->x1; x <= area->x2; x++)
  {
    int y1 = floor(yt);
    int y2 = floor(yb);

    yt += area->yt_step;
    yb += area->yb_step;

    if (x < 0 || x >= SCREENWIDTH)
      continue;

    if (y2 < y1 || y2 < 0 || y1 >= viewheight)
      continue;

    if (y1 < 0)
      y1 = 0;

    if (y2 >= viewheight) 
      y2 = viewheight-1;

    BOGUS_Column(main_scr, x, y1, y2-y1+1, col);
  }
}
#endif

//
// R2_DrawMaskedColumn
//
// Render a single column of a texture with transparent parts.  Used
// for both sprites and mid-masked textures on 2S lines.
//
// -AJA- Pretty un-optimised for now.  Wanna get it working first.

static float_t mask_top;
static float_t mask_scale;
static float_t mask_yl;
static float_t mask_yh;
static float_t mask_translucency;

static void R2_DrawMaskedColumn(const column_t * column)
{
  float_t raw_yl;
  float_t raw_yh;
  float_t raw_top;
  void (*draw)(void);

  // moved away from inner loop
  if (mask_translucency >= 0.98)
    draw = colfunc;
  else
    draw = R_DrawTranslucentColumn;

  CHECKVAL(mask_scale);
  dc_iscale = M_FloatToFixed(1.0 / mask_scale) - 1;
  dc_translucency = M_FloatToFixed(mask_translucency);

  // handle each post in turn
  for (; column->topdelta != 0xFF;
       column = (column_t *) ((byte *) column + column->length + 4))
  {
    // calculate unclipped screen coordinates for post
    raw_yl = mask_top + column->topdelta * mask_scale;
    raw_yh = raw_yl   + column->length * mask_scale;
    raw_top = raw_yl;

    // clip post to given area (which must lie within the screen)
    if (raw_yl < mask_yl)
      raw_yl = mask_yl;

    if (raw_yh > mask_yh)
      raw_yh = mask_yh;

    dc_yl = floor(raw_yl);
    dc_yh = floor(raw_yh) - 1;

    if (dc_yh >= dc_yl)
    {
      dc_source = (const byte *) column + 3;
      dc_texturemid = -floor(raw_top) * dc_iscale;  /// (floor(raw_yl-raw_top) - dc_yl) * dc_iscale;

      (* draw)();
    }
  }
}


//
// R2_DrawWall
//
// Render a single wall of a subsector.

void R2_DrawWall(drawsub_t *sub, drawwall_t *wall)
{
#if (SHOW_WALLS == 2)
  int x;
  int index;
  int offset;
  int column;
  int tex_num;
  const cached_tex_t *tex;
  void (*draw)(void);

  float_t yt, yb, yp, scale, lightlev;

  lighttable_t *lights;
  angle_t angle;

  // calculate lighting
  index = wall->props->lightlevel >> LIGHTSEGSHIFT;

  if (index < 0)
    lights = scalelight[0];
  else if (index >= LIGHTLEVELS)
    lights = scalelight[LIGHTLEVELS - 1];
  else
    lights = scalelight[index];

  tex_num = texturetranslation[wall->part->tex];

  tex = W_CacheTextureNum(tex_num);

  dc_translucency = M_FloatToFixed(wall->part->translucency);

  if (wall->part->translucency >= 0.99)
    draw = colfunc;
  else
    draw = R_DrawTranslucentColumn;

  // draw the columns
  for (x = wall->area->x1; x <= wall->area->x2; x++)
  {
    offset = x - wall->area->x1;

    yt = wall->area->yt + offset * wall->area->yt_step; 
    yb = wall->area->yb + offset * wall->area->yb_step; 
    yp = wall->area->yp + offset * wall->area->yp_step; 

    dc_x  = x;
    dc_yl = floor(yt);
    dc_yh = floor(yb);
    
    if (dc_yh < dc_yl || dc_yh < 0 || dc_yl >= viewheight)
      continue;

    if (dc_yl < 0)
      dc_yl = 0;

    if (dc_yh >= viewheight) 
      dc_yh = viewheight-1;

    CHECKVAL(y_distunit);
    scale = wall->scale1 + offset * wall->scale_step;
    lightlev = scale * 160.0 / y_distunit;

    // calculate texture column
    angle = viewangle - wall->angle + xtoviewangle[x] - ANG90;

    if (! (ANG90-ANG1  <= angle && angle <= ANG90+ANG1) &&
        ! (ANG270-ANG1 <= angle && angle <= ANG270+ANG1))
    {
      column = (int) (wall->x_offset - M_Tan(angle) * wall->distance);
    }
    else
      column = (int) wall->x_offset;

    // determine colourmap
    index = (int) (lightlev * 16.0); //!!!

    if (index >= MAXLIGHTSCALE)
      index = MAXLIGHTSCALE - 1;

    dc_colourmap = V_GetColtable(wall->props->colourmap, lights[index], 0);

    if (! wall->is_masked)
    {
      CHECKVAL(scale);

      dc_iscale = FLOAT_2_FIX(1.0 / scale);
      dc_texturemid = -(floor(yt) * dc_iscale) + FLOAT_2_FIX(yp);
      dc_source = W_GetColumn(column, tex);

      (* draw)();
    }
    else
    {
      const column_t *tex_col;

      mask_yl = (float_t) dc_yl;
      mask_yh = (float_t) dc_yh;
      mask_scale = scale;
      mask_top = yt - yp * mask_scale;
      mask_translucency = wall->part->translucency;

      // -AJA- FIXME: Major Ick !!
      // Will crash/medusa if the texture consists of more than one lump.
      tex_col = (const column_t *) (W_GetColumn(column, tex) - 3);
      
      R2_DrawMaskedColumn(tex_col);
    }
  }

  W_DoneWithTexture(tex);

#elif (SHOW_WALLS == 1)
  BOGUS_DrawArea(wall->area, wall->picnum);
#endif /* SHOW_WALLS */
}


//
// R2_DrawPlane
//
// Render a single plane of a subsector.

static lighttable_t *plane_lights;
static const colourmap_t *plane_colourmap;
static plane_info_t *plane_data;
static float plane_translucency;

static void R2_MapPlane(int y, int x1, int x2)
{
  angle_t angle;
  float_t distance;
  float_t length;
  int index;

#ifdef DEVELOPERS
  if (x2 < x1 || x1 < 0 || x2 >= viewwidth || y < 0 || y >= viewheight)
    I_Error("R2_MapPlane: y=%d x=%d..%d", y, x1, x2);
#endif

  distance = fabs(plane_data->h - viewz) * yslope[y];

  length = distance * distscale[x1];
  angle = viewangle + ANG90 + xtoviewangle[x1];

  ds_y  = y;
  ds_x1 = x1;
  ds_x2 = x2;

  ds_xstep = M_FloatToFixed(distance * basexscale);
  ds_ystep = M_FloatToFixed(distance * baseyscale);

  // -AJA- 1999/07/16: Add offsets (thanks Lee).
  // -AJA- NB: Using M_Cos/M_Sin here really slows things down.
  ds_xfrac =  M_FloatToFixed(viewx + plane_data->x_offset +
      FIX_2_FLOAT(finesine[angle >> ANGLETOFINESHIFT]) * length);
  ds_yfrac = M_FloatToFixed(-(viewy - plane_data->y_offset -
      FIX_2_FLOAT(finecosine[angle >> ANGLETOFINESHIFT]) * length));

  // -AJA- 1999/07/10: new stuff for colmap.ddf.
  index = 1 << (LIGHTZSHIFT - FRACBITS);
  CHECKVAL(index);
  index = distance / index;

  if (index < 0)
    index = 0;

  if (index >= MAXLIGHTZ)
    index = MAXLIGHTZ - 1;

  ds_colourmap = V_GetColtable(plane_colourmap, plane_lights[index],
      /* plane_sky */ false ? VCOL_Sky : 0);

  // -AJA- FIXME: move this test to a higher function
  if (plane_translucency >= 0.98)
  {
    spanfunc();
  }
  else
  {
    dc_translucency = FLOAT_2_FIX(plane_translucency);
    trans_spanfunc();
  }
}

static void R2_MakeSpans(int x, int t1, int b1, int t2, int b2)
{
  for (; t1 < t2 && t1 <= b1; t1++)
    R2_MapPlane(t1, spanstart[t1], x-1);

  for (; b1 > b2 && t1 <= b1; b1--)
    R2_MapPlane(b1, spanstart[b1], x-1);

  for (; t2 < t1 && t2 <= b2; t2++)
    spanstart[t2] = x;

  for (; b2 > b1 && t2 <= b2; b2--)
    spanstart[b2] = x;
}

void R2_DrawPlane(drawsub_t *sub, drawplane_t *plane)
{
#if (SHOW_PLANES == 2)
  int x;
  int lightlev;
  int flat;
  
  int L_yt = 9999, L_yb = 0;
  int R_yt, R_yb;

  lightlev = plane->props->lightlevel >> LIGHTSEGSHIFT;

  if (lightlev >= LIGHTLEVELS)
    lightlev = LIGHTLEVELS - 1;

  if (lightlev < 0)
    lightlev = 0;

  CHECKVAL(x_distunit);
  basexscale = viewsin / x_distunit;
  baseyscale = viewcos / x_distunit;

  plane_data = plane->info;
  plane_lights = zlight[lightlev];
  plane_colourmap = plane->props->colourmap;
  plane_translucency = plane_data->translucency;

  flat = flattranslation[plane_data->pic];
  ds_source = W_CacheLumpNum(flatlist[flat]);

  // draw the spans
  for (x = plane->area->x1; x <= plane->area->x2; x++)
  {
    int offset = x - plane->area->x1;

    R_yt = floor(plane->area->yt + offset * plane->area->yt_step);
    R_yb = floor(plane->area->yb + offset * plane->area->yb_step);

    if (R_yt < 0)
      R_yt = 0;

    if (R_yb >= viewheight)
      R_yb = viewheight-1;

    if (R_yt > R_yb)
    {
      R_yt = 9999;
      R_yb = 0;
    }
    
    R2_MakeSpans(x, L_yt, L_yb, R_yt, R_yb);

    L_yt = R_yt;
    L_yb = R_yb;
  }

  R2_MakeSpans(x, L_yt, L_yb, 9999, 0);

  W_DoneWithLump(ds_source);

#elif (SHOW_PLANES == 1)
  BOGUS_DrawArea(plane->area, plane->picnum);
#endif /* SHOW_PLANES */
}

//
// R2_DrawThing
//
// Render a single thing of a subsector.

extern lighttable_t *spritezlights;

void R2_DrawThing(drawsub_t *sub, drawthing_t *thing)
{
#if (SHOW_THINGS == 2)
  int x;
  int index;
  int offset;
  int column;

  column_t *tex_col;
  const patch_t *patch;
  int patch_width;

  float_t yt, yb, yp;

  patch = W_CacheLumpNum(thing->patch);
  patch_width = SHORT(patch->width);

  // calculate lighting
  if (thing->bright)
  {
    dc_colourmap = V_GetColtable(thing->props->colourmap, 255, 0);
  }
  else if (thing->mo->flags & MF_SHADOW)
  {
    dc_colourmap = fuzz_coltable;
  }
  else
  {
    index = (thing->props->lightlevel >> LIGHTSEGSHIFT);

    if (index < 0)
      spritezlights = zlight[0];
    else if (index >= LIGHTLEVELS)
      spritezlights = zlight[LIGHTLEVELS - 1];
    else
      spritezlights = zlight[index];

    index = thing->tz / (1 << (LIGHTZSHIFT - FRACBITS));

    if (index >= MAXLIGHTZ)
      index = MAXLIGHTZ - 1;

    dc_colourmap = V_GetColtable(thing->props->colourmap, 
        spritezlights[index], 0);
  }

  // handle translucency, translation & shadow effect

  index = 0;
  mask_translucency = 1.0;

  if (thing->mo->visibility < VISIBLE && level_flags.trans)
  {
    index |= 1;
    mask_translucency = thing->mo->visibility;
  }

  if (thing->trans_table)
  {
    index |= 2;
    dc_translation = thing->trans_table;
  }
  
  if (thing->mo->flags & MF_SHADOW)
  {
    colfunc = R_DrawFuzzColumn;
  } 
  else
  {
    DEV_ASSERT2(index <= 3);
    
    switch (index)
    {
      case 0: colfunc = R_DrawColumn; break;
      case 1: colfunc = R_DrawTranslucentColumn; break;
      case 2: colfunc = R_DrawTranslatedColumn; break;
      case 3: colfunc = R_DrawTranslucentTranslatedColumn; break;
    }
  }

  // draw the columns
  for (x = thing->area->x1; x <= thing->area->x2; x++)
  {
    offset = x - thing->area->x1;

    yt = thing->area->yt + offset * thing->area->yt_step; 
    yb = thing->area->yb + offset * thing->area->yb_step; 
    yp = thing->area->yp + offset * thing->area->yp_step; 

    dc_x  = x;
    dc_yl = floor(yt);
    dc_yh = floor(yb);
    
    if (dc_yh < dc_yl || dc_yh < 0 || dc_yl >= viewheight)
      continue;

    if (dc_yl < 0)
      dc_yl = 0;

    if (dc_yh >= viewheight) 
      dc_yh = viewheight-1;

    column = floor(thing->xfrac + offset * thing->ixscale);

    if (thing->flip)
      column = patch_width - 1 - column;

    if (column < 0 || column >= patch_width)
      continue;

    mask_yl = (float_t) dc_yl;
    mask_yh = (float_t) dc_yh;
    mask_scale = thing->yscale;
    mask_top = yt - yp * mask_scale;

    // -AJA- pretty icky...
    tex_col = (column_t *) ((byte *) patch +
        LONG(patch->columnofs[column]));
    
    R2_DrawMaskedColumn(tex_col);
  }

  W_DoneWithLump(patch);

  colfunc = basecolfunc;

#elif (SHOW_THINGS == 1)
  BOGUS_DrawArea(thing->area, thing->picnum);
#endif /* SHOW_THINGS */
}

//
// R2_DrawSortThings
//
static drawthing_t **mapdrawthings = NULL;
static int mapdrawthing_num = 0;

static void R2_DrawSortThings(drawsub_t *sub, drawfloor_t *floor)
{
  drawthing_t *thing;

  int i, count=0;

  for (thing=floor->things; thing; thing=thing->next)
    count++;

  if (count == 0)
    return;

  // maybe reallocate array
  if (count > mapdrawthing_num)
  {
    Z_Resize(mapdrawthings, drawthing_t *, count);
    mapdrawthing_num = count;
  }

  // fill array
  count=0;

  for (thing=floor->things; thing; thing=thing->next)
    mapdrawthings[count++] = thing;

  // sort array
  #define CMP(a,b)  ((a)->dist_scale < (b)->dist_scale)
  QSORT(drawthing_t *, mapdrawthings, count, CUTOFF);
  #undef CMP

  // draw array
  for (i=0; i < count; i++)
    R2_DrawThing(sub, mapdrawthings[i]);
}

//
// R2_DrawFloor
//
// Draw a single floor of a subsector.

void R2_DrawFloor(drawsub_t *sub, drawfloor_t *floor)
{
  drawwall_t  *wall;
  drawplane_t *plane;

  for (wall=floor->walls; wall; wall=wall->next)
  {
    R2_DrawWall(sub, wall);
  }

  for (plane=floor->planes; plane; plane=plane->next)
  {
    R2_DrawPlane(sub, plane);
  }

  R2_DrawSortThings(sub, floor);
}

//
// R2_DrawSubsector
//
// Draw a subsector based on all the information collected when
// traversing the BSP tree.

void R2_DrawSubsector(drawsub_t *sub)
{
  drawfloor_t *floor;

  for (floor=sub->floors; floor; floor=floor->next)
  {
    R2_DrawFloor(sub, floor);
  }
}

//
// R2_DrawSky
//
void R2_DrawSky(void)
{
  int x;
  int column;
  int dc_yl2, dc_yh2;
  const cached_tex_t *tex;

  // -AJA- FIXME: this sucks...
  static const colourmap_t *normal = NULL;
  if (! normal)
    normal = DDF_ColmapLookup("NORMAL");

  dc_iscale = M_FloatToFixed(skytexturescale);

  dc_colourmap = V_GetColtable(normal, 255, VCOL_Sky);

  dc_texturemid = M_FloatToFixed(skytexturemid);

  // -KM- 1998/07/31 Give skytextures the possibility of being
  //  animated...
  tex = W_CacheTextureNum(texturetranslation[skytexture]);

  for (x=0; x < viewwindowwidth; x++)
  {
    dc_x = x;

    dc_yl = 0;
    dc_yh = MIN(viewheight-1, ceil(sky_clip_top[x]));

    dc_yl2 = MAX(0, floor(sky_clip_bottom[x]));
    dc_yh2 = viewheight-1;

    if (dc_yl >= dc_yh && dc_yl2 >= dc_yh2)
      continue;
    
    if (dc_yh >= dc_yl2)
    {
      dc_yh = dc_yh2;
      dc_yh2 = -2;
    }
    
    if (dc_yl <= dc_yh)
    {
#ifndef NOSMOOTHING
      dc_xfrac = ((viewangle + xtoviewangle[x]) >> 6) & 0xffff;
#endif
      column = (viewangle + xtoviewangle[x]) >> ANGLETOSKYSHIFT;
      dc_source = W_GetColumn(column, tex);
#ifndef NOSMOOTHING
      dc_source2 = W_GetColumn(column + 1, tex);
#endif
      colfunc();
    }

    dc_yl = dc_yl2;
    dc_yh = dc_yh2;
    
    if (dc_yl <= dc_yh)
    {
#ifndef NOSMOOTHING
      dc_xfrac = ((viewangle + xtoviewangle[x]) >> 6) & 0xffff;
#endif
      column = (viewangle + xtoviewangle[x]) >> ANGLETOSKYSHIFT;
      dc_source = W_GetColumn(column, tex);
#ifndef NOSMOOTHING
      dc_source2 = W_GetColumn(column + 1, tex);
#endif
      colfunc();
    }
  }
  W_DoneWithTexture(tex);
}


// DEBUG routines

void BOGUS_Clear(void)
{
}

void BOGUS_Line(float_t x1, float_t y1, float_t x2, float_t y2, int col)
{
}

