//----------------------------------------------------------------------------
//  EDGE True BSP Rendering (Polygonisation)
//----------------------------------------------------------------------------
// 
//  Copyright (c) 1999-2000  The EDGE Team.
// 
//  This program is free software; you can redistribute it and/or
//  modify it under the terms of the GNU General Public License
//  as published by the Free Software Foundation; either version 2
//  of the License, or (at your option) any later version.
//
//  This program is distributed in the hope that it will be useful,
//  but WITHOUT ANY WARRANTY; without even the implied warranty of
//  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//  GNU General Public License for more details.
//
//----------------------------------------------------------------------------
//
//  Based on the DOOM source code, released by Id Software under the
//  following copyright:
//
//    Copyright (C) 1993-1996 by id Software, Inc.
//
//----------------------------------------------------------------------------
//
// -AJA- 1999/09/16: Wrote this file.
//

#include "i_defs.h"

#include "dm_defs.h"
#include "dm_state.h"
#include "m_bbox.h"

#include "p_local.h"
#include "p_mobj.h"
#include "r_defs.h"
#include "r_main.h"
#include "r_plane.h"
#include "r_things.h"
#include "r_state.h"
#include "v_res.h"
#include "z_zone.h"

#include "r2_defs.h"


#define DEBUG  0


#define ANGLE_ACCURACY  (ANGLEBITS - 14)
#define POINT_ACCURACY  0.9


static int num_errors;
static int num_done;


typedef struct convex_poly_s
{
  float_t mid_x, mid_y;

  halfplane_t *planes;
  cut_point_t *points;
  cut_point_t *lines;
}
convex_poly_t;


// Here we keep a record of the nodes visited while walking the normal
// BSP tree.

#define MAX_BSP_DEPTH  256

static node_t *visited_nodes[MAX_BSP_DEPTH];
static int visited_sides[MAX_BSP_DEPTH];
static int visited_node_num;


static float_t RoughPerp(halfplane_t *A, float_t x, float_t y)
{
  float_t px = x - A->x1;
  float_t py = y - A->y1;

  float_t dx = A->x2 - A->x1;
  float_t dy = A->y2 - A->y1;

  return (px*dy) - (py*dx);
}

static float_t ExactPerp(halfplane_t *A, float_t x, float_t y)
{
  float_t px = x - A->x1;
  float_t py = y - A->y1;

  float_t dx = A->x2 - A->x1;
  float_t dy = A->y2 - A->y1;

  float_t len = sqrt(dx*dx + dy*dy);

  return (px * dy - py * dx) / len;
}

static void IntersectLines(halfplane_t * A, halfplane_t * B, 
    float_t *x, float_t *y)
{
  // use rough perp here, since the length cancels out
  float_t perp1 = RoughPerp(A, B->x1, B->y1);
  float_t perp2 = RoughPerp(A, B->x2, B->y2);

  DEV_ASSERT2(perp1 != perp2);

  *x = B->x1 + (B->x2 - B->x1) * perp1 / (perp1 - perp2);
  *y = B->y1 + (B->y2 - B->y1) * perp1 / (perp1 - perp2);
}

static void PolyAddPoint(convex_poly_t * poly, float_t x, float_t y)
{
  cut_point_t *cut;

  // check if point already present
  for (cut=poly->points; cut != NULL; cut=cut->next)
  {
    float_t x_diff = ABS(x - cut->x);
    float_t y_diff = ABS(y - cut->y);

    if (x_diff < POINT_ACCURACY && y_diff < POINT_ACCURACY)
    {
      // close enough to be the same
      return;
    }
  }

  // check if point can be "snapped" to a vertex
  // !!! disabled for now, really slows it down
#if 0
  {
    int i;

    for (i=0; i < numvertexes; i++)
    {
      float_t vx = vertexes[i].x;
      float_t vy = vertexes[i].y;

      float_t x_diff = ABS(x - vx);
      float_t y_diff = ABS(y - vy);

      if (MAX(x_diff, y_diff) < POINT_ACCURACY)
      {
        // Snap ! :-)
        x = vx;
        y = vy;
        break;
      }
    }
  }
#endif

  // link in new point
  cut = R2_NewCutPoint();

  cut->x = x;
  cut->y = y;

  cut->what_seg = NULL;

  cut->x2 = cut->y2 = cut->angle2 = 0;

  cut->next = poly->points;
  poly->points = cut;
}

// returns false if lines are parallel
static boolean_t PolyIntersectHalfplane(convex_poly_t * poly, 
    halfplane_t * A, halfplane_t * B)
{
  angle_t angle1 = R_PointToAngle(A->x1, A->y1, A->x2, A->y2);
  angle_t angle2 = R_PointToAngle(B->x1, B->y1, B->x2, B->y2);

  angle_t diff = angle2 - angle1;

  float_t x, y;
  
  if (diff >= ANG180)
    diff = 0 - diff;
 
  if (diff <= (1 << ANGLE_ACCURACY) ||
      diff >= (ANG180 - (1 << ANGLE_ACCURACY)))
  {
    // parallel lines
    return false;
  }

  IntersectLines(A, B, &x, &y);

  PolyAddPoint(poly, x, y);

  return true;
}

#define HPSIGN(f)  ((f) < 0 ? 1 : 0)

// returns true if the seg lies along the halfplane but is facing the
// wrong way.
static boolean_t SegWrongOnHalfplane(halfplane_t *seg, halfplane_t *hp)
{
  float_t dx1 = seg->x2 - seg->x1;
  float_t dy1 = seg->y2 - seg->y1;
  float_t dx2 = hp->x2 - hp->x1;
  float_t dy2 = hp->y2 - hp->y1;

  if (ABS(ExactPerp(hp, seg->x1, seg->y1)) > POINT_ACCURACY)
    return false;

  if (ABS(ExactPerp(hp, seg->x2, seg->y2)) > POINT_ACCURACY)
    return false;

  return (HPSIGN(dx1) != HPSIGN(dx2)) ||
         (HPSIGN(dy1) != HPSIGN(dy2));
}

// returns true if the other seg runs along the main seg and is facing
// the opposite way.

static boolean_t SegAdjacentOnSeg(seg_t *main, seg_t *other)
{
  float_t dx1 = main->x2 - main->x1;
  float_t dy1 = main->y2 - main->y1;
  float_t dx2 = other->x2 - other->x1;
  float_t dy2 = other->y2 - other->y1;
  
  halfplane_t h;
  
  // Bounding box check
  
  if (MIN(other->x1,other->x2) > MAX(main->x1,main->x2) + POINT_ACCURACY ||
      MIN(other->y1,other->y2) > MAX(main->y1,main->y2) + POINT_ACCURACY ||
      MAX(other->x1,other->x2) < MIN(main->x1,main->x2) + POINT_ACCURACY ||
      MAX(other->y1,other->y2) < MIN(main->y1,main->y2) + POINT_ACCURACY)
  {
    return false;
  }
  
  h.x1 = main->x1;
  h.y1 = main->y1;
  h.x2 = main->x2;
  h.y2 = main->y2;
  
  if (ABS(ExactPerp(&h, other->x1, other->y1)) > POINT_ACCURACY)
  {
    return false;
  }
  if (ABS(ExactPerp(&h, other->x2, other->y2)) > POINT_ACCURACY)
  {
    return false;
  }

  // Sanity check

  h.x1 = other->x1;
  h.y1 = other->y1;
  h.x2 = other->x2;
  h.y2 = other->y2;
  
  if (ABS(ExactPerp(&h, main->x1, main->y1)) > POINT_ACCURACY)
  {
    return false;
  }
  if (ABS(ExactPerp(&h, main->x2, main->y2)) > POINT_ACCURACY)
  {
    return false;
  }

  return (HPSIGN(dx1) != HPSIGN(dx2)) ||
         (HPSIGN(dy1) != HPSIGN(dy2));
}

static void PolyAddHalfplane(convex_poly_t * poly,
    float_t ax, float_t ay, float_t bx, float_t by, boolean_t is_seg)
{
  halfplane_t *hp, *cur;

#if (DEBUG >= 2)
  I_WriteDebug("  AddHalfplane (%1.1f, %1.1f) -> (%1.1f, %1.1f) %s\n",
ax, ay, bx, by, is_seg ? "SEG" : "NODE");
#endif

  hp = R2_NewHalfPlane();

  hp->x1 = ax;
  hp->y1 = ay;
  hp->x2 = bx;
  hp->y2 = by;

  // create points
  for (cur=poly->planes; cur != NULL; cur=cur->next)
  {
    boolean_t par;

    par = !PolyIntersectHalfplane(poly, cur, hp);

    //
    // Sanity check: if this seg lines along one of the halfplanes,
    // but going in the opposite direction, then something is wrong
    // with the WAD.
    //
    if (par && is_seg && SegWrongOnHalfplane(hp, cur))
    {
      I_WriteDebug("PolyAddHalfplane: bad seg in subsector.\n");
      num_errors++;
      return;
    }
  }

  // link in new halfplane

  hp->next = poly->planes;
  poly->planes = hp;
}

static boolean_t PointInHalfplane(halfplane_t * hp, cut_point_t * p)
{
  float_t perp = ExactPerp(hp, p->x, p->y);

  // Some fudging here, so that points that lie directly on the
  // halfplane line don't get discarded.
  
  return (perp > -POINT_ACCURACY / 2.0);
}

static boolean_t PolyDudPoint(convex_poly_t * poly, cut_point_t * p)
{
  halfplane_t *hp;

  for (hp=poly->planes; hp != NULL; hp=hp->next)
  {
    if (! PointInHalfplane(hp, p))
      return true;
  }

  return false;
}

static void PolyRemoveDuds(convex_poly_t * poly)
{
  cut_point_t *cur;
  cut_point_t *next;
  cut_point_t *prev = NULL;

  for (cur=poly->points; cur != NULL; cur=next)
  {
    next = cur->next;
    
    if (PolyDudPoint(poly, cur))
    {
      // unlink it
      if (prev)
        prev->next = next;
      else
        poly->points = next;
    }
    else
    {
      prev = cur;
    }
  }
}

static void PolyComputeMid(convex_poly_t * poly)
{
  cut_point_t *cur;

  float_t avg_x = 0;
  float_t avg_y = 0;

  int avg_num = 0;

  for (cur=poly->points; cur != NULL; cur=cur->next)
  {
    avg_x += cur->x;
    avg_y += cur->y;

    avg_num++;
  }

  if (avg_num == 0)
  {
    I_WriteDebug("PolyComputeMid: no points !\n");
    num_errors++;
    return;
  }
  
  poly->mid_x = avg_x / (float_t) avg_num;
  poly->mid_y = avg_y / (float_t) avg_num;

#if (DEBUG >= 2)
  I_WriteDebug("  POLY MID = (%1.1f, %1.1f)\n", poly->mid_x, poly->mid_y);
#endif
}

static void PolyComputeAngles(convex_poly_t * poly)
{
  cut_point_t *cur;

  for (cur=poly->points; cur != NULL; cur=cur->next)
  {
    cur->angle = R_PointToAngle(poly->mid_x, poly->mid_y, cur->x, cur->y);

#if (DEBUG >= 3)
  I_WriteDebug("    CUT POINT (%1.1f, %1.1f) angle %1.2f\n",
      cur->x, cur->y, ANG_2_FLOAT(cur->angle));
#endif
  }
}

static void PolySortByAngle(convex_poly_t * poly)
{
  cut_point_t *cur;
  cut_point_t *prev;

  cut_point_t *min;
  cut_point_t *min_prev;

  cut_point_t *sort_head;

  // do insertion sort

  sort_head = poly->lines;
  poly->lines = NULL;

#if (DEBUG >= 3)
  I_WriteDebug("    SORTING:\n");
#endif

  while (sort_head != NULL)
  {
    // find minimum from old list
    min = sort_head;
    min_prev = NULL;
  
    for (prev=sort_head, cur=sort_head->next;
         cur != NULL; 
         prev=cur, cur=cur->next)
    {
      if (cur->angle < min->angle)
      {
        min = cur;
        min_prev = prev;
      }
    }

    // remove it from old list
    if (min_prev)
      min_prev->next = min->next;
    else
      sort_head = min->next;

    // add it to head of new list
    min->next = poly->lines;
    poly->lines = min;

#if (DEBUG >= 3)
    if (min->what_seg)
      I_WriteDebug("      SEG %p  angles %1.2f -> %1.2f  "
          "(%1.1f, %1.1f) -> (%1.1f, %1.1f)\n",
          min->what_seg, ANG_2_FLOAT(min->angle), ANG_2_FLOAT(min->angle2),
          min->x, min->y, min->x2, min->y2);
#elif (DEBUG >= 2)
    I_WriteDebug("      POINT   angles %1.2f  (%1.1f, %1.1f)\n", 
        ANG_2_FLOAT(min->angle), min->x, min->y);
#endif
  }

  // Lines have now been sorted from maximum angle to minimum angle,
  // which ensures that they are in ANTI-clockwise order.
}

#define ANGLE_IN_RANGE(L,x,H)  \
        ((L) <= (H) ? ((L) <= (x) && (x) <= (H))  \
                    : ((x) <= (H) || (x) >= (L)))
        
static void PolyAddSeg(convex_poly_t * poly, seg_t *seg)
{
  float_t x1 = seg->x1;
  float_t y1 = seg->y1;
  float_t x2 = seg->x2;
  float_t y2 = seg->y2;

  cut_point_t *cur;

  cur = R2_NewCutPoint();

  cur->what_seg = seg;

  cur->x = x1;
  cur->y = y1;

  cur->x2 = x2;
  cur->y2 = y2;

  cur->angle  = R_PointToAngle(poly->mid_x, poly->mid_y, x1, y1);
  cur->angle2 = R_PointToAngle(poly->mid_x, poly->mid_y, x2, y2);

  if (cur->angle - cur->angle2 >= ANG180)
  {
    I_WriteDebug("PolyAddSeg: Bad range: first=0x%04x second=0x%04x\n",
        cur->angle >> 16, cur->angle2 >> 16);
    num_errors++;
    return;
  }
  
  // link it in
  cur->next = poly->lines;
  poly->lines = cur;
}

static boolean_t PolyPointShadowed(convex_poly_t * poly, cut_point_t *p)
{
  cut_point_t *cur_s;

  for (cur_s=poly->lines; cur_s != NULL; cur_s=cur_s->next)
  {
    angle_t first, second;

    if (! cur_s->what_seg)
      continue;

    // Some fudging, to allow for small roundoff errors that may have
    // occurred when computing the points / angles, and to ignore
    // points that lie directly at the leading/trailing points of the
    // seg.
    
    first  = cur_s->angle  + (2 << ANGLE_ACCURACY);
    second = cur_s->angle2 - (2 << ANGLE_ACCURACY);

    if (ANGLE_IN_RANGE(second, p->angle, first))
      return true;
  }

  return false;
}

static void PolyTransferNonShadowed(convex_poly_t * poly)
{
  cut_point_t *cur;
  
  while (poly->points)
  {
    cur = poly->points;

    // remove from point list
    poly->points = cur->next;

    if (PolyPointShadowed(poly, cur))
    {
      // forget it
    }
    else
    {
      // add it to lines list
      cur->next = poly->lines;
      poly->lines = cur;
    }
  }
}

//
// Allocated Segs
//
// Most seg_ts are part of a global seg_t chunk, which will be Z_Freed
// with one single call. But some segs are added by this module, which
// are not part of that list. Therefore we need to maintain a list of which
// segs we have to Z_Free individually at level exit.
//
typedef struct allocated_seg_s
{
  struct allocated_seg_s *next;
  seg_t seg;
} allocated_seg_t;

static allocated_seg_t *allocated_segs = NULL;
static seg_t *AllocateSeg(void)
{
  allocated_seg_t *newseg = Z_New(allocated_seg_t, 1);
  newseg->next = allocated_segs;
  allocated_segs = newseg;
  return &newseg->seg;
}

void R2_FreeAllocatedSegs(void)
{
  allocated_seg_t *cur;

  while (allocated_segs)
  {
    cur = allocated_segs;
    allocated_segs = allocated_segs->next;
    Z_Free(cur);
  }
}

static seg_t * MakeMiniseg(subsector_t * sub,
    float_t x1, float_t y1, float_t x2, float_t y2)
{
  seg_t *mini = AllocateSeg();

  mini->x1 = x1;  mini->y1 = y1;
  mini->x2 = x2;  mini->y2 = y2;

  mini->angle  = R_PointToAngle(x1, y1, x2, y2);
  mini->length = R_PointToDist (x1, y1, x2, y2);

  mini->front_sub = sub;
  mini->back_sub  = NULL;
  mini->partner   = NULL;

  mini->miniseg = true;

  return mini;
}

static void LastDitchBackOfSeg(subsector_t *sub, seg_t *seg)
{
  float_t x, y;
  float_t dx, dy;

  x = seg->x1/2 + seg->x2/2;
  y = seg->y1/2 + seg->y2/2;

  dx = -(seg->y2 - seg->y1) / 2;
  dy =  (seg->x2 - seg->x1) / 2;

  // ouch, what a hack...
  while (ABS(dx) > 2.0 || ABS(dy) > 2.0)
  {
    dx /= 2;
    dy /= 2;
  }

  seg->back_sub = R_PointInSubsector(x+dx, y+dy);
}

static void ComputeBoundingBox(subsector_t *sub, float_t *bbox)
{
  seg_t *seg;

  M_ClearBox(bbox);

  for (seg=sub->segs; seg != NULL; seg=seg->sub_next)
  {
    M_AddToBox(bbox, seg->x1, seg->y1);
    M_AddToBox(bbox, seg->x2, seg->y2);
  }
}

static void PolygoniseSubsector(int subsec, float_t *bbox)
{ 
  subsector_t *sub = &subsectors[subsec];
  seg_t *seg;

  convex_poly_t poly;

  cut_point_t *cut;
  cut_point_t *cut_start;
  cut_point_t *cut_prev;
  
  int i;

  poly.planes = NULL;
  poly.points = NULL;
  poly.lines  = NULL;

  R2_ClearPoly();

  num_errors = 0;

#if (DEBUG >= 1)
  I_WriteDebug("PolygoniseSubsector %d\n", subsec);
#endif

  // add halfplane from each node
  for (i=0; i < visited_node_num; i++)
  {
    node_t *node = visited_nodes[i];
    
    int side = visited_sides[i];

    float_t x1 = node->line.x;
    float_t y1 = node->line.y;
    float_t x2 = node->line.x + node->line.dx;
    float_t y2 = node->line.y + node->line.dy;

    if (side == 0)
    {
      PolyAddHalfplane(&poly, x1, y1, x2, y2, false);
    }
    else
    {
      PolyAddHalfplane(&poly, x2, y2, x1, y1, false);
    }
  }

  // add halfplane from each seg in subsector

  for (seg=sub->segs; seg != NULL; seg=seg->sub_next)
  {
    float_t x1 = seg->x1;
    float_t y1 = seg->y1;
    float_t x2 = seg->x2;
    float_t y2 = seg->y2;

    PolyAddHalfplane(&poly, x1, y1, x2, y2, true);
  }

  PolyRemoveDuds(&poly);
  PolyComputeMid(&poly);
  PolyComputeAngles(&poly);

  for (seg=sub->segs; seg != NULL; seg=seg->sub_next)
  {
    PolyAddSeg(&poly, seg);
  }

  PolyTransferNonShadowed(&poly);
  PolySortByAngle(&poly);

  // Finally !  Now we can add the minisegs to the subsector, by going
  // ANTI-clockwise around the points and checking for non-seg points.
  //
  // PS: we re-create the subsector's seg-list here too.  Thus we get
  // a nice convex polygon with all the vertices in clockwise order.

  if (poly.lines == NULL || poly.lines->next == NULL)
  {
    I_WriteDebug("PolygoniseSubsector: bad subsector.\n");
    num_errors++;
    return;
  }

  sub->segs = NULL;

  cut_prev = poly.lines;
  cut = cut_start = cut_prev->next;

  do
  {
    boolean_t gap;
    
    float_t x1, y1;
    float_t x2, y2;
    angle_t ang1, ang2;

    if (cut_prev->what_seg)
    {
      x1 = cut_prev->x2;
      y1 = cut_prev->y2;
      ang1 = cut_prev->angle2;
    }
    else
    {
      x1 = cut_prev->x;
      y1 = cut_prev->y;
      ang1 = cut_prev->angle;
    }

    x2 = cut->x;
    y2 = cut->y;
    ang2 = cut->angle;

    gap = true;

    // length check, ignore really short gaps
    if (ABS(x2 - x1) < POINT_ACCURACY &&
        ABS(y2 - y1) < POINT_ACCURACY)
    {
      gap = false;
    }
    else if (cut_prev->what_seg && cut->what_seg)
    {
      // check if segs touch
      int diff = (int) (ang1 - ang2);

      if (ABS(diff) <= (1 << ANGLE_ACCURACY))
        gap = false;
    }
    
    if (gap)
    {
      // link in new miniseg
      seg_t *mini = MakeMiniseg(sub, x1, y1, x2, y2);

      mini->sub_next = sub->segs;
      sub->segs = mini;

#if (DEBUG >= 2)
  I_WriteDebug("  ADDING MINISEG  angles %1.2f -> %1.2f  "
      "(%1.1f, %1.1f) -> (%1.1f, %1.1f)\n", 
      ANG_2_FLOAT(ang1), ANG_2_FLOAT(ang2), x1, y1, x2, y2);
#endif
    }

    if (cut->what_seg)
    {
      // re-link it in
      cut->what_seg->sub_next = sub->segs;
      sub->segs = cut->what_seg;

#if (DEBUG >= 2)
  I_WriteDebug("  ADDING SEG %p  angles %1.2f -> %1.2f  "
      "(%1.1f, %1.1f) -> (%1.1f, %1.1f)\n", 
      cut->what_seg, ANG_2_FLOAT(cut->angle), ANG_2_FLOAT(cut->angle2),
      cut->what_seg->x1, cut->what_seg->y1, 
      cut->what_seg->x2, cut->what_seg->y2);
#endif
    }

    cut_prev = cut;
    cut = (cut->next) ? cut->next : poly.lines;
  }
  while (cut != cut_start);
  
  ComputeBoundingBox(sub, bbox);
  
#if (DEBUG == 0)
  if (++num_done % 10 == 0)
    I_Printf(num_errors ? "!" : ".");
#else
  I_WriteDebug("\n");
#endif
}
 
static void PolygoniseBSPNode(int bspnum, float_t *bbox)
{
  node_t *node;

//  I_Error("ANDY IN TROUBLE at PolygoniseBSPNode, %d",numnodes);

  // found a subsector ?
  if (bspnum & NF_SUBSECTOR)
  {
    PolygoniseSubsector(bspnum & ~NF_SUBSECTOR, bbox);
    return;
  }

  node = &nodes[bspnum];

  if (visited_node_num == MAX_BSP_DEPTH)
    I_Error("PolygoniseSubsector: BSP Tree too deep !\n");
   
  visited_nodes[visited_node_num++] = node;

  // recursively walk the sub-trees

  visited_sides[visited_node_num-1] = 0;
  PolygoniseBSPNode(node->children[0], node->bbox[0]);

  visited_sides[visited_node_num-1] = 1;
  PolygoniseBSPNode(node->children[1], node->bbox[1]);

  // compute new bounding box
  M_CopyBox(bbox, node->bbox[0]);
  M_UnionBox(bbox, node->bbox[1]);

  visited_node_num--;
}

// Returns RHS.
static seg_t * SplitOneSeg(subsector_t * sub, seg_t * seg,  
    float_t end_x, float_t end_y)
{
  seg_t *new_seg = AllocateSeg();

  float_t dx = seg->x2 - end_x;
  float_t dy = seg->y2 - end_y;
  float_t len = R_PointToDist(0.0, 0.0, dx, dy);

  new_seg[0] = seg[0];

  seg->x2 = new_seg->x1 = end_x;
  seg->y2 = new_seg->y1 = end_y;

  new_seg->offset += R_PointToDist(0.0, 0.0, dx, dy);

  // compute lengths
  new_seg->length = len;
  seg->length    -= len;

  // link it in
  new_seg->sub_next = seg->sub_next;
  seg->sub_next = new_seg;

  return new_seg;
}

static boolean_t AdjacentSegPair(subsector_t *main_sub, seg_t *main,
    subsector_t *other_sub, seg_t *other)
{
//  float_t x, y; // UNUSED
  float_t dx, dy;
  float_t m1, m2, o1, o2;
  
  // ignore one sided segs
  if (!other->miniseg && other->backsector == NULL)
  {
    return false;
  }

  // check if the segs exactly match
  if (ABS(main->x1 - other->x2) < POINT_ACCURACY &&
      ABS(main->y1 - other->y2) < POINT_ACCURACY &&
      ABS(main->x2 - other->x1) < POINT_ACCURACY &&
      ABS(main->y2 - other->y1) < POINT_ACCURACY)
  {
    main->partner  = other;
    main->back_sub = other_sub;

    other->partner = main;
    other->back_sub = main_sub;

    return true;
  }

  // check if they are on the same line
  if (! SegAdjacentOnSeg(main, other))
  {
    return false;
  }

  // calculate base point and delta x & y
// --UNUSED--
//  x = main->x1;
//  y = main->y1;

  dx = main->x2 - main->x1;
  dy = main->y2 - main->y1;

  // determine offsets (in deltas) from base point
  // Note that we swap o1 & o2, which normally face the opposite way.
  m1 = 0.0;
  m2 = 1.0;

  if (ABS(dy) > ABS(dx))
  {
    m1 *= ABS(dy);
    m2 *= ABS(dy);

//  --UNUSED--
//    if (ABS(dy))
//      dx /= ABS(dy);

    if (ABS(dy))
      dy /= ABS(dy);

    o2 = (other->y1 - main->y1) * dy;
    o1 = (other->y2 - main->y1) * dy;
  }
  else
  {
    m1 *= ABS(dx);
    m2 *= ABS(dx);

//  --UNUSED--
//    if (ABS(dx))
//      dy /= ABS(dx);

    if (ABS(dx))
      dx /= ABS(dx);

    o2 = (other->x1 - main->x1) * dx;
    o1 = (other->x2 - main->x1) * dx;
  }

  if (o2 < o1)
  {
    // -AJA- shouldn't happen, but really short segs can pass the
    // SegAdjacentOnSeg() perpendicular tests when they shouldn't.
    return false;
  }

  // OK, now we're talking 1D coords -- much easier.

  // check for overlap
  if (m1 > o2 - POINT_ACCURACY || m2 < o1 + POINT_ACCURACY)
    return false;

  // check if main seg is totally contained in other seg
  if (o1 - POINT_ACCURACY < m1 && m2 < o2 + POINT_ACCURACY)
  {
    main->back_sub = other_sub;
    return true;
  }

  // Damn, need to split main seg

  // split to the right ?
  if (m2 > o2 + POINT_ACCURACY)
  {
    SplitOneSeg(main_sub, main, other->x1, other->y1);
  }
  
  // split to the left ?
  if (m1 < o1 - POINT_ACCURACY)
  {
    main = SplitOneSeg(main_sub, main, other->x2, other->y2);
  }
  
  main->partner = other;
  main->back_sub = other_sub;

  // make sure we don't miss any of the new seg pieces
  return false;
}

static void FixupSeg(subsector_t *sub, seg_t *seg)
{
  int i;

  // ignore one sided segs
  if (!seg->miniseg && seg->backsector == NULL)
    return;

  for (i=0; i < numsubsectors && seg->back_sub == NULL; i++)
  {
    subsector_t *test_sub = & subsectors[i];
    seg_t *other;

    // ignore ourselves
    if (test_sub == sub)
      continue;

    // ignore non-adjacent subsectors
    if (seg->miniseg && test_sub->sector != sub->sector)
      continue;

    if (!seg->miniseg && test_sub->sector != seg->backsector)
      continue;

    for (other=test_sub->segs; 
         other != NULL && seg->back_sub == NULL; 
         other=other->sub_next)
    {
      if (AdjacentSegPair(sub, seg, test_sub, other))
        break;
    }
  }

  if (seg->back_sub == NULL)
  {
    I_WriteDebug("FixupSeg: Couldn't pair off seg %p.\n", seg);
    num_errors++;
    LastDitchBackOfSeg(sub, seg);
  }
}

static void FixupAdjacentSegs(void)
{
  int i;

  for (i=0; i < numsubsectors; i++)
  {
    subsector_t *sub = & subsectors[i];
    seg_t *seg;

    num_errors = 0;

    for (seg=sub->segs; seg != NULL; seg=seg->sub_next)
    {
      FixupSeg(sub, seg);
    }

#if (DEBUG == 0)
    if (++num_done % 10 == 0)
      I_Printf(num_errors ? "!" : ".");
#endif
  }
}

//
// R2_Polygonise
//
// Polygonise all the subsectors, by walking the BSP tree.
//
void R2_Polygonise(void)
{
  float_t dummy_bbox[4];

  I_Printf("Polygonising");
  num_done = 0;

  visited_node_num = 0;

  if (numnodes == 0)
    PolygoniseSubsector(0, dummy_bbox);
  else
    PolygoniseBSPNode(numnodes - 1, dummy_bbox);

  I_Printf("\nFixing up");
  num_done = 0;

  FixupAdjacentSegs();

  I_Printf("\n");

  R2_FreeupPoly();
}

