//----------------------------------------------------------------------------
//  EDGE True BSP Rendering (Utility routines)
//----------------------------------------------------------------------------
// 
//  Copyright (c) 1999-2000  The EDGE Team.
// 
//  This program is free software; you can redistribute it and/or
//  modify it under the terms of the GNU General Public License
//  as published by the Free Software Foundation; either version 2
//  of the License, or (at your option) any later version.
//
//  This program is distributed in the hope that it will be useful,
//  but WITHOUT ANY WARRANTY; without even the implied warranty of
//  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//  GNU General Public License for more details.
//
//----------------------------------------------------------------------------
//
//  Based on the DOOM source code, released by Id Software under the
//  following copyright:
//
//    Copyright (C) 1993-1996 by id Software, Inc.
//
//----------------------------------------------------------------------------
//
// -AJA- 1999/08/31: Wrote this file.
//

#include "i_defs.h"

#include "dm_state.h"
#include "m_fixed.h"

#include "r_data.h"
#include "r_defs.h"
#include "r_state.h"
#include "r_main.h"
#include "v_res.h"
#include "z_zone.h"

#include "r2_defs.h"


#define ALLOC_Q  16

#define TDEBUG  0

#define F12UNIT  (1 << 12)


// arrays of stuff

static  halfplane_t ** arr_halves    = NULL;
static  cut_point_t ** arr_points    = NULL;
static screenline_t ** arr_lines     = NULL;
static   drawwall_t ** arr_walls     = NULL;
static  drawplane_t ** arr_planes    = NULL;
static  drawthing_t ** arr_things    = NULL;
static  drawfloor_t ** arr_floors    = NULL;
static  planeback_t ** arr_backs     = NULL;
static    clipseg_t ** arr_clipsegs  = NULL;
          drawsub_t ** arr_drawsubs  = NULL;

static int num_halves   = 0, max_halves   = 0;
static int num_points   = 0, max_points   = 0;
static int num_lines    = 0, max_lines    = 0;
static int num_walls    = 0, max_walls    = 0;
       int num_planes   = 0, max_planes   = 0;
static int num_things   = 0, max_things   = 0;
static int num_floors   = 0, max_floors   = 0;
static int num_backs    = 0, max_backs    = 0;
static int num_clipsegs = 0, max_clipsegs = 0;
       int num_drawsubs = 0, max_drawsubs = 0;


screenline_t *empty_list;

drawsub_t ** drawsub_array = NULL;
int drawsub_array_size = -1;

static int sky_clip_size = 0;
float_t *sky_clip_top = NULL;
float_t *sky_clip_bottom = NULL;


// poly clear function

void R2_ClearPoly(void)
{
  num_halves = 0;
  num_points = 0;
}

// bsp clear function

void R2_ClearBSP(void)
{
  int x;

  num_lines    = 0;
  num_walls    = 0;
  num_planes   = 0;
  num_things   = 0;
  num_floors   = 0;
  num_drawsubs = 0;
  num_clipsegs = 0;

  empty_list = NULL;

  if (drawsub_array_size != numsubsectors)
  {
    drawsub_array_size = numsubsectors;

    Z_Resize(drawsub_array, drawsub_t *, drawsub_array_size);
  }

  Z_Clear(drawsub_array, drawsub_t *, numsubsectors);

  // realloc clip arrays, if needed
  // -ES- FIXME: use stack array!
  if (sky_clip_size < viewwidth)
  {
    sky_clip_size = viewwidth;

    Z_Resize(sky_clip_top,    float_t, sky_clip_size);
    Z_Resize(sky_clip_bottom, float_t, sky_clip_size);
  }

  for (x=0; x < sky_clip_size; x++)
  {
    sky_clip_top[x]    = (float_t) 0;
    sky_clip_bottom[x] = (float_t) viewheight;
  }
}

// plane clear function

void R2_ClearPlanes(void)
{
  num_backs = 0;
}


// allocate functions

// -ES- FIXME: Use stack array
#define ALLOCATOR(TYPE, BASEVAR, NUMVAR, MAXVAR)  \
{  \
  TYPE *cur; int i;  \
  \
  if (NUMVAR >= MAXVAR)  \
  {  \
    MAXVAR += ALLOC_Q;  \
    Z_Resize(BASEVAR, TYPE *, MAXVAR);  \
  \
    cur = Z_New(TYPE, ALLOC_Q);  \
  \
    for (i=0; i < ALLOC_Q; i++)  \
    {  \
      BASEVAR[MAXVAR-ALLOC_Q + i] = & cur[i];  \
    }  \
  }  \
  \
  NUMVAR++; return BASEVAR[NUMVAR - 1];  \
}

#define DEALLOCATOR(TYPE, BASEVAR, NUMVAR, MAXVAR)  \
do {  \
  int i;  \
  \
  if (BASEVAR)  \
  {  \
    for (i=0; i < MAXVAR; i += ALLOC_Q)  \
    {  \
      Z_Free(BASEVAR[i]);  \
    }  \
    \
    Z_Free(BASEVAR);  \
  }  \
  \
  NUMVAR = MAXVAR = 0;  \
  BASEVAR = NULL;  \
\
} while (0)


halfplane_t *R2_NewHalfPlane(void)
  ALLOCATOR(halfplane_t, arr_halves, num_halves, max_halves)

cut_point_t *R2_NewCutPoint(void)
  ALLOCATOR(cut_point_t, arr_points, num_points, max_points)

screenline_t *R2_NewScreenLine(void)
  ALLOCATOR(screenline_t, arr_lines, num_lines, max_lines)

drawwall_t *R2_NewDrawWall(void)
  ALLOCATOR(drawwall_t, arr_walls, num_walls, max_walls)

drawplane_t *R2_NewDrawPlane(void)
  ALLOCATOR(drawplane_t, arr_planes, num_planes, max_planes)

drawthing_t *R2_NewDrawThing(void)
  ALLOCATOR(drawthing_t, arr_things, num_things, max_things)

drawfloor_t *R2_NewDrawFloor(void)
  ALLOCATOR(drawfloor_t, arr_floors, num_floors, max_floors)

drawsub_t *R2_NewDrawSub(void)
  ALLOCATOR(drawsub_t, arr_drawsubs, num_drawsubs, max_drawsubs)

planeback_t *R2_NewPlaneBack(void)
  ALLOCATOR(planeback_t, arr_backs, num_backs, max_backs)

clipseg_t *R2_NewClipSeg(void)
  ALLOCATOR(clipseg_t, arr_clipsegs, num_clipsegs, max_clipsegs)


void R2_FreeupPoly(void)
{
  DEALLOCATOR(halfplane_t, arr_halves, num_halves, max_halves);
  DEALLOCATOR(cut_point_t, arr_points, num_points, max_points);
}

static void FreeupBSP(void)
{
  DEALLOCATOR(screenline_t, arr_lines, num_lines, max_lines);
  DEALLOCATOR(drawwall_t, arr_walls, num_walls, max_walls);
  DEALLOCATOR(drawplane_t, arr_planes, num_planes, max_planes);
  DEALLOCATOR(drawthing_t, arr_things, num_things, max_things);
  DEALLOCATOR(drawfloor_t, arr_floors, num_floors, max_floors);
  DEALLOCATOR(drawsub_t, arr_drawsubs, num_drawsubs, max_drawsubs);
  DEALLOCATOR(planeback_t, arr_backs, num_backs, max_backs);
  DEALLOCATOR(clipseg_t, arr_clipsegs, num_clipsegs, max_clipsegs);
}

#if 0
static void DEBUG_Lines(void)
{
  screenline_t *cur;
  
  for (cur=empty_list; cur != NULL; cur=cur->next)
  {
    float_t yt2 = cur->yt + cur->yt_step * (cur->x2 - cur->x1);
    float_t yb2 = cur->yb + cur->yb_step * (cur->x2 - cur->x1);

    fprintf(stderr, "EMPTY %d..%d : [%1.1f -> %1.1f] .. [%1.1f -> %1.1f]\n",
      cur->x1, cur->x2, cur->yt, cur->yb, yt2, yb2);
  }

  fprintf(stderr, "END\n\n");
}
#endif


//
// R2_AddInitialLines
//
// Creates the initial empty line list which defines the borders of
// the screen.
//

void R2_AddInitialLines(void)
{
  screenline_t *line = R2_NewScreenLine();

  line->next = NULL;
  line->prev = NULL;

  line->x1 = 0;
  line->x2 = viewwidth - 1;

  line->yt = 0;
  line->yt_step = 0;

  line->yb = (float_t) viewheight;
  line->yb_step = 0;

  empty_list = line;
}

//
// R2_GetDrawsubFromSubsec
//
drawsub_t *R2_GetDrawsubFromSubsec(subsector_t *subsec)
{
  int index = subsec - subsectors;
  drawsub_t *dsub;

  if (! drawsub_array[index])
  {
    drawsub_array[index] = dsub = R2_NewDrawSub();

    // initialise it
    Z_Clear(dsub, drawsub_t, 1);

    dsub->subsec = subsec;
  }

  return drawsub_array[index];
}

//
// R2_CheckOccluded
//
// Checks to see if the inclusive range of columns is occluded by the
// current set of screen lines.  Return true if occluded, false
// otherwise.  Note that the 1D occlusion buffer is kept in sorted
// order (by x).

boolean_t R2_CheckOccluded(int x1, int x2)
{
  screenline_t *cur;

  for (cur=empty_list; cur != NULL; cur=cur->next)
  {
    if (cur->x1 > x2)
      break;

    if (cur->x2 >= x1)
      return false;
  }
  
  return true;
}

//
// R2_CheckOccludedShrink
//
// Checks to see if the inclusive range of columns is occluded by the
// current set of screen lines.  Return true if occluded, false
// otherwise (whereupon x1 & x2 are shrunk as much as possible).

boolean_t R2_CheckOccludedShrink(int *x1, int *x2)
{
  screenline_t *cur;

  int min_x1 = +9999;
  int max_x2 = -9999;

  for (cur=empty_list; cur != NULL; cur=cur->next)
  {
    if (cur->x1 <= *x2 && cur->x2 >= *x1)
    {
      // some overlap
      min_x1 = MIN(min_x1, cur->x1);
      max_x2 = MAX(max_x2, cur->x2);
    }
  }
  
  if (max_x2 < 0)
    return true;

  *x1 = MAX(*x1, min_x1);
  *x2 = MIN(*x2, max_x2);

  return false;
}

//
// SplitScreenLine
//
// Split one screen line into two horizontal pieces.  The given
// screenline becomes the LHS, and we return the RHS.

static INLINE screenline_t *SplitScreenLine(screenline_t *s, int new_x1)
{
  screenline_t *tmp = R2_NewScreenLine();

  int offset = new_x1 - s->x1;
  
  tmp->x1 = new_x1;
  tmp->x2 = s->x2;

  s->x2 = new_x1-1;

  tmp->yt = s->yt + s->yt_step * offset;
  tmp->yb = s->yb + s->yb_step * offset;

  tmp->yt_step = s->yt_step;
  tmp->yb_step = s->yb_step;

  // link in newbie
  tmp->next = s->next;
  tmp->prev = s;

  if (s->next)
    s->next->prev = tmp;

  s->next = tmp;

  return tmp;
}

//
// IntersectAbove
//
// Adjust the given area such that it contains the mathematical
// intersection upwards from the given line.  Returns true if the
// resulting intersection is empty.

static INLINE boolean_t IntersectAbove(screenline_t *area,
    float_t y1, float_t y2, float_t y_step)
{
  int x_diff = area->x2 - area->x1;

  float_t a_yt1 = area->yt;
  float_t a_yb1 = area->yb;

  float_t a_yt2 = a_yt1 + area->yt_step * x_diff;
  float_t a_yb2 = a_yb1 + area->yb_step * x_diff;

  // completely covered ?
  if (y1 >= a_yb1 && y2 >= a_yb2)
    return false;

  // completely missed ?
  if (y1 <= a_yt1 && y2 <= a_yt2)
    return true;

  // handle extreme possibilites
  // (small cop out)
  if (y1 > a_yb1 || y2 > a_yb2)
    return false;

  // check whether we cross the top line
  if (y1 >= a_yt1 && y2 >= a_yt2)
  {
    area->yb = y1;
    area->yb_step = y_step;
    return false;
  }

  // Damn, crosses top line.  Resulting area will be a triangle.  Now
  // we must find the intersection point and shrink the area
  // horizontally as well.

  // LARGE COP OUT: for now, we just raise the area's bottom line up
  // as much as possible.
  
  y1 = MAX(y1, a_yt1);
  y2 = MAX(y2, a_yt2);

  area->yb = y1;

  if (x_diff > 0)
    area->yb_step = (y2 - y1) / x_diff;
  else
    area->yb_step = 0;

  return false;
}

//
// IntersectBelow
//
// Adjust the given area such that it contains the mathematical
// intersection downwards from the given line.  Returns true if the
// resulting intersection is empty.

static INLINE boolean_t IntersectBelow(screenline_t *area,
    float_t y1, float_t y2, float_t y_step)
{
  int x_diff = area->x2 - area->x1;

  float_t a_yt1 = area->yt;
  float_t a_yb1 = area->yb;

  float_t a_yt2 = a_yt1 + area->yt_step * x_diff;
  float_t a_yb2 = a_yb1 + area->yb_step * x_diff;

  // completely covered ?
  if (y1 <= a_yt1 && y2 <= a_yt2)
    return false;

  // completely missed ?
  if (y1 >= a_yb1 && y2 >= a_yb2)
    return true;

  // handle extreme possibilites
  // (small cop out)
  if (y1 < a_yt1 || y2 < a_yt2)
    return false;

  // check whether we cross the bottom line
  if (y1 <= a_yb1 && y2 <= a_yb2)
  {
    area->yt = y1;
    area->yt_step = y_step;
    return false;
  }

  // Damn, crosses bottom line.  Resulting area will be a triangle.
  // Now we must find the intersection point and shrink the area
  // horizontally as well.

  // LARGE COP OUT: for now, we just lower the area's top line down as
  // much as possible.
  
  y1 = MIN(y1, a_yb1);
  y2 = MIN(y2, a_yb2);

  area->yt = y1;

  if (x_diff > 0)
    area->yt_step = (y2 - y1) / x_diff;
  else
    area->yt_step = 0;

  return false;
}

//
// R2_RemoveEmptyArea
//
// Removes the empty area of the 1D occlusion buffer.

void R2_RemoveEmptyArea(screenline_t * empty)
{
  if (empty->next)
    empty->next->prev = empty->prev;

  if (empty->prev)
    empty->prev->next = empty->next;
  else
    empty_list = empty->next;
}

//
// R2_ClipEmptyHoriz
//
// Clips the empty area from the 1D occlusion buffer against the given
// x range (which must lie within the area).  This can split the empty
// into multiple horizontal pieces.  Returns the middle piece that
// corresponds to the given x range.

screenline_t * R2_ClipEmptyHoriz(screenline_t *empty, int x1, int x2)
{
  // split area to the right ?
  if (x2 < empty->x2)
  {
    SplitScreenLine(empty, x2+1);
  }

  // split area to the left ?
  if (x1 > empty->x1)
  {
    empty = SplitScreenLine(empty, x1);
  }

  return empty;
}

//
// R2_ClipSolidArea
//
// Clips the solid area against the empty area.  The empty area must
// completely cover the solid area horizontally.  Returns true if the
// solid area was totally clipped.

boolean_t R2_ClipSolidArea(screenline_t *solid, screenline_t *empty)
{
  int x1 = solid->x1;
  int x2 = solid->x2;

  // edge points
  float_t e_yt1, e_yt2, e_yb1, e_yb2;
  
  e_yt1 = empty->yt + (x1 - empty->x1) * empty->yt_step;
  e_yt2 = empty->yt + (x2 - empty->x1) * empty->yt_step;

  e_yb1 = empty->yb + (x1 - empty->x1) * empty->yb_step;
  e_yb2 = empty->yb + (x2 - empty->x1) * empty->yb_step;
  
  if (IntersectAbove(solid, e_yb1, e_yb2, empty->yb_step))
    return true;

  if (IntersectBelow(solid, e_yt1, e_yt2, empty->yt_step))
    return true;

  return false;
}

//
// R2_ClipEmptyArea
//
// Clips the empty area against the new empty area.
//
// Returns true if the empty area was totally clipped (in which case,
// it has been removed from the 1D occlusion buffer).

boolean_t R2_ClipEmptyArea(screenline_t *empty, screenline_t *new_e)
{
  int x1 = empty->x1;
  int x2 = empty->x2;

  // edge points
  float_t e_yt1, e_yt2, e_yb1, e_yb2;
  
  e_yt1 = new_e->yt + (x1 - new_e->x1) * new_e->yt_step;
  e_yt2 = new_e->yt + (x2 - new_e->x1) * new_e->yt_step;

  e_yb1 = new_e->yb + (x1 - new_e->x1) * new_e->yb_step;
  e_yb2 = new_e->yb + (x2 - new_e->x1) * new_e->yb_step;
  
  if (IntersectAbove(empty, e_yb1, e_yb2, new_e->yb_step))
    return true;

  if (IntersectBelow(empty, e_yt1, e_yt2, new_e->yt_step))
    return true;

  return false;
}

void R2_AddSkyTop(int x1, int x2, float_t y, float_t y_step)
{
  for (; x1 <= x2; x1++, y += y_step)
    sky_clip_top[x1] = MAX(y, sky_clip_top[x1]);
}

void R2_AddSkyBottom(int x1, int x2, float_t y, float_t y_step)
{
  for (; x1 <= x2; x1++, y += y_step)
    sky_clip_bottom[x1] = MIN(y, sky_clip_bottom[x1]);
}

