//----------------------------------------------------------------------------
//  EDGE BSP Handling Code
//----------------------------------------------------------------------------
// 
//  Copyright (c) 1999-2000  The EDGE Team.
// 
//  This program is free software; you can redistribute it and/or
//  modify it under the terms of the GNU General Public License
//  as published by the Free Software Foundation; either version 2
//  of the License, or (at your option) any later version.
//
//  This program is distributed in the hope that it will be useful,
//  but WITHOUT ANY WARRANTY; without even the implied warranty of
//  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//  GNU General Public License for more details.
//
//----------------------------------------------------------------------------
//
//  Based on the DOOM source code, released by Id Software under the
//  following copyright:
//
//    Copyright (C) 1993-1996 by id Software, Inc.
//
//----------------------------------------------------------------------------
//
// -KM- 1998/09/27 Special sector colourmap changing
//

#include "i_defs.h"
#include "r_bsp.h"

#include "dm_defs.h"
#include "dm_state.h"
#include "m_bbox.h"
#include "r_main.h"
#include "r_plane.h"
#include "r_segs.h"
#include "r_state.h"
#include "r_things.h"
#include "z_zone.h"

seg_t *curline;
side_t *sidedef;
line_t *linedef;
sector_t *frontsector;
sector_t *backsector;
region_properties_t *front_properties;

stack_array_t drawsegs_a;
drawseg_t **drawsegs = NULL;
int ds_p;

float_t extrafl_offset;

//
// R_ClearDrawSegs
//
void R_ClearDrawSegs(void)
{
  ds_p = 0;
}

//
// ClipWallSegment
// Clips the given range of columns
// and includes it in the new clip list.
//
typedef struct
{
  int first;
  int last;

}
cliprange_t;

// -ES- 98/12/19 stack_array for solidsegs
// newend is one past the last valid seg
int numsolidsegs;
stack_array_t solidsegs_a;
cliprange_t **solidsegs = NULL;

//
// ClipSolidWallSegment
//
// Does handle solid walls,
//  e.g. single sided LineDefs (middle texture)
//  that entirely block the view.
// 
static void ClipSolidWallSegment (int first, int last)
{
  int next;
  int start;
  int end;
  boolean_t skip;

  // -ES- 98/12/19 stack_array for solidsegs
  Z_SetArraySize(&solidsegs_a, numsolidsegs + 1);

  // Find the first range that touches the range
  //  (adjacent pixels are touching).
  start = 0;
  end = numsolidsegs;
  while (solidsegs[start]->last < first - 1)
    start++;

  if (first < solidsegs[start]->first)
  {
    if (last < solidsegs[start]->first - 1)
    {
      // Post is entirely visible (above start),
      //  so insert a new clippost.
      R_StoreWallRange(first, last);
      next = end;
      numsolidsegs++;

      while (next != start)
      {
        *solidsegs[next] = *solidsegs[next - 1];
        next--;
      }
      solidsegs[next]->first = first;
      solidsegs[next]->last = last;

      return;
    }
    else
    {
      // There is a fragment above *start.
      R_StoreWallRange(first, solidsegs[start]->first - 1);
      // Now adjust the clip size.
      solidsegs[start]->first = first;
    }
  }

  // Bottom contained in start?
  if (last <= solidsegs[start]->last)
    return;

  skip = false;
  next = start;
  while (last >= solidsegs[next + 1]->first - 1)
  {
    // There is a fragment between two posts.
    R_StoreWallRange(solidsegs[next]->last + 1, solidsegs[next + 1]->first - 1);
    next++;

    if (last <= solidsegs[next]->last)
    {
      // Bottom is contained in next.
      // Adjust the clip size.
      solidsegs[start]->last = solidsegs[next]->last;
      skip = true;
      break;
    }
  }

  if (!skip)
  {
    // There is a fragment after *next.
    R_StoreWallRange(solidsegs[next]->last + 1, last);
    // Adjust the clip size.
    solidsegs[start]->last = last;
  }

  // Remove start+1 to next from the clip list,
  // because start now covers their area.
  if (next == start)
  {
    // Post just extended past the bottom of one post.
    return;
  }

  while (next++ != end)
  {
    // Remove a post.
    *solidsegs[++start] = *solidsegs[next];
  }

  numsolidsegs = start + 1;
  Z_SetArraySize(&solidsegs_a, numsolidsegs);
}

//
// ClipPassWallSegment
//
// Clips the given range of columns,
//  but does not includes it in the clip list.
// Does handle windows,
//  e.g. LineDefs with upper and lower texture.
//
static void ClipPassWallSegment (int first, int last)
{
  int start;

  // Find the first range that touches the range
  //  (adjacent pixels are touching).
  start = 0;
  while (solidsegs[start]->last < first - 1)
    start++;

  if (first < solidsegs[start]->first)
  {
    if (last < solidsegs[start]->first - 1)
    {
      // Post is entirely visible (above start).
      R_StoreWallRange(first, last);
      return;
    }

    // There is a fragment above *start.
    R_StoreWallRange(first, solidsegs[start]->first - 1);
  }

  // Bottom contained in start?
  if (last <= solidsegs[start]->last)
    return;

  while (last >= solidsegs[start + 1]->first - 1)
  {
    // There is a fragment between two posts.
    R_StoreWallRange(solidsegs[start]->last + 1, solidsegs[start + 1]->first - 1);
    start++;

    if (last <= solidsegs[start]->last)
      return;
  }

  // There is a fragment after *next.
  R_StoreWallRange(solidsegs[start]->last + 1, last);
}

//
// R_ClearClipSegs
//
void R_ClearClipSegs(void)
{
  static boolean_t firsttime = true;

  if (firsttime)
  {
    firsttime = false;
    Z_InitStackArray(&solidsegs_a, (void ***)&solidsegs, sizeof(cliprange_t), 0);
  }
  
  Z_SetArraySize(&solidsegs_a, numsolidsegs = 2);

  solidsegs[0]->first = -0x7fffffff;
  solidsegs[0]->last = -1;
  solidsegs[1]->first = viewwidth;
  solidsegs[1]->last = 0x7fffffff;
}

//
// R_AddLine
// Clips the given segment
// and adds any visible pieces to the line list.
//
static void AddLine(seg_t * seg)
{
  int x1;
  int x2;
  angle_t angle1;
  angle_t angle2;
  angle_t span;
  angle_t tspan;

  curline = seg;

  // OPTIMISE: quickly reject orthogonal back sides.
  angle1 = R_PointToAngle(viewx, viewy, seg->x1, seg->y1);
  angle2 = R_PointToAngle(viewx, viewy, seg->x2, seg->y2);

  // Clip to view edges.

  // -ES- 1999/03/20 Replaced clipangle with clipscope/leftclipangle/rightclipangle

  span = angle1 - angle2;

  // Back side? I.e. backface culling?
  if (span >= ANG180)
    return;

  // Global angle needed by segcalc.
  rw_angle1 = angle1;
  angle1 -= viewangle;
  angle2 -= viewangle;

  tspan = angle1 - rightclipangle;
  if (tspan > clipscope)
  {
    tspan -= clipscope;

    // Totally off the left edge?
    if (tspan >= span)
      return;

    angle1 = leftclipangle;
  }
  tspan = leftclipangle - angle2;
  if (tspan > clipscope)
  {
    tspan -= clipscope;

    // Totally off the left edge?
    if (tspan >= span)
      return;
    angle2 = rightclipangle;
  }

  // The seg is in the view range,
  // but not necessarily visible.
  angle1 = (angle1 + ANG90) >> ANGLETOFINESHIFT;
  angle2 = (angle2 + ANG90) >> ANGLETOFINESHIFT;
  x1 = viewangletox[angle1];
  x2 = viewangletox[angle2];

  // Does not cross a pixel?
  if (x1 == x2)
    return;

  backsector = seg->backsector;

  // Single sided line?
  if ((!backsector)
  // Closed door.
      || backsector->ceil.h <= frontsector->floor.h
      || backsector->floor.h >= frontsector->ceil.h)
  {
    ClipSolidWallSegment(x1, x2 - 1);
    return;
  }

  // Window.
  if (backsector->ceil.h != frontsector->ceil.h
      || backsector->floor.h != frontsector->floor.h
  // Reject empty lines used for triggers
  //  and special events.
  // Identical floor and ceiling on both sides,
  // identical light levels on both sides,
  // and no middle texture.
      || backsector->ceil.pic != frontsector->ceil.pic
      || backsector->floor.pic != frontsector->floor.pic
      || backsector->p.lightlevel != frontsector->p.lightlevel
      || backsector->p.colourmap != frontsector->p.colourmap
      || frontsector->regions->sec_next != NULL
      || backsector->regions->sec_next != NULL
      || seg->sidedef->middle.tex != 0)
  {
    ClipPassWallSegment(x1, x2 - 1);
  }
}

//
// R_CheckBBox
//
// Checks BSP node/subtree bounding box.
// Returns true if some part of the bbox might be visible.

int checkcoord[12][4] =
{
    {3, 0, 2, 1},
    {3, 0, 2, 0},
    {3, 1, 2, 0},
    {0},
    {2, 0, 2, 1},
    {0, 0, 0, 0},
    {3, 1, 3, 0},
    {0},
    {2, 0, 3, 1},
    {2, 1, 3, 1},
    {2, 1, 3, 0}
};

static boolean_t R_CheckBBox(float_t * bspcoord)
{
  int boxx;
  int boxy;
  int boxpos;

  float_t x1;
  float_t y1;
  float_t x2;
  float_t y2;

  angle_t angle1;
  angle_t angle2;
  angle_t span;
  angle_t tspan;

  int start;

  int sx1;
  int sx2;

  // Find the corners of the box
  // that define the edges from current viewpoint.
  if (viewx <= bspcoord[BOXLEFT])
    boxx = 0;
  else if (viewx < bspcoord[BOXRIGHT])
    boxx = 1;
  else
    boxx = 2;

  if (viewy >= bspcoord[BOXTOP])
    boxy = 0;
  else if (viewy > bspcoord[BOXBOTTOM])
    boxy = 1;
  else
    boxy = 2;

  boxpos = (boxy << 2) + boxx;
  if (boxpos == 5)
    return true;

  x1 = bspcoord[checkcoord[boxpos][0]];
  y1 = bspcoord[checkcoord[boxpos][1]];
  x2 = bspcoord[checkcoord[boxpos][2]];
  y2 = bspcoord[checkcoord[boxpos][3]];

  // check clip list for an open space
  angle1 = R_PointToAngle(viewx, viewy, x1, y1) - viewangle;
  angle2 = R_PointToAngle(viewx, viewy, x2, y2) - viewangle;

  span = angle1 - angle2;

  // Sitting on a line?
  if (span >= ANG180)
    return true;

  // -ES- 1999/03/20 Replaced clipangle with clipscope/leftclipangle/rightclipangle

  tspan = angle1 - rightclipangle;

  if (tspan > clipscope)
  {
    tspan -= clipscope;

    // Totally off the left edge?
    if (tspan >= span)
      return false;

    angle1 = leftclipangle;
  }
  tspan = leftclipangle - angle2;
  if (tspan > clipscope)
  {
    tspan -= clipscope;

    // Totally off the left edge?
    if (tspan >= span)
      return false;

    angle2 = rightclipangle;
  }

  // Find the first clippost that touches the source post
  //  (adjacent pixels are touching).

  angle1 = (angle1 + ANG90) >> ANGLETOFINESHIFT;
  angle2 = (angle2 + ANG90) >> ANGLETOFINESHIFT;
  sx1 = viewangletox[angle1];
  sx2 = viewangletox[angle2];

  // Does not cross a pixel.
  if (sx1 == sx2)
    return false;
  sx2--;

  start = 0;
  while (solidsegs[start]->last < sx2)
    start++;

  if (sx1 >= solidsegs[start]->first
      && sx2 <= solidsegs[start]->last)
  {
    // The clippost contains the new span.
    return false;
  }

  return true;
}

//
// Subsector
//
// Determine floor/ceiling planes.
// Add sprites of things in sector.
// Draw one or more line segments.

static void Subsector(int num)
{
  seg_t *seg;
  subsector_t *sub, *spr_sub;
  mobj_t *thing;

#ifdef DEVELOPERS
  if (num >= numsubsectors)
    I_Error("Subsector: ss %i with numss = %i",
        num, numsubsectors);
#endif

  sscount++;
  sub = &subsectors[num];

  frontsector = sub->sector;
  front_properties = R_PointInVertRegion(sub, viewz)->p;

  // -AJA- 1999/07/25: Need this for moving deep water.
  extrafl_offset = sub->sector->ceil.h - frontsector->ceil.h;

  floorplane = ceilingplane = NULL;

  if (frontsector->floor.h < viewz)
  {
    floorplane = R_FindPlane(&frontsector->floor, front_properties,
        VPF_FLOOR);
  }

  if (frontsector->ceil.h > viewz ||
      frontsector->ceil.pic == skyflatnum)
  {
    ceilingplane = R_FindPlane(&frontsector->ceil, front_properties,
        VPF_CEILING);
  }

  // -AJA- 1999/08/16: Support for rendering extra floors.
  extraplane = NULL;

  // -AJA- FIXME: very hacked together
  if (frontsector->regions->sec_next &&
      frontsector->regions->extrafloor)
  {
    if (frontsector->regions->sec_next->floor->h < viewz)
    {
      extraheight = frontsector->regions->sec_next->floor->h;
      extraplane = R_FindPlane(frontsector->regions->sec_next->floor,
          front_properties, VPF_FLOOR | VPF_EXTRA);
    }
    else if (frontsector->regions->ceil->h > viewz)
    {
      extraheight = frontsector->regions->ceil->h;
      extraplane = R_FindPlane(frontsector->regions->ceil,
          front_properties, VPF_CEILING | VPF_EXTRA);
    }
  }

  // Add sprites.  These MUST be done on a per-sector basis, since the
  // normal node information has bboxes which do not cover the whole
  // subsector, which means that a subsector that is actually visible
  // can be skipped by the R_CheckBBox() test in the BSP traversal.

  if (sub->sector->validcount != validcount)
  {
    sub->sector->validcount = validcount;

    for (spr_sub = sub->sector->subsectors; spr_sub; 
         spr_sub = spr_sub->sec_next)
    {
      for (thing = spr_sub->thinglist; thing; thing = thing->snext)
        R_AddSprite(thing);
    }
  }

  for (seg = sub->segs; seg != NULL; seg = seg->sub_next)
  {
    if (! seg->miniseg)
      AddLine(seg);
  }
}

//
// RenderBSPNode
//
// Renders all subsectors below a given node,
//  traversing subtree recursively.
// Just call with BSP root.

void R_RenderBSPNode(int bspnum)
{
  node_t *node;
  int side;

  // Found a subsector?
  if (bspnum & NF_SUBSECTOR)
  {
    if (bspnum == -1)
      Subsector(0);
    else
      Subsector(bspnum & (~NF_SUBSECTOR));
    return;
  }

  node = &nodes[bspnum];

  // Decide which side the view point is on.
  side = R_PointOnSide(viewx, viewy, node);

  // Recursively divide front space.
  // -AJA- NOTE: we don't check the bbox for the front space, the
  //       normal node info is not reliable enough.
  R_RenderBSPNode(node->children[side]);

  // Recursively divide back space.
  if (R_CheckBBox(node->bbox[side ^ 1]))
    R_RenderBSPNode(node->children[side ^ 1]);
}
