//----------------------------------------------------------------------------
//  EDGE Rendering Data Handling Code
//----------------------------------------------------------------------------
// 
//  Copyright (c) 1999-2000  The EDGE Team.
// 
//  This program is free software; you can redistribute it and/or
//  modify it under the terms of the GNU General Public License
//  as published by the Free Software Foundation; either version 2
//  of the License, or (at your option) any later version.
//
//  This program is distributed in the hope that it will be useful,
//  but WITHOUT ANY WARRANTY; without even the implied warranty of
//  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//  GNU General Public License for more details.
//
//----------------------------------------------------------------------------
//
//  Based on the DOOM source code, released by Id Software under the
//  following copyright:
//
//    Copyright (C) 1993-1996 by id Software, Inc.
//
//----------------------------------------------------------------------------
//
// -ACB- 1998/09/09 Reformatted File Layout.
// -KM- 1998/09/27 Colourmaps can be dynamically changed.
// -ES- 2000/02/12 Moved most of this module to w_texture.c.

#include "i_defs.h"
#include "r_data.h"

#include "e_search.h"
#include "dm_state.h"
#include "dm_defs.h"
#include "m_swap.h"
#include "p_local.h"
#include "r_sky.h"
#include "w_wad.h"
#include "w_textur.h"
#include "z_zone.h"

// flats list
static int numflats;
int *flatlist = NULL;
int *wad2flatlist = NULL;

// sprite lists
int numspritelumps;
int *spritelist = NULL;

// for global animation
int *flattranslation = NULL;

// needed for pre rendering
float_t *spritewidth;
float_t *spriteheight;
float_t *spriteoffset;
float_t *spritetopoffset;

//
// InitFlats
//
static void InitFlats(void)
{
  int i;
  int pos;

  if (!flatlist)
  {
    flatlist = W_GetList('F', &numflats);

#define CMP(a, b) (strnicmp(lumpinfo[a].name, lumpinfo[b].name, 8) < 0)
    QSORT(int, flatlist, numflats, CUTOFF);
#undef CMP

    wad2flatlist = Z_New(int, numlumps);

    for (i = 0; i < numlumps; i++)
    {
#define CMP(a) (strnicmp(lumpinfo[flatlist[a]].name, lumpinfo[i].name, 8) < 0)
      BSEARCH(numflats, pos);
#undef CMP
      if (pos >= 0 && pos < numflats && !strnicmp(lumpinfo[flatlist[pos]].name, lumpinfo[i].name, 8))
        wad2flatlist[i] = pos;
      else
        wad2flatlist[i] = -1;
    }
  }
  // Create translation table for global animation.
  Z_Resize(flattranslation, int, numflats + 1);

  for (i = 0; i < numflats; i++)
    flattranslation[i] = i;
}

//
// InitSpriteLumps
//
// Finds the width and hoffset of all sprites in the wad, so the sprite does
// not need to be cached completely just for having the header info ready
// during rendering.
//
static void InitSpriteLumps(void)
{
  int i;
  const patch_t *patch;

  if (!spritelist)
    spritelist = W_GetList('S', &numspritelumps);

  // -AJA- FIXME: clean this mess up
  spritewidth = Z_New(float_t, numspritelumps);
  spriteheight = Z_New(float_t, numspritelumps);
  spriteoffset = Z_New(float_t, numspritelumps);
  spritetopoffset = Z_New(float_t, numspritelumps);

  for (i = 0; i < numspritelumps; i++)
  {
    if (!(i & 63))
      I_Printf(".");

    patch = W_CacheLumpNum(spritelist[i]);

    spritewidth[i] = SHORT(patch->width);
    spriteheight[i] = SHORT(patch->height);
    spriteoffset[i] = SHORT(patch->leftoffset);
    spritetopoffset[i] = SHORT(patch->topoffset);

    W_DoneWithLump(patch);
  }
}

//
// R_InitData
//
// Locates all the lumps that will be used by all views
//
// Must be called after W_Init.
//
void R_InitData(void)
{
  InitFlats();
  InitSpriteLumps();
}

//
// R_FlatNumForName
// Retrieval, get a flat number for a flat name.
//
int R_FlatNumForName(const char *name)
{
  int i;

#define CMP(a) (strnicmp(lumpinfo[flatlist[a]].name, name, 8) < 0)
  BSEARCH(numflats, i);
#undef CMP

  if (i >= 0 && i < numflats && strnicmp(name, lumpinfo[flatlist[i]].name, 8))
  {
    // 25-6-98 KM We have a backup plan...
    int j, l = W_CheckNumForName(name);

    if (l == -1)
      I_Error("R_FlatNumForName: \'%.8s\' not found", name);
    
    for (j = i; j < numflats; j++)
      wad2flatlist[flatlist[j]]++;

    Z_Resize(flatlist, int, numflats + 1);
    Z_Resize(flattranslation, int, numflats + 2);
    memmove(&flatlist[i + 1], &flatlist[i], (numflats - i) * sizeof(int));
    memmove(&flattranslation[i + 1], &flattranslation[i], (numflats - i) * sizeof(int));

    flatlist[i] = j;
    flattranslation[i] = j;
    wad2flatlist[l] = i;
    P_ReInitPicAnims(i);
    numflats++;
  }

  return i;
}

//
// R_PrecacheLevel
// Preloads all relevant graphics for the level.
//
void R_PrecacheLevel(void)
{
  char *flatpresent;
  char *texturepresent;
  char *spritepresent;
  int i;
  int j;
  int k;
  int numtextures;
  int lump;

  mobj_t *currmobj;

  spriteframe_t *sf;

  if (demoplayback)
    return;

  // Precache flats.
  flatpresent = Z_ClearNew(char, numflats);

  for (i = 0; i < numsectors; i++)
  {
    flatpresent[sectors[i].floor.pic] = 1;
    flatpresent[sectors[i].ceil.pic] = 1;
  }

  for (i = 0; i < numflats; i++)
  {
    if (flatpresent[i])
    {
      lump = flatlist[i];
      W_PreCacheLumpNum(lump);
    }
  }

  // Precache textures.
  texturepresent = NULL;
  numtextures = 0;

  // expands the texture array if needed, and marks the texture as used.
  // all used textures will be precached.
#define CHECK(num)\
if (num >= numtextures)\
{\
  Z_Resize(texturepresent, char, num + 1);\
  for (j = numtextures; j < num; j++)\
    texturepresent[j] = 0;\
  numtextures = num + 1;\
}\
texturepresent[num] = 1;
  for (i = 0; i < numsides; i++)
  {
    CHECK(sides[i].top.tex)
    CHECK(sides[i].middle.tex)
    CHECK(sides[i].bottom.tex)
  }
#undef CHECK

  //
  // Sky texture is always present.
  //
  // Note that F_SKY1 is the name used to indicate a sky floor/ceiling as a flat,
  // while the sky texture is stored like a wall texture, with a level depended name;
  // dictated to by the currentmap structure.
  //
  texturepresent[skytexture] = 1;

  for (i = 0; i < numtextures; i++)
  {
    if (!texturepresent[i])
      continue;

    W_PreCacheTextureNum(i);
  }

  // Precache sprites.
  spritepresent = Z_ClearNew(char, numsprites);

  for (currmobj = mobjlisthead; currmobj != NULL; currmobj = currmobj->next)
    spritepresent[currmobj->sprite] = 1;

  for (i = 0; i < numsprites; i++)
  {
    if (!spritepresent[i])
      continue;

    for (j = 0; j < sprites[i].numframes; j++)
    {
      sf = &sprites[i].spriteframes[j];

      for (k = 0; k < 8; k++)
      {
        lump = spritelist[sf->lump[k]];
        W_PreCacheLumpNum(lump);
      }

    }

  }
  
  Z_Free(flatpresent);
  Z_Free(texturepresent);
  Z_Free(spritepresent);
}
