//----------------------------------------------------------------------------
//  EDGE Floor and Ceiling Rendering Code
//----------------------------------------------------------------------------
// 
//  Copyright (c) 1999-2000  The EDGE Team.
// 
//  This program is free software; you can redistribute it and/or
//  modify it under the terms of the GNU General Public License
//  as published by the Free Software Foundation; either version 2
//  of the License, or (at your option) any later version.
//
//  This program is distributed in the hope that it will be useful,
//  but WITHOUT ANY WARRANTY; without even the implied warranty of
//  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//  GNU General Public License for more details.
//
//----------------------------------------------------------------------------
//
//  Based on the DOOM source code, released by Id Software under the
//  following copyright:
//
//    Copyright (C) 1993-1996 by id Software, Inc.
//
//----------------------------------------------------------------------------
//
// -ACB- 1998/07/13 - Changed RANGECHECK for DEVELOPERS
// -KM- 1998/09/27 Dynamic colourmaps.
// -AJA- 1999/07/12: Updated for colmap.ddf.
// -AJA- 1999/07/16: Added xoffset & yoffset (for scrolling flats).
// -AJA- 1999/08/16: Support for rendering extra floors.
//

#include "i_defs.h"
#include "r_plane.h"

#include "dm_defs.h"
#include "dm_state.h"
#include "m_inline.h"
#include "r_local.h"
#include "r_sky.h"
#include "r_state.h"
#include "v_colour.h"
#include "w_textur.h"
#include "w_wad.h"
#include "z_zone.h"

//
// opening
//

// Here comes the obnoxious "visplane".
static visplane_t **visplanes;
static stack_array_t visplane_a;
static int numvisplanes;

visplane_t *floorplane;
visplane_t *ceilingplane;

// -AJA- 1999/08/16.
visplane_t *extraplane;

float_t extraheight;

// Openings remember the floor/ceiling clipping values.  They are
// refered to by drawsegs, and are used to clip sprites.
stack_array_t openings_a;
short *openings = NULL;
int lastopening = 0;

//
// Clip values are the solid pixel bounding the range.
//  floorclip starts out SCREENHEIGHT
//  ceilingclip starts out -1
//
short *floorclip = NULL;
short *ceilingclip = NULL;

short *extra_F_clip = NULL;  // -AJA- 1999/08/16.

short *extra_C_clip = NULL;  // -AJA- 1999/08/16.

//
// spanstart holds the start of a plane span
// initialised to 0 at start
//
int *spanstart = NULL;

//
// texture mapping
//
static const lighttable_t *planezlight;
static const colourmap_t *planenominal;
static boolean_t planesky;
static float_t planeheight;
static fixed_t plane_xoffset;
static fixed_t plane_yoffset;
static fixed_t plane_transp;

// 1999/03/20 Removed the origyslope hack
float_t *yslope = NULL;

float_t *distscale = NULL;
float_t basexscale;
float_t baseyscale;

// -ES- 1999/03/23 Changed to cache struct
typedef struct planecache_s
{
  float_t height;
  float_t distance;
  fixed_t xstep;
  fixed_t ystep;
} planecache_t;
planecache_t *planecache = NULL;

//
// R_InitPlanes
// Called at resolution change.
//
// 98-7-10 KM Prepare for dynamic screen sizing
void R_InitPlanes(void)
{
  static boolean_t firsttime = true;

  if (firsttime)
  {
    firsttime = false;
    Z_InitStackArray(&drawsegs_a, (void ***)&drawsegs, sizeof(drawseg_t), 0);
    Z_InitStackArray(&openings_a, (void ***)&openings, sizeof(short), -16);
  }
  else
    Z_ClearStackArray(&visplane_a);

  Z_InitStackArray(&visplane_a, (void ***)&visplanes, sizeof(visplane_t) +
      (SCREENWIDTH + 2) * 2 * sizeof(unsigned short), 0);
  
  // -ES- 1999/03/23 Structified this caching system
  Z_Resize(planecache, planecache_t, SCREENHEIGHT);
  Z_Clear(planecache, planecache_t, SCREENHEIGHT);
}

//
// R_MapPlane
//
// Uses global vars:
//  planeheight
//  ds_source
//  basexscale
//  baseyscale
//  viewx
//  viewy
//
// BASIC PRIMITIVE
// 

void R_MapPlane(int y, int x1, int x2)
{
  angle_t angle;
  float_t distance;
  float_t length;
  unsigned index;

#ifdef DEVELOPERS
  if (x2 < x1 || x1 < 0 || x2 >= viewwidth || y < 0 || y > viewheight)
  {
    I_Error("R_MapPlane: %i, %i at %i", x1, x2, y);
  }
#endif

  if (planeheight != planecache[y].height)
  {
    planecache[y].height = planeheight;
    distance = planecache[y].distance = planeheight * yslope[y];
    ds_xstep = planecache[y].xstep = M_FloatToFixed(distance * basexscale);
    ds_ystep = planecache[y].ystep = M_FloatToFixed(distance * baseyscale);
  }
  else
  {
    distance = planecache[y].distance;
    ds_xstep = planecache[y].xstep;
    ds_ystep = planecache[y].ystep;
  }

  length = distance * distscale[x1];
  angle = viewangle + xtoviewangle[x1];

  // -AJA- 1999/07/16: Add offsets (thanks Lee).
  ds_xfrac = M_FloatToFixed( viewx + M_Cos(angle) * length) + plane_xoffset;
  ds_yfrac = M_FloatToFixed(-viewy - M_Sin(angle) * length) + plane_yoffset;

  // -AJA- 1999/07/10: new stuff for colmap.ddf.
  index = distance / (1 << (LIGHTZSHIFT - FRACBITS));

  if (index >= MAXLIGHTZ)
    index = MAXLIGHTZ - 1;

  ds_colourmap = V_GetColtable(planenominal, planezlight[index],
      planesky ? VCOL_Sky : 0);
  ds_y = y;
  ds_x1 = x1;
  ds_x2 = x2;

  if (plane_transp >= FRACUNIT)
  {
    // high or low detail
    spanfunc();
  }
  else
  {
    dc_translucency = plane_transp;
    trans_spanfunc();
  }
}

//
// R_ClearPlanes
// At begining of frame.
//
void R_ClearPlanes(void)
{
  int i;
  angle_t angle;

  // opening / clipping determination
  for (i = 0; i < viewwidth; i++)
  {
    floorclip[i] = viewheight;
    ceilingclip[i] = -1;

    extra_F_clip[i] = viewheight;
    extra_C_clip[i] = -1;
  }

  Z_SetArraySize(&visplane_a, numvisplanes = 0);
  Z_SetArraySize(&openings_a, lastopening = 2 * viewwidth);
  // these two conditions are only true if the array needs updating.
  if (negonearray != viewwidth || openings[0] != viewheight)
  {
    negonearray = viewwidth;
    for (i = 0; i < viewwidth; i++)
    {
      openings[i] = viewheight;
      openings[i + viewwidth] = -1;
    }
  }

  // texture calculation
  Z_Clear(planecache, planecache_t, viewheight);

  // left to right mapping
  angle = viewangle - ANG90;

  // -ES- 1999/03/07 Fixed X FOV
  // scale will be unit scale at distance x_distunit
  basexscale = M_Cos(angle) / x_distunit;
  baseyscale = -M_Sin(angle) / x_distunit;
}

//
// Allocate & return a new visplane.
//
// -AJA- 1999/09/20: written.
//
static visplane_t *NewVisplane(void)
{
  unsigned short *memory;
  visplane_t *vp;

  Z_SetArraySize(&visplane_a, ++numvisplanes);
  vp = visplanes[numvisplanes - 1];

  // The short arrays for top/bottom are included in the element size of the
  // visplane stack_array, so we do not have to allocate any extra memory.
  memory = (unsigned short *)(vp + 1);
  Z_Clear(memory, unsigned short, (SCREENWIDTH + 2) * 2);

  vp->top = memory + 1;
  vp->bottom = memory + SCREENWIDTH + 3;

  return vp;
}

//
// R_FindPlane
//
// 23-6-98 KM Visplanes fixed
// -AJA- 1999/07/09: Altered for colmap.ddf.
// -AJA- 1999/08/16: Added parameters for extra floor support.

visplane_t *R_FindPlane(plane_info_t *info, 
    region_properties_t *props, int flags)
{
  int check;
  visplane_t *pl;

  for (check=0; check < numvisplanes; check++)
  {
    pl = visplanes[check];

    // -KM- 1998/07/31 Little Optimisation here...
    if ((info->h == pl->height) &&
        (info->pic == pl->picnum) &&
        (info->x_offset == pl->xoffset) &&
        (info->y_offset == pl->yoffset) &&
        (info->translucency == pl->transp) &&
        (props->lightlevel == pl->lightlevel) &&
        (props->colourmap == pl->colourmap) &&
        (flags == pl->flags))
    {
      return pl;
    }
  }

  pl = NewVisplane();

  pl->height = info->h;
  pl->picnum = info->pic;
  pl->xoffset = info->x_offset;
  pl->yoffset = info->y_offset;
  pl->transp = info->translucency;
  pl->lightlevel = props->lightlevel;
  pl->colourmap = props->colourmap;
  pl->flags = flags;

  pl->minx = SCREENWIDTH;
  pl->maxx = -1;

  // -KM- 1998/07/31 Fix here... 0xff -> -1
  memset(pl->top, -1, SCREENWIDTH * sizeof(unsigned short));

  return pl;
}

//
// R_CheckPlane
//
// 23-6-98 KM Visplanes fixed
//
// -AJA- 1999/08/17: Added a check which means we _always_ generate a
//       brand new visplane for an extra floor/ceiling, thus we get a
//       on-screen strip that is the same width as the drawseg.
//

visplane_t *R_CheckPlane(visplane_t * pl, int start, int stop)
{
  int intrl;
  int intrh;
  int unionl;
  int unionh;
  int x;

  visplane_t *new_pl;

  if (start < pl->minx)
  {
    intrl = pl->minx;
    unionl = start;
  }
  else
  {
    unionl = pl->minx;
    intrl = start;
  }

  if (stop > pl->maxx)
  {
    intrh = pl->maxx;
    unionh = stop;
  }
  else
  {
    unionh = pl->maxx;
    intrh = stop;
  }

  for (x = intrl; x <= intrh; x++)
    if (pl->top[x] != 0xffff)
      break;

  if (x > intrh && !(pl->flags & VPF_EXTRA))
  {
    pl->minx = unionl;
    pl->maxx = unionh;

    // use the same one
    return pl;
  }

  // make a new visplane
  new_pl = NewVisplane();

  new_pl->height = pl->height;
  new_pl->picnum = pl->picnum;
  new_pl->lightlevel = pl->lightlevel;
  new_pl->colourmap = pl->colourmap;
  new_pl->xoffset = pl->xoffset;
  new_pl->yoffset = pl->yoffset;
  new_pl->transp = pl->transp;
  new_pl->flags = pl->flags;

  new_pl->minx = start;
  new_pl->maxx = stop;

  memset(new_pl->top, -1, SCREENWIDTH * sizeof(unsigned short));

  return new_pl;
}

//
// R_MakeSpans
//
void R_MakeSpans(int x, int t1, int b1, int t2, int b2)
{

#ifdef DEVELOPERS
  // This is will crash when passed to R_MapPlane when using big Heapsizes
  if (b1 == 65535 && x == 0)
    I_Error("R_MakeSpans: Andy's MapPlane Error \n");
#endif

  while (t1 < t2 && t1 <= b1)
  {
#ifdef DEVELOPERS
  if (t1 >= viewwindowheight)
    I_Error("R_MakeSpans: t1 = %d\n", t1);
#endif
    R_MapPlane(t1, spanstart[t1], x - 1);
    t1++;
  }
  while (b1 > b2 && b1 >= t1)
  {
#ifdef DEVELOPERS
  if (b1 >= viewwindowheight)
    I_Error("R_MakeSpans: b1 = %d\n", b1);
#endif
    R_MapPlane(b1, spanstart[b1], x - 1);
    b1--;
  }

  while (t2 < t1 && t2 <= b2)
  {
#ifdef DEVELOPERS
  if (t2 >= viewwindowheight)
    I_Error("R_MakeSpans: t2 = %d\n", t2);
#endif
    spanstart[t2] = x;
    t2++;
  }
  while (b2 > b1 && b2 >= t2)
  {
#ifdef DEVELOPERS
  if (b2 >= viewwindowheight)
    I_Error("R_MakeSpans: b2 = %d\n", b2);
#endif
    spanstart[b2] = x;
    b2--;
  }
}

//
// R_DrawSinglePlane
//
// -AJA- 1999/08/17: Moved this code out of R_DrawPlanes, for extra
//       floor support.

void R_DrawSinglePlane(visplane_t * pl)
{
  int light;
  int x;
  int stop;
  int angle;
  const cached_tex_t *tex;

  if (pl->minx > pl->maxx)
    return;

  light = (pl->lightlevel >> LIGHTSEGSHIFT);

  if (light >= LIGHTLEVELS)
    light = LIGHTLEVELS - 1;

  if (light < 0)
    light = 0;

  planezlight = zlight[light];
  planenominal = pl->colourmap;
  planesky = (pl->picnum == skyflatnum);
  
  // sky flat
  if (planesky)
  {
    tex = W_CacheTextureNum(texturetranslation[skytexture]);
    dc_iscale = M_FloatToFixed(skytexturescale);

    // -AJA- Damn. Drawing the sky vertically means that we can't
    // apply a FOG colourmap to it with correct results.

    dc_colourmap = V_GetColtable(pl->colourmap, 255, VCOL_Sky);

    dc_texturemid = M_FloatToFixed(skytexturemid);
    for (x = pl->minx; x <= pl->maxx; x++)
    {
      dc_yl = pl->top[x];
      dc_yh = pl->bottom[x];

      if (dc_yl <= dc_yh)
      {
#ifndef NOSMOOTHING
        dc_xfrac = ((viewangle + xtoviewangle[x]) >> 6) & 0xffff;
#endif
        angle = (viewangle + xtoviewangle[x]) >> ANGLETOSKYSHIFT;
        dc_x = x;
        // -KM- 1998/07/31 Give skytextures the possibility of being
        //  animated...
        dc_source = W_GetColumn(angle, tex);
#ifndef NOSMOOTHING
        dc_source2 = W_GetColumn(angle + 1, tex);
#endif
// -AJA- try to find DrawColumn bug:
#ifdef DEVELOPERS
        if (dc_yh > 0x70000000)
          I_Error("R_DrawSinglePlane: Bad Sky: yl=%d yh=%d\n", dc_yl, dc_yh);
#endif
        colfunc();
      }
    }

    W_DoneWithTexture(tex);
  }
  else
  {
    // regular flat
    ds_source = W_CacheLumpNum(flatlist[flattranslation[pl->picnum]]);

    planeheight = fabs(pl->height - viewz);
    plane_xoffset = M_FloatToFixed(pl->xoffset);
    plane_yoffset = M_FloatToFixed(pl->yoffset);
    plane_transp = M_FloatToFixed(pl->transp);

    if (pl->maxx > SCREENWIDTH || pl->maxx < 0 ||
        pl->minx > SCREENWIDTH || pl->minx < 0)
    {
      I_Error("INTERNAL PLANE ERROR: minx=%d maxx=%d\n",
          pl->minx, pl->maxx);

    }

    pl->top[pl->maxx + 1] = 0xffff;
    pl->top[pl->minx - 1] = 0xffff;

    stop = pl->maxx + 1;

    for (x = pl->minx; x <= stop; x++)
    {
      R_MakeSpans(x, pl->top[x - 1], pl->bottom[x - 1],
          pl->top[x], pl->bottom[x]);
    }

    W_DoneWithLump(ds_source);
  }
}

//
// R_DrawPlanes
// At the end of each frame.
//

void R_DrawPlanes(void)
{
  visplane_t *pl;
  int i;

  for (i=0; i < numvisplanes; i++)
  {
    pl = visplanes[i];

    if (pl->minx > pl->maxx)
      continue;

    // -AJA- Skip extra floors/ceilings. They are done later.
    if (pl->flags & VPF_EXTRA)
      continue;

    R_DrawSinglePlane(pl);
  }
}
