//----------------------------------------------------------------------------
//  EDGE Rendering things (objects as sprites) Code
//----------------------------------------------------------------------------
// 
//  Copyright (c) 1999-2000  The EDGE Team.
// 
//  This program is free software; you can redistribute it and/or
//  modify it under the terms of the GNU General Public License
//  as published by the Free Software Foundation; either version 2
//  of the License, or (at your option) any later version.
//
//  This program is distributed in the hope that it will be useful,
//  but WITHOUT ANY WARRANTY; without even the implied warranty of
//  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//  GNU General Public License for more details.
//
//----------------------------------------------------------------------------
//
//  Based on the DOOM source code, released by Id Software under the
//  following copyright:
//
//    Copyright (C) 1993-1996 by id Software, Inc.
//
//----------------------------------------------------------------------------
//
// -KM- 1998/07/26 Replaced #ifdef RANGECHECK with #ifdef DEVELOPERS
// -KM- 1998/09/27 Dynamic colourmaps
// -AJA- 1999/07/12: Now uses colmap.ddf.
//

#include "i_defs.h"
#include "r_things.h"

#include "e_search.h"
#include "dm_defs.h"
#include "dm_state.h"
#include "m_argv.h"
#include "m_inline.h"
#include "m_swap.h"
#include "r_local.h"
#include "v_colour.h"
#include "w_wad.h"
#include "z_zone.h"

// The minimum distance between player and a visible sprite.
// FIXME: Decrease value, lower values are valid when float_t is used.
#define MINZ        (4.0)

typedef struct
{
  int x1;
  int x2;

  int column;
  int topclip;
  int bottomclip;

}
maskdraw_t;

//
// Sprite rotation 0 is facing the viewer,
//  rotation 1 is one angle turn CLOCKWISE around the axis.
// This is not the same as the angle,
//  which increases counter clockwise (protractor).
// There was a lot of stuff grabbed wrong, so I changed it...
//
float_t pspritescale;
float_t pspriteiscale;
float_t pspritescale2;
float_t pspriteiscale2;
float_t masked_translucency;

lighttable_t *spritezlights;

// constant arrays
//  used for psprite clipping and initializing clipping
// -ES- 1998/08/20 init to 0
int negonearray = 0;

const coltable_t *fuzz_coltable;

//
// INITIALIZATION FUNCTIONS
//

// variables used to look up
//  and range check thing_t sprites patches
spritedef_t *sprites;
int numsprites;

spriteframe_t sprtemp[29];
int maxframe;
char *spritename;

//
// Define MHFX_LAXSPRITEROTATIONS for a temporary fix to the problem
// of PWADS that define sprite rotations in different orders. This
// should be regarded as a temporary fix only and might not work
// properly in all cases. See README.sprites for details and
// ESPECIALLY before using "-sprnolim" - you have been warned!
//
// Basically, this works by removing the checks on two WAD entries that
// map to the same sprite, having both a "no rotations" version of a
// sprite and various rotations of a sprite, and so forth. I cannot, so
// far, see any reason for Doom to check for this internally, as it seems
// to use the most recently loaded version of a sprite (which is, after all,
// what we want), so I assume that it is designed to protect against a
// corrupted IWAD or a buggy PWAD, or something like that.
//
#define MHFX_LAXSPRITEROTATIONS

#ifdef MHFX_LAXSPRITEROTATIONS
int lax_sprite_rotations = 0;

#endif

static boolean_t extra_planes;

//
// InstallSpriteLump
// Local function for R_InitSprites.
//
static void InstallSpriteLump(int lump, unsigned frame, unsigned rotation, boolean_t flipped)
{
  int r;

  if (frame >= 29 || rotation > 8)
    I_Error("InstallSpriteLump: "
        "Bad frame characters in lump %i", lump);

  if ((int)frame > maxframe)
    maxframe = frame;

  if (rotation == 0)
  {
    // the lump should be used for all rotations
    // false=0, true=1, but array initialised to -1
    // allows doom to have a "no value set yet" boolean_t value!

    if ((sprtemp[frame].rotate == false)
#ifdef MHFX_LAXSPRITEROTATIONS
        && (!lax_sprite_rotations)
#endif
        )
      I_Error("R_InitSprites: Sprite %s frame %c has "
          "multip rot=0 lump", spritename, 'A' + frame);

    if ((sprtemp[frame].rotate == true)
#ifdef MHFX_LAXSPRITEROTATIONS
        && (!lax_sprite_rotations)
#endif
        )
      I_Error("R_InitSprites: Sprite %s frame %c has rotations "
          "and a rot=0 lump", spritename, 'A' + frame);

    sprtemp[frame].rotate = false;
    for (r = 0; r < 8; r++)
    {
      sprtemp[frame].lump[r] = lump;
      sprtemp[frame].flip[r] = (byte) flipped;
    }
    return;
  }

  // the lump is only used for one rotation
  if ((sprtemp[frame].rotate == false)
#ifdef MHFX_LAXSPRITEROTATIONS
      && (!lax_sprite_rotations)
#endif
      )
    I_Error("R_InitSprites: Sprite %s frame %c has rotations "
        "and a rot=0 lump", spritename, 'A' + frame);

  sprtemp[frame].rotate = true;

  // make 0 based
  rotation--;
  if ((sprtemp[frame].lump[rotation] != -1)
#ifdef MHFX_LAXSPRITEROTATIONS
      && (!lax_sprite_rotations)
#endif
      )
    I_Error("R_InitSprites: Sprite %s : %c : %c "
        "has two lumps mapped to it",
        spritename, 'A' + frame, '1' + rotation);

  sprtemp[frame].lump[rotation] = lump;
  sprtemp[frame].flip[rotation] = (byte) flipped;
}

//
// R_InitSpriteDefs
// Pass a null terminated list of sprite names
//  (4 chars exactly) to be used.
// Builds the sprite rotation matrixes to account
//  for horizontally flipped sprites.
// Will report an error if the lumps are inconsistant. 
// Only called at startup.
//
// Sprite lump names are 4 characters for the actor,
//  a letter for the frame, and a number for the rotation.
// A sprite that is flippable will have an additional
//  letter/number appended.
// The rotation character can be 0 to signify no rotations.
//
static void InitSpriteDefs(char **namelist)
{
  char **check;
  int i;
  int l;
  int intname;
  int frame;
  int rotation;

  // count the number of sprite names
  check = namelist;
  while (*check != NULL)
    check++;

  numsprites = check - namelist;

  if (!numsprites)
    return;

  sprites = Z_New(spritedef_t, numsprites);

  // scan all the lump names for each of the names,
  //  noting the highest frame letter.
  // Just compare 4 characters as ints
  for (i = 0; i < numsprites; i++)
  {
    spritename = namelist[i];
    memset(sprtemp, -1, sizeof(sprtemp));

    maxframe = -1;
    intname = *(int *)namelist[i];

    // scan the lumps,
    //  filling in the frames for whatever is found
    for (l = 0; l < numspritelumps; l++)
    {
      if (*(int *)lumpinfo[spritelist[l]].name == intname)
      {
        frame = lumpinfo[spritelist[l]].name[4] - 'A';
        rotation = lumpinfo[spritelist[l]].name[5] - '0';

        InstallSpriteLump(l, frame, rotation, false);

        if (lumpinfo[spritelist[l]].name[6])
        {
          frame = lumpinfo[spritelist[l]].name[6] - 'A';
          rotation = lumpinfo[spritelist[l]].name[7] - '0';
          InstallSpriteLump(l, frame, rotation, true);
        }
      }
    }

    // check the frames that were found for completeness
    if (maxframe == -1)
    {
      sprites[i].numframes = 0;
      continue;
    }

    maxframe++;

    for (frame = 0; frame < maxframe; frame++)
    {
      switch ((int)sprtemp[frame].rotate)
      {
        case -1:
          // no rotations were found for that frame at all
          I_Error("R_InitSprites: No patches found "
              "for %s frame %c", namelist[i], frame + 'A');
          break;

        case 0:
          // only the first rotation is needed
          break;

        case 1:
          // must have all 8 frames
          for (rotation = 0; rotation < 8; rotation++)
            if (sprtemp[frame].lump[rotation] == -1)
              I_Error("R_InitSprites: Sprite %s frame %c "
                  "is missing rotations",
                  namelist[i], frame + 'A');
          break;
      }
    }

    // allocate space for the frames present and copy sprtemp to it
    sprites[i].numframes = maxframe;
    sprites[i].spriteframes = Z_New(spriteframe_t, maxframe);
    Z_MoveData(sprites[i].spriteframes, sprtemp, spriteframe_t, maxframe);
  }
}

//
// GAME FUNCTIONS
//
// -ES- 2000/02/04 Switched to stack_array
static stack_array_t vissprite_a;
static vissprite_t *vissprites;
static int vissprite_p;

//
// R_InitSprites
//
void R_InitSprites(char **namelist)
{
#ifdef MHFX_LAXSPRITEROTATIONS
  lax_sprite_rotations = (M_CheckParm("-gfixall")||M_CheckParm("-gignrot"));
#endif
  Z_InitStackArray(&vissprite_a, (void ***)&vissprites, sizeof(vissprite_t), -16);
  InitSpriteDefs(namelist);
}

//
// R_ClearSprites
// Called at frame start.
//
void R_ClearSprites(void)
{
  vissprite_p = 0;
  Z_SetArraySize(&vissprite_a, 0);

  // OPTIMISE ME:
  fuzz_coltable = V_GetRawColtable(DDF_ColmapLookup("FUZZSHADOW"), 127);
}

//
// NewVisSprite
//
// -ES- 1999/03/22 Changed scale parameter to distance
// -ES- 1999/09/28 Killed Parameter

static int NewVisSprite(void)
{
  Z_SetArraySize(&vissprite_a, ++vissprite_p);

  return vissprite_p - 1;
}

//
// R_DrawMaskedColumn
//
// Used for sprites and masked mid textures.
// Masked means: partly transparent, i.e. stored
//  in posts/runs of opaque pixels.
//
// -AJA- 1999/08/08: Updated for thing scaling.

int mfloorclip;
int mceilingclip;
float_t mheightmaximum;

float_t spryscale;
float_t sprtopscreen;

void R_DrawMaskedColumn(const column_t * column)
{
  fixed_t topscreen;
  fixed_t bottomscreen;
  fixed_t basetexturemid;
  fixed_t texture_yoffset;  // -ES- 1999/03/20 Added This
  void (*drawer)(void);
  int yl2;

  if (masked_translucency > 0.99)
  {
    drawer = colfunc;
  }
  else
  {
    dc_translucency = M_FloatToFixed(masked_translucency);
    drawer = R_DrawTranslucentColumn;
  }

  basetexturemid = dc_texturemid;
  texture_yoffset = FixedMul(dc_iscale - 1, M_FloatToFixed(focusyfrac));

#ifndef NOSMOOTHING
  dc_xfrac = 0;
#endif

  for (; column->topdelta != 0xff;)
  {
    // calculate unclipped screen coordinates
    //  for post
    topscreen = M_FloatToFixed(sprtopscreen + spryscale * column->topdelta);
    bottomscreen = topscreen + M_FloatToFixed(spryscale * column->length);

    dc_yl = (topscreen + FRACUNIT /*-1*/ ) >> FRACBITS;
    dc_yh = (bottomscreen - 1) >> FRACBITS;
    yl2 = dc_yl;

    if (dc_yh >= openings[mfloorclip + dc_x])
      dc_yh = openings[mfloorclip + dc_x] - 1;
    if (dc_yl <= openings[mceilingclip + dc_x])
      dc_yl = openings[mceilingclip + dc_x] + 1;

    // -AJA- 1999/08/16: allow a maximum height, mainly for putting
    //       arbitrary textures on the sides of thick extra floors.

    if (mheightmaximum != INT_MAX)
    {
      float_t maxnum = spryscale * mheightmaximum;
      int maxpixel;
      
      if (maxnum <= 0)
         return;
      
      maxpixel = (maxnum < 32767.0) ? (int) maxnum : 32767;
      
// -AJA- try to find DrawColumn bug:
#ifdef DEVELOPERS
    if (maxpixel > 99999)
      I_Error("R_DrawMaskedColumn: INTERAL ERROR: maxpix=%d "
        "maxnum=%1.2f mheightmax=%1.2f spryscale=%1.2f\n",
        maxpixel, maxnum, mheightmaximum, spryscale);
#endif
      if (dc_yh - yl2 >= maxpixel)
        dc_yh = yl2 + maxpixel - 1;

// -AJA- try to find DrawColumn bug:
#ifdef DEVELOPERS
    if (dc_yh > 0x70000000)
      I_Error("R_DrawMaskedColumn: Bad Side: yl=%d yh=%d "
        "yl2=%d maxpix=%d\n", dc_yl, dc_yh, yl2, maxpixel); 
#endif
    }

    if (dc_yl <= dc_yh)
    {
#ifndef NOSMOOTHING
      dc_source2 =
#endif
      dc_source = (const byte *) column + 3;

      dc_texturemid = basetexturemid - (column->topdelta<<FRACBITS)-texture_yoffset;
//      dc_texturemid = -(yl2 * dc_iscale);

      // Drawn by either R_DrawColumn
      //  or (SHADOW) R_DrawFuzzColumn.
// -AJA- try to find DrawColumn bug:
#ifdef DEVELOPERS
      if (dc_yh > 0x70000000)
        I_Error("R_DrawMaskedColumn: Bad Col: yl=%d yh=%d\n", dc_yl, dc_yh);
#endif
      drawer();
    }
    column = (const column_t *) ((const byte *) column + column->length + 4);
  }

  dc_texturemid = basetexturemid;
}

//
// DrawVisSprite
//
// -KM- 1998/11/25 Rewritten to take advantage of Erik's 65535 level trans routines.
// -AJA- 1999/08/08: Updated for thing scaling.

static void DrawVisSprite(vissprite_t * vis, int x1, int x2)
{
  const column_t *column;
  int texturecolumn;
  fixed_t frac;
  const patch_t *patch;
  int trans = 0;

  patch = W_CacheLumpNum(vis->patch);

  dc_colourmap = vis->coltable;

  if (vis->mobjflags & MF_SHADOW)
  {
    trans |= 1;
  }

  if (vis->visibility < VISIBLE && level_flags.trans)
  {
    trans |= 2;
    dc_translucency = M_FloatToFixed(vis->visibility);
  }

  if (vis->trans_table)
  {
    dc_translation = vis->trans_table;
    trans |= 4;
  }

  colfunc = R_DrawColumn;

  if (trans & 1)
    colfunc = R_DrawFuzzColumn;
  else
    switch (trans)
    {
      default:
        break;
      case 2:
        colfunc = R_DrawTranslucentColumn;
        break;
      case 4:
        colfunc = R_DrawTranslatedColumn;
        break;
      case 6:
        colfunc = R_DrawTranslucentTranslatedColumn;
        break;
    }
  // -ES- 1999/03/04 Fixed aspect ratio
  spryscale = vis->yscale;
  dc_iscale = M_FloatToFixed(1/spryscale);
  dc_texturemid = M_FloatToFixed(vis->texturemid);
  frac = M_FloatToFixed(vis->startfrac);
  sprtopscreen = focusyfrac - M_FixedToFloat(dc_texturemid) * vis->dist_scale;
  masked_translucency = 1.0;  // -AJA- 1999/06/30
  mheightmaximum = INT_MAX;  // -AJA- 1999/08/16.

  for (dc_x = vis->x1; dc_x <= vis->x2; dc_x++, frac += M_FloatToFixed(vis->xiscale))
  {
    texturecolumn = frac >> FRACBITS;

#ifdef DEVELOPERS
    if (texturecolumn < 0 || texturecolumn >= SHORT(patch->width))
      I_Error("R_DrawVisSprite: bad texturecolumn %d (patch %d, width %d)", texturecolumn, vis->patch, patch->width);
#endif

    column = (const column_t *) ((const byte *) patch +
        LONG(patch->columnofs[texturecolumn]));
    R_DrawMaskedColumn(column);
  }

  W_DoneWithLump(patch);

  colfunc = basecolfunc;
}

//
// R_AddSprite
//
// Generates a vissprite for a thing if it might be visible.
//
// -AJA- 1999/08/08: Updated for thing scaling.

void R_AddSprite(mobj_t * thing)
{
  float_t tr_x;
  float_t tr_y;

  float_t gzt;

  float_t tx;
  float_t tz;

  float_t xscale;
  float_t yscale;  // -ES- 1999/03/04 Added yscale

  float_t dist_scale;

  int x1;
  int x2;

  spritedef_t *sprdef;
  spriteframe_t *sprframe;
  int lump;

  angle_t ang;
  float_t iscale;
  unsigned rot;
  boolean_t flip;

  int index;

  vissprite_t *vis;
  vert_region_t *thing_reg;

  if (thing->visibility == INVISIBLE)
    return;  // not visible

  // transform the origin point
  tr_x = thing->x - viewx;
  tr_y = thing->y - viewy;

  tz = tr_x * viewcos + tr_y * viewsin;

  // thing is behind view plane?
  if (tz < MINZ)
    return;

  // -ES- 1999/03/14 Use distunit for y and x scale.
  xscale = x_distunit / tz;
  yscale = y_distunit / tz;
  dist_scale = yscale;

  tx = tr_x * viewsin - tr_y * viewcos;

  // too far off the side?
  // -ES- 1999/03/13 Fixed clipping to work with large FOVs (up to 176 deg)
  // rejects all sprites where angle>176 deg (arctan 32), since those
  // sprites would result in overflow in future calculations
  // -ES- 1999/09/28 FIXME: Boost Limit, we're using float now...
  if (fabs(tx) / 32 > tz)
    return;

  // decide which patch to use for sprite relative to player
#ifdef DEVELOPERS
  if ((unsigned int)thing->sprite >= (unsigned int)numsprites)
    I_Error("R_ProjectSprite: invalid sprite number %i ",
        thing->sprite);
#endif
  sprdef = &sprites[thing->sprite];

// -ACB- 1998/06/29 More Informative Error Message
#ifdef DEVELOPERS
  if ((thing->frame & FF_FRAMEMASK) >= sprdef->numframes)
  {
    if (thing->frame & FF_FULLBRIGHT)
      thing->frame -= FF_FULLBRIGHT;

    I_Error("R_ProjectSprite: Invalid sprite frame %s:%c",
        sprnames[thing->sprite], (thing->frame + 'A'));
  }
#endif

  sprframe = &sprdef->spriteframes[thing->frame & FF_FRAMEMASK];

  if (sprframe->rotate)
  {
    // choose a different rotation based on player view
    ang = R_PointToAngle(viewx, viewy, thing->x, thing->y);
    rot = (ang - thing->angle + (unsigned)(ANG45 / 2) * 9) >> 29;
    lump = sprframe->lump[rot];
    flip = (boolean_t) sprframe->flip[rot];
  }
  else
  {
    // use single rotation for all views
    lump = sprframe->lump[0];
    flip = (boolean_t) sprframe->flip[0];
  }

  // calculate edges of the shape
  tx -= spriteoffset[lump] * thing->info->xscale;
  x1 = focusxfrac + tx * xscale;

  // off the right side?
  if (x1 > viewwidth)
    return;

  tx += spritewidth[lump] * thing->info->xscale;
  x2 = focusxfrac + tx * xscale - 1;

  // off the left side
  if (x2 < 0)
    return;

  xscale *= thing->info->xscale;
  yscale *= thing->info->yscale;

  gzt = thing->z + spritetopoffset[lump] * thing->info->yscale;

  // work out the vertical region

  thing_reg = R_PointInVertRegion(thing->subsector,
      thing->z+thing->height/2);

  // store information in a vissprite
  vis = &vissprites[NewVisSprite()];
  Z_LockStackArray(&vissprite_a);

  vis->mobjflags = thing->flags;
  vis->visibility = thing->visibility;
  vis->yscale = yscale;
  vis->distance = tz;
  vis->dist_scale = dist_scale;
  vis->gx = thing->x;
  vis->gy = thing->y;
  vis->gz = thing->z;
  vis->gzt = gzt;
  vis->texturemid = vis->gzt - viewz;
  vis->x1 = x1 < 0 ? 0 : x1;
  vis->x2 = x2 >= viewwidth ? viewwidth - 1 : x2;
  vis->pass_num = 0;  // -AJA- 1999/08/24.

  iscale = 1 / xscale;

  // -AJA- 1999/06/21: needed later for clipping against extra floors.
  vis->sector = thing->subsector->sector;

  if (flip)
  {
    vis->startfrac = spritewidth[lump] - 0.0001;//iscale / 2;
    vis->xiscale = -iscale;
  }
  else
  {
    vis->startfrac = 0;//iscale / 2;
    vis->xiscale = iscale;
  }

  if (vis->x1 > x1)
  {
    vis->startfrac += vis->xiscale * (vis->x1 - x1);
  }

  vis->patch = spritelist[lump];

  // translation support
  if (thing->info->palremap)
    vis->trans_table = V_GetTranslationTable(thing->info->palremap);
  else
    vis->trans_table = NULL;
  
  // get light level
  // -AJA- 1999/07/10: Updated for colmap.ddf.

  if (thing->frame & FF_FULLBRIGHT)
  {
    // full bright
    vis->coltable = V_GetColtable(thing_reg->p->colourmap, 255, 0);
  }
  else if (thing->flags & MF_SHADOW)
  {
    // shadow colourmap
    vis->coltable = fuzz_coltable;
  }
  else
  {
    // diminished light
    // -ES- 1999/03/22 Use distance instead of xscale for lighting calc
    // -AJA- 1999/07/10: Now uses zlight[] instead of scalelight[].

    int lightnum = (thing_reg->p->lightlevel >> LIGHTSEGSHIFT);

    if (lightnum < 0)
      spritezlights = zlight[0];
    else if (lightnum >= LIGHTLEVELS)
      spritezlights = zlight[LIGHTLEVELS - 1];
    else
      spritezlights = zlight[lightnum];

    index = vis->distance / (1 << (LIGHTZSHIFT - FRACBITS));

    if (index >= MAXLIGHTZ)
      index = MAXLIGHTZ - 1;

    vis->coltable = V_GetColtable(thing_reg->p->colourmap, 
        spritezlights[index], 0);
  }

  Z_UnlockStackArray(&vissprite_a);
}


//
// DrawPSprite
//
// -ACB- 1998/09/02 Fixed Error Messages to display correct procedure.
//
static void DrawPSprite(pspdef_t * psp, player_t * viewplayer,
    region_properties_t *props)
{
  float_t tx;
  int x1;
  int x2;
  spritedef_t *sprdef;
  spriteframe_t *sprframe;
  int lump;
  boolean_t flip;
  vissprite_t vis;
  const colourmap_t *nominal;

  // decide which patch to use

#ifdef DEVELOPERS
  if ((unsigned int)psp->state->sprite >= (unsigned int)numsprites)
    I_Error("DrawPSprite: invalid sprite number %i ",
        psp->state->sprite);
#endif

  sprdef = &sprites[psp->state->sprite];

#ifdef DEVELOPERS
  if ((psp->state->frame & FF_FRAMEMASK) >= sprdef->numframes)
    I_Error("DrawPSprite: invalid sprite frame %d : %ld ",
        psp->state->sprite, psp->state->frame);
#endif

  sprframe = &sprdef->spriteframes[psp->state->frame & FF_FRAMEMASK];

  lump = sprframe->lump[0];
  flip = (boolean_t) sprframe->flip[0];

  // calculate edges of the shape
  tx = psp->sx - BASEXCENTER;

  tx -= spriteoffset[lump];
  x1 = focusxfrac + tx * pspritescale;

  // off the right side
  if (x1 > viewwidth)
    return;

  tx += spritewidth[lump];
  x2 = focusxfrac + tx * pspritescale - 1;

  // off the left side
  if (x2 < 0)
    return;

  // -ACB- 1998/09/02 Clear the vissprite mem before setup. Safety thing..
  Z_Clear(&vis, vissprite_t, 1);

  vis.texturemid = BASEYCENTER - psp->sy + spritetopoffset[lump];
  vis.x1 = x1 < 0 ? 0 : x1;
  vis.x2 = x2 >= viewwidth ? viewwidth - 1 : x2;
  vis.yscale = pspritescale2;
  vis.dist_scale = pspritescale2;
  vis.mobjflags = 0;

  if (flip)
  {
    vis.xiscale = -pspriteiscale;
    // -ES- Fixme
    vis.startfrac = spritewidth[lump] - 0.0001;
  }
  else
  {
    vis.xiscale = pspriteiscale;
    vis.startfrac = 0;
  }

  if (vis.x1 > x1)
    vis.startfrac += vis.xiscale * (vis.x1 - x1);

  vis.patch = spritelist[lump];
  vis.visibility = viewplayer->mo->visibility;

  // calculate lighting
  // -AJA- 1999/07/10: Updated for colmap.ddf.
  nominal = props->colourmap;

  if (viewplayer->powers[PW_PartInvis] > 4 * 32 ||
      fmod(viewplayer->powers[PW_PartInvis], 16) >= 8)
  {
    // shadow draw
    vis.mobjflags |= MF_SHADOW;
    vis.coltable = fuzz_coltable;
  }
  else if (psp->state->frame & FF_FULLBRIGHT)
  {
    // full bright
    vis.coltable = V_GetColtable(nominal, 255, 0);
  }
  else
  {
    // local light
    vis.coltable = V_GetColtable(nominal, spritezlights[20], 0);
  }

  DrawVisSprite(&vis, vis.x1, vis.x2);
}

//
// R_DrawPlayerSprites
//
void R_DrawPlayerSprites(player_t * p)
{
  int i;
  int lightnum;

  // get light level
  lightnum = view_vertregion->p->lightlevel >> LIGHTSEGSHIFT;

  // -AJA- 1999/07/10: Now uses zlight[] instead of scalelight[].
  if (lightnum < 0)
    spritezlights = zlight[0];
  else if (lightnum >= LIGHTLEVELS)
    spritezlights = zlight[LIGHTLEVELS - 1];
  else
    spritezlights = zlight[lightnum];

  // clip to screen bounds
  mfloorclip = screenheightarray;
  mceilingclip = negonearray;

  // add all active psprites
  // Note: order is significant
  for (i = 0; i < NUMPSPRITES; i++)
  {
    pspdef_t *psp = &p->psprites[i];

    if (psp->state != S_NULL)
      DrawPSprite(psp, p, view_vertregion->p);
  }
}

//
// R_SortVisSprites
//
static int *mapvissprites = NULL;

void R_SortVisSprites(void)
{
  int i;

  Z_Resize(mapvissprites, int, vissprite_p);
  for (i = 0; i < vissprite_p; i++)
    mapvissprites[i] = i;

#define CMP(a, b) (vissprites[a].distance > vissprites[b].distance)
  QSORT(int, mapvissprites, vissprite_p, CUTOFF);
#undef CMP

  MAPSORT(vissprite_t, vissprites, mapvissprites, vissprite_p);
}

//
// DrawSprite
// 23-6-98 KM Reworked ds into neater code

static void DrawSprite(vissprite_t * spr)
{
  int ds_index;
  int clipbot;
  int cliptop;
  int x;
  int r1;
  int r2;
  int silhouette;

  clipbot = lastopening;
  cliptop = lastopening + vb_w;
  Z_SetArraySize(&openings_a, lastopening += 2*vb_w);

  for (x = spr->x1; x <= spr->x2; x++)
    openings[clipbot + x] = openings[cliptop + x] = -2;

  // Scan drawsegs from end to start for obscuring segs.
  // The first drawseg that has a greater scale is the clip seg.

  for (ds_index = ds_p - 1; ds_index >= 0; ds_index--)
  {
    drawseg_t *ds = drawsegs[ds_index];

    float_t close_scale;
    float_t far_scale;

    // determine if the drawseg obscures the sprite
    if (ds->x1 > spr->x2 ||
        ds->x2 < spr->x1 ||
        (!ds->silhouette && ds->maskedtexturecol == -1 && !ds->extraplane))
    {
      // does not cover sprite
      continue;
    }

    r1 = ds->x1 < spr->x1 ? spr->x1 : ds->x1;
    r2 = ds->x2 > spr->x2 ? spr->x2 : ds->x2;

    close_scale = MAX(ds->scale1, ds->scale2);
    far_scale = MIN(ds->scale1, ds->scale2);

    if (close_scale < spr->dist_scale ||
        (far_scale < spr->dist_scale &&
        !R_PointOnSegSide(spr->gx, spr->gy, ds->curline)))
    {
      // seg is behind sprite.
      // masked mid texture?
      if (ds->maskedtexturecol != -1 && !ds->thick_side)
        R_RenderMaskedSegRange(ds, r1, r2);

      if (ds->extraplane || ds->thick_side)
        extra_planes = true;

      // -AJA- 1999/08/24: check if sprite intersects with
      //       an extra floor/ceiling on-screen.

      if (!ds->extraplane)
        continue;

      if (((ds->extraplane->flags & VPF_FLOOR) &&
              (spr->gz >= ds->extraplane->height)) ||
          ((ds->extraplane->flags & VPF_CEILING) &&
              (spr->gzt <= ds->extraplane->height)))
      {
        // Sprite needs to be drawn in the next pass, otherwise
        // it would be clobbered by this extra plane.

        spr->pass_num++;
        
        Z_SetArraySize(&openings_a, lastopening -= 2*vb_w);

        return;
      }

      continue;
    }

    // clip this piece of the sprite
    silhouette = ds->silhouette;

    if ((silhouette & SIL_BOTTOM) && spr->gz >= ds->bsilheight)
      silhouette &= ~SIL_BOTTOM;

    if ((silhouette & SIL_TOP) && spr->gzt <= ds->tsilheight)
      silhouette &= ~SIL_TOP;

    if (silhouette == 1)
    {
      // bottom sil
      for (x = r1; x <= r2; x++)
        if (openings[clipbot + x] == -2)
          openings[clipbot + x] = openings[ds->sprbottomclip + x];
    }
    else if (silhouette == 2)
    {
      // top sil
      for (x = r1; x <= r2; x++)
        if (openings[cliptop + x] == -2)
          openings[cliptop + x] = openings[ds->sprtopclip + x];
    }
    else if (silhouette == 3)
    {
      // both
      for (x = r1; x <= r2; x++)
      {
        if (openings[clipbot + x] == -2)
          openings[clipbot + x] = openings[ds->sprbottomclip + x];
        if (openings[cliptop + x] == -2)
          openings[cliptop + x] = openings[ds->sprtopclip + x];
      }
    }
  }

  // all clipping has been performed, so draw the sprite

  // check for unclipped columns
  for (x = spr->x1; x <= spr->x2; x++)
  {
    if (openings[clipbot + x] == -2)
      openings[clipbot + x] = viewheight;

    if (openings[cliptop + x] == -2)
      openings[cliptop + x] = -1;
  }

  mfloorclip = clipbot;
  mceilingclip = cliptop;
  DrawVisSprite(spr, spr->x1, spr->x2);

  Z_SetArraySize(&openings_a, lastopening -= 2*vb_w);
}

//
// R_DrawExtraFloors
//
// -AJA- 1999/08/24: Draw all the extra floors that are stored in the
//       drawsegs.  These visplanes will cover all sprites drawn in
//       pass #0, and be covered by sprites drawn in pass #1.  
//
//       Note the drawing sequence: last drawseg -> first.  The first
//       drawseg is the closest one (guaranteed by the BSP traversal),
//       so this should ensure that extra floors/ceilings will be
//       rendered properly against other extra floors/ceilings.
//
// -AJA- 1999/09/21: Now draws the thick sides as well. The same note
//       about the drawing sequence applies, so that extra floors
//       should no longer bleed into the extra sides.
//
static void DrawExtraFloors(void)
{
  int i;

  for (i = ds_p - 1; i >= 0; i--)
  {
    drawseg_t *ds = drawsegs[i];

    if (-1 != ds->maskedtexturecol && ds->thick_side)
    {
      R_RenderMaskedSegRange(ds, ds->x1, ds->x2);
      ds->maskedtexturecol = -1;
    }

    if (ds->extraplane)
    {
      R_DrawSinglePlane(ds->extraplane);
      ds->extraplane = NULL;
    }
  }
}

//
// DrawMaskedSet
//
// -AJA- 1999/08/24: Draw all the sprites with a particular
//       `pass_num'.  Used for drawing sprites and extra floors in the
//       correct order.  Pretty hack-ish, really.
//
static void DrawMaskedSet(int pass)
{
  int i;

  Z_LockStackArray(&vissprite_a);
  for (i = 0; i < vissprite_p; i++)
  {
    if (vissprites[i].pass_num == pass)
      DrawSprite(&vissprites[i]);
  }
  Z_UnlockStackArray(&vissprite_a);

  // render any remaining masked mid textures

  for (i = ds_p - 1; i >= 0; i--)
  {
    drawseg_t *ds = drawsegs[i];

    if (-1 != ds->maskedtexturecol && !ds->thick_side)
    {
      R_RenderMaskedSegRange(ds, ds->x1, ds->x2);
      ds->maskedtexturecol = -1;
    }

    if (ds->extraplane || ds->thick_side)
      extra_planes = true;
  }
}

//
// R_DrawMasked
// 23-6-98 KM Reworked ds into neater code
//
void R_DrawMasked(void)
{
  if (vissprite_p)
    R_SortVisSprites();

  // draw all vissprites back to front

  extra_planes = false;
  DrawMaskedSet(0);

  // was there any extra floors/ceilings ?

  if (extra_planes)
  {
    DrawExtraFloors();
    DrawMaskedSet(1);
  }
}
