//----------------------------------------------------------------------------
//  EDGE Radius Trigger Actions
//----------------------------------------------------------------------------
// 
//  Copyright (c) 1999-2000  The EDGE Team.
// 
//  This program is free software; you can redistribute it and/or
//  modify it under the terms of the GNU General Public License
//  as published by the Free Software Foundation; either version 2
//  of the License, or (at your option) any later version.
//
//  This program is distributed in the hope that it will be useful,
//  but WITHOUT ANY WARRANTY; without even the implied warranty of
//  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//  GNU General Public License for more details.
//
//----------------------------------------------------------------------------
//
// -AJA- 1999/10/24: Split these off from the rad_trig.c file.
//

#include "i_defs.h"
#include "rad_trig.h"

#include "dm_defs.h"
#include "dm_state.h"
#include "hu_lib.h"
#include "hu_stuff.h"
#include "g_game.h"
#include "m_argv.h"
#include "m_inline.h"
#include "m_menu.h"
#include "m_random.h"
#include "m_swap.h"
#include "p_local.h"
#include "p_spec.h"
#include "r_defs.h"
#include "s_sound.h"
#include "v_colour.h"
#include "v_res.h"
#include "w_wad.h"
#include "z_zone.h"

// Tip Displayer info.
// -AJA- FIXME: duplicates much in s_tip_t
typedef struct
{
  // tip text DOH!
  const char *tip_text;
  char *tip_graphic;

  // level time tip arrived
  int tip_time;

  // display time
  int tip_delay;

  // sound related
  boolean_t tip_do_once;

  // can be overidden?
  boolean_t tip_override;
}
tip_t;

// Current tip
static tip_t cur_tip = { NULL, NULL, 0, 0, false, false };

#define TIP_LINE_MAX  10
static hu_textline_t tip_hu[TIP_LINE_MAX];
static int tip_hu_lines;

//
// TIP Functions
// Remember a tip which emits a sound has a higher priority than one without,
// so if one in use emits no sound it is terminated abruptly.
//
// Tip Displayer info.

void TIP_SendTip(s_tip_t * tip)
{
  if (cur_tip.tip_override && !tip->playsound)
    return;

  cur_tip.tip_text = tip->tip_text;
  cur_tip.tip_graphic = tip->tip_graphic;
  cur_tip.tip_delay = tip->display_time;

  if (tip->tip_ldf)
  {
    cur_tip.tip_text = DDF_LanguageLookup(tip->tip_ldf);
  }

  cur_tip.tip_override = cur_tip.tip_do_once = tip->playsound;
}

// -AJA- 1999/09/07: Reworked to handle tips with multiple lines.

void TIP_DisplayTips(int y)
{
  const char *tipptr;

  int i;
  boolean_t inited;

  int font_height = SHORT(hu_font[0]->height) + 2;

  tipptr = cur_tip.tip_text;

  // Is there actually a tip to display
  if (cur_tip.tip_delay == 0)
    return;

  // If the display time is up reset the tip and erase it.
  if (cur_tip.tip_delay == -2)
  {
    cur_tip.tip_delay = 0;
    cur_tip.tip_override = false;

    for (i = 0; i < tip_hu_lines; i++)
    {
      HUlib_eraseTextLine(&tip_hu[i]);
    }
    return;
  }

  // Make a noise when the tip is first displayed.
  // Note: This happens only once. (do_once)
  if (cur_tip.tip_do_once)
  {
    cur_tip.tip_do_once = false;
    S_StartSound(NULL, sfx_tink);
  }

  if (cur_tip.tip_graphic != NULL)
  {
    // FIXME: need some control
    V_DrawPatchInDirectName(main_scr, 120, 70, cur_tip.tip_graphic);

    return;
  }

  // Dump it to the screen

  tip_hu_lines = 1;
  inited = false;

  while (tipptr[0])
  {
    if (!inited)
    {
      HUlib_initTextLine(&tip_hu[tip_hu_lines - 1],
          160, 100, hu_font, HU_FONTSTART);
      tip_hu[tip_hu_lines - 1].centre = true;
      inited = true;
    }

    if (tipptr[0] == '\n')
    {
      if (tip_hu_lines == TIP_LINE_MAX)
      {
        // Debug_Printf("Tip too tall !\n");
        break;
      }

      tip_hu_lines++;
      inited = false;

      tipptr++;
      continue;
    }

    HUlib_addCharToTextLine(&tip_hu[tip_hu_lines - 1], tipptr[0]);

    tipptr++;
  }

  for (i = 0; i < tip_hu_lines; i++)
  {
    tip_hu[i].y = 100 + (i * 2 - tip_hu_lines) * font_height / 2;

    // -ACB- 1998/06/10 Tryed this instead of HUlibdrawTextLine,
    //   Uses trans table and is scaled up if possible.

    HUlib_drawTextLineTrans(&tip_hu[i], false, text_colour_white);
  }
}

//
// RAD_ResetTips
//
// Used when changing levels to clear any tips.
//
// -AJA- 1999/09/07: written.
//

void RAD_ResetTips(void)
{
  if (cur_tip.tip_delay > 0)
  {
    cur_tip.tip_delay = -2;
  }
}

//
// RAD_Ticker
//
// Does any tic-related RTS stuff.  For now, just update the tips.
//
// -AJA- 1999/09/07: written.

void RAD_Ticker(void)
{
  // update the tips.

  if (cur_tip.tip_delay > 0)
  {
    cur_tip.tip_delay--;

    if (cur_tip.tip_delay == 0)
    {
      // it timed out. mark it as such.

      cur_tip.tip_delay = -2;
    }
  }
}


// --- Radius Trigger Actions -----------------------------------------------

void RAD_ActNOP(rad_trigger_t *R, mobj_t *actor, void *param)
{
  // No Operation
}

void RAD_ActTip(rad_trigger_t *R, mobj_t *actor, void *param)
{
  s_tip_t *tip = (s_tip_t *) param;

  // Only display the tip to the player that stepped into the radius
  // trigger.

  if (actor->player == consoleplayer)
  {
    TIP_SendTip(tip);
  }
}

void RAD_ActSpawnThing(rad_trigger_t *R, mobj_t *actor, void *param)
{
  s_thing_t *t = (s_thing_t *) param;

  mobj_t *mo;
  const mobjinfo_t *minfo;

  // These *MUST* happen to everyone to keep netgames consistent.
  // Spawn a new map object.

  if (t->thing_name)
    minfo = DDF_MobjLookup(t->thing_name);
  else
    minfo = DDF_MobjLookupNum(t->thing_type);

  if (minfo == NULL)
  {
    if (t->thing_name)
      I_Error("Thing type: %s in RTS trigger not found.\n", t->thing_name);
    else
      I_Error("Thing type: %d in RTS trigger not found.\n", t->thing_type);
    return;
  }

  // -AJA- 1999/10/02: -nomonsters check.
  if (level_flags.nomonsters && (minfo->extendedflags & EF_MONSTER))
    return;

  // -AJA- 1999/10/07: -noextra check.
  if (!level_flags.have_extra && (minfo->extendedflags & EF_EXTRA))
    return;

  // -AJA- 1999/09/11: Support for supplying Z value.

  if (t->spawn_effect)
  {
    mo = P_MobjCreateObject(t->x, t->y, t->z, minfo->respawneffect);
    S_StartSound(mo, sfx_itmbk);
  }

  mo = P_MobjCreateObject(t->x, t->y, t->z, minfo);

  // -ACB- 1998/07/10 New Check, so that spawned mobj's don't
  //                  spawn somewhere where they should not.
  if (!P_CheckAbsPosition(mo, mo->x, mo->y, mo->z))
  {
    P_RemoveMobj(mo);
    return;
  }

  P_SetMobjDirAndSpeed(mo, t->angle, t->slope, 0);

  mo->spawnpoint.x = SHORT(M_FloatToFixed(t->x) >> FRACBITS);
  mo->spawnpoint.y = SHORT(M_FloatToFixed(t->y) >> FRACBITS);
  mo->spawnpoint.angle = SHORT(t->angle >> FRACBITS);
  mo->spawnpoint.type = minfo->ddf.number;

  if (t->ambush)
    mo->flags |= MF_AMBUSH;

  // -AJA- 1999/09/25: If radius trigger is a path node, then
  //       setup the thing to follow the path.

  if (R->info->next_in_path)
    mo->path_trigger = R->info;
}

void RAD_ActDamagePlayers(rad_trigger_t *R, mobj_t *actor, void *param)
{
  s_damagep_t *damage = (s_damagep_t *) param;
  player_t *p;
  
  // Make sure these can happen to everyone within the radius.
  // Damage the player(s)
  for (p = players; p; p = p->next)
  {
    if (!p->in_game)
      continue;

    if (RAD_WithinRadius(p->mo, R->info))
    {
      P_DamageMobj(p->mo, NULL, NULL, damage->damage_amount);
    }
  }
}

void RAD_ActHealPlayers(rad_trigger_t *R, mobj_t *actor, void *param)
{
  s_healp_t *heal = (s_healp_t *) param;
  player_t *p;

  // Heal the player(s)
  for (p = players; p; p = p->next)
  {
    if (!p->in_game)
      continue;
      
    if (RAD_WithinRadius(p->mo, R->info))
    {
      if (p->health >= heal->limit)
        continue;

      if (p->health + heal->heal_amount >= heal->limit)
        p->health = heal->limit;
      else
        p->health += heal->heal_amount;

      p->mo->health = p->health;
    }
  }
}

void RAD_ActArmourPlayers(rad_trigger_t *R, mobj_t *actor, void *param)
{
  s_armour_t *armour = (s_armour_t *) param;
  player_t *p;

  // Armour for player(s)
  for (p = players; p; p = p->next)
  {
    if (!p->in_game)
      continue;

    if (RAD_WithinRadius(p->mo, R->info))
    {
      if (armour->type > p->armourtype)
        p->armourtype = armour->type;

      if (p->armourpoints >= armour->limit)
        continue;

      if (p->armourpoints + armour->armour_amount > armour->limit)
        p->armourpoints = armour->limit;
      else
        p->armourpoints += armour->armour_amount;
    }
  }
}

void RAD_ActDamageMonsters(rad_trigger_t *R, mobj_t *actor, void *param)
{
  s_damage_monsters_t *mon = (s_damage_monsters_t *) param;
 
  // -AJA- FIXME: this is _so_ non-optimal...

  mobj_t *mo;
  const mobjinfo_t *info = NULL;

  if (mon->thing_name)
  {
    info = DDF_MobjLookup(mon->thing_name);
  }
  else if (mon->thing_type >= 0)
  {
    info = DDF_MobjLookupNum(mon->thing_type);

    if (info == NULL)
      I_Error("RTS ONDEATH: Unknown thing type %d.\n",
          mon->thing_type);
  }
  
  // scan the mobj list
  for (mo=mobjlisthead; mo != NULL; mo=mo->next)
  {
    if (info && mo->info != info)
      continue;

    if (! (mo->extendedflags & EF_MONSTER) || mo->health <= 0)
      continue;
    
    if (! RAD_WithinRadius(mo, R->info))
      continue;
    
    P_DamageMobj(mo, NULL, actor, mon->damage_amount);
  }
}

void RAD_ActGotoMap(rad_trigger_t *R, mobj_t *actor, void *param)
{
  s_gotomap_t *go = (s_gotomap_t *) param;

  // Warp to level n
  G_ExitToLevel(go->map_name, 5);
}

void RAD_ActExitLevel(rad_trigger_t *R, mobj_t *actor, void *param)
{
  s_exit_t *exit = (s_exit_t *) param;

  G_ExitLevel(exit->exittime);
}

void RAD_ActSaveGame(rad_trigger_t *R, mobj_t *actor, void *param)
{
  s_savegame_t *save = (s_savegame_t *) param;

  // quicksave ?
  if (save->slot < 0)
  {
    if (quickSaveSlot < 0)
      M_QuickSave();
    else
      M_DoSave(quickSaveSlot);
  }
  else
  {
    M_DoSave(save->slot);
  }
}

void RAD_ActPlaySound(rad_trigger_t *R, mobj_t *actor, void *param)
{
  s_sound_t *ambient = (s_sound_t *) param;

  // Ambient sound
  R->soundorg.x = ambient->x;
  R->soundorg.y = ambient->y;

  if (ambient->z == ONFLOORZ)
    R->soundorg.z = R_PointInSubsector(ambient->x, ambient->y)->
        sector->floor.h;
  else
    R->soundorg.z = ambient->z;

  S_StartSound((mobj_t *) &R->soundorg, ambient->soundid);
}

void RAD_ActKillSound(rad_trigger_t *R, mobj_t *actor, void *param)
{
  S_StopSound((mobj_t *) &R->soundorg);
}

void RAD_ActSkill(rad_trigger_t *R, mobj_t *actor, void *param)
{
  s_skill_t *skill = (s_skill_t *) param;

  // Skill selection trigger function
  // -ACB- 1998/07/30 replaced respawnmonsters with respawnsetting.
  // -ACB- 1998/08/27 removed fastparm temporaryly.
  
  gameskill = skill->skill;

  level_flags.fastparm = skill->FastMonsters;
  level_flags.respawn = skill->Respawn;
}

void RAD_ActSectorMove(rad_trigger_t *R, mobj_t *actor, void *param)
{
  s_sectorv_t *t = (s_sectorv_t *) param;

  // SectorV
  if (t->secnum < 0 || t->secnum >= numsectors)
    I_Error("RTS SECTORV: no such sector %d.\n", t->secnum);

  if (t->corf)
  { 
    // Floor Height
    sectors[t->secnum].floor.h += t->increment;
    P_ChangeSector(&sectors[t->secnum], true);
  }
  else
  { 
    // Ceiling Height
    sectors[t->secnum].ceil.h += t->increment;
    P_ChangeSector(&sectors[t->secnum], true);
  }
}

void RAD_ActSectorLight(rad_trigger_t *R, mobj_t *actor, void *param)
{
  s_sectorl_t *t = (s_sectorl_t *) param;

  // SectorL (Lighting)

  if (t->secnum < 0 || t->secnum >= numsectors)
    I_Error("RTS SECTORL: no such sector %d.\n", t->secnum);

  sectors[t->secnum].p.lightlevel += t->increment;
}

void RAD_ActEnableScript(rad_trigger_t *R, mobj_t *actor, void *param)
{
  s_enabler_t *t = (s_enabler_t *) param;
  rad_trigger_t *other;

  // Enable/Disable Scripts
  if (t->script_name)
  {
    other = RAD_FindTriggerByName(t->script_name);

    if (! other)
      return;
    
    other->disabled = t->new_disabled;
  }
  else
  {
    RAD_EnableByTag(t->tag, t->new_disabled);
  }
}

void RAD_ActActivateLinetype(rad_trigger_t *R, mobj_t *actor, void *param)
{
  s_lineactivator_t *t = (s_lineactivator_t *) param;

  P_RemoteActivation(actor, t->typenum, t->tag, 0, line_Any);
}

void RAD_ActJump(rad_trigger_t *R, mobj_t *actor, void *param)
{
  s_jump_t *t = (s_jump_t *) param;

  if (t->random_chance < 255 && P_Random() > t->random_chance)
    return;

  if (! t->cache_state)
  {
    t->cache_state = RAD_FindStateByLabel(R->info, t->label);

    if (! t->cache_state)
      I_Error("RTS: No such label `%s' for JUMP primitive.\n", t->label);
  }

  R->state = t->cache_state;

  // Jumps have a one tic surcharge, to prevent accidental infinite
  // loops within radius scripts.
  R->wait_tics += 1;
}

void RAD_ActSleep(rad_trigger_t *R, mobj_t *actor, void *param)
{
  R->disabled = true;
}

