//----------------------------------------------------------------------------
//  EDGE Radius Trigger Parsing
//----------------------------------------------------------------------------
// 
//  Copyright (c) 1999-2000  The EDGE Team.
// 
//  This program is free software; you can redistribute it and/or
//  modify it under the terms of the GNU General Public License
//  as published by the Free Software Foundation; either version 2
//  of the License, or (at your option) any later version.
//
//  This program is distributed in the hope that it will be useful,
//  but WITHOUT ANY WARRANTY; without even the implied warranty of
//  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//  GNU General Public License for more details.
//
//----------------------------------------------------------------------------

#include "i_defs.h"
#include "rad_trig.h"

#include "dm_defs.h"
#include "dm_state.h"
#include "hu_lib.h"
#include "hu_stuff.h"
#include "g_game.h"
#include "m_argv.h"
#include "m_swap.h"
#include "p_local.h"
#include "p_spec.h"
#include "r_defs.h"
#include "s_sound.h"
#include "v_res.h"
#include "w_wad.h"
#include "z_zone.h"


typedef struct define_s
{
  // next in list
  struct define_s *next;

  char *name;
  char *value;
}
define_t;

typedef struct rts_parser_s
{
  // needed level:
  //   -1 : don't care
  //    0 : outside any blocks
  //    1 : within START_MAP block
  //    2 : within RADIUSTRIGGER block
  int level;

  // name
  char *name;

  // parser function
  void (* parser)(int pnum, const char ** pars);
}
rts_parser_t;


int rad_cur_linenum;
char *rad_cur_filename;
static char *rad_cur_linedata = NULL;

static char tokenbuf[4096];

// -AJA- 1999/09/12: Made all these static.  The variable `defines'
//       was clashing with the one in ddf_main.c.  ARGH !

// Define List
static define_t *defines;

// Determine whether the code blocks are started and terminated.
static int rad_cur_level = 0;

static const char *rad_level_names[3] =
{ "outer area", "map area", "trigger area" };

// Location of current script
static rad_script_t *this_rad;
static char *this_map = NULL;

// Pending state info for current script
static int pending_wait_tics = 0;
static char *pending_label = NULL;


int RAD_StringHashFunc(const char *s)
{
  int r = 0;
  int c;

  while (*s)
  {
    r *= 36;
    if (*s >= 'a' && *s <= 'z')
      c = *s - 'a';
    else if (*s >= 'A' && *s <= 'Z')
      c = *s - 'A';
    else if (*s >= '0' && *s <= '9')
      c = *s - '0' + 'Z' - 'A' + 1;
    else
      c = *s;
    r += c % 36;
    s++;
  }

  return r;
}

//
// RAD_Error
//
// -AJA- 2000/01/04: written.

void RAD_Error(const char *err, ...)
{
  va_list argptr;
  char buffer[1024];

  va_start(argptr, err);
  vsprintf(buffer, err, argptr);
  va_end(argptr);
  
  I_Printf("\n\nFound error near line %d of %s\n", rad_cur_linenum, 
        rad_cur_filename);
  
  if (rad_cur_linedata)
    I_Printf("with line contents: %s\n", rad_cur_linedata);
    
  I_Error("\n%s", buffer);
}

static void RAD_ErrorSetLineData(const char *data)
{
  if (rad_cur_linedata)
    Z_Free(rad_cur_linedata);
  
  rad_cur_linedata = Z_StrDup(data);
}

static void RAD_ErrorClearLineData(void)
{
  if (rad_cur_linedata)
  {
    Z_Free(rad_cur_linedata);
    rad_cur_linedata = NULL;
  }
}


// Searches through the #defines namespace for a match and returns
// its value if it exists.
static boolean_t CheckForDefine(const char *s, char ** val)
{
  define_t *tempnode = defines;

  for (; tempnode; tempnode = tempnode->next)
  {
    if (strcmp(s, tempnode->name) == 0)
    {
      *val = Z_StrDup(tempnode->value);
      return true;
    }
  }
  return false;
}

static void RAD_CheckForInt(const char *value, int *retvalue)
{
  const char *pos = value;
  int count = 0;
  int length = strlen(value);

  // Accomodate for "-" as you could have -5 or something like that.
  if (*pos == '-')
  {
    count++;
    pos++;
  }
  while (isdigit(*pos++))
    count++;

  // Is the value an integer?
  if (length != count)
    RAD_Error("Parameter is not of numeric type: %s\n", value);

  *retvalue = atoi(value);
}

static void RAD_CheckForFloat(const char *value, float_t *retvalue)
{
  if (sscanf(value, "%f", retvalue) != 1)
    RAD_Error("Parameter is not of numeric type: %s\n", value);
}

//
// RAD_CheckForPercent
//
// Reads percentages (0%..100%), storing the results as a 0-255 int
// FIXME: Globally change to 0..1 representation of percentages
static void RAD_CheckForPercent(const char *info, void *storage)
{
  char s[101];
  char *p;
  float_t f;

  // just check that the string is valid
  Z_StrNCpy(s, info, 100);
  for (p = s; isdigit(*p) || *p == '.'; p++)
  { }

  // the number must be followed by %
  if (*p != '%')
    RAD_Error("Parameter is not of percent type: %s\n", info);
  *p = 0;

  RAD_CheckForFloat(s, &f);
  if (f < 0.0 || f > 100.0)
    RAD_Error("Percentage out of range: %s\n", info);

  // convert to a 0-255 value
  *(int *)storage = (int)(f * 2.555);
}

// -ES- Copied from DDF_MainGetTime.
// FIXME: Collect all functions that are common to DDF and RTS,
// and move them to a new module for RTS+DDF common code.
static void RAD_CheckForTime(const char *info, void *storage)
{
  float val;
  int *dest = (int *)storage;
  int i;
  char *s;

  DEV_ASSERT2(info && storage);

  // -ES- 1999/09/14 MAXT means that time should be maximal.
  if (!stricmp(info, "maxt"))
  {
    *dest = INT_MAX; // -ACB- 1999/09/22 Standards, Please.
    return;
  }

  s = strchr(info, 'T');
  if (!s)
    s = strchr(info, 't');
    
  if (s)
  {
    i = s-info;
    s = I_TmpMalloc(i + 1);
    Z_StrNCpy(s, info, i);
    RAD_CheckForInt(s, storage);
    I_TmpFree(s);
    return;
  }

  if (sscanf(info, "%f", &val) != 1)
  {
    I_Warning("Bad time value `%s'.\n", info);
    return;
  }

  *dest = (int)(val * (float)TICRATE);
}

static armour_type_e RAD_CheckForArmourType(const char *info)
{
  if (DDF_CompareName(info, "GREEN") == 0)
    return ARMOUR_Green;
  else if (DDF_CompareName(info, "BLUE") == 0)
    return ARMOUR_Blue;
  else if (DDF_CompareName(info, "YELLOW") == 0)
    return ARMOUR_Yellow;
  else if (DDF_CompareName(info, "RED") == 0)
    return ARMOUR_Red;
  
  RAD_Error("Unknown armour type: %s\n", info);
  return 0;
}

// RAD_UnquoteString
//
// Remove the quotes from the given string, returning a newly
// allocated string.

static char *RAD_UnquoteString(const char *s)
{
  int tokenlen = 0;

  // skip initial quote
  s++;

  while (*s != '"')
  {
#ifdef DEVELOPERS
    if (*s == 0)
      I_Error("INTERNAL ERROR: bad string.\n");
#endif
    
    // -AJA- 1999/09/07: check for \n. Only temporary, awaiting bison...
    if (s[0] == '\\' && toupper(s[1]) == 'N')
    {
      tokenbuf[tokenlen++] = '\n';
      s += 2;
      continue;
    }
    
    tokenbuf[tokenlen++] = *s++;
  }

  tokenbuf[tokenlen] = 0;
  return Z_StrDup(tokenbuf);
}

static boolean_t CheckForBoolean(const char *s)
{
  if (strcmp(s, "TRUE") == 0 || strcmp(s, "1") == 0)
    return true;

  if (strcmp(s, "FALSE") == 0 || strcmp(s, "0") == 0)
    return false;

  // Nope, it's an error.
  RAD_Error("Bad boolean value (should be TRUE or FALSE): %s\n", s);
  return false;
}

// DoParseWhenAppear
//
// -AJA- 1999/09/25: written (rather quickly).
// -AJA- FIXME: duplicated code, use DDF_MainGetWhenAppear someday.

static void DoParseWhenAppear(rad_script_t *scr, const char *skill)
{
  scr->appear = 0;
  
  if (strstr(skill, "1")) scr->appear |= 1 << 0;
  if (strstr(skill, "2")) scr->appear |= 1 << 1;
  if (strstr(skill, "3")) scr->appear |= 1 << 2;
  if (strstr(skill, "4")) scr->appear |= 1 << 3;
  if (strstr(skill, "5")) scr->appear |= 1 << 4;

  if (strstr(skill, "SP") || strstr(skill, "sp"))
    scr->appear |= WNAP_Single;
  if (strstr(skill, "COOP") || strstr(skill, "coop"))
    scr->appear |= WNAP_Coop;
  if (strstr(skill, "DM") || strstr(skill, "dm"))
    scr->appear |= WNAP_DeathMatch;
}

// AddStateToScript
//
// Adds a new action state to the tail of the current set of states
// for the given radius trigger.
//
// -AJA- 1999/10/23: written.

static void AddStateToScript(rad_script_t *R, int tics,
    void (* action)(struct rad_trigger_s *R, mobj_t *actor, void *param), 
    void *param)
{
  rts_state_t *state;

  state = Z_ClearNew(rts_state_t, 1);

  state->tics = tics;
  state->action = action;
  state->param = param;

  state->tics += pending_wait_tics;
  state->label = pending_label;

  pending_wait_tics = 0;
  pending_label = NULL;

  // link it in
  state->next = NULL;
  state->prev = R->last_state;

  if (R->last_state)
    R->last_state->next = state;
  else
    R->first_state = state;

  R->last_state = state;
}

//
// ClearOneScripts
//
static void ClearOneScript(rad_script_t *scr)
{
  Z_Free(scr->mapid);

  while (scr->boss_trig)
  {
    s_ondeath_t *cur = scr->boss_trig;
    scr->boss_trig = cur->next;

    Z_Free(cur);
  }
  
  while (scr->height_trig)
  {
    s_onheight_t *cur = scr->height_trig;
    scr->height_trig = cur->next;

    Z_Free(cur);
  }

  // free all states
  while (scr->first_state)
  {
    rts_state_t *cur = scr->first_state;
    scr->first_state = cur->next;

    if (cur->param)
      Z_Free(cur->param);
    
    Z_Free(cur);
  }
}

// ClearPreviousScripts
//
// Removes any radius triggers for a given map when start_map is used.
// Thus triggers in later RTS files/lumps replace those in earlier RTS
// files/lumps in the specified level.
// 
// -AJA- 1999/10/05: written.

static void ClearPreviousScripts(const char *mapid)
{
  rad_script_t *scr, *next;

  for (scr=r_scripts; scr; scr=next)
  {
    next = scr->next;

    if (strcmp(scr->mapid, mapid) == 0)
    {
      // unlink and free it
      if (scr->next)
        scr->next->prev = scr->prev;

      if (scr->prev)
        scr->prev->next = scr->next;
      else
        r_scripts = scr->next;
      
      ClearOneScript(scr);

      Z_Free(scr);
    }
  }
}

// ClearAllScripts
//
// Removes all radius triggers from all maps.
// 
// -AJA- 1999/10/05: written.

static void ClearAllScripts(void)
{
  while (r_scripts)
  {
    rad_script_t *scr = r_scripts;
    r_scripts = scr->next;

    ClearOneScript(scr);

    Z_Free(scr);
  }
}

// RAD_CollectParameters
//
// Collect the parameters from the line into an array of strings
// `pars', which can hold at most `max' string pointers.
// 
// -AJA- 1999/10/23: written.
// -AJA- 2000/01/02: Moved #define handling to here.

static void RAD_CollectParameters(const char *line, int *pnum, 
    char ** pars, int max)
{
  int tokenlen = -1;
  boolean_t in_string = false;

  *pnum = 0;

  for (;;)
  {
    int ch = *line;

    if (in_string)
    {
      if (ch == 0)
        RAD_Error("Nonterminated string found.\n");

      if (ch == '"')
        in_string = false;

      tokenbuf[tokenlen++] = ch;
      line++;
      continue;
    }

    if (tokenlen >= 0)
    {
      // end of token ?
      if (ch == 0 || isspace(ch) ||
          ch == ';' || (line[0] == '/' && line[1] == '/'))
      {
        tokenbuf[tokenlen] = 0;
        tokenlen = -1;

        if (*pnum >= max)
          RAD_Error("Too many tokens on line\n");
          
        // check for defines 
        if (! CheckForDefine(tokenbuf, &pars[*pnum]))
          pars[*pnum] = Z_StrDup(tokenbuf);

        *pnum += 1;

        // end of line ?
        if (ch == 0 || ch == ';' || (line[0] == '/' && line[1] == '/'))
          break;

        line++;
        continue;
      }

      tokenbuf[tokenlen++] = ch;
      line++;
      continue;
    }
    
    // end of line ?
    if (ch == 0 || ch == ';' || (line[0] == '/' && line[1] == '/'))
      break;

    if (isspace(ch))
    {
      line++;
      continue;
    }

    // string ?
    if (ch == '"')
      in_string = true;

    // must be token
    tokenbuf[0] = ch;
    tokenlen = 1;
    line++;
    continue;
  }
}

// RAD_FreeParameters
//
// Free previously collected parameters.
// 
// -AJA- 1999/10/23: written.

static void RAD_FreeParameters(int pnum, char **pars)
{
  while (pnum > 0)
  {
    Z_Free(pars[--pnum]);
  }
}


// ---- Primitive Parsers ----------------------------------------------

static void RAD_ParseVersion(int pnum, const char **pars)
{
  // #Version <vers>

  float vers;

  if (pnum != 2)
    RAD_Error("%s: Wrong number of parameters.\n", pars[0]);

  RAD_CheckForFloat(pars[1], &vers);
  
  if (vers > (float_t) PARSERV / (float_t) PARSERVFIX)
    RAD_Error("This version of EDGE cannot handle this script\n");
}

static void RAD_ParseClearAll(int pnum, const char **pars)
{
  // #ClearAll

  if (pnum != 1)
    RAD_Error("%s: Wrong number of parameters.\n", pars[0]);

  ClearAllScripts();
}

static void RAD_ParseDefine(int pnum, const char **pars)
{
  // #Define <identifier> <num>

  define_t *newdef;

  if (pnum != 3)
    RAD_Error("%s: Wrong number of parameters.\n", pars[0]);

  newdef = Z_ClearNew(define_t, 1);

  newdef->name  = Z_StrDup(pars[1]);
  newdef->value = Z_StrDup(pars[2]);

  // link it in
  newdef->next = defines;
  defines = newdef;
}

static void RAD_ParseStartMap(int pnum, const char **pars)
{
  // Start_Map <map>

  if (pnum != 2)
    RAD_Error("%s: Wrong number of parameters.\n", pars[0]);

  // -AJA- 1999/08/02: New scripts replace old ones.
  ClearPreviousScripts(pars[1]);

  this_map = Z_StrDup(pars[1]);

  rad_cur_level++;
}

static void RAD_ParseRadiusTrigger(int pnum, const char **pars)
{
  // RadiusTrigger <x> <y> <radius>
  // RadiusTrigger <x> <y> <radius> <low z> <high z>

  // -AJA- 1999/09/12: Reworked for having Z-restricted triggers.

  if (pnum != 4 && pnum != 6)
    RAD_Error("%s: Wrong number of parameters.\n", pars[0]);

  // Set the node up, from now on we can use rscript as it points
  // to the new node.

  this_rad = Z_ClearNew(rad_script_t, 1);

  // set defaults
  this_rad->z1 = (float_t) INT_MIN;
  this_rad->z2 = (float_t) INT_MAX;
  this_rad->z_mid = (float_t) M_PI;
  this_rad->appear = DEFAULT_APPEAR;
  this_rad->min_players = 0;
  this_rad->max_players = INT_MAX;
  this_rad->repeat_count = -1;
  this_rad->repeat_delay = 0;

  pending_wait_tics = 0;
  pending_label = NULL;

  RAD_CheckForFloat(pars[1], &this_rad->x);
  RAD_CheckForFloat(pars[2], &this_rad->y);
  RAD_CheckForFloat(pars[3], &this_rad->radius);

  if (pnum >= 6)
  {
    RAD_CheckForFloat(pars[4], &this_rad->z1);
    RAD_CheckForFloat(pars[5], &this_rad->z2);

    this_rad->z_mid = (this_rad->z1 + this_rad->z2) / 2.0;
  }

  // link it in
  this_rad->next = r_scripts;
  this_rad->prev = NULL;

  if (r_scripts)
    r_scripts->prev = this_rad;

  r_scripts = this_rad;

  rad_cur_level++;
}

static void RAD_ParseEndRadiusTrigger(int pnum, const char **pars)
{
  // End_RadiusTrigger

  if (pnum != 1)
    RAD_Error("%s: Wrong number of parameters.\n", pars[0]);

  if (rad_cur_level != 2)
    RAD_Error("%s: Missing RADIUSTRIGGER !\n", pars[0]);

  // handle any pending WAIT or LABEL values
  if (pending_wait_tics > 0 || pending_label)
  {
    AddStateToScript(this_rad, 0, RAD_ActNOP, NULL);
  }
  
  this_rad->mapid = Z_StrDup(this_map);
  this_rad = NULL;

  rad_itemsread++;
  rad_cur_level--;
}

static void RAD_ParseEndMap(int pnum, const char **pars)
{
  // End_Map

  if (pnum != 1)
    RAD_Error("%s: Wrong number of parameters.\n", pars[0]);

  if (rad_cur_level == 2)
    RAD_Error("%s found, but END_RADIUSTRIGGER missing !\n", pars[0]);

  if (rad_cur_level != 1)
    RAD_Error("%s: Missing START_MAP !\n", pars[0]);
  
  this_map = NULL;

  rad_cur_level--;
}

static void RAD_ParseName(int pnum, const char **pars)
{
  // Name <name>
  //
  // -AJA- 1999/09/25: added this primitive.

  if (pnum != 2)
    RAD_Error("%s: Wrong number of parameters.\n", pars[0]);

  if (this_rad->script_name)
    RAD_Error("Script already has a name: `%s'\n", this_rad->script_name);
    
  this_rad->script_name = Z_StrDup(pars[1]);
}

static void RAD_ParseTag(int pnum, const char **pars)
{
  // Tag <number>
  //
  // -AJA- 2000/01/09: added this primitive.

  if (pnum != 2)
    RAD_Error("%s: Wrong number of parameters.\n", pars[0]);

  if (this_rad->tag != 0)
    RAD_Error("Script already has a tag: `%d'\n", this_rad->tag);
   
  RAD_CheckForInt(pars[1], &this_rad->tag);
}

static void RAD_ParseWhenAppear(int pnum, const char **pars)
{
  // When_Appear 1:2:3:4:5:sp:coop:dm:extra

  if (pnum != 2)
    RAD_Error("%s: Wrong number of parameters.\n", pars[0]);

  DoParseWhenAppear(this_rad, pars[1]);
}

static void RAD_ParseWhenPlayerNum(int pnum, const char **pars)
{
  // When_Player_Num <num>
  // When_Player_Num <min> <max>

  if (pnum < 2 || pnum > 3)
    RAD_Error("%s: Wrong number of parameters.\n", pars[0]);

  RAD_CheckForInt(pars[1], &this_rad->min_players);
  this_rad->max_players = this_rad->min_players;

  if (pnum >= 3)
    RAD_CheckForInt(pars[2], &this_rad->max_players);

  if (this_rad->min_players < 0 || 
      this_rad->min_players > this_rad->max_players)
  {
    RAD_Error("Illegal playernum range: %d..%d\n",
        this_rad->min_players, this_rad->max_players);
  }
}

static void RAD_ParseTaggedRepeatable(int pnum, const char **pars)
{
  // Tagged_Repeatable <num repetitions>
  // Tagged_Repeatable <num repetitions> <delay>

  if (pnum < 2 || pnum > 3)
    RAD_Error("%s: Wrong number of parameters.\n", pars[0]);

  if (this_rad->repeat_count >= 0)
    RAD_Error("%s: can only be used once.\n", pars[0]);

  RAD_CheckForInt(pars[1], &this_rad->repeat_count);

  // -ES- 2000/03/03 Changed to RAD_CheckForTime.
  if (pnum >= 3)
    RAD_CheckForTime(pars[2], &this_rad->repeat_delay);
  else
    this_rad->repeat_delay = 1;
}

static void RAD_ParseTaggedUse(int pnum, const char **pars)
{
  // Tagged_Use

  if (pnum != 1)
    RAD_Error("%s: Wrong number of parameters.\n", pars[0]);

  this_rad->tagged_use = true;
}

static void RAD_ParseTaggedIndependent(int pnum, const char **pars)
{
  // Tagged_Independent

  if (pnum != 1)
    RAD_Error("%s: Wrong number of parameters.\n", pars[0]);

  this_rad->tagged_independent = true;
}

static void RAD_ParseTaggedImmediate(int pnum, const char **pars)
{
  // Tagged_Immediate

  if (pnum != 1)
    RAD_Error("%s: Wrong number of parameters.\n", pars[0]);

  this_rad->tagged_immediate = true;
}

static void RAD_ParseTaggedDisabled(int pnum, const char **pars)
{
  // Tagged_Disabled

  if (pnum != 1)
    RAD_Error("%s: Wrong number of parameters.\n", pars[0]);

  this_rad->tagged_disabled = true;
}

static void RAD_ParseTaggedPath(int pnum, const char **pars)
{
  // Tagged_Path <next node>
  //
  // -AJA- 1999/09/25: added this primitive.

  if (pnum != 2)
    RAD_Error("%s: Wrong number of parameters.\n", pars[0]);

  if (this_rad->next_in_path)
    RAD_Error("%s: Only one allowed per trigger.\n", pars[0]);
   
  this_rad->next_in_path = Z_StrDup(pars[1]);
}

static void RAD_ParseOnDeath(int pnum, const char **pars)
{
  // OnDeath <thing type>
  // OnDeath <thing type> <threshhold>

  s_ondeath_t *cond;

  if (pnum < 2 || pnum > 3)
    RAD_Error("%s: Wrong number of parameters.\n", pars[0]);

  cond = Z_ClearNew(s_ondeath_t, 1);
  cond->threshhold = 0;

  // get map thing
  if (pars[1][0] == '-' || pars[1][0] == '+' || isdigit(pars[1][0]))
  {
    RAD_CheckForInt(pars[1], &cond->thing_type);
  }
  else
    cond->thing_name = Z_StrDup(pars[1]);

  if (pnum >= 3)
  {
    RAD_CheckForInt(pars[2], &cond->threshhold);
  }

  // link it into list of ONDEATH conditions
  cond->next = this_rad->boss_trig;
  this_rad->boss_trig = cond;
}

static void RAD_ParseOnHeight(int pnum, const char **pars)
{
  // OnHeight <low Z> <high Z>
  // OnHeight <low Z> <high Z> <sector num>

  s_onheight_t *cond;

  if (pnum < 3 || pnum > 4)
    RAD_Error("%s: Wrong number of parameters.\n", pars[0]);

  cond = Z_ClearNew(s_onheight_t, 1);

  cond->sec_num = -1;

  RAD_CheckForFloat(pars[1], &cond->z1);
  RAD_CheckForFloat(pars[2], &cond->z2);
  
  if (cond->z1 > cond->z2)
    RAD_Error("%s: bad height range %1.1f..%1.1f\n", pars[0],
      cond->z1, cond->z2);

  // get sector reference
  if (pnum >= 4)
  {
    RAD_CheckForInt(pars[3], &cond->sec_num);
  }

  // link it into list of ONHEIGHT conditions
  cond->next = this_rad->height_trig;
  this_rad->height_trig = cond;
}

static void RAD_ParseLabel(int pnum, const char **pars)
{
  // Label <label>

  if (pnum != 2)
    RAD_Error("%s: Wrong number of parameters.\n", pars[0]);

  if (pending_label)
    RAD_Error("State already has a label: `%s'\n",
        pending_label);

  pending_label = Z_StrDup(pars[1]);
}

static void RAD_ParseEnableScript(int pnum, const char **pars)
{
  // Enable_Script  <script name>
  // Disable_Script <script name>
  
  s_enabler_t *t;

  if (pnum != 2)
    RAD_Error("%s: Wrong number of parameters.\n", pars[0]);

  t = Z_ClearNew(s_enabler_t, 1);

  t->script_name = Z_StrDup(pars[1]);
  t->new_disabled = DDF_CompareName("DISABLE_SCRIPT", pars[0]) == 0;

  AddStateToScript(this_rad, 0, RAD_ActEnableScript, t);
}

static void RAD_ParseEnableTagged(int pnum, const char **pars)
{
  // Enable_Tagged  <tag num>
  // Disable_Tagged <tag num>
  
  s_enabler_t *t;

  if (pnum != 2)
    RAD_Error("%s: Wrong number of parameters.\n", pars[0]);

  t = Z_ClearNew(s_enabler_t, 1);

  RAD_CheckForInt(pars[1], &t->tag);

  if (t->tag <= 0)
    RAD_Error("Bad tag value: %s\n", pars[1]);

  t->new_disabled = DDF_CompareName("DISABLE_TAGGED", pars[0]) == 0;

  AddStateToScript(this_rad, 0, RAD_ActEnableScript, t);
}

static void RAD_ParseExitLevel(int pnum, const char **pars)
{
  // ExitLevel
  // ExitLevel <wait tics>

  s_exit_t *exit;
  
  if (pnum < 1 || pnum > 2)
    RAD_Error("%s: Wrong number of parameters.\n", pars[0]);

  exit = Z_ClearNew(s_exit_t, 1);
  exit->exittime = 10;

  if (pnum >= 2)
  {
    RAD_CheckForTime(pars[1], &exit->exittime);
  }

  AddStateToScript(this_rad, 0, RAD_ActExitLevel, exit);
}

static void RAD_ParseSaveGame(int pnum, const char **pars)
{
  // SaveGame
  // SaveGame <slot>

  s_savegame_t *save;
  
  if (pnum < 1 || pnum > 2)
    RAD_Error("%s: Wrong number of parameters.\n", pars[0]);

  save = Z_ClearNew(s_savegame_t, 1);
  save->slot = -2;

  if (pnum >= 2)
  {
    RAD_CheckForInt(pars[1], &save->slot);
  }

  AddStateToScript(this_rad, 0, RAD_ActSaveGame, save);
}

static void RAD_ParseTip(int pnum, const char **pars)
{
  // Tip "<text>"
  // Tip "<text>" <time>
  // Tip "<text>" <time> <has sound>
  //
  // (likewise for Tip_LDF)
  // (likewise for Tip_Graphic)

  s_tip_t *tip;

  if (pnum < 2 || pnum > 4)
    RAD_Error("%s: Wrong number of parameters.\n", pars[0]);

  tip = Z_ClearNew(s_tip_t, 1);

  tip->display_time = 3 * TICRATE;
  tip->playsound = false;

  if (DDF_CompareName(pars[0], "TIP_GRAPHIC") == 0)
    tip->tip_graphic = Z_StrDup(pars[1]);
  else if (DDF_CompareName(pars[0], "TIP_LDF") == 0)
    tip->tip_ldf = Z_StrDup(pars[1]);
  else if (pars[1][0] == '"')
    tip->tip_text = RAD_UnquoteString(pars[1]);
  else
    RAD_Error("Needed string for TIP primitive.\n");

  if (pnum >= 3)
    RAD_CheckForTime(pars[2], &tip->display_time);

  if (pnum >= 4)
    tip->playsound = CheckForBoolean(pars[3]);

  AddStateToScript(this_rad, 0, RAD_ActTip, tip);
}

static void RAD_ParseSpawnThing(int pnum, const char **pars)
{
  // SpawnThing <thingid>
  // SpawnThing <thingid> <angle>
  // SpawnThing <thingid> <x> <y>
  // SpawnThing <thingid> <x> <y> <angle>
  // SpawnThing <thingid> <x> <y> <angle> <z>
  // SpawnThing <thingid> <x> <y> <angle> <z> <slope>
  //
  // (likewise for SpawnThing_Ambush)
  // (likewise for SpawnThing_Flash)
  //
  // -ACB- 1998/08/06 Use mobjinfo_t linked list
  // -AJA- 1999/09/11: Extra fields for Z and slope.

  // -AJA- 1999/09/11: Reworked for spawning things at Z.

  s_thing_t *t;
  const char *angle_str;
  int val;

  if (pnum < 2 || pnum > 7)
    RAD_Error("%s: Wrong number of parameters.\n", pars[0]);

  t = Z_ClearNew(s_thing_t, 1);

  // set defaults
  t->x = this_rad->x;
  t->y = this_rad->y;

  if (this_rad->z1 == (float_t) INT_MIN)
    t->z = ONFLOORZ;
  else
    t->z = this_rad->z1;

  t->ambush = DDF_CompareName("SPAWNTHING_AMBUSH", pars[0]) == 0;
  t->spawn_effect = DDF_CompareName("SPAWNTHING_FLASH", pars[0]) == 0;

  // get map thing
  if (pars[1][0] == '-' || pars[1][0] == '+' || isdigit(pars[1][0]))
  {
    RAD_CheckForInt(pars[1], &t->thing_type);
  }
  else
    t->thing_name = Z_StrDup(pars[1]);

  // get angle
  angle_str = (pnum == 3) ? pars[2] : 
              (pnum >= 5) ? pars[4] : NULL;

  if (angle_str) 
  {
    RAD_CheckForInt(angle_str, &val);

    if (ABS(val) <= 360)
      t->angle = FLOAT_2_ANG((float_t) val);
    else
      t->angle = val << 16;
  }

  // check for x & y, z, slope

  if (pnum >= 4)
  {
    RAD_CheckForFloat(pars[2], &t->x);
    RAD_CheckForFloat(pars[3], &t->y);
  }
  if (pnum >= 6)
  {
    RAD_CheckForFloat(pars[5], &t->z);
  }
  if (pnum >= 7)
  {
    RAD_CheckForFloat(pars[6], &t->slope);

    // FIXME: Merge with DDF_MainGetSlope someday.
    t->slope /= 45.0;
  }

  AddStateToScript(this_rad, 0, RAD_ActSpawnThing, t);
}

static void RAD_ParsePlaySound(int pnum, const char **pars)
{
  // PlaySound <soundid>
  // PlaySound <soundid> <x> <y>
  // PlaySound <soundid> <x> <y> <z>
  //
  // -AJA- 1999/09/12: Reworked for playing sound at specific Z.

  s_sound_t *t;

  if (pnum < 2 || pnum == 3 || pnum > 5)
    RAD_Error("%s: Wrong number of parameters.\n", pars[0]);

  t = Z_ClearNew(s_sound_t, 1);

  t->soundid = DDF_SfxLookupSound(pars[1]);

  t->x = this_rad->x;
  t->y = this_rad->y;
  t->z = (this_rad->z_mid == (float_t)M_PI) ? ONFLOORZ : this_rad->z_mid;

  if (pnum >= 4)
  {
    RAD_CheckForFloat(pars[2], &t->x);
    RAD_CheckForFloat(pars[3], &t->y);
  }

  if (pnum >= 5)
  {
    RAD_CheckForFloat(pars[4], &t->z);
  }

  AddStateToScript(this_rad, 0, RAD_ActPlaySound, t);
}

static void RAD_ParseKillSound(int pnum, const char **pars)
{
  // KillSound

  if (pnum != 1)
    RAD_Error("%s: Wrong number of parameters.\n", pars[0]);

  AddStateToScript(this_rad, 0, RAD_ActKillSound, NULL);
}

static void RAD_ParseDamagePlayer(int pnum, const char **pars)
{
  // DamagePlayer <amount>

  s_damagep_t *t;

  if (pnum != 2)
    RAD_Error("%s: Wrong number of parameters.\n", pars[0]);

  t = Z_ClearNew(s_damagep_t, 1);

  RAD_CheckForFloat(pars[1], &t->damage_amount);

  AddStateToScript(this_rad, 0, RAD_ActDamagePlayers, t);
}

static void RAD_ParseHealPlayer(int pnum, const char **pars)
{
  // HealPlayer <amount>
  // HealPlayer <amount> <limit>

  s_healp_t *heal;

  if (pnum < 2 || pnum > 3)
    RAD_Error("%s: Wrong number of parameters.\n", pars[0]);

  heal = Z_ClearNew(s_healp_t, 1);

  RAD_CheckForFloat(pars[1], &heal->heal_amount);

  if (pnum < 3)
    heal->limit = MAXHEALTH;
  else
    RAD_CheckForFloat(pars[2], &heal->limit);

  if (heal->limit < 0 || heal->limit > MAXHEALTH)
    RAD_Error("Health limit out of range: %1.1f\n", heal->limit);

  if (heal->heal_amount < 0 || heal->heal_amount > heal->limit)
    RAD_Error("Health value out of range: %1.1f\n", heal->heal_amount);

  AddStateToScript(this_rad, 0, RAD_ActHealPlayers, heal);
}

static void RAD_ParseGiveArmour(int pnum, const char **pars)
{
  // GiveArmour <type> <amount>
  // GiveArmour <type> <amount> <limit>
  
  s_armour_t *armour;

  if (pnum < 3 || pnum > 4)
    RAD_Error("%s: Wrong number of parameters.\n", pars[0]);

  armour = Z_ClearNew(s_armour_t, 1);

  armour->type = RAD_CheckForArmourType(pars[1]);

  RAD_CheckForFloat(pars[2], &armour->armour_amount);

  if (pnum < 4)
    armour->limit = MAXARMOUR;
  else
    RAD_CheckForFloat(pars[3], &armour->limit);

  if (armour->limit < 0 || armour->limit > MAXARMOUR)
    RAD_Error("Armour limit out of range: %1.1f\n", armour->limit);

  if (armour->armour_amount < 0 || armour->armour_amount > armour->limit)
    RAD_Error("Armour value out of range: %1.1f\n", armour->armour_amount);

  AddStateToScript(this_rad, 0, RAD_ActArmourPlayers, armour);
}

static void RAD_ParseDamageMonsters(int pnum, const char **pars)
{
  // Damage_Monsters <monster> <amount>

  s_damage_monsters_t *mon;

  if (pnum != 3)
    RAD_Error("%s: Wrong number of parameters.\n", pars[0]);

  mon = Z_ClearNew(s_damage_monsters_t, 1);

  // get monster type
  if (pars[1][0] == '-' || pars[1][0] == '+' || isdigit(pars[1][0]))
  {
    RAD_CheckForInt(pars[1], &mon->thing_type);
  }
  else if (DDF_CompareName(pars[1], "ANY") == 0)
    mon->thing_type = -1;
  else
    mon->thing_name = Z_StrDup(pars[1]);

  RAD_CheckForFloat(pars[2], &mon->damage_amount);
  
  AddStateToScript(this_rad, 0, RAD_ActDamageMonsters, mon);
}

static void RAD_ParseSkill(int pnum, const char **pars)
{
  // Skill <skill> <respawn> <fastmonsters>

  s_skill_t *skill;
  int val;

  if (pnum != 4)
    RAD_Error("%s: Wrong number of parameters.\n", pars[0]);

  skill = Z_ClearNew(s_skill_t, 1);

  RAD_CheckForInt(pars[1], &val);

  skill->skill = (skill_t) (val - 1);
  skill->Respawn = CheckForBoolean(pars[2]);
  skill->FastMonsters = CheckForBoolean(pars[3]);

  AddStateToScript(this_rad, 0, RAD_ActSkill, skill);
}

static void RAD_ParseGotoMap(int pnum, const char **pars)
{
  // GotoMap <map name>

  s_gotomap_t *go;

  if (pnum != 2)
    RAD_Error("%s: Wrong number of parameters.\n", pars[0]);

  go = Z_ClearNew(s_gotomap_t, 1);

  go->map_name = Z_StrDup(pars[1]);

  AddStateToScript(this_rad, 0, RAD_ActGotoMap, go);
}

static void RAD_ParseSectorMove(int pnum, const char **pars)
{
  // SectorV <sector> <amount> <ceil or floor>

  s_sectorv_t *secv;

  if (pnum != 4)
  {
    RAD_Error("%s: Wrong number of parameters.\n", pars[0]);
    return;
  }
  
  secv = Z_ClearNew(s_sectorv_t, 1);

  RAD_CheckForInt(pars[1], &secv->secnum);
  RAD_CheckForFloat(pars[2], &secv->increment);

  if (DDF_CompareName(pars[3], "FLOOR") == 0)
    secv->corf = 1;
  else if (DDF_CompareName(pars[3], "CEILING") == 0)
    secv->corf = 0;
  else
    secv->corf = CheckForBoolean(pars[3]);

  AddStateToScript(this_rad, 0, RAD_ActSectorMove, secv);
}

static void RAD_ParseSectorLight(int pnum, const char **pars)
{
  // SectorL <sector> <amount>

  s_sectorl_t *secl;

  if (pnum != 3)
    RAD_Error("%s: Wrong number of parameters.\n", pars[0]);

  secl = Z_ClearNew(s_sectorl_t, 1);

  RAD_CheckForInt(pars[1], &secl->secnum);
  RAD_CheckForInt(pars[2], &secl->increment);

  AddStateToScript(this_rad, 0, RAD_ActSectorLight, secl);
}

static void RAD_ParseActivateLinetype(int pnum, const char **pars)
{
  // Activate_LineType <linetype> <tag>
  //
  // -AJA- 1999/10/21: added this primitive.

  s_lineactivator_t *lineact;

  if (pnum != 3)
    RAD_Error("%s: Wrong number of parameters.\n", pars[0]);

  lineact = Z_ClearNew(s_lineactivator_t, 1);

  RAD_CheckForInt(pars[1], &lineact->typenum);
  RAD_CheckForInt(pars[2], &lineact->tag);

  AddStateToScript(this_rad, 0, RAD_ActActivateLinetype, lineact);
}

static void RAD_ParseWait(int pnum, const char **pars)
{
  // Wait <time>
  //
  // -AJA- 2000/01/05: added this primitive.

  int tics;

  if (pnum != 2)
    RAD_Error("%s: Wrong number of parameters.\n", pars[0]);

  RAD_CheckForTime(pars[1], &tics);

  if (tics <= 0)
    RAD_Error("%s: Invalid time: %d\n", pars[0], tics);
    
  pending_wait_tics += tics;
}

static void RAD_ParseJump(int pnum, const char **pars)
{
  // Jump <label>
  // Jump <label> <random chance>
  //
  // -AJA- 2000/01/05: added this primitive.

  s_jump_t *jump;

  if (pnum < 2 || pnum > 3)
    RAD_Error("%s: Wrong number of parameters.\n", pars[0]);

  jump = Z_ClearNew(s_jump_t, 1);

  jump->label = Z_StrDup(pars[1]);

  if (pnum >= 3)
    RAD_CheckForPercent(pars[2], &jump->random_chance);

  AddStateToScript(this_rad, 0, RAD_ActJump, jump);
}

static void RAD_ParseSleep(int pnum, const char **pars)
{
  // Sleep
  //
  // -AJA- 2000/01/09: added this primitive.

  if (pnum != 1)
    RAD_Error("%s: Wrong number of parameters.\n", pars[0]);

  AddStateToScript(this_rad, 0, RAD_ActSleep, NULL);
}


//  PARSER TABLE

static rts_parser_t radtrig_parsers[] =
{
  // directives...
  {-1,"#DEFINE", RAD_ParseDefine},
  {0,"#VERSION", RAD_ParseVersion},
  {0,"#CLEARALL", RAD_ParseClearAll},

  // basics...
  {0,"START_MAP", RAD_ParseStartMap},
  {1,"RADIUSTRIGGER", RAD_ParseRadiusTrigger},
  {-1,"END_RADIUSTRIGGER", RAD_ParseEndRadiusTrigger},
  {-1,"END_MAP", RAD_ParseEndMap},
  
  // properties...
  {2,"NAME", RAD_ParseName},
  {2,"TAG", RAD_ParseTag},
  {2,"WHEN_APPEAR", RAD_ParseWhenAppear},
  {2,"WHEN_PLAYER_NUM", RAD_ParseWhenPlayerNum},
  {2,"TAGGED_REPEATABLE", RAD_ParseTaggedRepeatable},
  {2,"TAGGED_USE", RAD_ParseTaggedUse},
  {2,"TAGGED_INDEPENDENT", RAD_ParseTaggedIndependent},
  {2,"TAGGED_IMMEDIATE", RAD_ParseTaggedImmediate},
  {2,"TAGGED_PATH", RAD_ParseTaggedPath},
  {2,"TAGGED_DISABLED", RAD_ParseTaggedDisabled},
  {2,"ONDEATH", RAD_ParseOnDeath},
  {2,"ONHEIGHT", RAD_ParseOnHeight},
  {2,"LABEL", RAD_ParseLabel},

  // actions...
  {2,"TIP", RAD_ParseTip},
  {2,"TIP_LDF", RAD_ParseTip},
  {2,"TIP_GRAPHIC", RAD_ParseTip},
  {2,"EXITLEVEL", RAD_ParseExitLevel},
  {2,"SAVEGAME", RAD_ParseSaveGame},
  {2,"SPAWNTHING", RAD_ParseSpawnThing},
  {2,"SPAWNTHING_AMBUSH", RAD_ParseSpawnThing},
  {2,"SPAWNTHING_FLASH", RAD_ParseSpawnThing},
  {2,"PLAYSOUND", RAD_ParsePlaySound},
  {2,"KILLSOUND", RAD_ParseKillSound},
  {2,"DAMAGEPLAYER", RAD_ParseDamagePlayer},
  {2,"HEALPLAYER", RAD_ParseHealPlayer},
  {2,"GIVEARMOUR", RAD_ParseGiveArmour},
  {2,"DAMAGE_MONSTERS", RAD_ParseDamageMonsters},
  {2,"SKILL", RAD_ParseSkill},
  {2,"GOTOMAP", RAD_ParseGotoMap},
  {2,"SECTORV", RAD_ParseSectorMove},
  {2,"SECTORL", RAD_ParseSectorLight},
  {2,"ENABLE_SCRIPT", RAD_ParseEnableScript},
  {2,"DISABLE_SCRIPT", RAD_ParseEnableScript},
  {2,"ENABLE_TAGGED", RAD_ParseEnableTagged},
  {2,"DISABLE_TAGGED", RAD_ParseEnableTagged},
  {2,"ACTIVATE_LINETYPE", RAD_ParseActivateLinetype},
  {2,"WAIT", RAD_ParseWait},
  {2,"JUMP", RAD_ParseJump},
  {2,"SLEEP", RAD_ParseSleep},

  // that's all, folks.
  {0, NULL, NULL}
};

// Primitive Parser
//
void RAD_ParseLine(char *s)
{
  int pnum;
  char *pars[16];
  rts_parser_t *cur;

  RAD_ErrorSetLineData(s);

  RAD_CollectParameters(s, &pnum, pars, 16);

  if (pnum == 0)
  {
    RAD_ErrorClearLineData();
    return;
  }

  for (cur = radtrig_parsers; cur->name != NULL; cur++)
  {
    if (DDF_CompareName(pars[0], cur->name) != 0)
      continue;

    // check level
    if (cur->level >= 0)
    {
      if (cur->level != rad_cur_level)
        RAD_Error("Primitive `%s' used in wrong place "
          "(found in %s, but should be in %s).\n", pars[0],
          rad_level_names[rad_cur_level],
          rad_level_names[cur->level]);
    }
    
    // found it, invoke the parser function
    (* cur->parser)(pnum, (const char **) pars);

    RAD_FreeParameters(pnum, pars);
    RAD_ErrorClearLineData();

    return;
  }

  RAD_Error("Unknown primitive: %s\n", pars[0]);
}

void RAD_ParserBegin(void)
{
  rad_cur_level = 0;
}

void RAD_ParserDone(void)
{
  if (rad_cur_level >= 2)
    RAD_Error("RADIUSTRIGGER: block not terminated !\n");

  if (rad_cur_level == 1)
    RAD_Error("START_MAP: block not terminated !\n");
}

