//
// EDGE Sound & Music Utility File
//
// Currently holds the MUS -> MIDI Conversion Code
//
// -ACB- 1999/10/21 Written
//

#include "i_defs.h"
#include "s_sound.h"

#include "z_zone.h"

#define MIDINOTE_VOLECITY 0x64
#define MIDEVENTMAXSIZE 4
#define MIDCHANDRUM 0x9
#define MUSCHANDRUM 0xF
#define MIDIDELTATICS 89

// MIDI Header MAGIC
const byte midihead[8] = { 0x4D, 0x54, 0x68, 0x64, 0x00, 0x00, 0x00, 0x06 };

// MIDI Header positions
typedef enum
{
  midi_string = 0,
  midi_type   = 8,
  midi_tracks = 10,
  midi_dtics  = 12,
  midi_data   = 14
}
midiheader_e;

// MIDI Header Data Struct
typedef struct midiinfo_s
{
  int type;
  int tracknum;
  int dtics;
  struct miditrack_s *track;
}
midiinfo_t;

// MIDI Track MAGIC
const byte trackhead[4] = { 0x4D, 0x54, 0x72, 0x6B };

// MIDI Track Header Position
typedef enum
{
  midi_track_string = 0,
  midi_track_length = 4
}
miditrackhead_e;

// MIDI Track Data Struct
typedef struct miditrack_s
{
  int length;
  byte *data;
  int volecity;
  int deltatime;
}
miditrack_t;

// MUS Header MAGIC
const char musmagic[4] = { 0x4D, 0x55, 0x53, 0x1A };

// MUS Header positions
typedef enum
{
  mus_string     = 0,
  mus_scorelen   = 4,
  mus_scorestart = 6,
  mus_channels   = 8,
  mus_instrnum   = 12,
  mus_instrlist  = 16
}
mushead_e;

// MUS Data Struct
typedef struct musinfo_s
{
  int scorelen;
  int scorestart;
  int channels;
  int instrnum;
  byte *instrlist;
  int currentpos;
}
musinfo_t;

typedef enum
{
  MUS_NOTERELEASE   = 0,
  MUS_NOTEPRESS     = 1,
  MUS_PITCHWHEEL    = 2,
  MUS_COMMANDCHANGE = 3,
  MUS_CONTROLCHANGE = 4,
  MUS_UNKNOWNEVENT1 = 5,
  MUS_ENDOFSCORE    = 6,
  MUS_UNKNOWNEVENT2 = 7,
  ENDOFMUS
}
musevent_e;

typedef enum
{
  MIDI_NOTERELEASE   = 0x80,
  MIDI_NOTEPRESS     = 0x90,
  MIDI_COMMANDCHANGE = 0xB0,
  MIDI_PATCHCHANGE   = 0xC0,
  MIDI_PITCHWHEEL    = 0xE0,
  ENDOFMIDI
}
midevent_e;

musinfo_t *musinfo;
midiinfo_t *midiinfo;
int *mus2midichannel;
int *midichan2track;
int miditrackcount;

// ===================Internal====================

//
// GetTimeInfo
//
// This retreives time from the data. Each value is a 7-bit, the
// 8th bit is not set on the last 7-bit number. This is calc-ed by
// by:
//
// 1) Set time value to 0
//
// 2) Read a byte value
//
// 3) Multiply previous time (if any) by 128 and added the byte value - ignoring
//    the MSB.
//
// 4) Check the MSB, if set read th next byte - go to step 2.
//
// The value in your time value is the time returned.
//
//
static int GetTimeInfo(byte *data, int *offset)
{
  int time;

  time=0;

  while (data[*offset] & 0x80)
  {
    time = (time*0x80) + (data[*offset] & ~0x80);
    (*offset)++;
  }

  time=(time*0x80)+data[*offset];
  (*offset)++;

  return time;
}

//
// WriteTrackTimeInfo
//
// Write to a track a number using the method of storage above.
//
static void WriteTrackTimeInfo(miditrack_t *track)
{
  byte buffer;
  int time;

  time = track->deltatime;

  if (!time)
  {
    // Increase track length
    track->length++;

    // Add to data
    track->data[track->length-1] = 0x80;

    // Clear track delta time
    track->deltatime = 0;
    return;
  }

  // We need to break down time
  while (time)
  {
    // Get the 7-bit chunk
    buffer = time&0x7F;

    // Increase track length
    track->length++;

    // Add to data
    track->data[track->length-1] = (buffer+0x80);

    // bitshift out the read data
    time = time >> 7;
  }

  // Get rid of the last MSB: this is our last byte.
  track->data[track->length-1] -= 0x80;

  // Clear track delta time
  track->deltatime = 0;
}

//
// ReadMUSHeaderFile
//
static boolean_t ReadMUSHeaderFile(byte *data)
{
  int i;

  for (i=0; i<4; i++)
  {
    if (data[mus_string+i] != musmagic[i])
      return false;
  }

  musinfo->scorelen   = data[mus_scorelen] + (data[mus_scorelen+1]<<8);
  musinfo->scorestart = data[mus_scorestart] + (data[mus_scorestart+1]<<8);
  musinfo->channels   = data[mus_channels] + (data[mus_channels+1]<<8);
  musinfo->instrnum   = data[mus_instrnum] + (data[mus_instrnum+1]<<8);

  // Check for too many channels
  if (musinfo->channels > 15)
    return false;

  musinfo->instrlist = Z_ClearNew(byte, musinfo->instrnum);

  i=0;
  while (i<musinfo->instrnum)
  {
    musinfo->instrlist[i] = data[mus_instrlist+(i*2)];
    i++;
  }

  return true;
}

//
// GetAvailMIDIChan
//
static int GetAvailMIDIChan(void)
{
  int allocchans;
  int olddrumval;
  int i;

  allocchans = -1;

  olddrumval = mus2midichannel[15];
  for (i=0; i<musinfo->channels; i++)
  {
    if (mus2midichannel[i] > allocchans)
      allocchans = mus2midichannel[i];
  }
  mus2midichannel[15] = olddrumval;

  return (allocchans==8)?(10):(allocchans+1);
}

//
// AddMIDIEvent
//
static void AddMIDIEvent(int tracknum, byte *data, int datalen)
{
  miditrack_t *track;
  int i;

  track = &midiinfo->track[tracknum];

  // Write the time for this event
  WriteTrackTimeInfo(track);

  // Add the data
  for (i=0; i<datalen; i++)
    track->data[track->length+i] = data[i];

  // set the length
  track->length += datalen;

  return;
}

//
// GetMUSEvent
//
static boolean_t GetMUSEvent(int *offset, byte *data)
{
  boolean_t last;
  int channelnum;
  int midichan, miditrackhandle;
  byte eventtype;
  byte *museventdata;
  byte mideventdata[MIDEVENTMAXSIZE];

  //
  // MUS Event Descriptor:
  //
  // bits 0,1,2,3 - Channel Number
  // bits 4,5,6   - Event type
  // bit 7        - Last Event before time delay
  //

  // Get the last event bit
  last = (data[*offset]&0x80) ? true : false;

  // Get the Event type
  eventtype = (data[*offset]&0x70)>>4;

  // Get the Channel number
  channelnum = data[*offset]&0xF;
  if (mus2midichannel[channelnum] == -1)
  {
    I_Printf("Called for %d\n",channelnum);
    mus2midichannel[channelnum] = (channelnum==15?9:GetAvailMIDIChan());
    midichan2track[mus2midichannel[channelnum]] = miditrackcount++;
  }

  midichan = mus2midichannel[channelnum];
  miditrackhandle = midichan2track[midichan];

  // Get event data
  (*offset)++;
  museventdata = &data[*offset];
  Z_Clear(mideventdata, byte, MIDEVENTMAXSIZE);

  switch(eventtype)
  {
    //
    // Release this note
    //
    case MUS_NOTERELEASE:
    {
      mideventdata[0] = MIDI_NOTERELEASE | midichan;
      mideventdata[1] = (museventdata[0] & 0x7F); // just in case, kill MSB
      mideventdata[2] = MIDINOTE_VOLECITY;
      AddMIDIEvent(miditrackhandle, mideventdata, 3);
      (*offset)++; // we read a byte - move on.
      break;
    }

    //
    // Press this note
    //
    case MUS_NOTEPRESS:
    {
      mideventdata[0] = MIDI_NOTEPRESS | midichan;
      mideventdata[1] = (museventdata[0] & 0x7F);

      // Check for volecity change
      if (museventdata[0] >= 0x80)
      {
        midiinfo->track[miditrackhandle].volecity = (museventdata[1] & 0x7F);
        (*offset)++;
      }

      mideventdata[2] = midiinfo->track[miditrackhandle].volecity;
      AddMIDIEvent(miditrackhandle, mideventdata, 3);
      (*offset)++;
      break;
    }

    //
    // Pitch Changes
    //
    case MUS_PITCHWHEEL:
    {
      mideventdata[0] = MIDI_PITCHWHEEL | midichan;
      mideventdata[1] = (museventdata[0]&0x01)<<6;
      mideventdata[2] = (museventdata[0])>>1;
      AddMIDIEvent(miditrackhandle, mideventdata, 3);
      (*offset)++;
      break;
    }

    //
    // Command Changes
    //
    case MUS_COMMANDCHANGE:
    {
      mideventdata[0] = MIDI_COMMANDCHANGE | midichan;

      switch(museventdata[0])
      {
        case 10: mideventdata[1] = 0x78; break; // All sounds off
        case 11: mideventdata[1] = 0x7B; break; // All notes off
        case 12: mideventdata[1] = 0x7E; break; // Mono
        case 13: mideventdata[1] = 0x7F; break; // Poly
        case 14: mideventdata[1] = 0x79; break; // Reset all controllers
        default: I_Error("GetMUSEvent: Corrupt MUS File - MUS_COMMANDCHANGE"); break;
      }
      mideventdata[2]=(museventdata[0]==12)?15:0;
      AddMIDIEvent(miditrackhandle, mideventdata, 3);
      (*offset)++;
      break;
    }

    //
    // Instrument, Pan or Volume Change
    //
    case MUS_CONTROLCHANGE:
    {
      if (!museventdata[0]) // Instrument Patch
      {
        mideventdata[0] = MIDI_PATCHCHANGE | midichan;
        mideventdata[1] = museventdata[1];

        AddMIDIEvent(miditrackhandle, mideventdata, 2);
        (*offset)+=2;
      }
      else // control change
      {
        mideventdata[0] = MIDI_COMMANDCHANGE | midichan;

        switch(museventdata[0])
        {
          case 1: mideventdata[1] = 0x00; break; // Bank Select
          case 2: mideventdata[1] = 0x01; break; // Modulation 
          case 3: mideventdata[1] = 0x07; break; // Volume
          case 4: mideventdata[1] = 0x0A; break; // Pan 
          case 5: mideventdata[1] = 0x0B; break; // Expression 
          case 6: mideventdata[1] = 0x5B; break; // Reverb Depth
          case 7: mideventdata[1] = 0x5D; break; // Chorus Depth
          case 8: mideventdata[1] = 0x40; break; // Sustain Pedal
          case 9: mideventdata[1] = 0x43; break; // Soft Pedal
          default: I_Error("GetMUSEvent: Corrupt MUS File - MUS_CONTROLCHANGE"); break;
        }

        mideventdata[2] = museventdata[1];
        AddMIDIEvent(miditrackhandle, mideventdata, 3);
        (*offset)+=2;
      }

      break;
    }

    // End of Score
    case MUS_ENDOFSCORE:
      return false;

    case MUS_UNKNOWNEVENT1:
    case MUS_UNKNOWNEVENT2:
    default:
      I_Error("GetMUSEvent: MUS Event %d Unknown - corrupt data assumed\n", eventtype);
      break;
  }

  return last;
}
// ===============End of Internals================

//
// S_UtilConvertMUStoMIDI
//
byte *S_UtilConvertMUStoMIDI(byte *data)
{
  int i,j;
  int totaltime;
  int deltatime;
  FILE *midifile;
  byte tracks;
  miditrack_t *track;
  int pos;
  byte *mididata;

  musinfo = Z_ClearNew(musinfo_t, 1);

  if (!ReadMUSHeaderFile(data))
    return NULL;

  midiinfo = Z_ClearNew(midiinfo_t, 1);
  midiinfo->track = Z_ClearNew(miditrack_t, 16);
  mus2midichannel = Z_ClearNew(int, musinfo->channels);
  midichan2track = Z_ClearNew(int, 16);
  miditrackcount=0;

  for (i=0; i<musinfo->channels; i++)
    mus2midichannel[i] = -1;

  for (i=0; i<16; i++)
  {
    // set volecity
    midiinfo->track[i].volecity = MIDINOTE_VOLECITY;

    // Alloc for maximum possible size
    midiinfo->track[i].data = Z_New(byte, musinfo->scorelen);
  }

  musinfo->currentpos = musinfo->scorestart;

  while (musinfo->currentpos < (musinfo->scorestart+musinfo->scorelen))
  {
    // returns true, we need to get time info
    if (GetMUSEvent(&musinfo->currentpos, data))
    {
      deltatime = GetTimeInfo(data, &musinfo->currentpos);
      totaltime += deltatime;

      for (i=0; i<16; i++)
        midiinfo->track[i].deltatime += deltatime;
    }
  }

  tracks=0;
  for (i=0; i<16; i++)
  {
    if (midiinfo->track[i].length)
      tracks++;
  }

  mididata = Z_ClearNew(byte, 65536*2);

  // Write Header
  // String
  for(i=0; i<8; i++)
    mididata[midi_string+i] = midihead[i];

  // Type - Always synchronous (1)
  mididata[midi_type] = 0;
  mididata[midi_type+1] = 1;

  // Track count
  mididata[midi_tracks] = 0;
  mididata[midi_tracks+1] = tracks;

  // Deltatics
  mididata[midi_dtics] = 0;
  mididata[midi_dtics+1] = MIDIDELTATICS;
  pos = midi_data; //always 14 or bigger...

  for(i=0; i<16; i++)
  {
    track = &midiinfo->track[i];

    if (track->length)
    {
      // Track Header
      for (j=0; j<4; j++)
        mididata[pos+j] = trackhead[j];

      pos +=4;

      // Track size - only limited to 64k so we call ignore the first two bytes
      mididata[pos] = 0; pos++;
      mididata[pos] = 0; pos++;
      mididata[pos] = ((track->length&0xFF00)>>8); pos++;
      mididata[pos] = (track->length&0xFF); pos++;

      // Track Data
      Z_MoveData(&mididata[pos], track->data, byte, track->length);
      pos+=track->length;

      I_Printf("length: %d\n", pos);

      // Free Memory for midi track - unneeded now
      Z_Free(track->data);
    }
  }

  // Free Memory for midi tracks - unneeded now
  Z_Free(midiinfo->track);

  I_Printf("length: %d\n", pos);
  midifile = fopen("E:\\EDGE\\DJGPP\\NEW.MID","wb+");
  if (!midifile)
    I_Error("Can't create file");

  fwrite(mididata, sizeof(byte), pos, midifile);
  fclose(midifile);

  return mididata;
}
