//----------------------------------------------------------------------------
//  EDGE New SaveGame Handling (Utility routines)
//----------------------------------------------------------------------------
// 
//  Copyright (c) 1999-2000  The EDGE Team.
// 
//  This program is free software; you can redistribute it and/or
//  modify it under the terms of the GNU General Public License
//  as published by the Free Software Foundation; either version 2
//  of the License, or (at your option) any later version.
//
//  This program is distributed in the hope that it will be useful,
//  but WITHOUT ANY WARRANTY; without even the implied warranty of
//  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//  GNU General Public License for more details.
//
//----------------------------------------------------------------------------
//
//  Based on the DOOM source code, released by Id Software under the
//  following copyright:
//
//    Copyright (C) 1993-1996 by id Software, Inc.
//
//----------------------------------------------------------------------------
//
// -AJA- 1999/09/26: Wrote this file.
//

#include "i_defs.h"
#include "sv_game.h"

#include "dm_state.h"
#include "p_local.h"
#include "p_spec.h"
#include "r_state.h"
#include "w_wad.h"
#include "wi_stuff.h"
#include "z_zone.h"

#include <math.h>


static byte *save_buffer;
static int save_p, save_size;

#define STRING_MARKER   0xAA
#define NULLSTR_MARKER  0xDE


// -------------------------------------------------------------
//
// initialisation & termination
//

void SV_BeginRead(byte *start, int len) 
{ 
  save_buffer = start;
  save_p = 0;
  save_size = len;
}

void SV_EndRead(void) 
{ 
  // nothing to do
}

void SV_BeginWrite(void) 
{ 
  save_buffer = NULL;
  save_p = save_size = 0;
} 

void SV_QueryWrite(byte **start, int *len) 
{ 
  *start = save_buffer;
  *len = save_size;
}

void SV_EndWrite(void)
{ 
  Z_Free(save_buffer);

  save_buffer = NULL;
}

// -------------------------------------------------------------
//
// basic R/W primitives
//

void SV_PutByte(unsigned char value) 
{
  if (save_p == save_size)
  {
    save_size += 1024;
    Z_Resize(save_buffer, byte, save_size);
  }

  save_buffer[save_p++] = value;
}

unsigned char SV_GetByte(void) 
{ 
  if (save_p == save_size)
    I_Error("Bad savegame (out of data).\n");

  return save_buffer[save_p++];
}

void SV_PutMarker(char id[4]) 
{
  int i;

  for (i=0; i < 4; i++)
    SV_PutByte((unsigned char) id[i]);
}

void SV_CheckMarker(char id[4]) 
{ 
  int i;

  for (i=0; i < 4; i++)
    if (SV_GetByte() != (unsigned char) id[i])
      I_Error("Bad savegame (invalid marker).\n");
}

// -------------------------------------------------------------
//
// basic datatypes
//

void SV_PutShort(unsigned short value) 
{
  SV_PutByte(value & 0xff);
  SV_PutByte(value >> 8);
}

void SV_PutInt(unsigned int value) 
{
  SV_PutShort(value & 0xffff);
  SV_PutShort(value >> 16);
}

#ifdef HAVE_64BIT_TYPE
void SV_PutLong(unsigned long long value) 
{
  SV_PutInt(value & 0xffffffff);
  SV_PutInt(value >> 32);
}
#endif

unsigned short SV_GetShort(void) 
{ 
  return SV_GetByte() | (SV_GetByte() << 8);
}

unsigned int SV_GetInt(void) 
{ 
  return SV_GetShort() | (SV_GetShort() << 16);
}

#ifdef HAVE_64BIT_TYPE
unsigned long long SV_GetLong(void) 
{
  return SV_GetInt() | (SV_GetInt() << 32);
}
#endif

// -------------------------------------------------------------
//
// fixed point & angles
//

void SV_PutFixed(fixed_t value) 
{ 
  SV_PutInt((unsigned int) value);
}

void SV_PutAngle(angle_t value) 
{ 
  SV_PutInt((unsigned int) value);
}

fixed_t SV_GetFixed(void) 
{
  return (fixed_t) SV_GetInt();
}

angle_t SV_GetAngle(void) 
{
  return (angle_t) SV_GetInt();
}

// -------------------------------------------------------------
//
// floating point
//

void SV_PutFloat(float value) 
{ 
  int exp;
  int mant;
  boolean_t neg;

  neg = (value < 0.0);
  value = fabs(value);

  mant = (int) ldexp(frexp(value, &exp), 30);

  SV_PutShort(exp);
  SV_PutInt((unsigned int) (neg ? -mant : mant));
}

float_t SV_GetFloat(void) 
{ 
  int exp;
  int mant;

  exp = SV_GetShort();
  mant = (int) SV_GetInt();

  return ldexp((float) mant, -30 + exp);
}

// -------------------------------------------------------------
//
// strings
//

void SV_PutString(char *str) 
{ 
  if (str == NULL)
  {
    SV_PutByte(NULLSTR_MARKER);
    return;
  }
  
  SV_PutByte(STRING_MARKER);
  SV_PutShort(strlen(str));

  for (; *str; str++)
    SV_PutByte(*str);
}

char *SV_GetString(void) 
{
  int len, i;
  char *result;
  
  int type = SV_GetByte();

  if (type == NULLSTR_MARKER)
    return NULL;
  
  if (type != STRING_MARKER)
    I_Error("Bad savegame (invalid string).\n");

  len = SV_GetShort();

  result = Z_New(char, len + 1);
  result[len] = 0;

  for (i = 0; i < len; i++)
    result[i] = (char) SV_GetByte();

  return result;
}

void SV_FreeString(char *str)
{
  Z_Free(str);
}
