//----------------------------------------------------------------------------
//  Texture Conversion and Caching code
//----------------------------------------------------------------------------
// 
//  Copyright (c) 1999-2000  The EDGE Team.
// 
//  This program is free software; you can redistribute it and/or
//  modify it under the terms of the GNU General Public License
//  as published by the Free Software Foundation; either version 2
//  of the License, or (at your option) any later version.
//
//  This program is distributed in the hope that it will be useful,
//  but WITHOUT ANY WARRANTY; without even the implied warranty of
//  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//  GNU General Public License for more details.
//
//----------------------------------------------------------------------------
//
// This module converts image lumps on disk to usable structures, and also
// provides a caching system for these.
//
// -ES- 2000/02/12 Written.

#include "i_defs.h"
#include "w_textur.h"

#include "e_search.h"
#include "dm_state.h"
#include "dm_defs.h"
#include "m_argv.h"
#include "m_swap.h"
#include "p_local.h"
#include "r_local.h"
#include "r_sky.h"
#include "r2_defs.h"
#include "v_colour.h"
#include "w_wad.h"
#include "z_zone.h"

//---------------------------------------------------------------------
// Graphics:
// ^^^^^^^^^
// DOOM graphics for walls and sprites is stored in vertical runs of
// opaque pixels (posts).
//
// A column is composed of zero or more posts, a patch or sprite is
// composed of zero or more columns.
//---------------------------------------------------------------------

struct texturedef_s;

typedef struct tex_cache_header_s
{
  int users;
  struct texturedef_s *texture;
  byte *composite;
  struct tex_cache_header_s *prev, *next;
}
texcache_header_t;

//
// Texture definition.
//
// Each texture is composed of one or more patches,
// with patches being lumps stored in the WAD.
//
// The lumps are referenced by number, and patched into the rectangular
// texture space using origin and possibly other attributes.
//
typedef struct
{
  short originx;
  short originy;
  short patch;
  short stepdir;
  short colourmap;
}
mappatch_t;

//
// Texture definition.
//
// A DOOM wall texture is a list of patches which are to be combined in a
// predefined order.
//
// Removing the obsolete columndirectory fails because this defines how a
// texture is stored in the wad file.
//
typedef struct
{
  char name[8];
  boolean_t masked;
  short width;
  short height;
  int columndirectory;
  short patchcount;
  mappatch_t patches[1];
}
maptexture_t;

//
// A single patch from a texture definition, basically a rectangular area
// within the texture rectangle.
//
// Note: Block origin (always UL), which has already accounted
// for the internal origin of the patch.
//
typedef struct
{
  int originx;
  int originy;
  int patch;
  boolean_t cache;
  // if the texture is cached, and cache==true, this points to the patch lump.
  const byte *lump;
}
texpatch_t;

//
// A texturedef_t describes a rectangular texture, which is composed of
// one or more mappatch_t structures that arrange graphic patches.
//
typedef struct texturedef_s
{
  // Keep name for switch changing, etc.
  char name[8];
  short width;
  short height;

  int compositesize;
  int widthmask;

  // shows which patch to draw for each column (index to patches[]).
  // 254 means that GenerateLookup must be called, 255 means that a composite
  // should be used.
  byte *columnlump;
  unsigned short *columnofs;

  texcache_header_t *cache;
  
  // All the patches[patchcount] are drawn back to front into the cached texture.
  short patchcount;
  texpatch_t patches[1];
}
texturedef_t;

static texcache_header_t texcachehead;

// Texture List
static int *texturelist = NULL;

// textures list
static int numtextures = 0;
static texturedef_t **textures = NULL;

float_t *textureheight = NULL;

int *texturetranslation = NULL;

//
// MAPTEXTURE_T CACHING
//
// When a texture is first needed, it counts the number of composite columns
// required in the texture and allocates space for a column directory and any
// new columns.
//
// The directory will simply point inside other patches if there is only one
// patch in a given column, but any columns with multiple patches will have
// new column_ts generated.
//

//
// DrawColumnInCache
//
// Clip and draw a column from a patch into a cached post.
//
static void DrawColumnInCache(const column_t * patch, byte * cache, int originy, int cacheheight)
{
  int count;
  int position;
  byte *source;

  while (patch->topdelta != 0xff)
  {
    source = (byte *) patch + 3;
    count = patch->length;
    position = originy + patch->topdelta;

    if (position < 0)
    {
      count += position;
      position = 0;
    }

    if (position + count > cacheheight)
      count = cacheheight - position;

    if (count > 0)
      Z_MoveData(cache + position, source, byte, count);

    patch = (const column_t *) ((const byte *) patch + patch->length + 4);
  }
}

//
// GenerateComposite
//
// Using the texture definition, the composite texture is created from the patches,
// and each column is cached.
//
static void GenerateComposite(texturedef_t *texture)
{
  int x;
  int x1;
  int x2;
  int i;

  byte *block;
  byte *collump;
  const patch_t *realpatch;
  const column_t *patchcol;
  texpatch_t *patch;
  unsigned short *colofs;

  block = texture->cache->composite;

  collump = texture->columnlump;
  colofs = texture->columnofs;

  // Composite the columns together.
  for (i = 0, patch = texture->patches; i < texture->patchcount; i++, patch++)
  {
    realpatch = W_CacheLumpNum(patch->patch);
    x1 = patch->originx;
    x2 = x1 + SHORT(realpatch->width);

    if (x1 < 0)
      x = 0;
    else
      x = x1;

    if (x2 > texture->width)
      x2 = texture->width;

    for (; x < x2; x++)
    {
      // Column does not have multiple patches?
      if (collump[x] != 255)
        continue;

      patchcol = (const column_t *) ((const byte *) realpatch + LONG(realpatch->columnofs[x - x1]));

      DrawColumnInCache(patchcol, block + colofs[x], patch->originy, texture->height);
    }

    W_DoneWithLump(realpatch);
  }
}

//
// GenerateLookup
//
static void GenerateLookup(texturedef_t *texture)
{
  int x;
  int x1;
  int x2;
  int i;

  byte *patchcount;  // patchcount[texture->width]

  const patch_t *realpatch;
  texpatch_t *patch;

  byte *collump;
  unsigned short *colofs;

  texture->compositesize = 0;
  collump = texture->columnlump;
  colofs = texture->columnofs;

  // Now count the number of columns that are covered by more than one patch.
  // Fill in the lump / offset, so columns with only a single patch are all done.
  patchcount = Z_ClearNew(byte, texture->width);

  for (i = 0; i < texture->patchcount; i++)
  {
    patch = &texture->patches[i];
    
    realpatch = W_CacheLumpNum(patch->patch);
    x1 = patch->originx;
    x2 = x1 + SHORT(realpatch->width);
    // -ES- Default is to not put all patches into the cache. We will change
    // it to true wherever needed during the second pass.
    patch->cache = false;
    patch->lump = NULL;

    if (x1 < 0)
      x = 0;
    else
      x = x1;

    if (x2 > texture->width)
      x2 = texture->width;

    for (; x < x2; x++)
    {
      int ofs = LONG(realpatch->columnofs[x - x1]);
      const column_t *c = (const column_t *)(((const byte *)realpatch) + ofs);
      // -ES- Skip if the column is empty in a complex texture
      if (c->topdelta != 0xff || texture->patchcount == 1)
      {
        patchcount[x]++;
        collump[x] = i;
        colofs[x] = ofs + 3;
      }
    }
    W_DoneWithLump(realpatch);
  }

  for (x = 0; x < texture->width; x++)
  {
    if (patchcount[x] == 1)
    {
      texture->patches[collump[x]].cache = true;
    }
    else
    {
      // -ES- 2000/03/26 If there is no patch, warn and replace it with
      // rubbish, instead of warning and then crashing.
      if (patchcount[x] == 0)
        I_Printf("GenerateLookup: column without a patch \'%.8s\'\n", texture->name);
        
      // Use the cached block.
      collump[x] = -1;
      colofs[x] = texture->compositesize;

      if (texture->compositesize > 0x10000 - texture->height)
        I_Error("GenerateLookup: texture '%.8s' is >64k", texture->name);

      texture->compositesize += texture->height;
    }

  }

  Z_Free(patchcount);
}

// tiny ll helpers
static INLINE void InsertAtTail(texcache_header_t *h)
{
  h->prev = texcachehead.prev;
  h->next = &texcachehead;
  h->prev->next = h;
  h->next->prev = h;
}
static INLINE void Unlink(texcache_header_t *h)
{
  h->prev->next = h->next;
  h->next->prev = h->prev;
}

// the number of bytes of the texture cache that currently can be freed.
// useful when we decide how much to flush.
// this changes when the number of users of a block changes between 1 and 0.
static int cache_size;

// The number of bytes the cache item of a texture needs
#define TEXCACHESIZE(t) (sizeof(texcache_header_t) + sizeof(cached_tex_t) + \
    (t)->compositesize * sizeof(byte))


//
// LoadTexture
//
static texcache_header_t *LoadTexture(texturedef_t *t)
{
  int i;
  texcache_header_t *h;
  cached_tex_t *l;
  texpatch_t *p;
  
  // Do some boring first time work
  if (t->columnlump[0] == 254)
    GenerateLookup(t);

  DEV_ASSERT(t->cache == NULL, ("LoadTexture: Already cached"));

  h = (texcache_header_t *) Z_Malloc(TEXCACHESIZE(t));
  l = (cached_tex_t *)(h + 1);
  if (t->compositesize)
    h->composite = (byte *)(l + 1);
  else
    h->composite = NULL;

  h->users = 1;
  h->texture = t;
  t->cache = h;

  InsertAtTail(h);

  if (h->composite)
    GenerateComposite(t);

  for (i = 0, p = t->patches; i < t->patchcount; i++, p++)
    if (p->cache)
      p->lump = W_CacheLumpNum(p->patch);
  
  return h;
}

//
// UnloadTexture
//
// Removes the texture from the cache.
static void UnloadTexture(texcache_header_t *h)
{
  texpatch_t *p;
  texturedef_t *t = h->texture;
  int i;

  DEV_ASSERT(h->users == 0, ("UnloadTexture: %d users", h->users));

  cache_size -= TEXCACHESIZE(t);

  // First unregister all lumps.
  for (i = 0, p = t->patches; i < t->patchcount; i++, p++)
    if (p->cache)
    {
      W_DoneWithLump(p->lump);
      p->lump = NULL;
    }

  // Now free the rest of the mem, and remove all references.
  Unlink(h);
  Z_Free(h);
  t->cache = NULL;
}

const cached_tex_t *W_CacheTextureNum(int texnum)
{
  texturedef_t *tex;

  DEV_ASSERT(texnum < numtextures, ("W_CacheTextureNum: texnum (%d) >= numtextures (%d)", texnum, numtextures));
  tex = textures[texnum];

  if (tex->cache)
  {
    // already cached
    if (tex->cache->users == 0)
      // no longer freeable
      cache_size -= TEXCACHESIZE(tex);
    tex->cache->users++;
  }
  else
  {
    // load into cache
    tex->cache = LoadTexture(tex);
  }

  return (const cached_tex_t *)(tex->cache + 1);
}

void W_PreCacheTextureNum(int texnum)
{
  W_DoneWithTexture(W_CacheTextureNum(texnum));
}

void W_DoneWithTexture(const cached_tex_t *l)
{
  texcache_header_t *h;

  DEV_ASSERT(l, ("W_DoneWithTexture: NULL parm"));
  h = ((texcache_header_t *)l) - 1; // Intentional Const Override

  DEV_ASSERT(h->users > 0, ("W_DoneWithTexture: No users"));

  h->users--;
  if (h->users == 0)
  {
    cache_size += TEXCACHESIZE(h->texture);
    Unlink(h);
    InsertAtTail(h);
  }
}

static void FlushTextureCaches(z_urgency_e urge)
{
  int bytes_to_free = 0;
  texcache_header_t *h;

  switch (urge)
  {
    case Z_UrgencyLow: bytes_to_free = cache_size / 16; break;
    case Z_UrgencyMedium: bytes_to_free = cache_size / 8; break;
    case Z_UrgencyHigh: bytes_to_free = cache_size / 2; break;
    case Z_UrgencyExtreme: bytes_to_free = INT_MAX; break;
    default: I_Error("W_FlushLumpCache: Invalid urgency level %d", urge);
  }

  for (h = &texcachehead; h->next != &texcachehead && bytes_to_free > 0;)
  {
    if (h->next->users == 0)
    {
      bytes_to_free -= TEXCACHESIZE(h->texture);
      UnloadTexture(h->next);
    }
    else
      h = h->next;
  }
}

//
// W_GetColumn
//
// Returns the start of column #col of the texture.
const byte *W_GetColumn(int col, const cached_tex_t *t)
{
  const texcache_header_t *h;
  texturedef_t *tex;
  byte lump;
  
  DEV_ASSERT(t, ("W_GetColumn: texlump == NULL"));
  h = ((const texcache_header_t *)t) - 1;

  tex = h->texture;
  col &= tex->widthmask;
  lump = tex->columnlump[col];
  
  if (lump == 255)
    return h->composite + tex->columnofs[col];
  else
    return tex->patches[lump].lump + tex->columnofs[col];
}

//
// InitTextureLumps
//
// -ACB- 1998/09/09 Removed the Doom II SkyName change: unnecessary and not DDF.
//                  Reformatted and cleaned up.
//
static void InitTextureLumps(int pnames, int texture1, int texture2)
{
  const maptexture_t *mtexture;
  texturedef_t *texture;
  const mappatch_t *mpatch;
  texpatch_t *patch;

  int i;
  int j;
  int nummappatches;
  int offset;
  int maxoff;
  int maxoff2;
  int numtextures1;
  int numtextures2;
  int counttextures;
  int patchcount;

  const int *maptex;
  const int *maptex1;
  const int *maptex2;
  int *patchlookup;
  const int *directory;

  char name[9];
  const char *names;
  static int lastpnames = -1;
  const char *name_p;

  int base_size;
  int width;
  byte *base;

  // Load the patch names from pnames.lmp.
  name[8] = 0;
  // -KM- 1999/01/31 Fix certain maps that don't have a pnames, but have
  //  eg texture2
  if (pnames != -1)
    lastpnames = pnames;
  else
    pnames = lastpnames;

  if (pnames == -1)
    I_Error("InitTextureLumps: Missing Texture resource 'PNAMES'\n");

  names = W_CacheLumpNum(pnames);
  nummappatches = LONG(*((const int *)names));
  name_p = names + 4;

  patchlookup = Z_New(int, nummappatches);

  for (i = 0; i < nummappatches; i++)
  {
    Z_StrNCpy(name, name_p + i * 8, 8);
    patchlookup[i] = W_CheckNumForName(name);
  }

  W_DoneWithLump(names);

  //
  // Load the map texture definitions from textures.lmp.
  //
  // The data is contained in one or two lumps:
  //   TEXTURE1 for shareware
  //   TEXTURE2 for commercial.
  //
  maptex = maptex1 = W_CacheLumpNum(texture1);
  numtextures1 = LONG(*maptex);
  maxoff = W_LumpLength(texture1);
  directory = maptex + 1;

  if (texture2 != -1)
  {
    maptex2 = W_CacheLumpNum(texture2);
    numtextures2 = LONG(*maptex2);
    maxoff2 = W_LumpLength(texture2);
  }
  else
  {
    maptex2 = NULL;
    numtextures2 = 0;
    maxoff2 = 0;
  }

  counttextures = numtextures;
  numtextures += numtextures1 + numtextures2;

  Z_Resize(textures, texturedef_t *, numtextures);
  Z_Resize(textureheight, float_t, numtextures);

  for (i = counttextures; i < numtextures; i++, directory++)
  {
    if (i == (numtextures1 + counttextures))
    {
      // Start looking in second texture file.
      maptex = maptex2;
      maxoff = maxoff2;
      directory = maptex + 1;
    }

    offset = LONG(*directory);
    if (offset > maxoff)
      I_Error("R_InitTextures: bad texture directory");

    mtexture = (const maptexture_t *) ((const byte *) maptex + offset);

    // -ES- 2000/02/10 Texture must have patches.
    patchcount = SHORT(mtexture->patchcount);
    if (!patchcount)
      I_Error("R_InitTextures: Texture '%.8s' has no patches", mtexture->name);

    width = SHORT(mtexture->width);
    if (width == 0)
      I_Error("R_InitTextures: Texture '%.8s' has zero width", mtexture->name);

    // -ES- Allocate texture, patches and columnlump/ofs in one big chunk
    base_size = sizeof(texturedef_t) + sizeof(texpatch_t) * (patchcount - 1);
    texture = textures[i] = (texturedef_t *) Z_Malloc(base_size + width * (sizeof(byte) + sizeof(short)));
    base = (byte *)texture + base_size;

    texture->columnofs = (unsigned short *)base;
    
    texture->columnlump = (byte *)(texture->columnofs + width);
    // means uninitialised
    texture->columnlump[0] = 254;

    texture->width = width;
    texture->height = SHORT(mtexture->height);
    texture->patchcount = patchcount;

    texture->cache = NULL;

    Z_MoveData(texture->name, mtexture->name, char, 8);

    mpatch = &mtexture->patches[0];
    patch = &texture->patches[0];

    for (j = 0; j < texture->patchcount; j++, mpatch++, patch++)
    {
      patch->originx = SHORT(mpatch->originx);
      patch->originy = SHORT(mpatch->originy);
      patch->patch = patchlookup[SHORT(mpatch->patch)];

      if (patch->patch == -1)
        I_Error("R_InitTextures: Missing patch in texture \'%.8s\'", texture->name);
    }

    j = 1;

    while (j * 2 <= texture->width)
      j <<= 1;

    texture->widthmask = j - 1;
    textureheight[i] = texture->height;
  }

  W_DoneWithLump(maptex1);

  if (maptex2)
    W_DoneWithLump(maptex2);
  
  Z_Free(patchlookup); // finished uning so free
}

//
// R_InitTextures
//
// Initialises the texture list with the textures from the world map.
//
// -ACB- 1998/09/09 Fixed the Display routine from display rubbish.
//
void W_InitTextures(void)
{
  wadtex_resource_t *tex;
  int maxtex;
  int i, j, l;
  const int *texlump;

  // This is the only initialisation we need
  texcachehead.next = texcachehead.prev = &texcachehead;
  Z_RegisterCacheFlusher(FlushTextureCaches);

  tex = W_GetTextureResources();

  for (maxtex = 0; tex[maxtex].pnames != -2; maxtex++)
  {
    for (i = 0; i < 2; i++)
    {
      if (tex[maxtex].texture[i] != -1)
      {
        texlump = W_CacheLumpNum(tex[maxtex].texture[i]);
        W_DoneWithLump(texlump);
      }
    }
  }

  spritelist = W_GetList('S', &numspritelumps);

  // -ACB- 1998/09/09 Removed the Doom II SkyName change: unnecessary and not DDF.
  for (i = 0; i < maxtex; i++)
  {
    if (tex[i].texture[0] != -1)
      InitTextureLumps(tex[i].pnames, tex[i].texture[0], tex[i].texture[1]);
    else if (tex[i].texture[1] != -1)
      InitTextureLumps(tex[i].pnames, tex[i].texture[1], tex[i].texture[0]);
  }

  // Fixme: Clean up system. W_GetTextureResources can only be used once.
  Z_Free(tex);

  Z_Resize(texturelist, int, numtextures);

  // Create translation table for global animation.
  Z_Resize(texturetranslation, int, numtextures + 1);

  for (i = 0; i < numtextures; i++)
    texturetranslation[i] = texturelist[i] = i;

#define CMP(a, b) (strnicmp(textures[a]->name, textures[b]->name, 8) < 0)
  QSORT(int, texturelist, numtextures, CUTOFF);
#undef CMP

  for (i = 0; i < numtextures - 1; i = j)
  {
    j = i + 1;
    l = i;
    while (j < numtextures && !strnicmp(textures[texturelist[i]]->name, textures[texturelist[j]]->name, 8))
    {
      if (texturelist[j] > texturelist[l])
        l = j;
      j++;
    }
    j = i;
    while (j < numtextures && !strnicmp(textures[texturelist[i]]->name, textures[texturelist[j]]->name, 8))
      texturelist[j++] = texturelist[l];
  }
}

//
// R_CheckTextureNumForName
//
// Check whether texture is available. Filter out NoTexture indicator.
//
int W_CheckTextureNumForName(const char *name)
{
  int i;

  // "NoTexture" marker.
  if (!name)
    return 0;
  if (!name[0] || name[0] == '-')
    return 0;

#define CMP(a) (strnicmp(textures[texturelist[a]]->name, name, 8) < 0)
  BSEARCH(numtextures, i);
#undef CMP

  if (i >= 0 && i < numtextures &&
      !strnicmp(textures[texturelist[i]]->name, name, 8))
    return texturelist[i];

  return -1;
}

//
// R_TextureNumForName
//
// Calls R_CheckTextureNumForName, aborts with error message.
//
int W_TextureNumForName(const char *name)
{
  int i;

  i = W_CheckTextureNumForName(name);

  if (i == -1)
    I_Error("R_TextureNumForName: \'%.8s\' not found", name);

  return i;
}

//
//  OPENGL SUPPORT
//

#ifdef USE_GL

// -AJA- this func for the GL code -- remove later
static int W_TextureGetWidth(const cached_tex_t *t)
{
  const texcache_header_t *h;
  DEV_ASSERT2(t);
  h = ((const texcache_header_t *)t) - 1;
  return h->texture->width;
}

// -AJA- this func for the GL code -- remove later
static int W_TextureGetHeight(const cached_tex_t *t)
{
  const texcache_header_t *h;
  DEV_ASSERT2(t);
  h = ((const texcache_header_t *)t) - 1;
  return h->texture->height;
}

#define GAMMA_RED(src)  (gammatable[usegamma][playpal_data[0][src][0]])
#define GAMMA_GRN(src)  (gammatable[usegamma][playpal_data[0][src][1]])
#define GAMMA_BLU(src)  (gammatable[usegamma][playpal_data[0][src][2]])

//
// W_ConvertFlatRGBA
//
byte *W_ConvertFlatRGBA(int flat, int *width, int *height)
{
  int x, y;

  byte *rgba_data;

  ds_source = W_CacheLumpNum(flatlist[flat]);

  rgba_data = Z_New(byte, 64 * 64 * 4);

  // glmax_tex_size is guaranteed to be >= 64 by the specs
  
  for (y=0; y < 64; y++)
  for (x=0; x < 64; x++)
  {
    const byte src_pix = ds_source[y*64 + x];
    byte *dest = &rgba_data[((63-y)*64 + x) * 4];

    dest[0] = GAMMA_RED(src_pix);
    dest[1] = GAMMA_GRN(src_pix);
    dest[2] = GAMMA_BLU(src_pix);
    dest[3] = 255;
  }

  W_DoneWithLump(ds_source);

  *width  = 64;
  *height = 64;

  return rgba_data;
}

static int MakeValidSize(int value)
{
  if (value <=    4) return    4;
  if (value <=    8) return    8;
  if (value <=   16) return   16;
  if (value <=   32) return   32;
  if (value <=   64) return   64;
  if (value <=  128) return  128;
  if (value <=  256) return  256;
  if (value <=  512) return  512;
  if (value <= 1024) return 1024;
  if (value <= 2048) return 2048;

  I_Error("Texture size (%d) too large !\n", value);
  return -1;
}

static void WriteColumnIntoRGBA(byte *rgba_data, int width,
    int bottom, int y_step, const column_t * column)
{
  int y, h;

  while (column->topdelta != 0xff)
  {
    y = column->topdelta;
    h = column->length;

    DEV_ASSERT2((y+h)/y_step <= bottom);

    dc_source = (const byte *) column + 3;
    
    for (; h > 0; h -= y_step, y += y_step, dc_source += y_step)
    {
      byte src_pix = dc_source[0];
      byte *dest = &rgba_data[(bottom - 1 - y/y_step) * width * 4];

      dest[0] = GAMMA_RED(src_pix);
      dest[1] = GAMMA_GRN(src_pix);
      dest[2] = GAMMA_BLU(src_pix);
      dest[3] = 255;
    }

    column = (const column_t *) ((const byte *) column + column->length + 4);
  }
}

//
// W_ConvertTextureRGBA
//
byte *W_ConvertTextureRGBA(int tex, int *width, int *height,
    float_t *right, float_t *bottom, int is_masked)
{
  int x, y, w, h;
  int x_step, y_step;

  byte *rgba_data;

  const cached_tex_t *ct;

  ct = W_CacheTextureNum(tex);

  (*width)  = w = W_TextureGetWidth(ct);
  (*height) = h = W_TextureGetHeight(ct);
  
  // determine next power-of-two size
  
  (*width)  = MakeValidSize(*width);
  (*height) = MakeValidSize(*height);

  (*right)  = (float_t)w / (float_t)(*width);
  (*bottom) = (float_t)h / (float_t)(*height);
  
  // compute horizontal shrink
  for (x_step=1; (*width) > glmax_tex_size; (*width) /= 2, x_step *= 2)
  { }

  // compute vertical shrink
  for (y_step=1; (*height) > glmax_tex_size; (*height) /= 2, y_step *= 2)
  { }

  // clear data, thus default alpha is 0, i.e. transparent
  rgba_data = Z_ClearNew(byte, (*width) * (*height) * 4);

  if (is_masked)
  {
    // handle masked textures specially
    
    for (x=0; x < w/x_step; x++)
    {
      const column_t *col;

      // -ES- Fixme: Medusa if texture consists of >1 patch
      col = (const column_t *) (W_GetColumn(x * x_step, ct) - 3);

      WriteColumnIntoRGBA(&rgba_data[x*4], (*width), h/y_step,
          y_step, col);
    }
  }
  else
  {
    for (x=0; x < w/x_step; x++)
    {
      dc_source = W_GetColumn(x * x_step, ct);
    
      for (y=0; y < h/y_step; y++)
      {
        const byte src_pix = dc_source[y * y_step];
        byte *dest = &rgba_data[((h/y_step - 1 - y) * (*width) + x) * 4];

        dest[0] = GAMMA_RED(src_pix);
        dest[1] = GAMMA_GRN(src_pix);
        dest[2] = GAMMA_BLU(src_pix);
        dest[3] = 255;
      }
    }
  }

  W_DoneWithTexture(ct);

  return rgba_data;
}

//
// W_ConvertSpriteRGBA
//
byte *W_ConvertSpriteRGBA(int patnum, int *width, int *height,
    float_t *right, float_t *bottom)
{
  int x, w, h;

  byte *rgba_data;

  const patch_t *patch = W_CacheLumpNum(patnum);

  (*width)  = w = patch->width;
  (*height) = h = patch->height;
  
  // determine next power-of-two size
  
  (*width)  = MakeValidSize(*width);
  (*height) = MakeValidSize(*height);

  (*right)  = (float_t)w / (float_t)(*width);
  (*bottom) = (float_t)h / (float_t)(*height);

  // -AJA- FIXME: handle sprites bigger than glmax_tex_size.
  
  // clear data, thus default alpha is 0, i.e. transparent
  rgba_data = Z_ClearNew(byte, (*width) * (*height) * 4);

  for (x=0; x < w; x++)
  {
    const column_t *column;

    column = (const column_t *) ((const byte *) patch +
        LONG(patch->columnofs[x]));

    WriteColumnIntoRGBA(&rgba_data[x*4], (*width), h, 1, column);
  }
    
  W_DoneWithLump(patch);

  return rgba_data;
}

#endif USE_GL
