//----------------------------------------------------------------------------
//  EDGE WAD Support Code
//----------------------------------------------------------------------------
// 
//  Copyright (c) 1999-2000  The EDGE Team.
// 
//  This program is free software; you can redistribute it and/or
//  modify it under the terms of the GNU General Public License
//  as published by the Free Software Foundation; either version 2
//  of the License, or (at your option) any later version.
//
//  This program is distributed in the hope that it will be useful,
//  but WITHOUT ANY WARRANTY; without even the implied warranty of
//  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//  GNU General Public License for more details.
//
//----------------------------------------------------------------------------
//
//  Based on the DOOM source code, released by Id Software under the
//  following copyright:
//
//    Copyright (C) 1993-1996 by id Software, Inc.
//
//----------------------------------------------------------------------------
//
// This file contains various levels of support 
// for using sprites and flats directly from a PWAD as well as some minor
// optimisations for patches. Because there are some PWADs that do arcane
// things with sprites, it is possible that this feature may not always
// work (at least, not until I become aware of them and support them) and
// so this feature can be turned off from the command line if necessary.
//
// -MH- 1998/03/04
//

#include "i_defs.h"
#include "w_wad.h"

#include "dm_type.h"
#include "e_main.h"
#include "e_search.h"
#include "m_argv.h"
#include "m_misc.h"
#include "m_swap.h"
#include "e_player.h"
#include "rad_trig.h"
#include "w_textur.h"
#include "z_zone.h"

// Create the start and end markers
static char startn[9] =
  {0, '_', 'S', 'T', 'A', 'R', 'T', 0, 0};  // normal ....... S_START/S_START

static char starta[9] =
  {0, 0, '_', 'S', 'T', 'A', 'R', 'T', 0};  // alternate .... SS_START/FF_START

static char start1[9] =
  {0, '1', '_', 'S', 'T', 'A', 'R', 'T', 0};  // bad1 ......... S1_START/F1_START

static char start2[9] =
  {0, '2', '_', 'S', 'T', 'A', 'R', 'T', 0};  // bad2 ......... S2_START/F2_START

static char start3[9] =
  {0, '3', '_', 'S', 'T', 'A', 'R', 'T', 0};  // bad3 ......... S3_START/F3_START

static char endn[9] =
  {0, '_', 'E', 'N', 'D', 0, 0, 0, 0};  // normal ....... S_END/S_END

static char enda[9] =
  {0, 0, '_', 'E', 'N', 'D', 0, 0, 0};  // alternate .... SS_END/FF_END

static char end1[9] =
  {0, '1', '_', 'E', 'N', 'D', 0, 0, 0};  // bad1 ......... S1_END/F1_END

static char end2[9] =
  {0, '2', '_', 'E', 'N', 'D', 0, 0, 0};  // bad2 ......... S2_END/F2_END

static char end3[9] =
  {0, '3', '_', 'E', 'N', 'D', 0, 0, 0};  // bad3 ......... S3_END/F3_END

//
//  GLOBALS
//

// Location of each lump on disk.
lumpinfo_t *lumpinfo;
static int *lumpmap = NULL;
int numlumps;

typedef struct lumpheader_s
{
#ifdef DEVELOPERS
#define LUMPID ((int)0xAC45197e)
  // Should be LUMPID
  int id;
#endif
  // number of users.
  int users;
  // index in lumplookup
  int lumpindex;
  struct lumpheader_s *next, *prev;
}
lumpheader_t;

static lumpheader_t **lumplookup;
static lumpheader_t lumphead;
// number of freeable bytes in cache (excluding headers).
// Used to decide how many bytes we should flush.
static int cache_size = 0;

typedef struct
{
  char *file_name;
  int reloadlump;
  int handle;
  enum {FILE_WAD, FILE_SINGLELUMP} type;
} data_file_t;

static data_file_t *data_files = NULL;

// Sprites, Flats, Patches
int sprite_lists = 0;  // number of sprite begin..end sections found

int flat_lists = 0;  // number of flat begin..end sections found

int patch_lists = 0;  // number of patch begin..end sections found

int lax_list_syntax = 1;  // use commonly found begin..end syntax

int lax_sprite_rot = 0;  // disengage sprite rotation/limit checks

// Raw filenames
typedef struct raw_filename_s
{
  char *name;
  boolean_t allow_ddf;
}
raw_filename_t;

int addwadnum = 0;
static int maxwadfiles = 0;
static raw_filename_t *wadfiles = NULL;

//
// Is the name a sprite list start flag?
// If lax syntax match, fix up to standard syntax.
//
static int IsS_START(char *name)
{
  int result, lax;

  result = ((name[0] == 'S') &&
      (name[1] == '_') &&
      (name[2] == 'S') &&
      (name[3] == 'T') &&
      (name[4] == 'A') &&
      (name[5] == 'R') &&
      (name[6] == 'T') &&
      (name[7] == 0));
  lax = ((!result) &&
      (name[0] == 'S') &&
      (name[1] == 'S') &&
      (name[2] == '_') &&
      (name[3] == 'S') &&
      (name[4] == 'T') &&
      (name[5] == 'A') &&
      (name[6] == 'R') &&
      (name[7] == 'T'));
  if (lax)
  {
    if (!lax_list_syntax)
      I_Error("IsS_START: flag SS_START encountered in sprites list");

    //fix up flag to standard syntax
    result = 1;
    name[0] = 'S';
    name[1] = '_';
    name[2] = 'S';
    name[3] = 'T';
    name[4] = 'A';
    name[5] = 'R';
    name[6] = 'T';
    name[7] = 0;
  }

  return result;

}

//
// Is the name a sprite list end flag?
// If lax syntax match, fix up to standard syntax.
//
static int IsS_END(char *name)
{
  int result, lax;

  result = ((name[0] == 'S') &&
      (name[1] == '_') &&
      (name[2] == 'E') &&
      (name[3] == 'N') &&
      (name[4] == 'D') &&
      (name[5] == 0) &&
      (name[6] == 0) &&
      (name[7] == 0));
  lax = ((!result) &&
      (name[0] == 'S') &&
      (name[1] == 'S') &&
      (name[2] == '_') &&
      (name[3] == 'E') &&
      (name[4] == 'N') &&
      (name[5] == 'D') &&
      (name[6] == 0) &&
      (name[7] == 0));
  if (lax)
  {
    if (!lax_list_syntax)
      I_Error("W_IsS_END: flag SS_END encountered in sprites list");

    //fix up flag to standard syntax
    result = 1;
    name[0] = 'S';
    name[1] = '_';
    name[2] = 'E';
    name[3] = 'N';
    name[4] = 'D';
    name[5] = 0;
    name[6] = 0;
    name[7] = 0;
  }

  return result;

}

//
// Is the name a flat list start flag?
// If lax syntax match, fix up to standard syntax.
//
static int IsF_START(char *name)
{
  int result, lax;

  result = ((name[0] == 'F') &&
      (name[1] == '_') &&
      (name[2] == 'S') &&
      (name[3] == 'T') &&
      (name[4] == 'A') &&
      (name[5] == 'R') &&
      (name[6] == 'T') &&
      (name[7] == 0));
  lax = ((!result) &&
      (name[0] == 'F') &&
      (name[1] == 'F') &&
      (name[2] == '_') &&
      (name[3] == 'S') &&
      (name[4] == 'T') &&
      (name[5] == 'A') &&
      (name[6] == 'R') &&
      (name[7] == 'T'));
  if (lax)
  {
    if (!lax_list_syntax)
      I_Error("W_IsF_START: flag FF_START encountered in flats list");

    //fix up flag to standard syntax
    result = 1;
    name[0] = 'F';
    name[1] = '_';
    name[2] = 'S';
    name[3] = 'T';
    name[4] = 'A';
    name[5] = 'R';
    name[6] = 'T';
    name[7] = 0;
  }

  return result;

}

//
// Is the name a flat list end flag?
// If lax syntax match, fix up to standard syntax.
//
static int IsF_END(char *name)
{
  int result, lax;

  result = ((name[0] == 'F') &&
      (name[1] == '_') &&
      (name[2] == 'E') &&
      (name[3] == 'N') &&
      (name[4] == 'D') &&
      (name[5] == 0) &&
      (name[6] == 0) &&
      (name[7] == 0));
  lax = ((!result) &&
      (name[0] == 'F') &&
      (name[1] == 'F') &&
      (name[2] == '_') &&
      (name[3] == 'E') &&
      (name[4] == 'N') &&
      (name[5] == 'D') &&
      (name[6] == 0) &&
      (name[7] == 0));
  if (lax)
  {
    if (!lax_list_syntax)
      I_Error("W_IsF_END: flag FF_END encountered in flats list");

    //fix up flag to standard syntax
    result = 1;
    name[0] = 'F';
    name[1] = '_';
    name[2] = 'E';
    name[3] = 'N';
    name[4] = 'D';
    name[5] = 0;
    name[6] = 0;
    name[7] = 0;
  }

  return result;

}

//
// Is the name a patch list start flag?
// If lax syntax match, fix up to standard syntax.
//
static int IsP_START(char *name)
{
  int result, lax;

  result = ((name[0] == 'P') &&
      (name[1] == '_') &&
      (name[2] == 'S') &&
      (name[3] == 'T') &&
      (name[4] == 'A') &&
      (name[5] == 'R') &&
      (name[6] == 'T') &&
      (name[7] == 0));
  lax = ((!result) &&
      (name[0] == 'P') &&
      (name[1] == 'P') &&
      (name[2] == '_') &&
      (name[3] == 'S') &&
      (name[4] == 'T') &&
      (name[5] == 'A') &&
      (name[6] == 'R') &&
      (name[7] == 'T'));
  if (lax)
  {
    if (!lax_list_syntax)
      I_Error("W_IsP_START: flag PP_START encountered in patches list");

    //fix up flag to standard syntax
    result = 1;
    name[0] = 'P';
    name[1] = '_';
    name[2] = 'S';
    name[3] = 'T';
    name[4] = 'A';
    name[5] = 'R';
    name[6] = 'T';
    name[7] = 0;
  }

  return result;

}

//
// Is the name a patches list end flag?
// If lax syntax match, fix up to standard syntax.
//
static int IsP_END(char *name)
{
  int result, lax;

  result = ((name[0] == 'P') &&
      (name[1] == '_') &&
      (name[2] == 'E') &&
      (name[3] == 'N') &&
      (name[4] == 'D') &&
      (name[5] == 0) &&
      (name[6] == 0) &&
      (name[7] == 0));
  lax = ((!result) &&
      (name[0] == 'P') &&
      (name[1] == 'P') &&
      (name[2] == '_') &&
      (name[3] == 'E') &&
      (name[4] == 'N') &&
      (name[5] == 'D') &&
      (name[6] == 0) &&
      (name[7] == 0));
  if (lax)
  {
    if (!lax_list_syntax)
      I_Error("W_IsP_END: flag PP_END encountered in patches list");

    //fix up flag to standard syntax
    result = 1;
    name[0] = 'P';
    name[1] = '_';
    name[2] = 'E';
    name[3] = 'N';
    name[4] = 'D';
    name[5] = 0;
    name[6] = 0;
    name[7] = 0;
  }

  return result;

}

//
// Return 1 if the name is null
//
static int IsNullName(char *name)
{
  return ((name[0] == 0) &&
      (name[1] == 0) &&
      (name[2] == 0) &&
      (name[3] == 0) &&
      (name[4] == 0) &&
      (name[5] == 0) &&
      (name[6] == 0) &&
      (name[7] == 0));
}

//
// Return 1 if the name is one of the redundant sprite list flags S1_START,
// S1_END, S2_START or S2_END. These are oft-encountered hacks that some
// editors need and are not recognised or needed by Doom.
//
static int IsBadSpriteFlag(char *name)
{
  return (((name[0] == 'S') &&
          (name[1] == '1') &&
          (name[2] == '_') &&
          (name[3] == 'S') &&
          (name[4] == 'T') &&
          (name[5] == 'A') &&
          (name[6] == 'R') &&
          (name[7] == 'T'))
      ||
      ((name[0] == 'S') &&
          (name[1] == '1') &&
          (name[2] == '_') &&
          (name[3] == 'E') &&
          (name[4] == 'N') &&
          (name[5] == 'D') &&
          (name[6] == 0) &&
          (name[7] == 0))
      ||
      ((name[0] == 'S') &&
          (name[1] == '2') &&
          (name[2] == '_') &&
          (name[3] == 'S') &&
          (name[4] == 'T') &&
          (name[5] == 'A') &&
          (name[6] == 'R') &&
          (name[7] == 'T'))
      ||
      ((name[0] == 'S') &&
          (name[1] == '2') &&
          (name[2] == '_') &&
          (name[3] == 'E') &&
          (name[4] == 'N') &&
          (name[5] == 'D') &&
          (name[6] == 0) &&
          (name[7] == 0)));
}

//
// Return 1 if the name is one of the redundant flat list flags F1_START,
// F1_END, F2_START or F2_END. These are oft-encountered hacks that some
// editors need and are not recognised or needed by Doom.
//
static int IsBadFlatFlag(char *name)
{
  return (((name[0] == 'F') &&
          (name[1] == '1') &&
          (name[2] == '_') &&
          (name[3] == 'S') &&
          (name[4] == 'T') &&
          (name[5] == 'A') &&
          (name[6] == 'R') &&
          (name[7] == 'T'))
      ||
      ((name[0] == 'F') &&
          (name[1] == '1') &&
          (name[2] == '_') &&
          (name[3] == 'E') &&
          (name[4] == 'N') &&
          (name[5] == 'D') &&
          (name[6] == 0) &&
          (name[7] == 0))
      ||
      ((name[0] == 'F') &&
          (name[1] == '2') &&
          (name[2] == '_') &&
          (name[3] == 'S') &&
          (name[4] == 'T') &&
          (name[5] == 'A') &&
          (name[6] == 'R') &&
          (name[7] == 'T'))
      ||
      ((name[0] == 'F') &&
          (name[1] == '2') &&
          (name[2] == '_') &&
          (name[3] == 'E') &&
          (name[4] == 'N') &&
          (name[5] == 'D') &&
          (name[6] == 0) &&
          (name[7] == 0)));
}

//
// Return 1 if the name is one of the redundant patch list flags P1_START,
// P1_END, P2_START or P2_END. These are oft-encountered hacks that some
// editors need and are not recognised or needed by Doom.
//
static int IsBadPatchFlag(char *name)
{
  return (((name[0] == 'P') &&
          (name[1] == '1') &&
          (name[2] == '_') &&
          (name[3] == 'S') &&
          (name[4] == 'T') &&
          (name[5] == 'A') &&
          (name[6] == 'R') &&
          (name[7] == 'T'))
      ||
      ((name[0] == 'P') &&
          (name[1] == '1') &&
          (name[2] == '_') &&
          (name[3] == 'E') &&
          (name[4] == 'N') &&
          (name[5] == 'D') &&
          (name[6] == 0) &&
          (name[7] == 0))
      ||
      ((name[0] == 'P') &&
          (name[1] == '2') &&
          (name[2] == '_') &&
          (name[3] == 'S') &&
          (name[4] == 'T') &&
          (name[5] == 'A') &&
          (name[6] == 'R') &&
          (name[7] == 'T'))
      ||
      ((name[0] == 'P') &&
          (name[1] == '2') &&
          (name[2] == '_') &&
          (name[3] == 'E') &&
          (name[4] == 'N') &&
          (name[5] == 'D') &&
          (name[6] == 0) &&
          (name[7] == 0)));
}

static int FileLength(int handle)
{
  struct stat fileinfo;

  if (fstat(handle, &fileinfo) == -1)
    I_Error("Error fstating");

  return fileinfo.st_size;
}

static void ExtractFileBase(const char *path, char *dest)
{
  const char *src;
  int length;

  src = path + strlen(path) - 1;

  // back up until a \ or the start
  while (src != path
      && *(src - 1) != '\\'
      && *(src - 1) != '/'
      && *(src - 1) != ':')  // Kester added :

  {
    src--;
  }

  // copy up to eight characters
  Z_Clear(dest, char, 8);
  length = 0;

  while (*src && *src != '.')
  {
    if (++length == 9)
      I_Error("Filename base of %s >8 chars", path);

    *dest++ = toupper((int)*src++);
  }
}

static wadtex_resource_t *wadtex = NULL;
static int maxwadtex = 0;

static void RegisterTexLumps(int pnames, int texture1, int texture2)
{
  Z_Resize(wadtex, wadtex_resource_t, ++maxwadtex);

  wadtex[maxwadtex - 1].pnames = pnames;
  wadtex[maxwadtex - 1].texture[0] = texture1;
  wadtex[maxwadtex - 1].texture[1] = texture2;
}

//
// W_GetTextureResources
//
// When all wadtex resources have been loaded, this one is called
// to finish it, so the texture initialisation can use it.
// It can only be used once, and it is the caller's responsibility to
// Z_Free the returned mem. Hence todo: Clean up & dynamicify texture usage
wadtex_resource_t *W_GetTextureResources(void)
{
  Z_Resize(wadtex, wadtex_resource_t, maxwadtex + 1);
  wadtex[maxwadtex].pnames = -2;
  return wadtex;
}

static void MapSetup(void)
{
  int i, j, l;

  Z_Resize(lumpmap, int, numlumps);

  for (i = 0; i < numlumps; i++)
    lumpmap[i] = i;
#define CMP(a, b) (strncmp(lumpinfo[a].name, lumpinfo[b].name, 8) < 0)
  QSORT(int, lumpmap, numlumps, CUTOFF);
#undef CMP

  // Handle patch wads
  for (i = numlumps; --i;)
  {
    if (!strncmp(lumpinfo[lumpmap[i]].name, lumpinfo[lumpmap[i - 1]].name, 8))
    {
      j = i - 1;
      l = -1;
      while (j >= 0 && !strncmp(lumpinfo[lumpmap[i]].name, lumpinfo[lumpmap[j]].name, 8))
      {
        if (l == -1)
        {
          if (lumpinfo[lumpmap[i]].file > lumpinfo[lumpmap[j]].file)
            l = i;
          else
            l = j;
        }
        else
        {
          if (lumpinfo[lumpmap[j]].file > lumpinfo[lumpmap[l]].file)
            l = j;
        }
        j--;
      }

      j = i;
      while (j >= 0 && !strncmp(lumpinfo[lumpmap[i]].name, lumpinfo[lumpmap[j]].name, 8))
        lumpmap[j--] = lumpmap[l];
    }
  }
}


//
// LUMP BASED ROUTINES.
//

//
// W_AddFile
//
// All files are optional, but at least one file must be
//  found (PWAD, if all required lumps are present).
// Files with a .wad extension are wadlink files
//  with multiple lumps.
// Other files are single lumps with the base filename
//  for the lump name.
//
// If filename starts with a tilde, the file is handled
//  specially to allow map reloads.
// But: the reload feature is a fragile hack...
// -ES- 2000/03/11 Made it less fragile.

static int wadfile = -1;

// -KM- 1999/01/31 Order is important, Languages are loaded before sfx, etc...
static struct
{
  char *name;
  void (* func)(void *data, int size);
  int lump;
}
DDF_Readers[] =
{
  { "DDFLANG", DDF_ReadLangs, 0 } ,
  { "DDFSFX",  DDF_ReadSFX, 0 } ,
  { "DDFCOLM", DDF_ReadColourMaps, 0 } ,  // -AJA- 1999/07/09.
  { "DDFATK",  DDF_ReadAtks, 0 } ,
  { "DDFWEAP", DDF_ReadWeapons, 0 } ,
  { "DDFTHING",DDF_ReadThings, 0 } ,
  { "DDFPLAY", DDF_ReadMusicPlaylist, 0 } ,
  { "DDFLINE", DDF_ReadLines, 0 } ,
  { "DDFSECT", DDF_ReadSectors, 0 } ,
  { "DDFSWTH", DDF_ReadSW, 0 } ,
  { "DDFANIM", DDF_ReadAnims, 0 } ,
  { "DDFGAME", DDF_ReadGames, 0 } ,
  { "DDFLEVL", DDF_ReadLevels, 0 },
  { "RSCRIPT", RAD_LoadLump, 0 }       // -AJA- 2000/04/21.
};

static void FreeLump(lumpheader_t *h)
{
  int lumpnum = h->lumpindex;

  cache_size -= W_LumpLength(lumpnum);
#ifdef DEVELOPERS
  if (h->id != LUMPID)
    I_Error("FreeLump: id != LUMPID");
  h->id = 0;
  if (h == NULL)
    I_Error("FreeLump: NULL lump");
  if (h->users)
    I_Error("FreeLump: lump %d has %d users!", lumpnum, h->users);
  if (lumplookup[lumpnum] != h)
    I_Error("FreeLump: Internal error, lump %d", lumpnum);
#endif
  lumplookup[lumpnum] = NULL;
  h->prev->next = h->next;
  h->next->prev = h->prev;
  Z_Free(h);
}

static void MarkAsCached(lumpheader_t *item)
{
#ifdef DEVELOPERS
  if (item->id != LUMPID)
    I_Error("MarkAsCached: id != LUMPID");
  if (!item)
    I_Error("MarkAsCached: lump %d is NULL", item->lumpindex);
  if (item->users)
    I_Error("MarkAsCached: lump %d has %d users!", item->lumpindex, item->users);
  if (lumplookup[item->lumpindex] != item)
    I_Error("MarkAsCached: Internal error, lump %d", item->lumpindex);
#endif

  cache_size += W_LumpLength(item->lumpindex);
}

static void AddFile(const char *filename, boolean_t allow_ddf)
{
  wadinfo_t header;
  lumpinfo_t *lump_p;
  int i;
  unsigned int j;
  int handle;
  int length;
  int startlump;
  int pnames = -1;
  int texture1 = -1;
  int texture2 = -1;
  int reloadlump = -1;
  boolean_t fileinfoalloced;
  filelump_t *fileinfo;
  filelump_t *startfileinfo;
  filelump_t singleinfo;

  fileinfoalloced = false;

  // open the file and add to directory
  for (i = sizeof(DDF_Readers) / sizeof(DDF_Readers[0]); i--;)
    DDF_Readers[i].lump = -1;

  // handle reload indicator.
  if (filename[0] == '~')
  {
    filename++;
    reloadlump = numlumps;
  }

  if ((handle = open(filename, O_RDONLY | O_BINARY)) == -1)
  {
    I_Printf(" couldn't open %s\n", filename);
    return;
  }

  I_Printf("  adding %s", filename);
  startlump = numlumps;
  wadfile++;
  Z_Resize(data_files, data_file_t, wadfile + 1);
  data_files[wadfile].reloadlump = reloadlump;
  data_files[wadfile].file_name = Z_StrDup(filename);

  if (M_CheckExtension("wad", filename) == EXT_MATCHING ||
      M_CheckExtension("gwa", filename) == EXT_MATCHING)
  {
    // WAD file
    data_files[wadfile].type = FILE_WAD;
    read(handle, &header, sizeof(header));
    if (strncmp(header.identification, "IWAD", 4))
    {
      // Homebrew levels?
      if (strncmp(header.identification, "PWAD", 4))
      {
        I_Error("Wad file %s doesn't have IWAD or PWAD id\n", filename);
      }
    }
    header.numlumps = LONG(header.numlumps);
    header.infotableofs = LONG(header.infotableofs);
    length = header.numlumps * sizeof(filelump_t);
    fileinfo = Z_New(filelump_t, header.numlumps);
    fileinfoalloced = true; // cheap, but functional
    lseek(handle, header.infotableofs, SEEK_SET);
    read(handle, fileinfo, length);
    numlumps += header.numlumps;
  }
  else
  {
    // single lump file
    data_files[wadfile].type = FILE_SINGLELUMP;
    fileinfo = &singleinfo;
    singleinfo.filepos = 0;
    singleinfo.size = LONG(FileLength(handle));
    ExtractFileBase(filename, singleinfo.name);
    numlumps++;
  }

  // Fill in lumpinfo
  Z_Resize(lumpinfo, lumpinfo_t, numlumps);

  lump_p = &lumpinfo[startlump];

  data_files[wadfile].handle = reloadlump == -1 ? -1 : handle;

  startfileinfo = fileinfo; // keep for de-alloc
  for (i = startlump; i < numlumps; i++, lump_p++, fileinfo++)
  {
    lump_p->position = LONG(fileinfo->filepos);
    lump_p->size = LONG(fileinfo->size);
    lump_p->file = wadfile;
    strncpy(lump_p->name, fileinfo->name, 8);

    if (!strncmp(lump_p->name, "PNAMES", 8))
      pnames = i;
    else if (!strncmp(lump_p->name, "TEXTURE1", 8))
      texture1 = i;
    else if (!strncmp(lump_p->name, "TEXTURE2", 8))
      texture2 = i;
    // -KM- 1998/12/16 Load DDF/RSCRIPT file from wad.
    else if (allow_ddf)
    {
      for (j = 0; j < sizeof(DDF_Readers) / sizeof(DDF_Readers[0]); j++)
      {
        if (!strncmp(lump_p->name, DDF_Readers[j].name, 8))
        {
          DDF_Readers[j].lump = i;
          break;
        }
      }
    }

    sprite_lists += (IsS_START(lump_p->name));
    sprite_lists += (IsS_END(lump_p->name));
    flat_lists += (IsF_START(lump_p->name));
    flat_lists += (IsF_END(lump_p->name));
    patch_lists += (IsP_START(lump_p->name));
    patch_lists += (IsP_END(lump_p->name));

  }

  if (reloadlump == -1)
    close(handle);

  MapSetup();

  // -KM- 1999/01/31 Load lumps in correct order.
  for (j = 0; j < sizeof(DDF_Readers) / sizeof(DDF_Readers[0]); j++)
  {
    if (DDF_Readers[j].lump >= 0)
    {
      char *data;

      data = Z_New(char, W_LumpLength(DDF_Readers[j].lump) + 1);
      W_ReadLump(DDF_Readers[j].lump, data);
      data[W_LumpLength(DDF_Readers[j].lump)] = 0;

      // call read function
      (* DDF_Readers[j].func)(data, W_LumpLength(DDF_Readers[j].lump));

      Z_Free(data);
    }
  }

  RegisterTexLumps(pnames, texture1, texture2);
  I_Printf("\n");

  if (fileinfoalloced)
    Z_Free(startfileinfo);

  // -AJA- 1999/12/25: What did Santa bring EDGE ?  Just some support
  //       for "GWA" files (part of the "GL-Friendly Nodes" specs).
  
  if (M_CheckExtension("wad", filename) == EXT_MATCHING)
  {
    char namebuf[1024];

    int len = strlen(filename);

    // replace WAD extension with GWA
    strcpy(namebuf, filename);

    Z_StrNCpy(namebuf + len - 3, EDGEGWAEXT, 3);

    if (access(namebuf, R_OK) == 0)
    {
      // OK, the GWA file exists.  Load it.  This recursion bit is
      // rather sneaky, hopefully it doesn't break anything...

      AddFile(namebuf, false);
    }
  }
}

int *W_GetList(char name, int *num)
{
  int i, j, *list;
  int in_list = 0, count = 0;
  boolean_t flag;

  // -ACB- 23/02/2000 Proper ANSI C++ Implementation Init
  startn[0] = name;
  starta[0] = name; starta[1] = name;
  start1[0] = name;
  start2[0] = name;
  start3[0] = name;

  endn[0] = name;
  enda[0] = name; enda[1] = name;
  end1[0] = name;
  end2[0] = name;
  end3[0] = name;

  for (i = 0; i < numlumps; i++)
  {
    if (!strncmp(lumpinfo[i].name, startn, 8))  // Check for X_START
    {
      in_list++;
      continue;
    }
    if (!strncmp(lumpinfo[i].name, starta, 8))  // Check for XX_START

    {
      if (!lax_list_syntax)
        I_Error("W_GetList: flag %s encountered in sprites list", starta);
      in_list++;
      continue;
    }
    else if (!strncmp(lumpinfo[i].name, start1, 8)  // Check for X1_START
           || !strncmp(lumpinfo[i].name, start2, 8)  // Check for X2_START
         || !strncmp(lumpinfo[i].name, start3, 8))  // Check for X3_START

    {
      continue;
    }
    else if (!strncmp(lumpinfo[i].name, endn, 8))  // Check for X_END

    {
      in_list--;
      continue;
    }
    if (!strncmp(lumpinfo[i].name, enda, 8))  // Check for XX_END

    {
      if (!lax_list_syntax)
        I_Error("W_GetList: flag %s encountered in sprites list", enda);
      in_list--;
      continue;
    }
    else if (!strncmp(lumpinfo[i].name, end1, 8)  // Check for X1_END
           || !strncmp(lumpinfo[i].name, end2, 8)  // Check for X2_END
         || !strncmp(lumpinfo[i].name, end3, 8))  // Check for X3_END

    {
      continue;
    }
    else if (lumpinfo[i].size == 0)
    {
      continue;
    }
    else if (in_list > 0)  // If we are currently in the list then add lumps

    {
      count++;
    }
  }

  in_list = 0;
  list = Z_New(int, count);

  count = 0;

  // Search all lumps, forwards, cause backwards screws up
  // animations.
  for (i = 0; i < numlumps; i++)
  {
    // If it is a marker, begin adding lumps
    if (!strncmp(lumpinfo[i].name, startn, 8))  // Check for X_START

    {
      in_list++;
      continue;
    }
    if (!strncmp(lumpinfo[i].name, starta, 8))  // Check for XX_START

    {
      in_list++;
      continue;
    }
    else if (!strncmp(lumpinfo[i].name, start1, 8)  // Check for X1_START
           || !strncmp(lumpinfo[i].name, start2, 8)  // Check for X2_START
         || !strncmp(lumpinfo[i].name, start3, 8))  // Check for X3_START

    {
      continue;
    }
    else if (!strncmp(lumpinfo[i].name, endn, 8))  // Check for X_END

    {
      in_list--;
      continue;
    }
    if (!strncmp(lumpinfo[i].name, enda, 8))  // Check for XX_END

    {
      in_list--;
      continue;
    }
    else if (!strncmp(lumpinfo[i].name, end1, 8)  // Check for X1_END
           || !strncmp(lumpinfo[i].name, end2, 8)  // Check for X2_END
         || !strncmp(lumpinfo[i].name, end3, 8))  // Check for X3_END

    {
      continue;
    }
    else if (lumpinfo[i].size == 0)
    {
      continue;
    }
    else if (in_list > 0)  // If we are currently in the list then add lumps

    {
      flag = true;

      for (j = 0; (j < count) && flag; j++)  // Check to make sure we have no duplicates

      {
        if (!strnicmp(lumpinfo[i].name, lumpinfo[list[j]].name, 8))
        {
          list[j] = i;
          flag = false;
        }
      }

      // Increase the count
      if (flag)
        list[count++] = i;
    }
  }

  *num = count;
  Z_Resize(list, int, count);

  return list;
}

//
// W_Reload
//
// Flushes any of the reloadable lumps in memory
// and reloads the directory.
//
void W_Reload(void)
{
  wadinfo_t header;
  int lumpcount;
  lumpinfo_t *lump_p;
  int i;
  int handle;
  int length;
  int filenum;
  filelump_t *fileinfo;
  data_file_t *f;

  for (filenum = 0; filenum <= wadfile; filenum++)
  {
    f = &data_files[filenum];
    if (f->reloadlump == -1)
      continue;

    if ((handle = open(f->file_name, O_RDONLY | O_BINARY)) == -1)
      I_Error("W_Reload: couldn't open %s", f->file_name);

    read(handle, &header, sizeof(header));
    lumpcount = LONG(header.numlumps);
    header.infotableofs = LONG(header.infotableofs);
    length = lumpcount * sizeof(filelump_t);
    fileinfo = Z_New(filelump_t, lumpcount);
    lseek(handle, header.infotableofs, SEEK_SET);
    read(handle, fileinfo, length);

    // Fill in lumpinfo
    lump_p = &lumpinfo[f->reloadlump];

    for (i=f->reloadlump; i<(f->reloadlump+lumpcount); i++, lump_p++, fileinfo++)
    {
      if (lump_p->file != filenum)
        // only allow the number of lumps we had at startup.
        continue;
      
      if (lumplookup[i])
      {
        if (lumplookup[i]->users)
          I_Error("W_Reload: Leak detected!");
        FreeLump(lumplookup[i]);
      }
      
      lump_p->position = LONG(fileinfo->filepos);
      lump_p->size = LONG(fileinfo->size);
    }

    close(handle);

    Z_Free(fileinfo);
  }
}

//
// FlushLumpCache
//
// Flushes parts of the cache. High urgency means more is flushed.
static void FlushLumpCache(z_urgency_e urgency)
{
  int bytes_to_free = 0;
  lumpheader_t *h;

  switch (urgency)
  {
    case Z_UrgencyLow: bytes_to_free = cache_size / 16; break;
    case Z_UrgencyMedium: bytes_to_free = cache_size / 8; break;
    case Z_UrgencyHigh: bytes_to_free = cache_size / 2; break;
    case Z_UrgencyExtreme: bytes_to_free = INT_MAX; break;
    default: I_Error("FlushLumpCache: Invalid urgency level %d", urgency);
  }

  for (h = &lumphead; h->next != &lumphead && bytes_to_free > 0;)
  {
    if (h->next->users == 0)
    {
      bytes_to_free -= W_LumpLength(h->next->lumpindex);
      FreeLump(h->next);
    }
    else
      h = h->next;
  }
}

static void InitCaches(void)
{
  lumphead.next = lumphead.prev = &lumphead;
  Z_RegisterCacheFlusher(FlushLumpCache);
}

//
// W_AddRawFilename
//
void W_AddRawFilename(const char *file, boolean_t allow_ddf)
{
  raw_filename_t *r;

  I_WriteDebug("Added: %s\n", file);

  if (addwadnum == maxwadfiles)
    Z_Resize(wadfiles, raw_filename_t, ++maxwadfiles + 1);
  
  r = &wadfiles[addwadnum++];

  r->name = Z_StrDup(file);
  r->allow_ddf = allow_ddf;
}

//
// W_InitMultipleFiles
//
// Pass a null terminated list of files to use.
// All files are optional, but at least one file must be found.
// Files with a .wad extension are idlink files with multiple lumps.
// Other files are single lumps with the base filename for the lump name.
// Lump names can appear multiple times.
// The name searcher looks backwards, so a later file
//   does override all earlier ones.
//
boolean_t W_InitMultipleFiles(void)
{
  int r;

  InitCaches();

  // command line options for grouping sprites, flats & patches
  if (M_CheckParm("-gfixall"))
    lax_sprite_rot = 1;

  lax_list_syntax = (lax_list_syntax && (!M_CheckParm("-gstrict")));
  lax_sprite_rot = (lax_sprite_rot || (M_CheckParm("-gignrot")));

  // open all the files, load headers, and count lumps
  numlumps = 0;

  // will be realloced as lumps are added
  lumpinfo = NULL;

  for (r=0; r < addwadnum; r++)
    AddFile(wadfiles[r].name, wadfiles[r].allow_ddf);

  if (!numlumps)
  {
    I_Error("W_InitMultipleFiles: no files found");
    return false;
  }

  // check for errors
  if (sprite_lists % 2)
    I_Error("W_InitMultipleFiles: sprite begin..end flags don't match");
  
  if (flat_lists % 2)
    I_Error("W_InitMultipleFiles: flat begin..end flags don't match");
  
  if (patch_lists % 2)
    I_Error("W_InitMultipleFiles: patch begin..end flags don't match");
  
  if (lax_sprite_rot)
    I_Printf("Disengaging sprite rotation/limit checks\n");

  // set up caching
  lumplookup = Z_ClearNew(lumpheader_t *, numlumps);

  W_InitTextures();

  return true;
}

//
// W_InitFile
// Just initialise from a single file.
//
/* UNUSED
void W_InitFile(char *filename)
{
  char *names[2];

  names[0] = filename;
  names[1] = NULL;
  W_InitMultipleFiles(names);
}
*/

static int SearchNumForName(const char *name, int wadfile, int from, int to)
{
  union
  {
    char s[9];
    int x[2];

  }
  name8;

  int v1;
  int v2;
  int dir = from > to ? -1 : 1;
  lumpinfo_t *lump_p;

  // make the name into two integers for easy compares
  Z_StrNCpy(name8.s, name, 8);

  // case insensitive
  strupr(name8.s);

  v1 = name8.x[0];
  v2 = name8.x[1];

  // scan backwards so patch lump files take precedence
  for (lump_p = &lumpinfo[from]; lump_p != &lumpinfo[to]; lump_p += dir)
  {
    if ((wadfile >= 0) && (lump_p->file != wadfile))
      continue;
    if (*(int *)lump_p->name == v1
        && *(int *)&lump_p->name[4] == v2)
      return lump_p - lumpinfo;
  }

  // TFB. Not found.
  return -1;
}

//
// W_CheckNumForName
// Returns -1 if name not found.
//
// -ACB- 1998/08/09 Removed ifdef 0 stuff.
// -ACB- 1999/09/18 Added name to error message 
//
int W_CheckNumForName2(const char *name)
{
  int i;
  char buf[9];

  for (i = 0; name[i]; i++)
  {
#ifdef DEVELOPERS
    if (i > 8)
      I_Error("W_CheckNumForName: Name '%s' longer than 8 chars!", name);
#endif
    buf[i] = toupper(name[i]);
  }
  buf[i] = 0;

#define CMP(a) (strncmp(lumpinfo[lumpmap[a]].name, buf, 8) < 0)
  BSEARCH(numlumps, i);
#undef CMP

  if (!strncmp(lumpinfo[lumpmap[i]].name, buf, 8))
    return lumpmap[i];

  return -1;
}

//
// W_GetNumForName
// Calls W_CheckNumForName, but bombs out if not found.
//
int W_GetNumForName2(const char *name)
{
  int i;

  if ((i = W_CheckNumForName2(name)) == -1)
    I_Error("W_GetNumForName: \'%.8s\' not found!", name);

  return i;
}

//
// W_VerifyLumpName
//
// Verifies that the given lump number is valid and has the given
// name.
//
// -AJA- 1999/11/26: written.

boolean_t W_VerifyLumpName(int lump, const char *name)
{
  if (lump >= numlumps)
    return false;
  
  return (strncmp(lumpinfo[lump].name, name, 8) == 0);
}

//
// W_LumpLength
// Returns the buffer size needed to load the given lump.
//
int W_LumpLength(int lump)
{
  if (lump >= numlumps)
    I_Error("W_LumpLength: %i >= numlumps", lump);

  return lumpinfo[lump].size;
}

//
// W_ReadLump
//
// Loads the lump into the given buffer,
// which must be >= W_LumpLength().
//
void W_ReadLump(int lump, void *dest)
{
  int c;
  int handle;
  lumpinfo_t *l;
  data_file_t *f;

  if (lump >= numlumps)
    I_Error("W_ReadLump: %i >= numlumps", lump);

  l = lumpinfo + lump;
  f = &data_files[l->file];

  // -KM- 1998/07/31 This puts the loading icon in the corner of the screen :-)
  M_DisplayDisk();

  if (f->handle == -1)
  {
    // reloadable file, so use open / read / close
    if ((handle = open(f->file_name, O_RDONLY | O_BINARY)) == -1)
      I_Error("W_ReadLump: couldn't open %s", f->file_name);
  }
  else
  {
    handle = f->handle;
  }

  lseek(handle, l->position, SEEK_SET);
  c = read(handle, dest, l->size);

  if (c < l->size)
    I_Error("W_ReadLump: only read %i of %i on lump %i", c, l->size, lump);

  if (f->handle == -1)
    close(handle);
}

void W_DoneWithLump(const void *ptr)
{
  lumpheader_t *h = ((lumpheader_t *)ptr); // Intentional Const Override

#ifdef DEVELOPERS
  if (h == NULL)
    I_Error("W_DoneWithLump: NULL pointer");
  if (h[-1].id != LUMPID)
    I_Error("W_DoneWithLump: id != LUMPID");
  if (h[-1].users == 0)
    I_Error("W_DoneWithLump: lump %d has no users!", h[-1].lumpindex);
#endif
  h--;
  h->users--;
  if (h->users == 0)
  {
    // Move the item to the tail.
    h->prev->next = h->next;
    h->next->prev = h->prev;
    h->prev = lumphead.prev;
    h->next = &lumphead;
    h->prev->next = h;
    h->next->prev = h;
    MarkAsCached(h);
  }
}

//
// W_DoneWithLump_Flushable
//
// Call this if the lump probably won't be used for a while, to hint the
// system to flush it early.
//
// Useful if you are creating a cache for e.g. some kind of lump
// conversions (like the sound cache).
//
void W_DoneWithLump_Flushable(const void *ptr)
{
  lumpheader_t *h = ((lumpheader_t *)ptr); // Intentional Const Override

#ifdef DEVELOPERS
  if (h == NULL)
    I_Error("W_DoneWithLump: NULL pointer");
  h--;
  if (h->id != LUMPID)
    I_Error("W_DoneWithLump: id != LUMPID");
  if (h->users == 0)
    I_Error("W_DoneWithLump: lump %d has no users!", h->lumpindex);
#endif
  h->users--;
  if (h->users == 0)
  {
    // Move the item to the head of the list.
    h->prev->next = h->next;
    h->next->prev = h->prev;
    h->next = lumphead.next;
    h->prev = &lumphead;
    h->prev->next = h;
    h->next->prev = h;
    MarkAsCached(h);
  }
}

//
// W_CacheLumpNum
//
const void *W_CacheLumpNum2 (int lump)
{
  lumpheader_t *h;

#ifdef DEVELOPERS
  if ((unsigned int)lump >= (unsigned int)numlumps)
    I_Error("W_CacheLumpNum: %i >= numlumps", lump);
#endif

  h = lumplookup[lump];

  if (h)
  {
    // cache hit
    if (h->users == 0)
      cache_size -= W_LumpLength(h->lumpindex);
    h->users++;
  }
  else
  {
    // cache miss. load the new item.
    h = (lumpheader_t *) Z_Malloc(sizeof(lumpheader_t) + W_LumpLength(lump));
    lumplookup[lump] = h;
#ifdef DEVELOPERS
    h->id = LUMPID;
#endif
    h->lumpindex = lump;
    h->users = 1;
    h->prev = lumphead.prev;
    h->next = &lumphead;
    h->prev->next = h;
    lumphead.prev = h;

    W_ReadLump(lump, (void *)(h + 1));
  }

  return (void *)(h + 1);
}

//
// W_CacheLumpName
//
const void *W_CacheLumpName2(const char *name)
{
  return W_CacheLumpNum2(W_GetNumForName2(name));
}

//
// W_PreCacheLumpNum
//
// Attempts to load lump into the cache, if it isn't already there
void W_PreCacheLumpNum(int lump)
{
  W_DoneWithLump(W_CacheLumpNum(lump));
}

//
// W_PreCacheLumpName
//
void W_PreCacheLumpName(const char *name)
{
  W_DoneWithLump(W_CacheLumpName(name));
}

int W_CacheInfo(int level)
{
  lumpheader_t *h;
  int value = 0;

  for (h = lumphead.next; h != &lumphead; h = h->next)
  {
    if ((level & 1) && h->users)
      value += W_LumpLength(h->lumpindex);
    if ((level & 2) && !h->users)
      value += W_LumpLength(h->lumpindex);
  }
  return value;
}

//
// W_LoadLumpNum
//
// Returns a copy of the lump (it is your responsibility to free it)
void *W_LoadLumpNum(int lump)
{
  void *p;
  const void *cached;
  int length = W_LumpLength(lump);
  p = (void *) Z_Malloc(length);
  cached = W_CacheLumpNum2(lump);
  memcpy(p, cached, length);
  W_DoneWithLump(cached);
  return p;
}

//
// W_LoadLumpName
//
void *W_LoadLumpName(const char *name)
{
  return W_LoadLumpNum(W_GetNumForName2(name));
}

//
// Profile
//
static int info[2500][10];
static int profilecount;

static void Profile(void)
{
  int i;
  lumpheader_t *ptr;
  char ch;
  FILE *f;
  int j;
  char name[9];

  for (i = 0; i < numlumps; i++)
  {
    ptr = lumplookup[i];
    if (!ptr)
    {
      ch = ' ';
      // -AJA- is the following line bogus ???
      // continue;
    }
    else
    {
      if (ptr->users)
        ch = 'S';
      else
        ch = 'P';
    }
    info[i][profilecount] = ch;
  }
  profilecount++;

  f = fopen("waddump.txt", "w");
  name[8] = 0;

  for (i = 0; i < numlumps; i++)
  {
    fprintf(f, "%.8s ", lumpinfo[i].name);

    for (j = 0; j < profilecount; j++)
      fprintf(f, "    %c", info[i][j]);

    fprintf(f, "\n");
  }
  fclose(f);
}
