//----------------------------------------------------------------------------
//  EDGE WIN32 DirectMusic MIDI Handler Functions 
//----------------------------------------------------------------------------
// 
//  Copyright (c) 1999-2000 The Edge Team.
// 
//  This program is free software; you can redistribute it and/or
//  modify it under the terms of the GNU General Public License
//  as published by the Free Software Foundation; either version 2
//  of the License, or (at your option) any later version.
//
//  This program is distributed in the hope that it will be useful,
//  but WITHOUT ANY WARRANTY; without even the implied warranty of
//  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//  GNU General Public License for more details.
//
//----------------------------------------------------------------------------
//

#include "..\i_defs.h"
#include "i_sysinc.h"

#include "..\m_misc.h"

typedef enum
{
  DM_NOTINITED = 0,
  DM_PLAYING   = 2,
  DM_PAUSED    = 3,
  DM_STOPPED   = 4
}
dmsegstate_e;

typedef struct
{
  IDirectMusicSegment *seg;           // DirectSound Segment
  IDirectMusicSegmentState *segstate; // DirectSound Segment State
  MUSIC_TIME startmtime;              // Start time
  MUSIC_TIME mtimeoffset;             // Offset time
  REFERENCE_TIME startrtime;
  REFERENCE_TIME rtimeoffset;
  dmsegstate_e state;
  boolean_t looping;
}
dirmusseg_t;

// ==== PRIVATE VARIABLES ====
// HWND handle
static HWND winhandle;

// Direct Music Interfaces
static IDirectMusic *musicobject          = NULL;
static IDirectMusicLoader *musloader      = NULL;
static IDirectMusicPerformance *performer = NULL;
static IDirectMusicPort *musicport        = NULL;

// Port ID of desired DirectMusic Port
static DMUS_PORTCAPS dmportcaps;

// DirectMusic Segment (song) playing
static dirmusseg_t *currmusic = NULL;

static boolean_t inited = false; // System started?
static boolean_t active = false; // System running?

// ===========================

// ============= INTERNALS =============

//
// WideConvert
//
// Converts a Unicode (Wide) string into a
// C type string.
//
// -ACB- 2000/01/08
//
char *WideConvert(WCHAR *widestr)
{
  int count;
  char *retstr;

  if (!widestr)
    return NULL;

  count = 0;
  while (widestr[count] != 0)
    count++;

  retstr = calloc(sizeof(char), (count+1));
  if (!retstr)
    return NULL;

  if (!WideCharToMultiByte(CP_ACP, 0, widestr, count, retstr, count, NULL, NULL))
  {
    free(retstr);
    return NULL;
  }

  return retstr;
}

//
// CStrConcert
//
// This converts a C-type null terminated string
// into a Unicode (Wide) string.
//
// -ACB- 2000/01/08 
// 
WCHAR *CStrConvert(char *cstr)
{
  int count;
  WCHAR *retwstr;

  if (!cstr)
    return NULL;

  count=strlen(cstr);

  retwstr = calloc(sizeof(WCHAR), (count+1));
  if (!retwstr)
    return NULL;

  if (!MultiByteToWideChar(CP_ACP, MB_PRECOMPOSED, cstr, count, retwstr, count))
  {
    free(retwstr);
    return NULL;
  }

  return retwstr;
}

//
// GetDirName
//
// This function gets the directory info from the 'srcname'. 'srcname' is
// a filename which may or may not have directory info. This directory info
// is placed the 'deststr' - it is assumes to previously allocated and have
// a maximum length 'maxlen'.
//
// Notes:
//   If 'srcname' has no directory info the full path name
//   of the current directory is returned.
//
//   All directory info is considered to be relative to the current directory.
//
// -ACB- 2000/01/09 
//
boolean_t GetDirName(char *srcname, char *destdir, unsigned int maxlen)
{
  int srcstrlen;
  int count;
  int firstsep;
  int lastsep;
  char currdir[_MAX_PATH];

  // Sanity checks...
  if (!srcname || !destdir)
    return false;

  if (_getcwd(currdir, _MAX_PATH) == NULL)
    return false; // we need the current directory

  if (strlen(currdir) > maxlen)
    return false; // too big

  srcstrlen = strlen(srcname);

  count=srcstrlen;
  while(count >= 0 && srcname[count] != '\\')
    count--;

  if (count == -1 || srcstrlen < 2)
  {
    strcpy(destdir, currdir);
    return true; 
  }

  lastsep = count;

  count=0;
  while(count < srcstrlen && srcname[count] != '\\')
    count++;

  firstsep = count;

  // First and last directory seperator are one and the same.
  if (firstsep == lastsep)
  {
   if (firstsep == 0 || srcname[0] == '.')
    {
      strcpy(destdir, currdir);
      return true; 
    }

    if (strlen(currdir) > (maxlen-(1+firstsep)))
      return false; // too big

    strcpy(destdir, currdir);
    strcat(destdir, "\\");
    strncat(destdir, srcname, firstsep);
    return true;
  }
  else if (srcname[0] == '.' && firstsep == 1)
  {
    if (strlen(currdir) > (1+(maxlen-(lastsep-firstsep))))
      return false; // too big

    strcpy(destdir, currdir);
    strncat(destdir, &srcname[firstsep], lastsep-firstsep);
    return true;
  }
  else if (firstsep == 0)
  {
    if (strlen(currdir) > (maxlen-lastsep))
      return false; // too big

    strcpy(destdir, currdir);
    strncat(destdir, srcname, lastsep);
    return true;
  }

  if (strlen(currdir) > (maxlen-(lastsep+1)))
    return false; // too big

  strcpy(destdir, currdir);
  strcat(destdir, "\\");
  strncat(destdir, srcname, lastsep);

  return true;
}

//
// StripDirInfo
//
boolean_t StripDirInfo(char *filename)
{
  char *tempptr;
  int filenamlen;
  int count;

  if (!filename)
    return false;

  filenamlen = strlen(filename);

  count=filenamlen;
  while(count > -1 && filename[count] != '\\')
    count--;

  if (count == -1)
    return true; // OK: no changes need be made

  tempptr = &filename[count+1];

  count=0;
  while (*tempptr)
  {
    filename[count] = *tempptr;
    tempptr++;
    count++;
  }

  filename[count] = '\0';

  return true;
}

//
// AddDirectMusicPort
//
boolean_t CheckDirectMusicPort(DMUS_PORTCAPS *pcaps)
{
  char *desc;

  // Add only output ports
  if (pcaps->dwClass != DMUS_PC_OUTPUTCLASS)
    return false;

  // Software Synthesis is a big performance eater
  if (pcaps->dwFlags & DMUS_PC_SOFTWARESYNTH)
    return false;

  // Can't do shit with legacy stuff
  if (pcaps->dwFlags & DMUS_PORT_WINMM_DRIVER)
    return false;

  desc = WideConvert(pcaps->wszDescription);

  if (stricmp(desc, "MIDI Mapper"))
  {
    free(desc);
    return false;
  }
  
  free(desc);
  return true;
}

//
// CreatePerformance
//
// Creates the performance manager for DirectMusic (if possible)
// and sets the 'performer' pointer. If this fails 'performer' is
// set to NULL.
//
// This does more than just init the performance, it also adds
// the DirectMusic port.
//
void CreatePerformance(void)
{
  HRESULT result;
  DMUS_PORTPARAMS portparams; 

  if (!inited) // gotta have a music object
    return;

  // Create the performance
  result = CoCreateInstance(&CLSID_DirectMusicPerformance, NULL, CLSCTX_INPROC, &IID_IDirectMusicPerformance, &performer);
  if (result != S_OK)
  {
    I_PostMusicError("CreatePerformance: Unable to create DirectMusicPerformance");
    performer = NULL;
    return;
  }
  
  result = performer->lpVtbl->Init(performer, &musicobject, NULL, winhandle);
  
  if (result != S_OK)
  {
    I_PostMusicError("CreatePerformance: Unable to Init DirectMusic Performance");
    performer->lpVtbl->Release(performer);
    performer = NULL;
    return;
  }

  musicobject->lpVtbl->SetDirectSound(musicobject, I_SoundReturnObject(), NULL);

  memset(&portparams, 0, sizeof(DMUS_PORTPARAMS));
  portparams.dwSize = sizeof(DMUS_PORTPARAMS);

  // Lets get ourselves a port...
  result = musicobject->lpVtbl->CreatePort(musicobject, &dmportcaps.guidPort, &portparams, &musicport, NULL);   
  if (result != S_OK)
  {
    I_PostMusicError("CreatePerformance: Unable to create DirectMusic Port");
    performer->lpVtbl->Release(performer);
    performer = NULL;
    return;
  }

  if (performer->lpVtbl->AddPort(performer, musicport) != S_OK)
  {
    I_PostMusicError("I_DirectMusicPlayback: Failed to add output port");
    musicport->lpVtbl->Release(musicport);
    musicport = NULL;
    performer->lpVtbl->Release(performer);
    performer = NULL;
    return;
  }

  if (performer->lpVtbl->AssignPChannelBlock(performer, 0, musicport, 1) != S_OK)
  {
    I_PostMusicError("I_DirectMusicPlayback: Failed to assign pchanblock");
    musicport->lpVtbl->Release(musicport);
    musicport = NULL;
    performer->lpVtbl->Release(performer);
    performer = NULL;
    return;
  }

 

  active = true;
  return;
}

// =====================================

//
// I_StartupDirectMusic
//
// Starts up as much of DirectMusic as we need.
//
boolean_t I_StartupDirectMusic(void *sysinfo)
{
  HRESULT result;
  DWORD index;
  DMUS_PORTCAPS portcaps;

  // Allow for re-start
  if (inited)
    I_ShutdownDirectMusic();

  // Get main window
  winhandle = *((HWND*)sysinfo);

  // Create the music object
  result = CoCreateInstance(&CLSID_DirectMusic, NULL, CLSCTX_INPROC, &IID_IDirectMusic, &musicobject);
  if (result != S_OK)
  {
    I_PostMusicError( "I_StartupDirectMusic: Unable to create DirectMusic");
    return false;
  }

  // Enumerate ports and get a useable port
  memset(&portcaps, 0, sizeof(DMUS_PORTCAPS));
  portcaps.dwSize = sizeof(DMUS_PORTCAPS);
  
  index=0;
  result = S_OK;
  while(result != S_FALSE)
  {
    result = musicobject->lpVtbl->EnumPort(musicobject, index, &portcaps);
    if (CheckDirectMusicPort(&portcaps))
    {
      result = S_OK;
      break;
    }
    index++;
  }

  // if 'result' is not S_OK then CheckDirectMusicPort() didn't like any ports
  if (result != S_OK)
  {
    I_PostMusicError( "I_StartupDirectMusic: Unable to get a decent DirectMusic port");
    return false;
  }

  memcpy(&dmportcaps, &portcaps, sizeof(DMUS_PORTCAPS));
  
  inited = true;
  currmusic = NULL;

  return true;

}

//
// I_DirectMusicPlayback
//
// Plays back file given.
//
int I_DirectMusicPlayback(char *filename, boolean_t loops)
{
  if (!inited)
    return -1;

  // This checks that a song is not currently active, which means we have to create the performer
  if (!active)
  {
    CreatePerformance(); // Get this to setup 'performer' if possible

    if (!performer)
      return -1;
  }

  return -1;
}

//
// I_DirectMusicTicker
//
void I_DirectMusicTicker(int *handle)
{
  if (!inited)
  {
    *handle = -1;
    return;
  }
}

//
// I_DirectMusicPause
//
void I_DirectMusicPause(int *handle)
{
  if (!inited)
  {
    *handle = -1;
    return;
  }
}

//
// I_DirectMusicResume
//
void I_DirectMusicResume(int *handle)
{
  if (!inited)
  {
    *handle = -1;
    return;
  }
}

//
// I_DirectMusicStop
//
void I_DirectMusicStop(void)
{
  if (!inited || !active)
    return;
  
  if (musicport)
  {
    musicport->lpVtbl->Release(musicport);
    musicport = NULL;
  }

  // If we are inited and active, we must have a performer
  if (performer)
  {
    performer->lpVtbl->Release(performer);
    performer = NULL;
  }

  active = false;
  return;
}

//
// I_ShutdownDirectMusic
//
void I_ShutdownDirectMusic(void)
{
  if (!inited)
    return;

  if (active)
    I_DirectMusicStop();

  if (musicobject)
  {
    musicobject->lpVtbl->Release(musicobject);
    musicobject = NULL;
  }

  inited = false;
  return;
}
