//----------------------------------------------------------------------------
//  EDGE WIN32 DirectSound Handler Functions 
//----------------------------------------------------------------------------
// 
//  Copyright (c) 1999-2000 The Edge Team.
// 
//  This program is free software; you can redistribute it and/or
//  modify it under the terms of the GNU General Public License
//  as published by the Free Software Foundation; either version 2
//  of the License, or (at your option) any later version.
//
//  This program is distributed in the hope that it will be useful,
//  but WITHOUT ANY WARRANTY; without even the implied warranty of
//  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//  GNU General Public License for more details.
//
//----------------------------------------------------------------------------
//

#include "..\i_defs.h"
#include "i_sysinc.h"

// Channel flags
typedef enum
{
  _DXS_LOOPING  = 0x01, // Is looping
  _DXS_NOBUFFER = 0x02, // Channel has no DirectSound buffer
  _DXS_PAUSED   = 0x04, // Is paused
  _DXS_PLAYING  = 0x08, // Is playing
  _DXS_REUSABLE = 0x10  // Reusable short length buffer
}
_dxsound_e;

// List and data struct of sfx for playback
typedef struct storesfx_s 
{
  unsigned int length;
  unsigned int samplerate;
  unsigned char *data;
}
storesfx_t;

// Channel for internal sound control
typedef struct channel_s 
{
  LPDIRECTSOUNDBUFFER buffer; // DirectSound buffer
  int flags;                  // Behaviour Flags
  unsigned int sfxhandle;     // Sound Effect Handle
  DWORD pausedpos;            // Position for restart
}
channel_t;

//
// Short Buffers:
//   These are inited to a certain size and used for
//   relatively short sounds that do not repeat. These are 
//   used to prevent the need to create new buffers for
//   every sound.
//
static const unsigned int SHORTBUFFLEN = 22050;

// ==== PRIVATE VARIABLES ====
// HWND handle
static HWND winhandle;

// Direct Sound Interfaces
static LPDIRECTSOUND soundobject;
static LPDIRECTSOUNDBUFFER primarybuffer;

// Stored Sound Effects
static storesfx_t **storedsfx = NULL;
static unsigned int numstoredsfx = 0;

// Channels
static channel_t **channels;
static unsigned int numchannels;

// Error Description
static char errordesc[256];
static char scratcherror[256];

// System Started?
static boolean_t inited;
// ===========================

// ==== PRIVATE FUNCTIONS ====
static LPDIRECTSOUNDBUFFER CreateBuffer(int length);
static channel_t *AddChannel(void);
static channel_t *GetDisposableChannel(int *channelid, storesfx_t *sfx);
static channel_t *GetChannel(int *channelid);
// ===========================

// ================ INTERNALS =================

void DS_Error( HRESULT hresult, char *msg )
{
  char errmsg[128];
  switch(hresult)
  {
    case DS_OK:
      strcpy(errmsg,"The request completed successfully.");
      break;
    case DSERR_ALLOCATED:
      strcpy(errmsg,"The request failed because resources, such as a priority level, were already in use by another caller.");
      break;

    case DSERR_ALREADYINITIALIZED:
      strcpy(errmsg,"The object is already initialized.");
      break;

    case DSERR_BADFORMAT:
      strcpy(errmsg,"The specified wave format is not supported.");
      break;

    case DSERR_BUFFERLOST:
      strcpy(errmsg,"The buffer memory has been lost and must be restored.");
      break;

    case DSERR_CONTROLUNAVAIL:
      strcpy(errmsg,"The control (volume, pan, and so forth) requested by the caller is not available.");
      break;

    case DSERR_GENERIC:
      strcpy(errmsg,"An undetermined error occurred inside the DirectSound subsystem.");
      break;

    case DSERR_INVALIDCALL:
      strcpy(errmsg,"This function is not valid for the current state of this object.");
      break;

    case DSERR_INVALIDPARAM:
      strcpy(errmsg,"An invalid parameter was passed to the returning function.");
      break;

    case DSERR_NOAGGREGATION:
      strcpy(errmsg,"The object does not support aggregation.");
      break;

    case DSERR_NODRIVER:
      strcpy(errmsg,"No sound driver is available for use.");
      break;

    case DSERR_OTHERAPPHASPRIO:
      strcpy(errmsg,"This value is obsolete and is not used.");
      break;

    case DSERR_OUTOFMEMORY:
      strcpy(errmsg,"The DirectSound subsystem could not allocate sufficient memory to complete the caller's request.");
      break;

    case DSERR_PRIOLEVELNEEDED:
      strcpy(errmsg,"The caller does not have the priority level required for the function to succeed.");
      break;

    case DSERR_UNINITIALIZED:
      strcpy(errmsg,"The IDirectSound::Initialize method has not been called or has not been called successfully before other methods were called.");
      break;

    case DSERR_UNSUPPORTED:
      strcpy(errmsg,"The function called is not supported at this time.");
      break;

    default:
      sprintf(errmsg,"Unknown Error Code: %04X", hresult);
      break;
  }

  sprintf(msg, "DSERR: %s\n", errmsg);
}



//
// CalculatePanning
//
// This is a conversion of seperation value from
// the range 0(left)-128(mid)-255(right) to the
// DirectSound Values
//
LONG CalculatePanning(int pan)
{
  return (LONG)(((pan-128)*DSBPAN_RIGHT)/128);
}

//
// CalculateVolume
//
// This is a conversion of seperation value from
// the range 0(quietest)-255(loudest) to the
// DirectSound Values. Due to DirectSound handling
// the calculation makes DirectSound half-vol the
// quietest (quality reasons).
//
LONG CalculateVolume(int vol)
{
  return (LONG)(-5000+((5000*vol)/256));
}

//
// CreateBuffer
//
// Creates a DirectSound Secondary Buffer of the correct length
//
LPDIRECTSOUNDBUFFER CreateBuffer(int length)
{
  LPDIRECTSOUNDBUFFER buffer;
  DSBUFFERDESC bufferdesc;
  PCMWAVEFORMAT pcmwf;

  // Set up wave format structure.
  memset(&pcmwf, 0, sizeof(PCMWAVEFORMAT));
  pcmwf.wf.wFormatTag         = WAVE_FORMAT_PCM;      
  pcmwf.wf.nChannels          = 1;
  pcmwf.wf.nSamplesPerSec     = 11025;
  pcmwf.wf.nBlockAlign        = 1; // ?
  pcmwf.wf.nAvgBytesPerSec    = 11025*1*1;
  pcmwf.wBitsPerSample        = (WORD)8;

  // Set up DSBUFFERDESC structure.
  memset(&bufferdesc, 0, sizeof(DSBUFFERDESC)); 
  bufferdesc.dwSize        = sizeof(DSBUFFERDESC);
  bufferdesc.dwFlags       = DSBCAPS_CTRLVOLUME|DSBCAPS_CTRLFREQUENCY|DSBCAPS_CTRLPAN|DSBCAPS_GETCURRENTPOSITION2;
  bufferdesc.dwBufferBytes = length;
  bufferdesc.lpwfxFormat   = (LPWAVEFORMATEX)&pcmwf;  
    
  if (soundobject->lpVtbl->CreateSoundBuffer(soundobject, &bufferdesc, &buffer, NULL) != DS_OK)
   return NULL;
 
  return buffer;
}

//
// AddChannel
//
// Add a new a new channel to the channel list
//
channel_t *AddChannel(void)
{
  channel_t **newlist;
  
  if (numchannels)
  {
    newlist = malloc(sizeof(channel_t*)*(numstoredsfx+1));
    
    if (!newlist)
    {
      strcpy(errordesc,"AddChannel: Malloc failure on channels list");
      return NULL;
    }

    memcpy(newlist, channels, sizeof(channel_t*)*numchannels);
   
    free(channels);
    channels = newlist;
  }
  else
  {
    channels = malloc(sizeof(channel_t*));
  }

  channels[numchannels] = malloc(sizeof(channel_t));
    
  if (!channels[numchannels])
  {
    strcpy(errordesc, "AddChannel: Unable to malloc for channel");
    return NULL;
  }

  channels[numchannels]->flags = _DXS_NOBUFFER;

  numchannels++;

  return channels[numchannels-1];
}

//
// GetDisposableChannel
//
// Returns a channel that will be destroyed after use
//
channel_t *GetDisposableChannel(int *channelid, storesfx_t *sfx)
{
  unsigned int i;
  channel_t *channel;

  // Search out spare channel
  for (i = 0; i < numchannels; i++)
  {
    if (channels[i]->flags & _DXS_NOBUFFER)
      break;
  }

  //
  // if 'i' is not equal to number of channels we found a channel with no
  // buffer. I think we'll take that.
  //
  if (i!=numchannels)
  {
    channel = channels[i];
  }
  else // if all spaces used, knock out a unused reusable
  {
    // Search out spare channel
    for (i = 0; i < numchannels; i++)
    {
      if (!(channels[i]->flags & _DXS_PLAYING) && (channels[i]->flags & _DXS_REUSABLE))
        break;
    }

    //
    // if 'i' is not equal to the number of channels, we have found a
    // free reusable channel. Destroy the current buffer and reset
    // the flags.
    //
    if (i != numchannels)
    {
      channel = channels[i];
      channel->buffer->lpVtbl->Release(channel->buffer);
      channel->buffer = NULL;
      channel->flags = _DXS_NOBUFFER;
    }
    else // None free? lets create one.
    {
      channel = AddChannel();

      // if no channel just return NULL: AddChannel would of set the error message
      if (!channel)
        return NULL;
    }
  }

  //
  // This looks odd, but given the fact that we got the channel variety of
  // ways, the code checks for the matching channel the old fashioned, but
  // fool proof way.
  //
  for (i = 0; i < numchannels && channels[i] != channel; i++)
    ;

  // Check for the seemingly impossible: channel not on list
  if (i==numchannels)
  {
    strcpy(errordesc, "GetDisposable: SERIOUS INTERNAL ERROR, Channel allocated did not appear on the list");
    return NULL;
  }

  *channelid = i;
  channel->buffer = CreateBuffer(sfx->length);

  // CreateBuffer failure
  if (!channel->buffer)
  {
    strcpy(errordesc, "GetDisposable: CreateBuffer() Failed");
    return NULL;
  }

  channel->flags = 0; 

  return channel;
}

//
// GetChannel
//
// Gets a reusable channel, which will not be deleted
// after use, only if needs be. This channel will be of the
// length of SHORTBUFFLEN and is only used for sound effects
// that are less than that length and do not loop.
// 
channel_t *GetChannel(int *channelid)
{
  unsigned int i,j;
  channel_t *channel;

  // Search out spare channel
  for (i = 0; i < numchannels; i++)
  {
    if (!(channels[i]->flags & _DXS_PLAYING) &&   // not playing
        !(channels[i]->flags & _DXS_NOBUFFER) &&  // has a DSBuffer
        (channels[i]->flags & _DXS_REUSABLE))     // is reusable
      break;
  }

  //
  // if 'i' is equal to numchannels, we failed to find a free reusable
  // channel.
  //
  if (i == numchannels)
  {
    //
    // Search through the current channel list for one with no
    // DSBuffer.
    //
    for (j = 0; j < numchannels && !(channels[j]->flags & _DXS_NOBUFFER); j++)
      ;

    //
    // if 'j' is not equal to the number of channels, we have found
    // a channel with no DSBuffer...
    //
    if (j != numchannels)
    {
      channel = channels[j];
    }
    else // ...otherwise we create one.
    {
      channel = AddChannel();

      // Just return NULL: AddChannel() sets the error message
      if (!channel)
        return NULL;
    }

    channel->buffer = CreateBuffer(SHORTBUFFLEN);

    if (!channel->buffer)
    {
      strcpy(errordesc,"GetChannel: CreateBuffer() Failed");
      return NULL;
    }

    channel->flags = _DXS_REUSABLE;
  }
  else
  {
    channel = channels[i];
  }

  //
  // This looks odd, but given the fact that we got the channel variety of
  // ways, the code checks for the matching channel the old fashioned, but
  // fool proof way.
  //
  for (i = 0; i < numchannels && channels[i] != channel; i++)
    ;

  // Check for the seemingly impossible: channel not on list
  if (i==numchannels)
  {
    strcpy(errordesc, "GetChannel: SERIOUS INTERNAL ERROR, Channel allocated did not appear on the list");
    return NULL;
  }

  *channelid = i;

  return channel;
}

// ============= END OF INTERNALS =============

//
// I_StartupSound
//
// Exactly what it says on the label
//
// -ACB- 1999/10/06
// 
boolean_t I_StartupSound(void *sysinfo)
{
  WAVEFORMATEX format;
  DSBUFFERDESC bufferdesc;
  HWND window;
  HRESULT result;

  winhandle     = NULL;
  soundobject   = NULL;
  primarybuffer = NULL;
  channels      = NULL;
  numchannels   = 0;
  inited        = false;
  memset(errordesc, '\0', 256);

  window = *(HWND*)sysinfo;

  // Create DirectSound     
  if (DirectSoundCreate(NULL, &soundobject, NULL) != DS_OK)
  {
    strcpy(errordesc, "I_StartupSound: Unable to Create Direct Sound Object");
    return false;
  }
    
  // Set co-op level 
  if (soundobject->lpVtbl->SetCooperativeLevel(soundobject, window, DSSCL_EXCLUSIVE) != DS_OK)
  {
    strcpy(errordesc, "I_StartupSound: SetCooperativeLevel Failed");
    soundobject->lpVtbl->Release(soundobject);
    soundobject = NULL;
    return false;
  }
    
  // Setup Primary Buffer
  memset(&bufferdesc, 0, sizeof(DSBUFFERDESC));
  bufferdesc.dwSize = sizeof(DSBUFFERDESC);
  bufferdesc.dwFlags = DSBCAPS_PRIMARYBUFFER;

  // Create Primary
  result = soundobject->lpVtbl->CreateSoundBuffer(soundobject, &bufferdesc, &primarybuffer, NULL);
  if (result != DS_OK)
  {
    DS_Error(result, errordesc);
    //strcpy(errordesc, "I_StartupSound: CreateSoundBuffer Failed");
    soundobject->lpVtbl->Release(soundobject);
    soundobject = NULL;
    return false;
  }
    
  // Set primary buffer format
  memset(&format, 0, sizeof(WAVEFORMATEX)); 
  format.wFormatTag = WAVE_FORMAT_PCM;      
  format.nChannels = 2; 
  format.nSamplesPerSec = 11025; 
  format.wBitsPerSample = 8; 
  format.nBlockAlign = 2;
  format.nAvgBytesPerSec = 11025*2*1;
  format.cbSize = 0;

  if (primarybuffer->lpVtbl->SetFormat(primarybuffer, &format) != DS_OK)
  {
    strcpy(errordesc, "I_StartupSound: primarybuffer->SetFormat Failed");
    soundobject->lpVtbl->Release(soundobject);
    soundobject = NULL;
    return false;
  }

  winhandle = *(HWND*)sysinfo;
  inited = true;

  I_Warning("I_StartupSound: DirectSound Init OK\n");
  return true;
}

//
// I_LoadSfx
//
// Loads a sound effect from data.
//
// Returns -1 on Error.
//
// -ACB- 1999/10/05 Written
//
boolean_t I_LoadSfx(const unsigned char *data, unsigned int length, unsigned int freq, unsigned int handle)
{
  unsigned int i;
  storesfx_t **oldlist;
  byte *snddata;
  
  if (!data)
  {
    strcpy(errordesc, "I_LoadSFX: Data Pointer is NULL");
    return false;
  }

  if (handle >= numstoredsfx)
  {
    i = numstoredsfx;
    numstoredsfx = handle + 1;

    if (storedsfx)
    {
      oldlist = storedsfx;
      storedsfx = (storesfx_t**)malloc(sizeof(storesfx_t*)*numstoredsfx);
      if (!storedsfx)
      {
        strcpy(errordesc, "I_LoadSFX: Alloc failed on first storedsfx_t**");
        return false;
      }
      memcpy(storedsfx, oldlist, sizeof(storesfx_t*)*i);
      free(oldlist);
    }
    else
    {
      storedsfx = (storesfx_t**)malloc(sizeof(storesfx_t*)*numstoredsfx);
      if (!storedsfx)
      {
        strcpy(errordesc, "I_LoadSFX: Alloc failed on first storedsfx_t**");
        return false;
      }
    }

    // clear any new elements
    for (; i < numstoredsfx; i++)
      storedsfx[i] = NULL;
  }

  storedsfx[handle] = malloc(sizeof(storesfx_t));
  if (!storedsfx[handle])
  {
    strcpy(errordesc, "I_LoadSFX: Unable to malloc for stored sound effect");
    return false;
  }

  // -ACB- 2000/05/17 Copy data into storedsfx structure
  snddata = malloc(sizeof(byte)*length);
  if (!snddata)
  {
    strcpy(errordesc, "I_LoadSFX: Unable to malloc for sound effect data");
    free(storedsfx[handle]);
    return false;
  }
 
  memcpy(snddata, data, sizeof(byte)*length);

  storedsfx[handle]->data       = snddata;
  storedsfx[handle]->length     = length;
  storedsfx[handle]->samplerate = freq;

  return true;
}

//
// I_UnloadSfx
//
// -ACB- 2000/02/19 Written
// -ACB- 2000/05/17 Free Sound Effects Data
//
boolean_t I_UnloadSfx(unsigned int handle)
{
  unsigned int i;

  if (handle < numstoredsfx)
  {
    strcpy(errordesc, "I_UnloadSfx: *WARNING* out of range");
    return false;
  }

  // Check handle
  if (!storedsfx[handle])
  {
    strcpy(errordesc, "I_UnloadSfx: *WARNING* NULL sample");
    return false;
  }

  // Kill playing sound effects
  for (i = 0; i < numchannels; i++)
  {
    // A paused sound effect is in use...
    if (channels[i]->sfxhandle == handle &&
        (channels[i]->flags&_DXS_PLAYING || channels[i]->flags&_DXS_PAUSED))
    {
      I_SoundKill(i);
    }
  }

  // Check handle data
  if(!storedsfx[handle]->data)
  {
    strcpy(errordesc, "I_UnloadSfx: *WARNING* NULL sample data");
    free(storedsfx[handle]);
    storedsfx[handle] = NULL;
    return false;
  }

  free(storedsfx[handle]->data);
  free(storedsfx[handle]);
  storedsfx[handle] = NULL;
  return true;
}

//
// I_SoundAlter
//
// Alters the parameters of a currently playing sound
//
// -ACB- 1999/10/06
//
boolean_t I_SoundAlter(unsigned int chanid, int pan, int vol)
{
  channel_t *channel;
  LONG actualpan;
  LONG actualvol;

  if (chanid >= numchannels)
  {
    strcpy(errordesc, "I_SoundAlter: Invalid Channel Number Given");
    return false;
  }

  channel = channels[chanid];

  if (pan < 0 || pan > 255)
  {
    strcpy(errordesc, "I_SoundAlter: Pan value given is out of range");
    return false;
  }

  if (vol < 0 || vol > 255)
  {
    strcpy(errordesc, "I_SoundAlter: Volume value given is out of range");
    return false;
  }

  actualpan = CalculatePanning(pan);
  actualvol = CalculateVolume(vol);

  if (!(channel->flags & _DXS_PLAYING))
  {
    strcpy(errordesc, "I_SoundAlter: Channel is not playing or active");
    return false;
  }

  if (channel->buffer->lpVtbl->SetVolume(channel->buffer, actualvol) != DS_OK)
  {
    strcpy(errordesc, "I_SoundAlter: Unable to set DirectSound buffer volume");
    return false;
  }

  if (channel->buffer->lpVtbl->SetPan(channel->buffer, actualpan) != DS_OK)
  {
    strcpy(errordesc, "I_SoundAlter: Unable to set DirectSound buffer pan");
    return false;
  }

  return true;
}

//
// I_SoundKill
//
// This kills a sound. Stops the sound playing and destroys
// the buffer if necessary.
//
boolean_t I_SoundKill(unsigned int chanid)
{
  channel_t *channel;

  if (chanid >= numchannels)
  {
    strcpy(errordesc, "I_SoundKill: Invalid Channel Number Given");
    return false;
  }

  channel = channels[chanid];

  if (channel->flags & _DXS_NOBUFFER)
  {
    strcpy(errordesc, "I_SoundKill: Channel has no DSBuffer");
    return false;
  }

  if (channel->flags & _DXS_PLAYING)
  {
    if (channel->buffer->lpVtbl->Stop(channel->buffer) != DS_OK)
    {
      strcpy(errordesc, "I_SoundKill: Unable to stop channel playback");
      return false;
    }
  }

  if (channel->flags & _DXS_REUSABLE)
  {
    channel->flags = _DXS_REUSABLE; 
  }
  else // dispose of the disposable
  {
    channel->buffer->lpVtbl->Release(channel->buffer);
    channel->buffer = NULL;
    channel->flags = _DXS_NOBUFFER;
  }

  channel->sfxhandle = -1;

  return true;
}

//
// I_SoundPlayback
//
// Sets a channel up for playback with given soundID. returns -1 on failure.
//
// -ACB- 1999/10/05 Written.
//
int I_SoundPlayback(unsigned int soundid, int pan, int vol, boolean_t looping)
{
  storesfx_t *sfx;
  channel_t *channel;
  LONG actualvol;
  LONG actualpan;
  LPVOID buff1;
  LPVOID buff2;
  const unsigned char *byteptr;
  DWORD len1;
  DWORD len2;
  int channelid;

  if (!inited)
  {
    strcpy(errordesc, "I_SoundPlayback: Sound System Not Inited");
    return -1;
  }

  if (pan < 0 || pan > 255)
  {
    strcpy(errordesc, "I_SoundPlayback: Pan value given is out of range");
    return -1;
  }

  if (vol < 0 || vol > 255)
  {
    strcpy(errordesc, "I_SoundPlayback: Volume value given is out of range");
    return -1;
  }

  if (soundid > numstoredsfx)
  {
    strcpy(errordesc, "I_SoundPlayback: Invalid SoundID Given");
    return -1;
  }

  actualvol = CalculateVolume(vol);
  actualpan = CalculatePanning(pan);
  sfx = storedsfx[soundid];

  //
  // For sounds that are bigger that a 'short' buffer or are
  // looping sounds, we create a buffer which will be marked
  // as disposable. The reason for this is for optimisation
  // reasons: we save time by re-using an existing buffer as
  // opposed to creating and destroying every time. We ONLY
  // do this here for looping, because DirectSound is better
  // at handling this; we also do it for long sounds because
  // streaming audio is messy and would prove to be poor
  // on slower systems.
  //
  if (looping || (sfx->length > SHORTBUFFLEN))
  {
    channel = GetDisposableChannel(&channelid, sfx);
    
    if (!channel)
      return -1;

    channel->buffer->lpVtbl->SetCurrentPosition(channel->buffer, 0);

    if (channel->buffer->lpVtbl->Lock(channel->buffer, 0, sfx->length, &buff1, &len1, &buff2, &len2, DSBLOCK_FROMWRITECURSOR) != DS_OK)
    {
      strcpy(errordesc, "I_SoundPlayback: Unable to lock DirectSound buffer");
      return -1;
    }

  }
  else
  {
    channel = GetChannel(&channelid);

    if (!channel)
      return -1;

    channel->buffer->lpVtbl->SetCurrentPosition(channel->buffer, 0);

    if (channel->buffer->lpVtbl->Lock(channel->buffer, 0, SHORTBUFFLEN, &buff1, &len1, &buff2, &len2, DSBLOCK_FROMWRITECURSOR) != DS_OK)
    {
      strcpy(errordesc, "I_SoundPlayback: Unable to lock DirectSound buffer");
      return -1;
    }
  }
  
  byteptr = sfx->data;

  memset(buff1, 0x7F, len1);

  if (buff2 && len2)
    memset(buff2, 0x7F, len2);

  // Copy the data across
  if (len1 >= sfx->length)
    memcpy(buff1, byteptr, sfx->length);
  else
    memcpy(buff1, byteptr, len1);
  
  if (channel->buffer->lpVtbl->Unlock(channel->buffer, buff1, len1, buff2, len2) != DS_OK)
  {
    strcpy(errordesc, "I_SoundPlayback: Unable to unlock DirectSound buffer");
    return -1;
  }

  if (channel->buffer->lpVtbl->SetVolume(channel->buffer, actualvol) != DS_OK)
  {
    strcpy(errordesc, "I_SoundPlayback: Unable to set DirectSound buffer volume");
    return -1;
  }

  if (channel->buffer->lpVtbl->SetPan(channel->buffer, actualpan) != DS_OK)
  {
    strcpy(errordesc, "I_SoundPlayback: Unable to set DirectSound buffer pan");
    return -1;
  }

  if (channel->buffer->lpVtbl->SetFrequency(channel->buffer, sfx->samplerate) != DS_OK)
  {
    strcpy(errordesc, "I_SoundPlayback: Unable to set DirectSound buffer freq");
    return -1;
  }

  if (looping)
  {
    if (channel->buffer->lpVtbl->Play(channel->buffer, 0, 0, DSBPLAY_LOOPING) != DS_OK)
    {
      strcpy(errordesc, "I_SoundPlayback: Unable to play sound");
      return -1;
    }

    channel->flags |= _DXS_LOOPING;
  }
  else
  {
    if (channel->buffer->lpVtbl->Play(channel->buffer, 0, 0, 0) != DS_OK)
    {
      strcpy(errordesc, "I_SoundPlayback: Unable to play sound");
      return -1;
    }
  }

  channel->flags |= _DXS_PLAYING;
  channel->sfxhandle = soundid;

  return channelid;
}

//
// I_SoundCheck
//
// Returns if a channel is currently active
//
// -ACB- 1999/10/06 Written
//
boolean_t I_SoundCheck(unsigned int chanid)
{
  if (chanid >= numchannels)
    return false;

  return ((channels[chanid]->flags & _DXS_PLAYING)?true:false);
}

//
// I_SoundPause
//
boolean_t I_SoundPause(unsigned int chanid)
{
  channel_t *channel;
  DWORD curpos;
  DWORD status;

  if (!inited)
  {
    strcpy(errordesc, "I_SoundPause: Sound System Not Inited");
    return false;
  }

  if (chanid >= numchannels)
  {
    strcpy(errordesc, "I_SoundPause: Invalid Channel Number Given");
    return false;
  }

  channel = channels[chanid];

  if (!(channel->flags & _DXS_PLAYING))
  {
    strcpy(errordesc, "I_SoundPause: Channel is not playing or active");
    return false;
  }

  channel->buffer->lpVtbl->GetStatus(channel->buffer, &status);

  // Kill any lost buffer
  if (status & DSBSTATUS_BUFFERLOST)
  {
    channel->buffer->lpVtbl->Release(channel->buffer);
    channel->buffer = NULL;
    channel->flags = _DXS_NOBUFFER;
    strcpy(errordesc, "I_SoundPause: Channel's DirectSound Buffer has been lost.");
    return false;
  }

  if (channel->buffer->lpVtbl->GetCurrentPosition(channel->buffer, &curpos, NULL) != DS_OK)
  {
    strcpy(errordesc, "I_SoundPause: Unable to get current position");
    return false;
  }

  channel->pausedpos = curpos;

  if (channel->buffer->lpVtbl->Stop(channel->buffer) != DS_OK)
  {
    strcpy(errordesc, "I_SoundPause: Unable to stop sound playing");
    return false;
  }

  channel->flags &= ~_DXS_PLAYING;
  channel->flags |= _DXS_PAUSED;

  I_Warning("Chanid: %d paused\n", chanid); 

  return true;
}

//
// I_SoundResume
//
boolean_t I_SoundResume(unsigned int chanid)
{
  channel_t *channel;
  DWORD status;

  if (!inited)
  {
    strcpy (errordesc,"I_SoundResume: Sound System Not Inited");
    return false;
  }

  if (chanid >= numchannels)
  {
    strcpy(errordesc,"I_SoundResume: Invalid Channel Num");
    return false;
  }

  channel = channels[chanid];

  if (!(channel->flags & _DXS_PAUSED))
  {
    strcpy(errordesc,"I_SoundResume: Channel is not paused");
    return false;
  }

  channel->buffer->lpVtbl->GetStatus(channel->buffer, &status);

  // Kill any lost buffer
  if (status & DSBSTATUS_BUFFERLOST)
  {
    channel->buffer->lpVtbl->Release(channel->buffer);
    channel->buffer = NULL;
    channel->flags = _DXS_NOBUFFER;
    strcpy (errordesc,"I_SoundResume: Channel's DirectSound Buffer has been lost");
    return false;
  }

  if (channel->buffer->lpVtbl->SetCurrentPosition(channel->buffer, channel->pausedpos) != DS_OK)
  {
    channel->buffer->lpVtbl->Release(channel->buffer);
    channel->buffer = NULL;
    channel->flags = _DXS_NOBUFFER;
    strcpy (errordesc,"I_SoundResume: Unable to set buffer position - effect NUKED");
    return false;
  }

  if (channel->flags & _DXS_LOOPING)
  {
    if (channel->buffer->lpVtbl->Play(channel->buffer,0,0,DSBPLAY_LOOPING) != DS_OK)
    {
      channel->buffer->lpVtbl->Release(channel->buffer);
      channel->buffer = NULL;
      channel->flags = _DXS_NOBUFFER;
      strcpy (errordesc,"I_SoundResume: Unable to restart looping playback - effect NUKED");
      return false;
    }
  }
  else
  {
    if (channel->buffer->lpVtbl->Play(channel->buffer,0,0,0) != DS_OK)
    {
      channel->buffer->lpVtbl->Release(channel->buffer);
      channel->buffer = NULL;
      channel->flags = _DXS_NOBUFFER;
      strcpy(errordesc, "I_SoundResume: Unable to restart playback - effect NUKED");
      return false;
    }
  }

  I_Warning("Chanid: %d resumed\n", chanid); 

  channel->flags &= ~_DXS_PAUSED;
  channel->flags |= _DXS_PLAYING;
  channel->pausedpos = 0;
  return true;
}

//
// I_SoundTicker
//
// Updates the sound effects
//
// -ACB- 1999/10/06 Written.
//
void I_SoundTicker(void)
{
  unsigned int i;
  channel_t *channel;
  DWORD status;

  if (!inited)
  {
    strcpy(errordesc, "I_SoundTicker: Sound System Not Inited");
    return;
  }

  for (i = 0; i < numchannels; i++)
  {
    channel = channels[i];

    if (!(channel->flags & _DXS_PLAYING) && (channel->flags & _DXS_PAUSED))
      continue;

    // -ACB- 2000/05/20 Don't process channel with no buffer....
    if (!channel->buffer)
      continue;

    channel->buffer->lpVtbl->GetStatus(channel->buffer, &status);

    // Kill any lost buffer
    if (status & DSBSTATUS_BUFFERLOST)
    {
      channel->buffer->lpVtbl->Release(channel->buffer);
      channel->buffer = NULL;
      channel->flags = _DXS_NOBUFFER; // -ACB- 2000/05/20 No buffer here!
      continue;
    }

    if (!(status & DSBSTATUS_PLAYING))
    {
      if (!(channel->flags & _DXS_REUSABLE))
      {
        channel->buffer->lpVtbl->Release(channel->buffer);
        channel->buffer = NULL;
        channel->flags = _DXS_NOBUFFER;
      }
      else
      {
        channel->flags = _DXS_REUSABLE;
      }
    }

  }

  return;
}

//
// I_ShutdownSound
//
// Shutdown of the sound output system
//
// -ACB- 1999/10/06 Written
// -ACB- 2000/05/17 Fixed to allow for sound caching
//
void I_ShutdownSound(void)
{
  unsigned int i;

  if (!inited)
    return;

  if (numchannels)
  {
    for (i = 0; i < numchannels; i++)
    {
      if (!(channels[i]->flags & _DXS_NOBUFFER))
        channels[i]->buffer->lpVtbl->Release(channels[i]->buffer);

      free(channels[i]);
    } 

    free(channels);
  }

  numchannels = 0;
  channels = NULL;

  if (numstoredsfx)
  {
    for (i=0; i<numstoredsfx; i++)
    {  
      if (storedsfx[i])
      { 
        if (storedsfx[i]->data)
          free(storedsfx[i]->data);

        free(storedsfx[i]);
      }
    }

    free(storedsfx);
  }

  numstoredsfx = 0;
  storedsfx = NULL;

  if (primarybuffer)
  {
    primarybuffer->lpVtbl->Release(primarybuffer);
    primarybuffer = NULL;
  }

  if (soundobject)
  {
    soundobject->lpVtbl->Release(soundobject);
    soundobject = NULL;
  }

  return;
}

//
// I_SoundReturnObject
//
// This really goes against the grain. DirectMusic requires
// that if a DirectSound system exists, we make it known. Trying
// to make modular code in WIN32 just got harder.
//
// -ACB- 2000/01/05
//
LPDIRECTSOUND I_SoundReturnObject(void)
{
  if (!inited)
    return NULL;

  return soundobject;
}

//
// I_SoundReturnError
//
const char *I_SoundReturnError(void)
{
  memcpy(scratcherror, errordesc, sizeof(char)*256);
  memset(errordesc, '\0', sizeof(char)*256);
  return scratcherror;
}
